/*
 *  This file is part of the Heritrix web crawler (crawler.archive.org).
 *
 *  Licensed to the Internet Archive (IA) by one or more individual 
 *  contributors. 
 *
 *  The IA licenses this file to You under the Apache License, Version 2.0
 *  (the "License"); you may not use this file except in compliance with
 *  the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.archive.modules.extractor;

import java.util.LinkedHashMap;

import org.archive.modules.CrawlURI;
import org.archive.net.UURI;
import org.archive.net.UURIFactory;
import org.archive.util.Recorder;

public class ExtractorMultipleRegexTest extends ContentExtractorTestBase {

    /*
     * Test data is actual content of
     * https://www.facebook.com/NorthCarolinaStateParks as retrieved in browser
     * on Nov 15 2012, and expected outlinks are actual urls retrieved as a
     * result of scrolling down.
     */
    protected static final String TEST_URI = "https://www.facebook.com/NorthCarolinaStateParks";
    
    protected static final String[] EXPECTED_OUTLINKS = {
        "https://www.facebook.com/ajax/pagelet/generic.php/ProfileTimelineSectionPagelet?ajaxpipe=1&ajaxpipe_token=AXiCHz3XOZzlpLVQ&no_script_path=1&data=%7B%22profile_id%22%3A143412869029%2C%22start%22%3A1349074800%2C%22end%22%3A1351753199%2C%22query_type%22%3A25%2C%22section_pagelet_id%22%3A%22pagelet_timeline_month_all_last%22%2C%22load_immediately%22%3Afalse%2C%22time_cutoff%22%3A1351254247%2C%22force_no_friend_activity%22%3Afalse%7D&__user=0&__a=1&__adt=2",
        "https://www.facebook.com/ajax/pagelet/generic.php/ProfileTimelineSectionPagelet?ajaxpipe=1&ajaxpipe_token=AXiCHz3XOZzlpLVQ&no_script_path=1&data=%7B%22profile_id%22%3A143412869029%2C%22start%22%3A1325404800%2C%22end%22%3A1357027199%2C%22query_type%22%3A8%2C%22filter_after_timestamp%22%3A1349074799%2C%22section_pagelet_id%22%3A%22pagelet_timeline_year_current%22%2C%22load_immediately%22%3Afalse%2C%22time_cutoff%22%3A1351254247%2C%22force_no_friend_activity%22%3Afalse%7D&__user=0&__a=1&__adt=3",
        "https://www.facebook.com/ajax/pagelet/generic.php/ProfileTimelineSectionPagelet?ajaxpipe=1&ajaxpipe_token=AXiCHz3XOZzlpLVQ&no_script_path=1&data=%7B%22profile_id%22%3A143412869029%2C%22start%22%3A1293868800%2C%22end%22%3A1325404799%2C%22query_type%22%3A8%2C%22section_pagelet_id%22%3A%22pagelet_timeline_year_last%22%2C%22load_immediately%22%3Afalse%2C%22time_cutoff%22%3A1351254247%2C%22force_no_friend_activity%22%3Afalse%7D&__user=0&__a=1&__adt=4",
        "https://www.facebook.com/ajax/pagelet/generic.php/ProfileTimelineSectionPagelet?ajaxpipe=1&ajaxpipe_token=AXiCHz3XOZzlpLVQ&no_script_path=1&data=%7B%22profile_id%22%3A143412869029%2C%22start%22%3A1262332800%2C%22end%22%3A1293868799%2C%22query_type%22%3A8%2C%22section_pagelet_id%22%3A%22pagelet_timeline_year_2010%22%2C%22load_immediately%22%3Afalse%2C%22time_cutoff%22%3A1351254247%2C%22force_no_friend_activity%22%3Afalse%7D&__user=0&__a=1&__adt=5",
        "https://www.facebook.com/ajax/pagelet/generic.php/ProfileTimelineSectionPagelet?ajaxpipe=1&ajaxpipe_token=AXiCHz3XOZzlpLVQ&no_script_path=1&data=%7B%22profile_id%22%3A143412869029%2C%22start%22%3A1230796800%2C%22end%22%3A1262332799%2C%22query_type%22%3A8%2C%22section_pagelet_id%22%3A%22pagelet_timeline_year_2009%22%2C%22load_immediately%22%3Afalse%2C%22time_cutoff%22%3A1351254247%2C%22force_no_friend_activity%22%3Afalse%7D&__user=0&__a=1&__adt=6",
        
        /*
         * these next 2 happened when I scolled far enough, but it appears 
         * they can't be found using the same technique as the earlier ones
         */
        // "https://www.facebook.com/ajax/pagelet/generic.php/ProfileTimelineSectionPagelet?ajaxpipe=1&ajaxpipe_token=AXiCHz3XOZzlpLVQ&no_script_path=1&data=%7B%22profile_id%22%3A143412869029%2C%22start%22%3A1291190400%2C%22end%22%3A1293868799%2C%22query_type%22%3A25%2C%22section_pagelet_id%22%3A%22pagelet_timeline_month_all_2010_12%22%2C%22load_immediately%22%3Afalse%2C%22time_cutoff%22%3A1351254247%2C%22parent_key%22%3A%22year_2010%22%2C%22force_no_friend_activity%22%3Afalse%7D&__user=0&__a=1&__adt=7",
        // "https://www.facebook.com/ajax/pagelet/generic.php/ProfileTimelineSectionPagelet?ajaxpipe=1&ajaxpipe_token=AXiCHz3XOZzlpLVQ&no_script_path=1&data=%7B%22profile_id%22%3A143412869029%2C%22start%22%3A1288594800%2C%22end%22%3A1291190399%2C%22query_type%22%3A25%2C%22section_pagelet_id%22%3A%22pagelet_timeline_month_all_2010_11%22%2C%22load_immediately%22%3Afalse%2C%22time_cutoff%22%3A1351254247%2C%22parent_key%22%3A%22year_2010%22%2C%22force_no_friend_activity%22%3Afalse%7D&__user=0&__a=1&__adt=8" 
    };

    /*
     * Actual content of https://www.facebook.com/NorthCarolinaStateParks as
     * retrieved in browser in Nov 15 2012. Java doesn't allow constants longer
     * than 65535 bytes, so we concatenate these chunks in code.
     */
    protected static final String[] TEST_CONTENT_CHUNKS = {
        "<!DOCTYPE html>\n", 
        "<html lang=\"en\" id=\"facebook\" class=\"no_js\">\n", 
        "<head><meta charset=\"utf-8\" /><script>function envFlush(a){function b(c){for(var d in a)c[d]=a[d];}if(window.requireLazy){requireLazy(['Env'],b);}else{Env=window.Env||{};b(Env);}}\n", 
        "envFlush({\"user\":\"0\",\"locale\":\"en_US\",\"method\":\"GET\",\"svn_rev\":671911,\"tier\":\"\",\"vip\":\"69.171.234.37\",\"static_base\":\"https:\\/\\/s-static.ak.facebook.com\\/\",\"www_base\":\"http:\\/\\/www.facebook.com\\/\",\"rep_lag\":2,\"fb_dtsg\":\"AQCld6oo\",\"ajaxpipe_token\":\"AXiCHz3XOZzlpLVQ\",\"lhsh\":\"1AQFaCi3m\",\"tracking_domain\":\"https:\\/\\/pixel.facebook.com\",\"retry_ajax_on_network_error\":\"1\",\"fbid_emoticons\":\"1\"});</script><script>envFlush({\"eagleEyeConfig\":{\"seed\":\"0ue1\",\"sessionStorage\":true}});CavalryLogger=false;window._script_path = \"\\/profile_page_timeline.php\";</script><noscript><meta http-equiv=\"refresh\" content=\"0; URL=/NorthCarolinaStateParks?_fb_noscript=1\" /></noscript><meta name=\"robots\" content=\"noodp, noydir\" /><meta name=\"referrer\" content=\"default\" id=\"meta_referrer\" /><meta name=\"description\" content=\"North Carolina State Parks and Recreation, Raleigh, NC. 8,945 likes &#xb7; 599 talking about this.\" /><link rel=\"alternate\" media=\"handheld\" href=\"https://www.facebook.com/NorthCarolinaStateParks\" /><title id=\"pageTitle\">North Carolina State Parks and Recreation | Facebook</title><meta property=\"og:title\" content=\"North Carolina State Parks and Recreation\" /><meta property=\"og:type\" content=\"government\" /><meta property=\"og:url\" content=\"http://www.facebook.com/NorthCarolinaStateParks\" /><meta property=\"og:site_name\" content=\"Facebook\" /><meta property=\"og:image\" content=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-ash3/174880_143412869029_133040642_n.jpg\" /><meta property=\"og:description\" content=\"The North Carolina State Parks System exists for the enjoyment, education, health and inspiration...\" /><link rel=\"shortcut icon\" href=\"https://s-static.ak.facebook.com/rsrc.php/yP/r/Ivn-CVe5TGK.ico\" /><noscript><meta http-equiv=\"X-Frame-Options\" content=\"deny\" /></noscript>\n", 
        "<link rel=\"stylesheet\" href=\"https://s-static.ak.fbcdn.net/rsrc.php/v2/yY/r/ypChFAdKEum.css\" />\n", 
        "<link rel=\"stylesheet\" href=\"https://s-static.ak.fbcdn.net/rsrc.php/v2/yz/r/bkO4dNVD63t.css\" />\n", 
        "<link rel=\"stylesheet\" href=\"https://s-static.ak.fbcdn.net/rsrc.php/v2/yi/r/RrI1xoxqHZV.css\" />\n", 
        "\n", 
        "<script src=\"https://s-static.ak.fbcdn.net/rsrc.php/v2/yl/r/GqLzR9PwXc2.js\"></script>\n", 
        "<script>window.Bootloader && Bootloader.done([\"j2Y1E\"]);</script><script>new (require(\"ServerJS\"))().handle({\"require\":[[\"lowerDomain\"],[\"QuicklingPrelude\"],[\"Primer\"]]})</script></head><body class=\"timelineLayoutLoggedOutProfile timelineLayoutLoggedOut timelineLayout pagesTimelineLayout fbx UIPage_LoggedOut webkit mac Locale_en_US\"><div id=\"FB_HiddenContainer\" style=\"position:absolute; top:-10000px; width:0px; height:0px;\"></div><div class=\"_li\"><div id=\"pagelet_bluebar\" data-referrer=\"pagelet_bluebar\"><div id=\"blueBarHolder\"><div id=\"blueBar\"><div><div class=\"loggedout_menubar_container\"><div class=\"clearfix loggedout_menubar\"><a class=\"lfloat\" href=\"/\" title=\"Go to Facebook Home\"><i class=\"fb_logo img sp_38839k sx_127aa9\"><u>Facebook logo</u></i></a><div class=\"menu_login_container rfloat\"><form id=\"login_form\" action=\"https://www.facebook.com/login.php?login_attempt=1\" method=\"post\" onsubmit=\"return window.Event &amp;&amp; Event.__inlineSubmit &amp;&amp; Event.__inlineSubmit(this,event)\"><input type=\"hidden\" name=\"lsd\" value=\"AVqaEiN6\" autocomplete=\"off\" /><table cellspacing=\"0\"><tr><td class=\"html7magic\"><label for=\"email\">Email or Phone</label></td><td class=\"html7magic\"><label for=\"pass\">Password</label></td></tr><tr><td><input type=\"text\" class=\"inputtext\" name=\"email\" id=\"email\" value=\"\" tabindex=\"1\" /></td><td><input type=\"password\" class=\"inputtext\" name=\"pass\" id=\"pass\" tabindex=\"2\" /></td><td><label class=\"uiButton uiButtonConfirm\" id=\"loginbutton\" for=\"u4kzrn22\"><input value=\"Log In\" tabindex=\"4\" type=\"submit\" id=\"u4kzrn22\" /></label></td></tr><tr><td class=\"login_form_label_field\"><div><div class=\"uiInputLabel clearfix\"><input id=\"persist_box\" type=\"checkbox\" name=\"persistent\" value=\"1\" tabindex=\"3\" class=\"uiInputLabelCheckbox\" /><label for=\"persist_box\">Keep me logged in</label></div><input type=\"hidden\" name=\"default_persistent\" value=\"0\" /></div></td><td class=\"login_form_label_field\"><a rel=\"nofollow\" href=\"http://www.facebook.com/recover/initiate\">Forgot your password?</a></td></tr></table><input type=\"hidden\" autocomplete=\"off\" id=\"next\" name=\"next\" value=\"https://www.facebook.com/NorthCarolinaStateParks\" /><input type=\"hidden\" name=\"charset_test\" value=\"&euro;,&acute;,\u20AC,\u00B4,\u6C34,\u0414,\u0404\" /><input type=\"hidden\" autocomplete=\"off\" name=\"timezone\" value=\"\" id=\"u4kzrn21\" /><input type=\"hidden\" name=\"lgnrnd\" value=\"090052_K5HA\" /><input type=\"hidden\" id=\"lgnjs\" name=\"lgnjs\" value=\"n\" /><input type=\"hidden\" autocomplete=\"off\" id=\"locale\" name=\"locale\" value=\"en_US\" /></form></div></div></div></div></div></div></div><div id=\"globalContainer\" class=\"uiContextualLayerParent\"><div id=\"content\" class=\"fb_content clearfix\"><div itemscope=\"itemscope\" itemtype=\"http://data-vocabulary.org/Organization\"><div id=\"toolbarContainer\"><div id=\"pagelet_toolbar\" data-referrer=\"pagelet_toolbar\"></div></div><div id=\"mainContainer\"><div id=\"leftColContainer\"><div id=\"leftCol\"></div></div><div id=\"contentCol\" class=\"clearfix hasRightCol\"><div id=\"rightCol\" role=\"complementary\"><div id=\"rightColContent\"></div></div><div id=\"contentArea\" role=\"main\" aria-describedby=\"pageTitle\"><div id=\"pagelet_timeline_main_column\" data-referrer=\"pagelet_timeline_main_column\" data-gt=\"&#123;&quot;profile_owner&quot;:&quot;143412869029&quot;,&quot;ref&quot;:&quot;timeline:timeline&quot;&#125;\"></div></div><div id=\"bottomContent\"></div></div></div></div></div><div id=\"pageFooter\" data-referrer=\"page_footer\"><div id=\"contentCurve\"></div><div class=\"clearfix\" id=\"footerContainer\"><div class=\"mrl lfloat\"><div class=\"fsm fwn fcg\"><span> Facebook \u00A9 2012</span> \u00B7 <a rel=\"dialog\" href=\"/ajax/intl/language_dialog.php?uri=https%3A%2F%2Fwww.facebook.com%2FNorthCarolinaStateParks\" title=\"Use Facebook in another language.\" role=\"button\">English (US)</a></div></div><div class=\"navigation fsm fwn fcg\" role=\"contentinfo\" aria-label=\"Facebook site links\"><a href=\"http://www.facebook.com/mobile/?ref=pf\" title=\"Check out Facebook Mobile.\">Mobile</a> \u00B7 <a href=\"http://www.facebook.com/find-friends?ref=pf\" title=\"Find anyone on the web.\">Find Friends</a> \u00B7 <a href=\"http://www.facebook.com/badges/?ref=pf\" title=\"Embed a Facebook badge on your website.\">Badges</a> \u00B7 <a href=\"http://www.facebook.com/directory/people/\" title=\"Browse our people directory.\">People</a> \u00B7 <a href=\"http://www.facebook.com/directory/pages/\" title=\"Browse our pages directory.\">Pages</a> \u00B7 <a href=\"http://www.facebook.com/appcenter/?ref=pf\" title=\"Check out Facebook App Center.\">Apps</a> \u00B7 <a href=\"http://www.facebook.com/appcenter/category/games/?ref=pf\" title=\"Check out Facebook games.\">Games</a> \u00B7 <a href=\"http://www.facebook.com/appcenter/category/music/?ref=pf\" title=\"Check out Facebook music apps.\">Music</a> \u00B7 <a href=\"http://www.facebook.com/facebook\" accesskey=\"8\" title=\"Read our blog, discover the resource center, and find job opportunities.\">About</a> \u00B7 <a href=\"http://www.facebook.com/campaign/landing.php?placement=pflo&amp;campaign_id=402047449186&amp;extra_1=auto\" title=\"Advertise on Facebook.\">Create an Ad</a> \u00B7 <a href=\"http://www.facebook.com/pages/create.php?ref_type=sitefooter\" title=\"Create a Page\">Create a Page</a> \u00B7 <a href=\"http://developers.facebook.com/?ref=pf\" title=\"Develop on our platform.\">Developers</a> \u00B7 <a href=\"http://www.facebook.com/careers/?ref=pf\" title=\"Make your next career move to our awesome company.\">Careers</a> \u00B7 <a href=\"http://www.facebook.com/privacy/explanation\" title=\"Learn about your privacy and Facebook.\">Privacy</a> \u00B7 <a href=\"http://www.facebook.com/help/cookies\" title=\"Learn about cookies and Facebook.\">Cookies</a> \u00B7 <a href=\"http://www.facebook.com/policies/?ref=pf\" accesskey=\"9\" title=\"Review our terms and policies.\">Terms</a> \u00B7 <a href=\"http://www.facebook.com/help/?ref=pf\" accesskey=\"0\" title=\"Visit our Help Center.\">Help</a></div></div></div></div></div><script type=\"text/javascript\">/*<![CDATA[*/function si_cj(m){setTimeout(function(){new Image().src=\"https:\\/\\/error.facebook.com\\/common\\/scribe_endpoint.php?c=si_clickjacking&t=7959\"+\"&m=\"+m;},5000);}if(top!=self && !false){try{if(parent!=top){throw 1;}var si_cj_d=[\"apps.facebook.com\",\"\\/pages\\/\",\"apps.beta.facebook.com\"];var href=top.location.href.toLowerCase();for(var i=0;i<si_cj_d.length;i++){if (href.indexOf(si_cj_d[i])>=0){throw 1;}}si_cj(\"3 \");}catch(e){si_cj(\"1 \\t\");window.document.write(\"\\u003Cstyle>body * {display:none !important;}\\u003C\\/style>\\u003Ca href=\\\"#\\\" onclick=\\\"top.location.href=window.location.href\\\" style=\\\"display:block !important;padding:10px\\\">\\u003Ci class=\\\"img sp_2vmry3 sx_ed8644\\\" style=\\\"display:block !important\\\">\\u003C\\/i>Go to Facebook.com\\u003C\\/a>\");/*iRpKrHjM*/}}/*]]>*/</script>\n", 
        "<script>Bootloader.setResourceMap({\"M7bAm\":{\"type\":\"css\",\"permanent\":1,\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/yY\\/r\\/ypChFAdKEum.css\"},\"jg0IV\":{\"type\":\"css\",\"permanent\":1,\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/yz\\/r\\/bkO4dNVD63t.css\"},\"zmU4p\":{\"type\":\"css\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/yi\\/r\\/RrI1xoxqHZV.css\"},\"X\\/Fq6\":{\"type\":\"css\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/y6\\/r\\/YlbIHaln_Rk.css\"},\"gYYej\":{\"type\":\"css\",\"permanent\":1,\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/ys\\/r\\/ltme-qREmd_.css\"},\"VDymv\":{\"type\":\"css\",\"permanent\":1,\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/ye\\/r\\/4AC1A2FNTqE.css\"}});Bootloader.setResourceMap({\"JU6Q1\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/ys\\/r\\/L0OIUBrvg93.js\"},\"MX2xp\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/yU\\/r\\/jlg9d7HhVQ1.js\"},\"mCr\\/S\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/yW\\/r\\/VWOQloMOlcl.js\"},\"UjXTQ\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/yr\\/r\\/2Mh25GaL3gk.js\"},\"T8gbI\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/yV\\/r\\/1nE5gvSZJjt.js\"},\"gM1Bs\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/y5\\/r\\/IAhlg6Orvch.js\"},\"oE4Do\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/yD\\/r\\/65c9OW-i481.js\"},\"rKoQM\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/yp\\/r\\/k0ISdhvY6BR.js\"},\"MqSmz\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/yH\\/r\\/ghlEJgSKAee.js\"},\"cNca2\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/yn\\/r\\/XQ1IPAI9NyR.js\"},\"So3Xq\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/yE\\/r\\/CcU63R6ET7_.js\"},\"bwsMw\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/y_\\/r\\/hwOyT9fmfZV.js\"},\"NAQYQ\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/yk\\/r\\/5j0Y3OPaGaT.js\"},\"zyFOp\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/y4\\/r\\/61VSNDvRJnS.js\"},\"XNAD8\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/yw\\/r\\/sJbfy4XdeB1.js\"},\"7yfEw\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/yu\\/r\\/4XVDrM9B9ui.js\"},\"XQYFa\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/yR\\/r\\/KLOTVWxf5IK.js\"},\"2ma5f\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/yo\\/r\\/hCn6u3fXLtg.js\"},\"wxq+C\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/ya\\/r\\/IWwHlykMyoD.js\"},\"DZmeF\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/y2\\/r\\/ZeZFU7zHXuC.js\"},\"H42Jh\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/y3\\/r\\/ppwOo4BAmlb.js\"},\"AtxWD\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/y1\\/r\\/fvXyHGM5qlP.js\"},\"q9ECf\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/y4\\/r\\/059HuoRuX3I.js\"},\"bs0gy\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/y9\\/r\\/KamU7KNCYdj.js\"}});\n", 
        "Bootloader.enableBootload({\"AsyncDOM\":{\"resources\":[\"JU6Q1\",\"gM1Bs\"],\"module\":true},\"ConfirmationDialog\":{\"resources\":[\"JU6Q1\",\"MX2xp\",\"oE4Do\"],\"module\":true},\"Dialog\":{\"resources\":[\"JU6Q1\",\"MX2xp\",\"M7bAm\"],\"module\":true},\"IframeShim\":{\"resources\":[\"JU6Q1\",\"MX2xp\",\"rKoQM\",\"MqSmz\"],\"module\":true},\"ErrorSignal\":{\"resources\":[\"JU6Q1\",\"cNca2\"],\"module\":true},\"event-extensions\":{\"resources\":[\"JU6Q1\"],\"module\":true},\"AsyncDialog\":{\"resources\":[\"JU6Q1\",\"MX2xp\",\"rKoQM\",\"M7bAm\"],\"module\":true},\"FbdDialogProvider\":{\"resources\":[\"So3Xq\",\"JU6Q1\",\"bwsMw\"],\"module\":true},\"React\":{\"resources\":[\"gM1Bs\"],\"module\":true},\"AsyncRequest\":{\"resources\":[\"JU6Q1\"],\"module\":true},\"PhotoSnowlift\":{\"resources\":[\"JU6Q1\",\"MX2xp\",\"rKoQM\",\"M7bAm\",\"NAQYQ\"],\"module\":true},\"SpotlightShareViewer\":{\"resources\":[\"JU6Q1\",\"MX2xp\",\"X\\/Fq6\",\"zyFOp\"],\"module\":true},\"PhotoTagger\":{\"resources\":[\"JU6Q1\",\"MX2xp\",\"M7bAm\",\"NAQYQ\",\"rKoQM\",\"XNAD8\"],\"module\":true},\"fb-photos-snowlift-css\":{\"resources\":[\"gYYej\",\"M7bAm\"]},\"Live\":{\"resources\":[\"JU6Q1\",\"gM1Bs\"],\"module\":true},\"PhotoTagApproval\":{\"resources\":[\"JU6Q1\",\"NAQYQ\",\"7yfEw\"],\"module\":true},\"PhotoTags\":{\"resources\":[\"JU6Q1\",\"NAQYQ\",\"M7bAm\",\"XNAD8\"],\"module\":true},\"TagTokenizer\":{\"resources\":[\"JU6Q1\",\"NAQYQ\",\"XQYFa\",\"7yfEw\",\"M7bAm\",\"rKoQM\"],\"module\":true},\"fb-photos-snowlift-fullscreen-css\":{\"resources\":[\"VDymv\"]},\"PhotoPivot\":{\"resources\":[\"JU6Q1\",\"MX2xp\",\"NAQYQ\",\"rKoQM\",\"M7bAm\",\"2ma5f\"],\"module\":true},\"SnowliftPicCropper\":{\"resources\":[\"JU6Q1\",\"MX2xp\",\"M7bAm\",\"NAQYQ\",\"rKoQM\",\"jg0IV\",\"XNAD8\",\"wxq+C\"],\"module\":true},\"PhotosButtonTooltips\":{\"resources\":[\"JU6Q1\",\"rKoQM\",\"MX2xp\",\"M7bAm\",\"DZmeF\"],\"module\":true},\"VideoRotate\":{\"resources\":[\"JU6Q1\",\"H42Jh\"],\"module\":true},\"AsyncResponse\":{\"resources\":[\"JU6Q1\"],\"module\":true},\"PhotoInlineEditor\":{\"resources\":[\"JU6Q1\",\"NAQYQ\",\"rKoQM\",\"MX2xp\",\"M7bAm\",\"7yfEw\",\"AtxWD\"],\"module\":true},\"Form\":{\"resources\":[\"JU6Q1\",\"MX2xp\"],\"module\":true},\"Toggler\":{\"resources\":[\"JU6Q1\",\"MX2xp\",\"M7bAm\",\"rKoQM\"],\"module\":true},\"Tooltip\":{\"resources\":[\"JU6Q1\",\"rKoQM\",\"MX2xp\",\"M7bAm\"],\"module\":true},\"DOM\":{\"resources\":[\"JU6Q1\"],\"module\":true},\"Input\":{\"resources\":[\"JU6Q1\"],\"module\":true},\"trackReferrer\":{\"resources\":[],\"module\":true},\"DimensionTracking\":{\"resources\":[\"JU6Q1\"],\"module\":true},\"HighContrastMode\":{\"resources\":[\"JU6Q1\"],\"module\":true},\"detect-broken-proxy-cache\":{\"resources\":[\"JU6Q1\"]},\"link-rel-preload\":{\"resources\":[\"JU6Q1\"]},\"legacy:dialog\":{\"resources\":[\"JU6Q1\",\"MX2xp\",\"M7bAm\"]},\"legacy:ajaxpipe\":{\"resources\":[\"JU6Q1\",\"MX2xp\",\"q9ECf\"]},\"legacy:async\":{\"resources\":[\"JU6Q1\"]},\"legacy:PhotoSnowlift\":{\"resources\":[\"JU6Q1\",\"MX2xp\",\"rKoQM\",\"M7bAm\",\"NAQYQ\",\"bs0gy\"]},\"legacy:Toggler\":{\"resources\":[\"JU6Q1\",\"MX2xp\",\"M7bAm\",\"rKoQM\"]}});</script>\n", 
        "<script>require(\"InitialJSLoader\").loadOnDOMContentReady([\"JU6Q1\",\"MX2xp\",\"mCr\\/S\",\"UjXTQ\",\"T8gbI\"]);</script>\n", 
        "<script>\n", 
        "Bootloader.configurePage([\"M7bAm\",\"jg0IV\",\"zmU4p\"]);\n", 
        "Bootloader.done([\"jDr+c\"]);\n", 
        "\n", 
        "\n", 
        "new (require(\"ServerJS\"))().handle({\"instances\":[[\"m885210_1\",[\"AsyncLayout\"],[\"contentArea\"],1]],\"elements\":[[\"m885211_2\",\"login_form\",2],[\"m885211_3\",\"loginbutton\",2],[\"m885211_4\",\"login_form\",2]],\"define\":[[\"BanzaiConfig\",[],{\"MAX_SIZE\":10000,\"MAX_WAIT\":60000,\"COMPRESSION_THRESHOLD\":800,\"gks\":{\"jslogger\":true,\"miny_compression\":true,\"notifications\":true,\"time_spent\":true,\"time_spent_debug\":true,\"useraction\":true}}]],\"require\":[[\"PixelRatio\",\"startDetecting\",[],[1]],[\"m885210_1\"],[\"Intl\",\"setPhonologicalRules\",[],[{\"meta\":{\"\\/_B\\/\":\"([.,!?\\\\s]|^)\",\"\\/_E\\/\":\"([.,!?\\\\s]|$)\"},\"patterns\":{\"\\/\\u0001(.*)('|&#039;)s\\u0001(?:'|&#039;)s(.*)\\/\":\"\\u0001$1$2s\\u0001$3\",\"\\/_\\u0001([^\\u0001]*)\\u0001\\/e\":\"mb_strtolower(\\\"\\u0001$1\\u0001\\\")\",\"\\/\\\\^\\\\x01([^\\\\x01])(?=[^\\\\x01]*\\\\x01)\\/e\":\"mb_strtoupper(\\\"\\u0001$1\\\")\",\"\\/_\\u0001([^\\u0001]*)\\u0001\\/\":\"javascript\"}}]],[\"PostLoadJS\",\"loadAndRequire\",[],[\"DimensionTracking\"]],[\"PostLoadJS\",\"loadAndCall\",[],[\"HighContrastMode\",\"init\",[\"0\"]]],[\"ClickRefLogger\"],[\"userAction\",\"setUATypeConfig\",[],[{\"uan\":false,\"uai\":false,\"uad\":false,\"uae\":false}]],[\"ScriptPathState\",\"setUserURISampleRate\",[],[0.0002]],[\"userAction\",\"setCustomSampleConfig\",[],[{\"uan\":{\"test\":{\"test\":true}}}]],[\"UserActionHistory\"],[\"TinyViewport\"],[\"WebStorageMonster\",\"schedule\",[],[false]],[\"AsyncRequestNectarLogging\"],[\"LoginFormController\",\"init\",[\"m885211_2\",\"m885211_3\"],[{\"__m\":\"m885211_2\"},{\"__m\":\"m885211_3\"}]],[\"FlipDirectionOnKeypress\"]]});\n", 
        "\n", 
        "onloadRegister_DEPRECATED(function (){$(\"u4kzrn21\").value = tz_calculate(1352998852)});\n", 
        "onloadRegister_DEPRECATED(function (){try { $(\"email\").focus(); } catch (_ignore) { }});\n", 
        "onafterloadRegister_DEPRECATED(function (){Bootloader.loadComponents([\"detect-broken-proxy-cache\"], function(){ detect_broken_proxy_cache(\"0\", \"c_user\") });});\n", 
        "onafterloadRegister_DEPRECATED(function (){Bootloader.loadComponents([\"link-rel-preload\"], function(){ link_rel_preload() });});\n", 
        "</script>\n", 
        "<script>var bigPipe = new (require('BigPipe'))({\"lid\":0,\"forceFinish\":true});</script>\n", 
        "\n", 
        "<script>bigPipe.onPageletArrive({\"is_last\":true,\"css\":[\"M7bAm\",\"jg0IV\",\"zmU4p\"],\"js\":[\"JU6Q1\",\"MX2xp\",\"mCr\\/S\",\"UjXTQ\",\"T8gbI\"],\"id\":\"first_response\",\"phase\":0})</script>\n", 
        "\n", 
        "<code class=\"hidden_elem\" id=\"u4kzrn23\"><!-- <div id=\"page_admin_panel\" data-referrer=\"page_admin_panel\"></div><div id=\"page_nax_callout\" data-referrer=\"page_nax_callout\"></div> --></code>\n", 
        "<script>bigPipe.onPageletArrive({\"content\":{\"pagelet_toolbar\":{\"container_id\":\"u4kzrn23\"}},\"id\":\"pagelet_toolbar\",\"phase\":1})</script>\n", 
        "\n", 
        "<code class=\"hidden_elem\" id=\"u4kzrn24\"><!-- <div id=\"pagelet_main_column_personal\" data-referrer=\"pagelet_main_column_personal_timeline\"></div> --></code>\n", 
        "<script>bigPipe.onPageletArrive({\"content\":{\"pagelet_timeline_main_column\":{\"container_id\":\"u4kzrn24\"}},\"jsmods\":{\"require\":[[\"TimelineController\",\"init\",[],[\"143412869029\",\"timeline\",{\"has_fixed_ads\":false,\"force_wide_ads\":0}]]]},\"js\":[\"JU6Q1\",\"MX2xp\"],\"id\":\"pagelet_timeline_main_column\",\"phase\":1})</script>\n", 
        "\n", 
        "<script>bigPipe.onPageletArrive({\"display_dependency\":[\"pagelet_toolbar\"],\"content\":{\"page_admin_panel\":\"\"},\"id\":\"page_admin_panel\",\"phase\":1})</script>\n", 
        "\n", 
        "<code class=\"hidden_elem\" id=\"u4kzrn29\"><!-- <div class=\"timelineLoggedOutSignUp\"><div id=\"pagelet_loggedout_sign_up\" data-referrer=\"pagelet_loggedout_sign_up\"></div></div><div class=\"fbTimelineTopSectionBase  fbTimelineLoggedOutTopSection\"><div id=\"pagelet_above_header_timeline\" data-referrer=\"pagelet_above_header_timeline\"></div><div id=\"above_header_timeline_placeholder\"></div><div class=\"fbTimelineSection mtm fbTimelineTopSection\"><div id=\"fbProfileCover\" class=\"\"><div class=\"cover\" id=\"u4kzrn25\"><a class=\"coverWrap coverImage\" href=\"http://www.facebook.com/photo.php?fbid=10151163086294030&amp;set=a.10150650858519030.412416.143412869029&amp;type=1\" rel=\"theater\" ajaxify=\"http://www.facebook.com/photo.php?fbid=10151163086294030&amp;set=a.10150650858519030.412416.143412869029&amp;type=1&amp;src=https%3A%2F%2Fsphotos-b.xx.fbcdn.net%2Fhphotos-ash4%2F292897_10151163086294030_673812349_n.jpg&amp;size=850%2C316\" id=\"fbCoverImageContainer\"><img class=\"coverPhotoImg photo img\" src=\"https://sphotos-b.xx.fbcdn.net/hphotos-ash4/292897_10151163086294030_673812349_n.jpg\" style=\"top:0px;width:100%;\" data-fbid=\"10151163086294030\" alt=\"Cover Photo\" /><div class=\"coverBorder\"></div><img class=\"coverChangeThrobber img\" src=\"https://s-static.ak.facebook.com/rsrc.php/v2/yk/r/LOOn0JtHNzb.gif\" alt=\"\" width=\"16\" height=\"16\" /></a></div><div id=\"fbTimelineHeadline\" class=\"clearfix\"><div class=\"actions\"><div class=\"actionsDropdown\" id=\"pagelet_timeline_page_actions\" data-referrer=\"pagelet_timeline_page_actions\"></div></div><div class=\"name\"><div class=\"photoContainer\"><a class=\"profilePicThumb\" id=\"profile_pic_education\" href=\"http://www.facebook.com/photo.php?fbid=10150651037894030&amp;set=a.441010379029.240088.143412869029&amp;type=1\" rel=\"theater\"><img class=\"profilePic img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-prn1/c10.10.160.160/421813_10150651037894030_1970507772_a.jpg\" alt=\"North Carolina State Parks and Recreation - Raleigh, NC\" id=\"u4kzrn28\" /></a></div><h2><span itemprop=\"name\">North Carolina State Parks and Recreation</span><div class=\"fsm fwn fcg\"><div class=\"fsm fwn fcg\">8,945 likes \u00B7 599 talking about this</div></div></h2></div></div><div id=\"timeline_sticky_header\" data-referrer=\"timeline_sticky_header\"></div></div><div class=\"fbTimelineNavigationPagelet\" id=\"pagelet_timeline_nav\" data-referrer=\"pagelet_timeline_nav\"></div></div><i class=\"topSectionBottomBorder img sp_2qck9m sx_2844d5\"></i></div><div id=\"timeline_tab_content\"><div id=\"pagelet_timeline_recent\" data-referrer=\"pagelet_timeline_recent\"></div><div id=\"timeline_tab_content_extra\"></div></div> --></code>\n", 
        "<script>bigPipe.onPageletArrive({\"display_dependency\":[\"pagelet_timeline_main_column\"],\"content\":{\"pagelet_main_column_personal\":{\"container_id\":\"u4kzrn29\"}},\"jsmods\":{\"instances\":[[\"m885216_7\",[\"TimelineCover\",\"m885216_6\"],[{\"__m\":\"m885216_6\"},315,false],1],[\"m885217_9\",[\"Spotlight\",\"LayerHideOnBlur\",\"LayerHideOnEscape\",\"m885217_10\"],[{\"addedBehaviors\":[{\"__m\":\"LayerHideOnBlur\"},{\"__m\":\"LayerHideOnEscape\"}],\"attributes\":{\"tabindex\":\"0\",\"role\":\"region\",\"aria-label\":\"Facebook Photo Theater\",\"aria-busy\":\"true\"},\"classNames\":[\"fbPhotoSnowlift\",\"fbxPhoto\",\"fbPhotoSnowliftPivotHover\"]},{\"__m\":\"m885217_10\"}],3]],\"markup\":[[\"m885217_10\",{\"__html\":\"\\u003Cdiv class=\\\"_n3\\\">\\u003Cdiv class=\\\"fbPhotoSnowliftContainer uiContextualLayerParent\\\">\\u003Cdiv class=\\\"clearfix fbPhotoSnowliftPopup\\\">\\u003Cdiv class=\\\"stageWrapper lfloat\\\">\\u003Cdiv class=\\\"fbPhotoSnowliftFullScreen fullScreenSwitch\\\" id=\\\"fullScreenSwitch\\\">\\u003Ca data-hover=\\\"tooltip\\\" aria-label=\\\"Enter Fullscreen\\\" data-tooltip-position=\\\"below\\\" data-tooltip-alignh=\\\"right\\\" href=\\\"#\\\" id=\\\"fbPhotoSnowliftFullScreenSwitch\\\">\\u003C\\/a>\\u003Ca data-hover=\\\"tooltip\\\" aria-label=\\\"Press Esc to exit fullscreen\\\" data-tooltip-position=\\\"below\\\" data-tooltip-alignh=\\\"right\\\" href=\\\"#\\\" id=\\\"fbPhotoSnowliftFullScreenClose\\\">\\u003C\\/a>\\u003C\\/div>\\u003Cdiv class=\\\"stage\\\">\\u003Cimg class=\\\"spotlight spotlight\\\" src=\\\"\\/images\\/spacer.gif\\\" \\/>\\u003Cdiv class=\\\"fbPhotosPhotoTagboxes tagContainer\\\" id=\\\"fbPhotoSnowliftTagBoxes\\\">\\u003C\\/div>\\u003Cdiv class=\\\"fbPhotoTagApproval\\\" id=\\\"fbPhotoSnowliftTagApproval\\\">\\u003C\\/div>\\u003C\\/div>\\u003Cdiv class=\\\"videoStage\\\">\\u003C\\/div>\\u003Cdiv class=\\\"stageActions\\\" id=\\\"snowliftStageActions\\\">\\u003Cdiv class=\\\"clearfix snowliftOverlay snowliftOverlayBar rightButtons\\\">\\u003Cdiv class=\\\"overlayBarButtons rfloat\\\">\\u003Cdiv class=\\\"_6a _6b fbPhotosPhotoActions\\\" id=\\\"fbPhotoSnowliftActions\\\">\\u003C\\/div>\\u003Cdiv class=\\\"fbPhotosPhotoButtons\\\" id=\\\"fbPhotoSnowliftButtons\\\">\\u003C\\/div>\\u003C\\/div>\\u003Cdiv class=\\\"mediaTitleInfo\\\">\\u003Cdiv class=\\\"mediaTitleBoxFlex\\\">\\u003Cdiv id=\\\"fbPhotoSnowliftMediaTitle\\\">\\u003C\\/div>\\u003Cspan class=\\\"mlm hidden_elem\\\" id=\\\"fbPhotoSnowliftPositionAndCount\\\">\\u003C\\/span>\\u003Cdiv class=\\\"mediaTitleInfoSpacer\\\">\\u003C\\/div>\\u003C\\/div>\\u003C\\/div>\\u003C\\/div>\\u003Cdiv class=\\\"fbPhotosPhotoTagboxBase newTagBox hidden_elem\\\" style=\\\"\\\">\\u003Cdiv class=\\\"borderTagBox\\\">\\u003Cdiv class=\\\"innerTagBox\\\">\\u003Cdiv class=\\\"ieContentFix\\\">\\u00a0\\u00a0\\u00a0\\u00a0\\u003C\\/div>\\u003C\\/div>\\u003C\\/div>\\u003Cdiv class=\\\"tag\\\">\\u003Cdiv class=\\\"tagPointer\\\">\\u003Ci class=\\\"tagArrow img sp_7uj4ry sx_8a61b2\\\">\\u003C\\/i>\\u003Cdiv class=\\\"tagName\\\">\\u003C\\/div>\\u003C\\/div>\\u003C\\/div>\\u003C\\/div>\\u003C\\/div>\\u003Ca class=\\\"snowliftPager prev\\\" title=\\\"Previous\\\">\\u003Ci>\\u003C\\/i>\\u003C\\/a>\\u003Ca class=\\\"snowliftPager next\\\" title=\\\"Next\\\">\\u003Ci>\\u003C\\/i>\\u003C\\/a>\\u003Cdiv class=\\\"photoError stageError\\\" id=\\\"fbPhotoSnowliftError\\\">\\u003C\\/div>\\u003C\\/div>\\u003Cdiv class=\\\"rhc photoUfiContainer\\\">\\u003Cdiv class=\\\"rhcHeader\\\">\\u003Cdiv class=\\\"fbPhotoSnowliftControls\\\">\\u003Ca data-hover=\\\"tooltip\\\" aria-label=\\\"Press Esc to close\\\" data-tooltip-position=\\\"below\\\" data-tooltip-alignh=\\\"right\\\" class=\\\"closeTheater\\\" href=\\\"#\\\">\\u003C\\/a>\\u003Ca data-hover=\\\"tooltip\\\" aria-label=\\\"Press Esc to exit fullscreen\\\" data-tooltip-position=\\\"below\\\" data-tooltip-alignh=\\\"right\\\" class=\\\"closeTheater fullscreenCommentClose\\\" href=\\\"#\\\">\\u003C\\/a>\\u003C\\/div>\\u003Cdiv id=\\\"fbPhotoSnowliftInlineEditor\\\">\\u003C\\/div>\\u003C\\/div>\\u003Cform rel=\\\"async\\\" class=\\\"fbPhotosSnowliftFeedbackForm rhcBody commentable_item autoexpand_mode\\\" method=\\\"post\\\" action=\\\"\\/ajax\\/ufi\\/modify.php\\\" onsubmit=\\\"return window.Event &amp;&amp; Event.__inlineSubmit &amp;&amp; Event.__inlineSubmit(this,event)\\\" id=\\\"u4kzrn27\\\">\\u003Cinput type=\\\"hidden\\\" name=\\\"charset_test\\\" value=\\\"&euro;,&acute;,\\u20ac,\\u00b4,\\u6c34,\\u0414,\\u0404\\\" \\/>\\u003Cinput type=\\\"hidden\\\" name=\\\"lsd\\\" value=\\\"AVqaEiN6\\\" autocomplete=\\\"off\\\" \\/>\\u003Cdiv class=\\\"uiScrollableArea rhcScroller native\\\" style=\\\"width:285px;\\\">\\u003Cdiv class=\\\"uiScrollableAreaWrap scrollable\\\" tabindex=\\\"0\\\">\\u003Cdiv class=\\\"uiScrollableAreaBody\\\">\\u003Cdiv class=\\\"uiScrollableAreaContent\\\">\\u003Cdiv class=\\\"clearfix fbPhotoSnowliftAuthorInfo\\\">\\u003Cdiv class=\\\"lfloat\\\" id=\\\"fbPhotoSnowliftAuthorPic\\\">\\u003C\\/div>\\u003Cdiv>\\u003Cdiv class=\\\"fbPhotoContributorName\\\" id=\\\"fbPhotoSnowliftAuthorName\\\">\\u003C\\/div>\\u003Cdiv class=\\\"mrs fsm fwn fcg\\\">\\u003Cspan id=\\\"fbPhotoSnowliftSubscribe\\\">\\u003C\\/span>\\u003Cspan id=\\\"fbPhotoSnowliftTimestamp\\\">\\u003C\\/span>\\u003Cdiv class=\\\"mls\\\" id=\\\"fbPhotoSnowliftAudienceSelector\\\">\\u003C\\/div>\\u003C\\/div>\\u003Cdiv class=\\\"fsm fwn fcg\\\">\\u003Cdiv class=\\\"fbPhotosOnProfile\\\" id=\\\"fbPhotoSnowliftOnProfile\\\">\\u003C\\/div>\\u003C\\/div>\\u003Cspan id=\\\"fbPhotoSnowliftViewOnApp\\\">\\u003C\\/span>\\u003Cspan id=\\\"fbPhotoSnowliftUseApp\\\">\\u003C\\/span>\\u003C\\/div>\\u003C\\/div>\\u003Cspan class=\\\"fbPhotosPhotoCaption\\\" tabindex=\\\"0\\\" aria-live=\\\"polite\\\" id=\\\"fbPhotoSnowliftCaption\\\">\\u003C\\/span>\\u003Cspan class=\\\"fbPhotoTagList\\\" id=\\\"fbPhotoSnowliftTagList\\\">\\u003C\\/span>\\u003Cdiv class=\\\"fbPhotoPagesTagList\\\" id=\\\"fbPhotoSnowliftPagesTagList\\\">\\u003C\\/div>\\u003Cdiv class=\\\"pts fbPhotoLegacyTagList\\\" id=\\\"fbPhotoSnowliftLegacyTagList\\\">\\u003C\\/div>\\u003Cdiv class=\\\"fbPhotosPhotoOwnerButtons stat_elem\\\" id=\\\"fbPhotoSnowliftOwnerButtons\\\">\\u003C\\/div>\\u003Cdiv class=\\\"fbPhotosSnowliftFeedback\\\" id=\\\"fbPhotoSnowliftFeedback\\\">\\u003C\\/div>\\u003C\\/div>\\u003C\\/div>\\u003C\\/div>\\u003C\\/div>\\u003Cdiv class=\\\"fbPhotosSnowboxFeedbackInput utn\\\" id=\\\"fbPhotoSnowliftFeedbackInput\\\">\\u003C\\/div>\\u003C\\/form>\\u003Cdiv class=\\\"_24q rhcFooter\\\">\\u003C\\/div>\\u003C\\/div>\\u003C\\/div>\\u003C\\/div>\\u003C\\/div>\"},3],[\"m885216_8\",{\"__html\":\"\\u003Cdiv>\\u003C\\/div>\"},2]],\"elements\":[[\"m885216_5\",\"pagelet_timeline_recent\",2],[\"m885216_6\",\"u4kzrn25\",2],[\"m885217_11\",\"u4kzrn27\",2,\"m885217_10\"],[\"m885218_12\",\"u4kzrn28\",2]],\"require\":[[\"TimelineContentLoader\",\"registerTimePeriod\",[\"m885216_5\"],[{\"__m\":\"m885216_5\"},\"recent\",{\"profile_id\":143412869029,\"start\":1351753200,\"end\":1354348799,\"query_type\":5,\"section_pagelet_id\":\"pagelet_timeline_recent\",\"load_immediately\":true},true,null,0,-1]],[\"m885216_7\"],[\"m885217_9\"],[\"Scrollable\"],[\"PhotoSnowlift\",\"initWithSpotlight\",[\"m885217_9\"],[{\"__m\":\"m885217_9\"},{\"refresh_fast\":false,\"refresh_rate\":20000,\"hw_rendering\":false,\"pivot_hover\":true,\"pivot_end_metric\":false,\"og_videos\":false,\"pivot_end\":false,\"snowlift_profile_pic_cropper\":false,\"min_ads\":2}]],[\"PhotoSnowlift\",\"touch\",[\"m885216_8\"],[{\"__m\":\"m885216_8\"}]],[\"BrokenProfilePicLogger\",\"checkAndLog\",[\"m885218_12\"],[{\"__m\":\"m885218_12\"}]]]},\"css\":[\"BbtG6\",\"M7bAm\",\"gYYej\",\"kXZht\"],\"bootloadable\":{\"PhotoCropper\":{\"resources\":[\"JU6Q1\",\"MX2xp\",\"M7bAm\",\"NAQYQ\",\"rKoQM\",\"jg0IV\",\"XNAD8\",\"7yfEw\"],\"module\":true}},\"resource_map\":{\"xjH7g\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/yJ\\/r\\/zjQYk3zi5gs.js\"},\"BbtG6\":{\"type\":\"css\",\"permanent\":1,\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/y2\\/r\\/hEpquSYmFoc.css\"},\"kXZht\":{\"type\":\"css\",\"permanent\":1,\"nonblocking\":1,\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/yF\\/r\\/bcFq6SQNjLm.css\"}},\"js\":[\"JU6Q1\",\"MX2xp\",\"xjH7g\",\"rKoQM\",\"NAQYQ\"],\"onafterload\":[\"Bootloader.loadComponents([\\\"fb-photos-snowlift-css\\\"], function(){  });\"],\"id\":\"pagelet_main_column_personal\",\"phase\":1})</script>\n", 
        "\n", 
        "<code class=\"hidden_elem\" id=\"u4kzrn219\"><!-- <div id=\"timelineNavContent\" class=\"fbTimelineNavigationWrapper\"><div class=\"fbTimelineNavigation\" id=\"u4kzrn218\"><ul id=\"fbTimelineNavTopRow\" class=\"clearfix\"><li class=\"about tile\"><div id=\"pagelet_timeline_summary\" data-referrer=\"pagelet_timeline_summary\"><div class=\"fbTimelineSummarySectionWrapper\"><div class=\"detail frame\"><div class=\"mat\"><div class=\"fbTimelineSummarySection\"><span class=\"fbProfileNonIconBylineFragment\"><div class=\"fsm fwn fcg\"><span>Government Organization</span></div></span><span class=\"fbLongBlurb\"><div>The North Carolina State Parks System exists for the enjoyment, education, health and inspiration of all our citizens and visitors.<br /> <br /> <a href=\"http://www.facebook.com/l.php?u=http%3A%2F%2Fwww.ncparks.gov&amp;h=9AQErLx8E&amp;s=1\" target=\"_blank\" rel=\"nofollow nofollow\" onmouseover=\"LinkshimAsyncLink.swap(this, &quot;http:\\\\/\\\\/www.ncparks.gov\\\\/&quot;);\" onclick=\"LinkshimAsyncLink.referrer_log(this, &quot;http:\\\\/\\\\/www.ncparks.gov\\\\/&quot;, &quot;http:\\\\/\\\\/www.facebook.com\\\\/si\\\\/ajax\\\\/l\\\\/render_linkshim_log\\\\/?u=http\\\\u00253A\\\\u00252F\\\\u00252Fwww.ncparks.gov&amp;h=9AQErLx8E&amp;s=1&quot;);\">www.ncparks.gov</a><br /> <br /> <a href=\"http://www.facebook.com/l.php?u=http%3A%2F%2Fwww.twitter.com%2FNCparks&amp;h=mAQF5-WQL&amp;s=1\" target=\"_blank\" rel=\"nofollow nofollow\" onmouseover=\"LinkshimAsyncLink.swap(this, &quot;http:\\\\/\\\\/www.twitter.com\\\\/NCparks&quot;);\" onclick=\"LinkshimAsyncLink.referrer_log(this, &quot;http:\\\\/\\\\/www.twitter.com\\\\/NCparks&quot;, &quot;http:\\\\/\\\\/www.facebook.com\\\\/si\\\\/ajax\\\\/l\\\\/render_linkshim_log\\\\/?u=http\\\\u00253A\\\\u00252F\\\\u00252Fwww.twitter.com\\\\u00252FNCparks&amp;h=mAQF5-WQL&amp;s=1&quot;);\">www.twitter.com/NCparks</a></div></span></div></div></div><a class=\"title\" href=\"https://www.facebook.com/NorthCarolinaStateParks/info\" aria-label=\"About North Carolina State Parks and Recreation\"><span class=\"text\">About</span></a></div></div></li><li class=\"photos tile firstTab\" id=\"pagelet_timeline_photos_nav_top\"><div id=\"pagelet_timeline_photos_nav_top\" data-referrer=\"pagelet_timeline_photos_nav_top_tc\"><div class=\"modify\" id=\"u4kzrn217\"><a class=\"tab\" href=\"https://www.facebook.com/NorthCarolinaStateParks/photos_stream\"><div class=\"detail\"><i style=\"background-image: url(&#039;https://fbcdn-photos-a.akamaihd.net/hphotos-ak-snc7/396736_10151257845269030_826810352_a.jpg&#039;);\" class=\"photo\"></i></div><div class=\"frame\"><div class=\"mat\"></div></div><div class=\"title\"><span class=\"icon\"><i class=\"img sp_2vmry3 sx_beccc5\"></i></span><span class=\"text\">Photos</span></div></a></div></div></li><li class=\"app_304039862941303 tile \" id=\"pagelet_timeline_app_nav_app_304039862941303_top\"><div id=\"pagelet_timeline_app_nav_app_304039862941303_top\" data-referrer=\"pagelet_timeline_app_nav_app_304039862941303_top_tc\"><div class=\"modify\" id=\"u4kzrn212\"><a class=\"tab\" href=\"https://www.facebook.com/NorthCarolinaStateParks/app_304039862941303\"><div class=\"detail\"><i style=\"background-image: url(&#039;https://fbcdn-photos-a.akamaihd.net/photos-ak-snc7/v85006/55/304039862941303/app_10_304039862941303_6478.gif&#039;)\" class=\"appimage\"></i></div><div class=\"frame\"><div class=\"mat\"></div></div><div class=\"title\"><span class=\"icon\"><img class=\"img\" src=\"https://fbcdn-photos-a.akamaihd.net/photos-ak-snc7/v85006/55/304039862941303/app_2_304039862941303_516769829.gif\" alt=\"\" /></span><span class=\"text\">Mobile App</span></div></a></div></div></li><li class=\"likes tile \" id=\"pagelet_timeline_likes_nav_top\"><div id=\"pagelet_timeline_likes_nav_top\" data-referrer=\"pagelet_timeline_likes_nav_top_tc\"><div class=\"modify\" id=\"u4kzrn211\"><a class=\"tab\" href=\"https://www.facebook.com/NorthCarolinaStateParks/likes\"><div class=\"detail\"><div><table class=\"uiGrid statsContainer\" cellspacing=\"0\" cellpadding=\"0\"><tbody><tr><td class=\"vMid hCent\"><span class=\"_wj stats fwb\">8,945</span></td></tr></tbody></table></div></div><div class=\"frame\"><div class=\"mat\"></div></div><div class=\"title\"><span class=\"icon\"><i class=\"img sp_bhkf7q sx_93ea6f\"></i></span><span class=\"text\">Likes</span></div></a></div></div></li><li class=\"app_4949752878 tile \" id=\"pagelet_timeline_app_nav_app_4949752878_top\"><div id=\"pagelet_timeline_app_nav_app_4949752878_top\" data-referrer=\"pagelet_timeline_app_nav_app_4949752878_top_tc\"><div class=\"modify\" id=\"u4kzrn213\"><a class=\"tab\" href=\"https://www.facebook.com/NorthCarolinaStateParks/app_4949752878\"><div class=\"detail\"><i style=\"background-image: url(&#039;https://fbcdn-photos-a.akamaihd.net/photos-ak-snc7/v85005/126/4949752878/app_10_4949752878_928164642.gif&#039;)\" class=\"appimage\"></i></div><div class=\"frame\"><div class=\"mat\"></div></div><div class=\"title\"><span class=\"icon\"><img class=\"img\" src=\"https://s-static.ak.facebook.com/rsrc.php/v2/yh/r/HxPMO4hvPzE.gif\" alt=\"\" /></span><span class=\"text\">Parks</span></div></a></div></div></li><li class=\"events tile \" id=\"pagelet_timeline_events_nav_top\"><div id=\"pagelet_timeline_events_nav_top\" data-referrer=\"pagelet_timeline_events_nav_top_tc\"><div class=\"modify\" id=\"u4kzrn216\"><a class=\"tab\" href=\"https://www.facebook.com/NorthCarolinaStateParks/events\"><div class=\"detail\"><i style=\"background-image: url(&#039;https://fbcdn-profile-a.akamaihd.net/hprofile-ak-snc6/187877_112501692159069_8065661_n.jpg&#039;)\" class=\"event\"></i></div><div class=\"frame\"><div class=\"mat\"></div></div><div class=\"title\"><span class=\"icon\"><i class=\"img sp_edi87m sx_8b2d66\"></i></span><span class=\"text\">Events</span></div></a></div></div></li><li class=\"videos tile \" id=\"pagelet_timeline_videos_nav_top\"><div id=\"pagelet_timeline_videos_nav_top\" data-referrer=\"pagelet_timeline_videos_nav_top_tc\"><div class=\"modify\" id=\"u4kzrn215\"><a class=\"tab\" href=\"http://www.facebook.com/media/set/?set=vb.143412869029&amp;type=2\"><div class=\"detail\"><i style=\"background-image: url(&#039;https://fbcdn-vthumb-a.akamaihd.net/hvthumb-ak-ash3/574070_399320430088869_399319873422258_5719_1424_t.jpg&#039;)\" class=\"video\"></i></div><div class=\"frame\"><div class=\"mat\"></div></div><div class=\"title\"><span class=\"icon\"><i class=\"img sp_2vmry3 sx_3591fe\"></i></span><span class=\"text\">Videos</span></div></a></div></div></li><li class=\"notes tile \" id=\"pagelet_timeline_notes_nav_top\"><div id=\"pagelet_timeline_notes_nav_top\" data-referrer=\"pagelet_timeline_notes_nav_top_tc\"><i id=\"u4kzrn214\"></i></div></li><li class=\"hidden_elem empty tile\"><div class=\"add\"><a class=\"tab\" href=\"#\"><div class=\"detail\"></div><div class=\"frame\"><div class=\"mat\"></div></div><div class=\"title\"></div></a></div></li><li class=\"hidden_elem empty tile\"><div class=\"add\"><a class=\"tab\" href=\"#\"><div class=\"detail\"></div><div class=\"frame\"><div class=\"mat\"></div></div><div class=\"title\"></div></a></div></li><li class=\"hidden_elem empty tile\"><div class=\"add\"><a class=\"tab\" href=\"#\"><div class=\"detail\"></div><div class=\"frame\"><div class=\"mat\"></div></div><div class=\"title\"></div></a></div></li><li class=\"hidden_elem empty tile\"><div class=\"add\"><a class=\"tab\" href=\"#\"><div class=\"detail\"></div><div class=\"frame\"><div class=\"mat\"></div></div><div class=\"title\"></div></a></div></li><li class=\"hidden_elem empty tile\"><div class=\"add\"><a class=\"tab\" href=\"#\"><div class=\"detail\"></div><div class=\"frame\"><div class=\"mat\"></div></div><div class=\"title\"></div></a></div></li><li class=\"hidden_elem empty tile\"><div class=\"add\"><a class=\"tab\" href=\"#\"><div class=\"detail\"></div><div class=\"frame\"><div class=\"mat\"></div></div><div class=\"title\"></div></a></div></li><li class=\"hidden_elem empty tile\"><div class=\"add\"><a class=\"tab\" href=\"#\"><div class=\"detail\"></div><div class=\"frame\"><div class=\"mat\"></div></div><div class=\"title\"></div></a></div></li><li class=\"hidden_elem empty tile\"><div class=\"add\"><a class=\"tab\" href=\"#\"><div class=\"detail\"></div><div class=\"frame\"><div class=\"mat\"></div></div><div class=\"title\"></div></a></div></li></ul><a class=\"fbTimelineMoreButton seemore stat_elem fbTimelineMoreButtonZero\" href=\"#\" id=\"u4kzrn210\"><div class=\"detail frame\"><div class=\"mat\"></div><div class=\"fbTimelineRibbon\"><div class=\"text\"> </div></div></div><div class=\"title\">More</div></a></div><div id=\"fbTimelineNavigationLoadContainer\"></div></div> --></code>\n", 
        "<script>bigPipe.onPageletArrive({\"display_dependency\":[\"pagelet_main_column_personal\"],\"content\":{\"pagelet_timeline_nav\":{\"container_id\":\"u4kzrn219\"}},\"jsmods\":{\"markup\":[[\"m885221_16\",{\"__html\":\"\\u003Cli class=\\\"uiMenuItem uiMenuItemRadio uiSelectorOption\\\" data-label=\\\"Likes\\\">\\u003Ca class=\\\"itemAnchor itemWithIcon\\\" role=\\\"menuitemradio\\\" tabindex=\\\"-1\\\" href=\\\"https:\\/\\/www.facebook.com\\/NorthCarolinaStateParks\\/likes\\\" aria-checked=\\\"false\\\">\\u003Ci class=\\\"mrs itemIcon img sp_bhkf7q sx_93ea6f\\\">\\u003C\\/i>\\u003Cspan class=\\\"itemLabel fsm\\\">Likes\\u003C\\/span>\\u003C\\/a>\\u003C\\/li>\"},2],[\"m885222_18\",{\"__html\":\"\\u003Cli class=\\\"uiMenuItem uiMenuItemRadio uiSelectorOption\\\" data-label=\\\"Mobile App\\\">\\u003Ca class=\\\"itemAnchor itemWithIcon\\\" role=\\\"menuitemradio\\\" tabindex=\\\"-1\\\" href=\\\"https:\\/\\/www.facebook.com\\/NorthCarolinaStateParks\\/app_304039862941303\\\" aria-checked=\\\"false\\\">\\u003Cimg class=\\\"mrs itemIcon img\\\" src=\\\"https:\\/\\/fbcdn-photos-a.akamaihd.net\\/photos-ak-snc7\\/v85006\\/55\\/304039862941303\\/app_2_304039862941303_516769829.gif\\\" alt=\\\"\\\" \\/>\\u003Cspan class=\\\"itemLabel fsm\\\">Mobile App\\u003C\\/span>\\u003C\\/a>\\u003C\\/li>\"},2],[\"m885222_20\",{\"__html\":\"\\u003Cli class=\\\"uiMenuItem uiMenuItemRadio uiSelectorOption\\\" data-label=\\\"Parks\\\">\\u003Ca class=\\\"itemAnchor itemWithIcon\\\" role=\\\"menuitemradio\\\" tabindex=\\\"-1\\\" href=\\\"https:\\/\\/www.facebook.com\\/NorthCarolinaStateParks\\/app_4949752878\\\" aria-checked=\\\"false\\\">\\u003Cimg class=\\\"mrs itemIcon img\\\" src=\\\"https:\\/\\/s-static.ak.facebook.com\\/rsrc.php\\/v2\\/yh\\/r\\/HxPMO4hvPzE.gif\\\" alt=\\\"\\\" \\/>\\u003Cspan class=\\\"itemLabel fsm\\\">Parks\\u003C\\/span>\\u003C\\/a>\\u003C\\/li>\"},2],[\"m885239_23\",{\"__html\":\"\\u003Cli class=\\\"uiMenuItem uiMenuItemRadio uiSelectorOption\\\" data-label=\\\"Videos\\\">\\u003Ca class=\\\"itemAnchor itemWithIcon\\\" role=\\\"menuitemradio\\\" tabindex=\\\"-1\\\" href=\\\"http:\\/\\/www.facebook.com\\/media\\/set\\/?set=vb.143412869029&amp;type=2\\\" aria-checked=\\\"false\\\">\\u003Ci class=\\\"mrs itemIcon img sp_2vmry3 sx_3591fe\\\">\\u003C\\/i>\\u003Cspan class=\\\"itemLabel fsm\\\">Videos\\u003C\\/span>\\u003C\\/a>\\u003C\\/li>\"},2],[\"m885239_25\",{\"__html\":\"\\u003Cli class=\\\"uiMenuItem uiMenuItemRadio uiSelectorOption\\\" data-label=\\\"Events\\\">\\u003Ca class=\\\"itemAnchor itemWithIcon\\\" role=\\\"menuitemradio\\\" tabindex=\\\"-1\\\" href=\\\"https:\\/\\/www.facebook.com\\/NorthCarolinaStateParks\\/events\\\" aria-checked=\\\"false\\\">\\u003Ci class=\\\"mrs itemIcon img sp_edi87m sx_8b2d66\\\">\\u003C\\/i>\\u003Cspan class=\\\"itemLabel fsm\\\">Events\\u003C\\/span>\\u003C\\/a>\\u003C\\/li>\"},2],[\"m885240_27\",{\"__html\":\"\\u003Cli class=\\\"uiMenuItem uiMenuItemRadio uiSelectorOption\\\" data-label=\\\"Photos\\\">\\u003Ca class=\\\"itemAnchor itemWithIcon\\\" role=\\\"menuitemradio\\\" tabindex=\\\"-1\\\" href=\\\"https:\\/\\/www.facebook.com\\/NorthCarolinaStateParks\\/photos_stream\\\" aria-checked=\\\"false\\\">\\u003Ci class=\\\"mrs itemIcon img sp_2vmry3 sx_beccc5\\\">\\u003C\\/i>\\u003Cspan class=\\\"itemLabel fsm\\\">Photos\\u003C\\/span>\\u003C\\/a>\\u003C\\/li>\"},2]],\"elements\":[[\"m885220_14\",\"u4kzrn210\",2],[\"m885221_15\",\"u4kzrn211\",2],[\"m885222_17\",\"u4kzrn212\",2],[\"m885222_19\",\"u4kzrn213\",2],[\"m885222_21\",\"u4kzrn214\",2],[\"m885239_22\",\"u4kzrn215\",2],[\"m885239_24\",\"u4kzrn216\",2],[\"m885240_26\",\"u4kzrn217\",2],[\"m885220_13\",\"u4kzrn218\",2]],\"require\":[[\"TimelineNavigation\",\"init\",[\"m885220_13\",\"m885220_14\"],[{\"__m\":\"m885220_13\"},{\"__m\":\"m885220_14\"},\"timeline\",false]],[\"TimelineNavigation\",\"initializeMorePager\",[],[]],[\"TimelineStickyHeaderNav\",\"initPageMenu\",[],[[\"photos\",\"app_304039862941303\",\"likes\",\"app_4949752878\",\"events\",\"videos\",\"notes\"],[]]],[\"TimelineNavigation\",\"registerNav\",[\"m885221_15\"],[\"likes\",{\"__m\":\"m885221_15\"},{\"recent\":false,\"async\":false}]],[\"TimelineStickyHeaderNav\",\"registerPageMenuItem\",[\"m885221_16\"],[\"likes\",{\"__m\":\"m885221_16\"}]],[\"TimelineNavigation\",\"registerNav\",[\"m885222_17\"],[\"app_304039862941303\",{\"__m\":\"m885222_17\"},{\"recent\":false,\"async\":false}]],[\"TimelineStickyHeaderNav\",\"registerPageMenuItem\",[\"m885222_18\"],[\"app_304039862941303\",{\"__m\":\"m885222_18\"}]],[\"TimelineNavigation\",\"registerNav\",[\"m885222_19\"],[\"app_4949752878\",{\"__m\":\"m885222_19\"},{\"recent\":false,\"async\":false}]],[\"TimelineStickyHeaderNav\",\"registerPageMenuItem\",[\"m885222_20\"],[\"app_4949752878\",{\"__m\":\"m885222_20\"}]],[\"TimelineNavigation\",\"registerNav\",[\"m885222_21\"],[\"notes\",{\"__m\":\"m885222_21\"},{\"recent\":false,\"async\":false,\"remove\":true}]],[\"TimelineNavigation\",\"registerNav\",[\"m885239_22\"],[\"videos\",{\"__m\":\"m885239_22\"},{\"recent\":false,\"async\":false}]],[\"TimelineStickyHeaderNav\",\"registerPageMenuItem\",[\"m885239_23\"],[\"videos\",{\"__m\":\"m885239_23\"}]],[\"TimelineNavigation\",\"registerNav\",[\"m885239_24\"],[\"events\",{\"__m\":\"m885239_24\"},{\"recent\":false,\"async\":false}]],[\"TimelineStickyHeaderNav\",\"registerPageMenuItem\",[\"m885239_25\"],[\"events\",{\"__m\":\"m885239_25\"}]],[\"TimelineNavigation\",\"registerNav\",[\"m885240_26\"],[\"photos\",{\"__m\":\"m885240_26\"},{\"recent\":false,\"async\":false}]],[\"TimelineStickyHeaderNav\",\"registerPageMenuItem\",[\"m885240_27\"],[\"photos\",{\"__m\":\"m885240_27\"}]]]},\"css\":[\"gYYej\",\"M7bAm\",\"FjEEx\",\"RqL3T\"],\"resource_map\":{\"M2kwM\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/ys\\/r\\/fcdCHvjp435.js\"},\"FjEEx\":{\"type\":\"css\",\"permanent\":1,\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/yb\\/r\\/paNGXWK_Oza.css\"},\"RqL3T\":{\"type\":\"css\",\"nonblocking\":1,\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/ye\\/r\\/wc_SWIGptat.css\"}},\"js\":[\"JU6Q1\",\"MX2xp\",\"M2kwM\",\"rKoQM\"],\"id\":\"pagelet_timeline_nav\",\"phase\":1})</script>\n", 
        "\n", 
        "<code class=\"hidden_elem\" id=\"u4kzrn220\"><!-- <div class=\"timeline\"></div> --></code>\n", 
        "<script>bigPipe.onPageletArrive({\"display_dependency\":[\"pagelet_main_column_personal\"],\"content\":{\"pagelet_above_header_timeline\":{\"container_id\":\"u4kzrn220\"}},\"id\":\"pagelet_above_header_timeline\",\"phase\":1})</script>\n", 
        "\n", 
        "<code class=\"hidden_elem\" id=\"u4kzrn223\"><!-- <div><span class=\"uiButtonGroup mtm mlm actionsContents profileHeaderButton uiButtonGroupOverlay\" id=\"u4kzrn221\"><span class=\"firstItem lastItem uiButtonGroupItem buttonItem\"><a class=\"uiButton uiButtonOverlay uiButtonLarge\" href=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2FNorthCarolinaStateParks\" role=\"button\" rel=\"dialog\" id=\"timelineHeadlineLikeButton\"><i class=\"mrs img sp_2qck9m sx_e3c960\"></i><span class=\"uiButtonText\">Like</span></a></span></span><span class=\"uiButtonGroup mtm mlm actionsContents profileHeaderButton uiButtonGroupOverlay\" id=\"u4kzrn222\"><span class=\"firstItem lastItem uiButtonGroupItem selectorItem\"><div class=\"uiSelector inlineBlock fbTimelineActionSelector uiSelectorRight\"><div class=\"wrap\"><a class=\"fbTimelineActionSelectorButton uiSelectorButton uiButton uiButtonOverlay uiButtonLarge uiButtonNoText\" href=\"#\" role=\"button\" aria-label=\"Other actions\" aria-haspopup=\"1\" rel=\"toggle\"><i class=\"mrs img sp_2qck9m sx_156e6d\"></i><span class=\"uiButtonText\"></span></a><div class=\"uiSelectorMenuWrapper uiToggleFlyout\"><div role=\"menu\" class=\"uiMenu uiSelectorMenu\"><ul class=\"uiMenuInner\"><li class=\"uiMenuItem\" data-label=\"Create a Page\" id=\"fbpage_createapage_action\"><a class=\"itemAnchor\" role=\"menuitem\" tabindex=\"0\" href=\"/pages/create.php?ref_id=143412869029\"><span class=\"itemLabel fsm\">Create a Page</span></a></li><li class=\"uiMenuSeparator\"></li><li class=\"uiMenuItem\" data-label=\"Privacy\"><a class=\"itemAnchor\" role=\"menuitem\" tabindex=\"-1\" href=\"/about/privacy\"><span class=\"itemLabel fsm\">Privacy</span></a></li><li class=\"uiMenuItem\" data-label=\"Terms\"><a class=\"itemAnchor\" role=\"menuitem\" tabindex=\"-1\" href=\"/legal/terms?ref=pf\"><span class=\"itemLabel fsm\">Terms</span></a></li></ul></div></div></div></div></span></span></div> --></code>\n", 
        "<script>bigPipe.onPageletArrive({\"display_dependency\":[\"pagelet_main_column_personal\"],\"is_last\":true,\"content\":{\"pagelet_timeline_page_actions\":{\"container_id\":\"u4kzrn223\"}},\"jsmods\":{\"elements\":[[\"m885242_28\",\"u4kzrn221\",2],[\"m885242_29\",\"u4kzrn222\",2]],\"require\":[[\"Selector\"]]},\"css\":[\"M7bAm\",\"gYYej\"],\"resource_map\":{\"vVPJj\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/yQ\\/r\\/XVGtKNKuPLA.js\"}},\"js\":[\"JU6Q1\",\"MX2xp\",\"vVPJj\",\"rKoQM\"],\"id\":\"pagelet_timeline_page_actions\",\"phase\":1})</script>\n", 
        "\n", 
        "<code class=\"hidden_elem\" id=\"u4kzrn229\"><!-- <div class=\"pam uiBoxOverlay bottomborder\"><div class=\"fsxl fwb\">North Carolina State Parks and Recreation<br /> is on Facebook.</div><div class=\"mvm fsl\">To connect with North Carolina State Parks and Recreation, sign up for Facebook today.</div><a class=\"uiButton uiButtonSpecial uiButtonLarge\" href=\"/r.php?profile_id=143412869029&amp;next%5B%00%2A%00protocol%5D=http&amp;next%5B%00%2A%00isGenericURI%5D=0&amp;next%5B%00%2A%00domain%5D=www.facebook.com&amp;next%5B%00%2A%00domainIsIp%5D=0&amp;next%5B%00%2A%00path%5D=%2FNorthCarolinaStateParks&amp;next%5B%00%2A%00failedRFC3986Parsing%5D=0&amp;next%5B%00%2A%00useRawURLEncode%5D=0&amp;next%5B%00%2A%00encodeNonASCIIPath%5D=0&amp;next%5B%00%2A%00originalString%5D=http%3A%2F%2Fwww.facebook.com%2FNorthCarolinaStateParks&amp;next%5B%00%2A%00pathArgsDelimiter%5D=%3F&amp;friend_or_subscriber=friend\" role=\"button\"><span class=\"uiButtonText\">Sign Up</span></a><a class=\"uiButton uiButtonConfirm uiButtonLarge\" href=\"/login.php?next=http%3A%2F%2Fwww.facebook.com%2FNorthCarolinaStateParks\" role=\"button\" name=\"login\"><span class=\"uiButtonText\">Log In</span></a></div> --></code>\n", 
        "<script>bigPipe.onPageletArrive({\"content\":{\"pagelet_loggedout_sign_up\":{\"container_id\":\"u4kzrn229\"}},\"jsmods\":{\"instances\":[[\"m885243_30\",[\"DialogX\",\"DialogHideOnSuccess\",\"LayerFadeOnHide\",\"LayerHideOnEscape\",\"LayerDestroyOnHide\",\"LayerHideOnTransition\",\"LayerRemoveOnHide\",\"m885243_31\"],[{\"width\":815,\"modal\":false,\"autohide\":null,\"titleID\":\"u4kzrn224\",\"addedBehaviors\":[{\"__m\":\"DialogHideOnSuccess\"},{\"__m\":\"LayerFadeOnHide\"},{\"__m\":\"LayerHideOnEscape\"},{\"__m\":\"LayerDestroyOnHide\"},{\"__m\":\"LayerHideOnTransition\"},{\"__m\":\"LayerRemoveOnHide\"}],\"attributes\":{\"id\":\"timelineSignUpDialog\"},\"classNames\":[\"timelineSignUpDialog\"]},{\"__m\":\"m885243_31\"}],3]],\"markup\":[[\"m885243_31\",{\"__html\":\"\\u003Cdiv>\\u003Cdiv class=\\\"pvs phm _1yw\\\" id=\\\"u4kzrn224\\\">To connect with North Carolina State Parks and Recreation, sign up for Facebook today.\\u003C\\/div>\\u003Cdiv class=\\\"pam _13 timelineSignUpDialogBody\\\">\\u003Cdiv class=\\\"clearfix\\\">\\u003Cdiv class=\\\"logInContainer lfloat\\\">\\u003Cdiv class=\\\"mbs mainTitle fsl fwb fcb\\\">Log In\\u003C\\/div>\\u003Cdiv>\\u003Cform id=\\\"login_form\\\" action=\\\"https:\\/\\/www.facebook.com\\/login.php?login_attempt=1\\\" method=\\\"post\\\" onsubmit=\\\"return window.Event &amp;&amp; Event.__inlineSubmit &amp;&amp; Event.__inlineSubmit(this,event)\\\">\\u003Cinput type=\\\"hidden\\\" name=\\\"lsd\\\" value=\\\"AVqaEiN6\\\" autocomplete=\\\"off\\\" \\/>\\u003Ctable class=\\\"login_form_standalone_labels\\\" cellspacing=\\\"4\\\">\\u003Ctr>\\u003Ctd class=\\\"html7magic\\\">\\u003Clabel for=\\\"email\\\">Email or Phone\\u003C\\/label>\\u003C\\/td>\\u003Ctd>\\u003Cinput type=\\\"text\\\" class=\\\"inputtext\\\" name=\\\"email\\\" id=\\\"email\\\" value=\\\"\\\" \\/>\\u003C\\/td>\\u003C\\/tr>\\u003Ctr>\\u003Ctd class=\\\"html7magic\\\">\\u003Clabel for=\\\"pass\\\">Password\\u003C\\/label>\\u003C\\/td>\\u003Ctd>\\u003Cinput type=\\\"password\\\" class=\\\"inputtext\\\" name=\\\"pass\\\" id=\\\"pass\\\" \\/>\\u003C\\/td>\\u003C\\/tr>\\u003Ctr>\\u003Ctd>\\u003C\\/td>\\u003Ctd class=\\\"login_form_label_field\\\">\\u003Cdiv>\\u003Cdiv class=\\\"uiInputLabel clearfix\\\">\\u003Cinput id=\\\"persist_box\\\" type=\\\"checkbox\\\" name=\\\"persistent\\\" value=\\\"1\\\" tabindex=\\\"3\\\" class=\\\"uiInputLabelCheckbox\\\" \\/>\\u003Clabel for=\\\"persist_box\\\">Keep me logged in\\u003C\\/label>\\u003C\\/div>\\u003Cinput type=\\\"hidden\\\" name=\\\"default_persistent\\\" value=\\\"0\\\" \\/>\\u003C\\/div>\\u003C\\/td>\\u003C\\/tr>\\u003Ctr>\\u003Ctd>\\u003C\\/td>\\u003Ctd>\\u003Clabel class=\\\"uiButton uiButtonConfirm\\\" id=\\\"loginbutton\\\" for=\\\"u4kzrn226\\\">\\u003Cinput value=\\\"Log In\\\" type=\\\"submit\\\" id=\\\"u4kzrn226\\\" \\/>\\u003C\\/label>\\u003C\\/td>\\u003C\\/tr>\\u003Ctr>\\u003Ctd>\\u003C\\/td>\\u003Ctd class=\\\"login_form_label_field\\\">\\u003Ca rel=\\\"nofollow\\\" href=\\\"http:\\/\\/www.facebook.com\\/recover\\/initiate\\\">Forgot your password?\\u003C\\/a>\\u003C\\/td>\\u003C\\/tr>\\u003C\\/table>\\u003Cinput type=\\\"hidden\\\" autocomplete=\\\"off\\\" id=\\\"next\\\" name=\\\"next\\\" value=\\\"https:\\/\\/www.facebook.com\\/NorthCarolinaStateParks\\\" \\/>\\u003Cinput type=\\\"hidden\\\" name=\\\"charset_test\\\" value=\\\"&euro;,&acute;,\\u20ac,\\u00b4,\\u6c34,\\u0414,\\u0404\\\" \\/>\\u003Cinput type=\\\"hidden\\\" autocomplete=\\\"off\\\" name=\\\"timezone\\\" value=\\\"\\\" id=\\\"u4kzrn225\\\" \\/>\\u003Cinput type=\\\"hidden\\\" name=\\\"lgnrnd\\\" value=\\\"090052_K5HA\\\" \\/>\\u003Cinput type=\\\"hidden\\\" id=\\\"lgnjs\\\" name=\\\"lgnjs\\\" value=\\\"n\\\" \\/>\\u003Cinput type=\\\"hidden\\\" autocomplete=\\\"off\\\" id=\\\"locale\\\" name=\\\"locale\\\" value=\\\"en_US\\\" \\/>\\u003C\\/form>\\u003C\\/div>\\u003C\\/div>\\u003Cdiv class=\\\"signUpContainer rfloat\\\">\\u003Cdiv class=\\\"mbs mainTitle fsl fwb fcb\\\">Sign Up\\u003C\\/div>\\u003Cdiv class=\\\"mbm subtitle fsm fwn fcg\\\">It&#039;s free and always will be.\\u003C\\/div>\\u003Cdiv data-referrer=\\\"simple_registration_form\\\">\\u003Cnoscript>\\u003Cdiv id=\\\"no_js_box\\\">\\u003Ch2>JavaScript is disabled on your browser.\\u003C\\/h2>\\u003Cp>Please enable JavaScript on your browser or upgrade to a JavaScript-capable browser to register for Facebook.\\u003C\\/p>\\u003C\\/div>\\u003C\\/noscript>\\u003Cdiv id=\\\"simple_registration_container\\\" class=\\\"simple_registration_container\\\">\\u003Cdiv id=\\\"reg_box\\\">\\u003Cform method=\\\"post\\\" id=\\\"reg\\\" name=\\\"reg\\\" action=\\\"https:\\/\\/m.facebook.com\\/r.php\\\" onsubmit=\\\"return function(event)&#123;return false;&#125;.call(this,event)!==false &amp;&amp; window.Event &amp;&amp; Event.__inlineSubmit &amp;&amp; Event.__inlineSubmit(this,event)\\\">\\u003Cinput type=\\\"hidden\\\" name=\\\"lsd\\\" value=\\\"AVqaEiN6\\\" autocomplete=\\\"off\\\" \\/>\\u003Cinput type=\\\"hidden\\\" autocomplete=\\\"off\\\" id=\\\"locale\\\" name=\\\"locale\\\" value=\\\"en_US\\\" \\/>\\u003Cinput type=\\\"hidden\\\" autocomplete=\\\"off\\\" id=\\\"terms\\\" name=\\\"terms\\\" value=\\\"on\\\" \\/>\\u003Cinput type=\\\"hidden\\\" autocomplete=\\\"off\\\" id=\\\"abtest_registration_group\\\" name=\\\"abtest_registration_group\\\" value=\\\"1\\\" \\/>\\u003Cinput type=\\\"hidden\\\" autocomplete=\\\"off\\\" id=\\\"referrer\\\" name=\\\"referrer\\\" value=\\\"\\\" \\/>\\u003Cinput type=\\\"hidden\\\" autocomplete=\\\"off\\\" id=\\\"md5pass\\\" name=\\\"md5pass\\\" value=\\\"\\\" \\/>\\u003Cinput type=\\\"hidden\\\" autocomplete=\\\"off\\\" id=\\\"validate_mx_records\\\" name=\\\"validate_mx_records\\\" value=\\\"1\\\" \\/>\\u003Cinput type=\\\"hidden\\\" autocomplete=\\\"off\\\" id=\\\"asked_to_login\\\" name=\\\"asked_to_login\\\" value=\\\"0\\\" \\/>\\u003Cinput type=\\\"hidden\\\" autocomplete=\\\"off\\\" id=\\\"ab_test_data\\\" name=\\\"ab_test_data\\\" value=\\\"\\\" \\/>\\u003Cdiv id=\\\"reg_form_box\\\" class=\\\"large_form\\\">\\u003Ctable class=\\\"uiGrid editor\\\" cellspacing=\\\"0\\\" cellpadding=\\\"1\\\">\\u003Ctbody>\\u003Ctr>\\u003Ctd class=\\\"label\\\">\\u003Clabel for=\\\"firstname\\\">Your First Name:\\u003C\\/label>\\u003C\\/td>\\u003Ctd>\\u003Cdiv class=\\\"field_container\\\">\\u003Cinput type=\\\"text\\\" class=\\\"inputtext\\\" id=\\\"firstname\\\" name=\\\"firstname\\\" aria-required=\\\"true\\\" \\/>\\u003C\\/div>\\u003C\\/td>\\u003C\\/tr>\\u003Ctr>\\u003Ctd class=\\\"label\\\">\\u003Clabel for=\\\"lastname\\\">Your Last Name:\\u003C\\/label>\\u003C\\/td>\\u003Ctd>\\u003Cdiv class=\\\"field_container\\\">\\u003Cinput type=\\\"text\\\" class=\\\"inputtext\\\" id=\\\"lastname\\\" name=\\\"lastname\\\" aria-required=\\\"true\\\" \\/>\\u003C\\/div>\\u003C\\/td>\\u003C\\/tr>\\u003Ctr>\\u003Ctd class=\\\"label\\\">\\u003Clabel for=\\\"reg_email__\\\">Your Email:\\u003C\\/label>\\u003C\\/td>\\u003Ctd>\\u003Cdiv class=\\\"field_container\\\">\\u003Cinput type=\\\"text\\\" class=\\\"inputtext\\\" id=\\\"reg_email__\\\" name=\\\"reg_email__\\\" aria-required=\\\"true\\\" autocomplete=\\\"off\\\" \\/>\\u003C\\/div>\\u003C\\/td>\\u003C\\/tr>\\u003Ctr>\\u003Ctd class=\\\"label\\\">\\u003Clabel for=\\\"reg_email_confirmation__\\\">Re-enter Email:\\u003C\\/label>\\u003C\\/td>\\u003Ctd>\\u003Cdiv class=\\\"field_container\\\">\\u003Cinput type=\\\"text\\\" class=\\\"inputtext\\\" id=\\\"reg_email_confirmation__\\\" name=\\\"reg_email_confirmation__\\\" aria-required=\\\"true\\\" autocomplete=\\\"off\\\" \\/>\\u003C\\/div>\\u003C\\/td>\\u003C\\/tr>\\u003Ctr>\\u003Ctd class=\\\"label\\\">\\u003Clabel for=\\\"reg_passwd__\\\">New Password:\\u003C\\/label>\\u003C\\/td>\\u003Ctd>\\u003Cdiv class=\\\"field_container\\\">\\u003Cinput type=\\\"password\\\" class=\\\"inputtext\\\" id=\\\"reg_passwd__\\\" name=\\\"reg_passwd__\\\" value=\\\"\\\" aria-required=\\\"true\\\" autocomplete=\\\"off\\\" \\/>\\u003C\\/div>\\u003C\\/td>\\u003C\\/tr>\\u003Ctr>\\u003Ctd class=\\\"label\\\">I am:\\u003C\\/td>\\u003Ctd>\\u003Cdiv class=\\\"field_container\\\">\\u003Cdiv class=\\\"hidden_elem\\\">\\u003Cselect>\\u003Coption>\\u003C\\/option>\\u003Coption>\\u003C\\/option>\\u003C\\/select>\\u003Cselect>\\u003Coption>\\u003C\\/option>\\u003Coption>\\u003C\\/option>\\u003C\\/select>\\u003C\\/div>\\u003Cselect class=\\\"select\\\" name=\\\"sex\\\" id=\\\"sex\\\" aria-required=\\\"true\\\">\\u003Coption value=\\\"0\\\">Select Sex:\\u003C\\/option>\\u003Coption value=\\\"1\\\">Female\\u003C\\/option>\\u003Coption value=\\\"2\\\">Male\\u003C\\/option>\\u003C\\/select>\\u003C\\/div>\\u003C\\/td>\\u003C\\/tr>\\u003Ctr>\\u003Ctd class=\\\"label\\\">Birthday:\\u003C\\/td>\\u003Ctd>\\u003Cdiv class=\\\"field_container\\\"> \\u003Cselect name=\\\"birthday_month\\\" id=\\\"birthday_month\\\" aria-label=\\\"Month\\\" class=\\\"\\\" onchange=\\\"return run_with(this, [&quot;editor&quot;], function() &#123;editor_date_month_change(this, &quot;birthday_day&quot;, &quot;birthday_year&quot;);&#125;);\\\">\\u003Coption value=\\\"-1\\\">Month:\\u003C\\/option>\\u003Coption value=\\\"1\\\">Jan\\u003C\\/option>\\u003Coption value=\\\"2\\\">Feb\\u003C\\/option>\\u003Coption value=\\\"3\\\">Mar\\u003C\\/option>\\u003Coption value=\\\"4\\\">Apr\\u003C\\/option>\\u003Coption value=\\\"5\\\">May\\u003C\\/option>\\u003Coption value=\\\"6\\\">Jun\\u003C\\/option>\\u003Coption value=\\\"7\\\">Jul\\u003C\\/option>\\u003Coption value=\\\"8\\\">Aug\\u003C\\/option>\\u003Coption value=\\\"9\\\">Sep\\u003C\\/option>\\u003Coption value=\\\"10\\\">Oct\\u003C\\/option>\\u003Coption value=\\\"11\\\">Nov\\u003C\\/option>\\u003Coption value=\\\"12\\\">Dec\\u003C\\/option>\\u003C\\/select> \\u003Cselect name=\\\"birthday_day\\\" id=\\\"birthday_day\\\" aria-label=\\\"Day\\\" class=\\\"\\\">\\u003Coption value=\\\"-1\\\">Day:\\u003C\\/option>\\u003Coption value=\\\"1\\\">1\\u003C\\/option>\\u003Coption value=\\\"2\\\">2\\u003C\\/option>\\u003Coption value=\\\"3\\\">3\\u003C\\/option>\\u003Coption value=\\\"4\\\">4\\u003C\\/option>\\u003Coption value=\\\"5\\\">5\\u003C\\/option>\\u003Coption value=\\\"6\\\">6\\u003C\\/option>\\u003Coption value=\\\"7\\\">7\\u003C\\/option>\\u003Coption value=\\\"8\\\">8\\u003C\\/option>\\u003Coption value=\\\"9\\\">9\\u003C\\/option>\\u003Coption value=\\\"10\\\">10\\u003C\\/option>\\u003Coption value=\\\"11\\\">11\\u003C\\/option>\\u003Coption value=\\\"12\\\">12\\u003C\\/option>\\u003Coption value=\\\"13\\\">13\\u003C\\/option>\\u003Coption value=\\\"14\\\">14\\u003C\\/option>\\u003Coption value=\\\"15\\\">15\\u003C\\/option>\\u003Coption value=\\\"16\\\">16\\u003C\\/option>\\u003Coption value=\\\"17\\\">17\\u003C\\/option>\\u003Coption value=\\\"18\\\">18\\u003C\\/option>\\u003Coption value=\\\"19\\\">19\\u003C\\/option>\\u003Coption value=\\\"20\\\">20\\u003C\\/option>\\u003Coption value=\\\"21\\\">21\\u003C\\/option>\\u003Coption value=\\\"22\\\">22\\u003C\\/option>\\u003Coption value=\\\"23\\\">23\\u003C\\/option>\\u003Coption value=\\\"24\\\">24\\u003C\\/option>\\u003Coption value=\\\"25\\\">25\\u003C\\/option>\\u003Coption value=\\\"26\\\">26\\u003C\\/option>\\u003Coption value=\\\"27\\\">27\\u003C\\/option>\\u003Coption value=\\\"28\\\">28\\u003C\\/option>\\u003Coption value=\\\"29\\\">29\\u003C\\/option>\\u003Coption value=\\\"30\\\">30\\u003C\\/option>\\u003Coption value=\\\"31\\\">31\\u003C\\/option>\\u003C\\/select> \\u003Cselect name=\\\"birthday_year\\\" id=\\\"birthday_year\\\" aria-label=\\\"Year\\\" class=\\\"\\\" onchange=\\\"return run_with(this, [&quot;editor&quot;], function() &#123;editor_date_month_change(&quot;birthday_month&quot;,&quot;birthday_day&quot;,this);&#125;);\\\">\\u003Coption value=\\\"-1\\\">Year:\\u003C\\/option>\\u003Coption value=\\\"2012\\\">2012\\u003C\\/option>\\u003Coption value=\\\"2011\\\">2011\\u003C\\/option>\\u003Coption value=\\\"2010\\\">2010\\u003C\\/option>\\u003Coption value=\\\"2009\\\">2009\\u003C\\/option>\\u003Coption value=\\\"2008\\\">2008\\u003C\\/option>\\u003Coption value=\\\"2007\\\">2007\\u003C\\/option>\\u003Coption value=\\\"2006\\\">2006\\u003C\\/option>\\u003Coption value=\\\"2005\\\">2005\\u003C\\/option>\\u003Coption value=\\\"2004\\\">2004\\u003C\\/option>\\u003Coption value=\\\"2003\\\">2003\\u003C\\/option>\\u003Coption value=\\\"2002\\\">2002\\u003C\\/option>\\u003Coption value=\\\"2001\\\">2001\\u003C\\/option>\\u003Coption value=\\\"2000\\\">2000\\u003C\\/option>\\u003Coption value=\\\"1999\\\">1999\\u003C\\/option>\\u003Coption value=\\\"1998\\\">1998\\u003C\\/option>\\u003Coption value=\\\"1997\\\">1997\\u003C\\/option>\\u003Coption value=\\\"1996\\\">1996\\u003C\\/option>\\u003Coption value=\\\"1995\\\">1995\\u003C\\/option>\\u003Coption value=\\\"1994\\\">1994\\u003C\\/option>\\u003Coption value=\\\"1993\\\">1993\\u003C\\/option>\\u003Coption value=\\\"1992\\\">1992\\u003C\\/option>\\u003Coption value=\\\"1991\\\">1991\\u003C\\/option>\\u003Coption value=\\\"1990\\\">1990\\u003C\\/option>\\u003Coption value=\\\"1989\\\">1989\\u003C\\/option>\\u003Coption value=\\\"1988\\\">1988\\u003C\\/option>\\u003Coption value=\\\"1987\\\">1987\\u003C\\/option>\\u003Coption value=\\\"1986\\\">1986\\u003C\\/option>\\u003Coption value=\\\"1985\\\">1985\\u003C\\/option>\\u003Coption value=\\\"1984\\\">1984\\u003C\\/option>\\u003Coption value=\\\"1983\\\">1983\\u003C\\/option>\\u003Coption value=\\\"1982\\\">1982\\u003C\\/option>\\u003Coption value=\\\"1981\\\">1981\\u003C\\/option>\\u003Coption value=\\\"1980\\\">1980\\u003C\\/option>\\u003Coption value=\\\"1979\\\">1979\\u003C\\/option>\\u003Coption value=\\\"1978\\\">1978\\u003C\\/option>\\u003Coption value=\\\"1977\\\">1977\\u003C\\/option>\\u003Coption value=\\\"1976\\\">1976\\u003C\\/option>\\u003Coption value=\\\"1975\\\">1975\\u003C\\/option>\\u003Coption value=\\\"1974\\\">1974\\u003C\\/option>\\u003Coption value=\\\"1973\\\">1973\\u003C\\/option>\\u003Coption value=\\\"1972\\\">1972\\u003C\\/option>\\u003Coption value=\\\"1971\\\">1971\\u003C\\/option>\\u003Coption value=\\\"1970\\\">1970\\u003C\\/option>\\u003Coption value=\\\"1969\\\">1969\\u003C\\/option>\\u003Coption value=\\\"1968\\\">1968\\u003C\\/option>\\u003Coption value=\\\"1967\\\">1967\\u003C\\/option>\\u003Coption value=\\\"1966\\\">1966\\u003C\\/option>\\u003Coption value=\\\"1965\\\">1965\\u003C\\/option>\\u003Coption value=\\\"1964\\\">1964\\u003C\\/option>\\u003Coption value=\\\"1963\\\">1963\\u003C\\/option>\\u003Coption value=\\\"1962\\\">1962\\u003C\\/option>\\u003Coption value=\\\"1961\\\">1961\\u003C\\/option>\\u003Coption value=\\\"1960\\\">1960\\u003C\\/option>\\u003Coption value=\\\"1959\\\">1959\\u003C\\/option>\\u003Coption value=\\\"1958\\\">1958\\u003C\\/option>\\u003Coption value=\\\"1957\\\">1957\\u003C\\/option>\\u003Coption value=\\\"1956\\\">1956\\u003C\\/option>\\u003Coption value=\\\"1955\\\">1955\\u003C\\/option>\\u003Coption value=\\\"1954\\\">1954\\u003C\\/option>\\u003Coption value=\\\"1953\\\">1953\\u003C\\/option>\\u003Coption value=\\\"1952\\\">1952\\u003C\\/option>\\u003Coption value=\\\"1951\\\">1951\\u003C\\/option>\\u003Coption value=\\\"1950\\\">1950\\u003C\\/option>\\u003Coption value=\\\"1949\\\">1949\\u003C\\/option>\\u003Coption value=\\\"1948\\\">1948\\u003C\\/option>\\u003Coption value=\\\"1947\\\">1947\\u003C\\/option>\\u003Coption value=\\\"1946\\\">1946\\u003C\\/option>\\u003Coption value=\\\"1945\\\">1945\\u003C\\/option>\\u003Coption value=\\\"1944\\\">1944\\u003C\\/option>\\u003Coption value=\\\"1943\\\">1943\\u003C\\/option>\\u003Coption value=\\\"1942\\\">1942\\u003C\\/option>\\u003Coption value=\\\"1941\\\">1941\\u003C\\/option>\\u003Coption value=\\\"1940\\\">1940\\u003C\\/option>\\u003Coption value=\\\"1939\\\">1939\\u003C\\/option>\\u003Coption value=\\\"1938\\\">1938\\u003C\\/option>\\u003Coption value=\\\"1937\\\">1937\\u003C\\/option>\\u003Coption value=\\\"1936\\\">1936\\u003C\\/option>\\u003Coption value=\\\"1935\\\">1935\\u003C\\/option>\\u003Coption value=\\\"1934\\\">1934\\u003C\\/option>\\u003Coption value=\\\"1933\\\">1933\\u003C\\/option>\\u003Coption value=\\\"1932\\\">1932\\u003C\\/option>\\u003Coption value=\\\"1931\\\">1931\\u003C\\/option>\\u003Coption value=\\\"1930\\\">1930\\u003C\\/option>\\u003Coption value=\\\"1929\\\">1929\\u003C\\/option>\\u003Coption value=\\\"1928\\\">1928\\u003C\\/option>\\u003Coption value=\\\"1927\\\">1927\\u003C\\/option>\\u003Coption value=\\\"1926\\\">1926\\u003C\\/option>\\u003Coption value=\\\"1925\\\">1925\\u003C\\/option>\\u003Coption value=\\\"1924\\\">1924\\u003C\\/option>\\u003Coption value=\\\"1923\\\">1923\\u003C\\/option>\\u003Coption value=\\\"1922\\\">1922\\u003C\\/option>\\u003Coption value=\\\"1921\\\">1921\\u003C\\/option>\\u003Coption value=\\\"1920\\\">1920\\u003C\\/option>\\u003Coption value=\\\"1919\\\">1919\\u003C\\/option>\\u003Coption value=\\\"1918\\\">1918\\u003C\\/option>\\u003Coption value=\\\"1917\\\">1917\\u003C\\/option>\\u003Coption value=\\\"1916\\\">1916\\u003C\\/option>\\u003Coption value=\\\"1915\\\">1915\\u003C\\/option>\\u003Coption value=\\\"1914\\\">1914\\u003C\\/option>\\u003Coption value=\\\"1913\\\">1913\\u003C\\/option>\\u003Coption value=\\\"1912\\\">1912\\u003C\\/option>\\u003Coption value=\\\"1911\\\">1911\\u003C\\/option>\\u003Coption value=\\\"1910\\\">1910\\u003C\\/option>\\u003Coption value=\\\"1909\\\">1909\\u003C\\/option>\\u003Coption value=\\\"1908\\\">1908\\u003C\\/option>\\u003Coption value=\\\"1907\\\">1907\\u003C\\/option>\\u003Coption value=\\\"1906\\\">1906\\u003C\\/option>\\u003Coption value=\\\"1905\\\">1905\\u003C\\/option>\\u003C\\/select>\\u003C\\/div>\\u003C\\/td>\\u003C\\/tr>\\u003Ctr>\\u003Ctd class=\\\"label\\\">\\u003C\\/td>\\u003Ctd>\\u003Cdiv id=\\\"birthday_warning\\\">\\u003Ca href=\\\"#\\\" ajaxify=\\\"\\/help\\/ajax\\/reg_birthday\\\" title=\\\"Click for more information\\\" rel=\\\"async\\\">Why do I need to provide my birthday?\\u003C\\/a>\\u003C\\/div>\\u003C\\/td>\\u003C\\/tr>\\u003Ctr>\\u003Ctd class=\\\"label\\\">\\u003C\\/td>\\u003Ctd>\\u003Cdiv class=\\\"fbRegistrationPPT\\\" id=\\\"ppt_container\\\">\\u003Cp class=\\\"privacy_policy_text text\\\">By clicking Sign Up, you agree to our \\u003Ca href=\\\"\\/terms.php\\\" target=\\\"_blank\\\" rel=\\\"nofollow\\\">Terms\\u003C\\/a> and that you have read our \\u003Ca href=\\\"\\/policy.php\\\" target=\\\"_blank\\\" rel=\\\"nofollow\\\">Data Use Policy\\u003C\\/a>, including our \\u003Ca href=\\\"\\/help\\/cookies\\/\\\" target=\\\"_blank\\\" rel=\\\"nofollow\\\">Cookie Use\\u003C\\/a>.\\u003C\\/p>\\u003C\\/div>\\u003C\\/td>\\u003C\\/tr>\\u003C\\/tbody>\\u003C\\/table>\\u003Cdiv class=\\\"reg_btn clearfix\\\">\\u003Clabel class=\\\"uiButton uiButtonSpecial\\\" for=\\\"u4kzrn227\\\">\\u003Cinput value=\\\"Sign Up\\\" name=\\\"websubmit\\\" onclick=\\\"setFormTypeABTest($(&quot;reg&quot;)); return run_with(this, [&quot;reg-util&quot;], function() &#123;RegUtil.getInstance().ajax_validate_data(&quot;https:\\\\\\/\\\\\\/www.facebook.com\\\\\\/ajax\\\\\\/register.php&quot;, &#123;&quot;ignore&quot;:[&quot;captcha&quot;,&quot;pc&quot;]&#125;, &quot;&quot;, &quot;&quot;);&#125;);\\\" type=\\\"submit\\\" id=\\\"u4kzrn227\\\" \\/>\\u003C\\/label>\\u003Cspan id=\\\"async_status\\\" class=\\\"async_status\\\" style=\\\"display: none\\\">\\u003Cimg class=\\\"img\\\" src=\\\"https:\\/\\/s-static.ak.facebook.com\\/rsrc.php\\/v2\\/yb\\/r\\/GsNJNwuI-UM.gif\\\" alt=\\\"\\\" width=\\\"16\\\" height=\\\"11\\\" \\/>\\u003C\\/span>\\u003C\\/div>\\u003C\\/div>\\u003Cdiv id=\\\"reg_captcha\\\" style=\\\"display: none\\\">\\u003Cdiv>\\u003Ch2 id=\\\"security_check_header\\\">Security Check\\u003C\\/h2>\\u003Cdiv id=\\\"outer_captcha_box\\\">\\u003Cdiv id=\\\"captcha_box\\\">\\u003Cdiv class=\\\"field_error\\\" id=\\\"captcha_response_error\\\" style=\\\"display:none;\\\">This field is required.\\u003C\\/div>\\u003Cdiv id=\\\"captcha\\\" class=\\\"captcha\\\">\\u003Cinput type=\\\"hidden\\\" autocomplete=\\\"off\\\" id=\\\"captcha_persist_data\\\" name=\\\"captcha_persist_data\\\" value=\\\"AZmlpB4dLJ4VyOobhP0Phl5pMmDAkhcaFP5Tg32TCmJs0iG3ODlFlvCzRZFGv3CXFQnquQCJ9uLTa_Vt3XDKaeBw9GGHYlcbqPQB_IokYq1MOby_6QsYIWw_sb_PGALalmcUs2pYgCAqnK3B28gdVFsdg7oCxjZlwWyACNO1h9OxQnY8oOcWVXgKOQpWEjdcyDgZaawKaQtdFI2ujY_qAxSWVugEBoKYPDskhYIYoJHQc3QsvtojKVYLw02QaO2JnyhtXN2MWQfkcBw_eXXua5ITNbJowFpEWT5smxsdGr2SUlv_eDUgTKzkUfpo9bPihfd2tsNC25ePuhZhXHAqNhOf5BQVRPvONhUNYBL7wBYA1Q\\\" \\/>\\u003Cdiv>\\u003Cdiv id=\\\"recaptcha_scripts\\\" style=\\\"display:none\\\">\\u003C\\/div>\\u003Cinput type=\\\"hidden\\\" autocomplete=\\\"off\\\" id=\\\"captcha_session\\\" name=\\\"captcha_session\\\" value=\\\"Rv6icB5w83WtHsFURgQ3Ng\\\" \\/>\\u003Cinput type=\\\"hidden\\\" autocomplete=\\\"off\\\" id=\\\"extra_challenge_params\\\" name=\\\"extra_challenge_params\\\" value=\\\"authp=nonce.tt.time.new_audio_default&amp;psig=Q4em2USGGlnHdZrklfvJ7EMyVSA&amp;nonce=Rv6icB5w83WtHsFURgQ3Ng&amp;tt=lv3R7DZbgGNLgCuMOxskOD9tkZI&amp;time=1352998852&amp;new_audio_default=1\\\" \\/>\\u003Cinput type=\\\"hidden\\\" autocomplete=\\\"off\\\" id=\\\"recaptcha_type\\\" name=\\\"recaptcha_type\\\" value=\\\"password\\\" \\/>\\u003Cdiv class=\\\"recaptcha_text\\\">\\u003Cdiv class=\\\"recaptcha_only_if_image\\\">Enter both words below, separated by a space.\\u003Cbr \\/>Can&#039;t read the words below? \\u003Ca href=\\\"#\\\" onclick=\\\"Recaptcha.reload(); return false\\\" id=\\\"recaptcha_reload_btn\\\">Try different words\\u003C\\/a> or \\u003Ca href=\\\"#\\\" onclick=\\\"Recaptcha.switch_type(&quot;audio&quot;); return false;\\\">an audio captcha\\u003C\\/a>.\\u003C\\/div>\\u003Cdiv class=\\\"recaptcha_only_if_audio\\\" style=\\\"display:none\\\">Please enter the words or numbers you hear.\\u003Cbr \\/>\\u003Ca href=\\\"#\\\" onclick=\\\"Recaptcha.reload(); return false\\\" id=\\\"recaptcha_reload_btn\\\">Try different words\\u003C\\/a> or \\u003Ca class=\\\"recaptcha_only_if_audio\\\" href=\\\"#\\\" onclick=\\\"Recaptcha.switch_type(&quot;image&quot;); return false;\\\">back to text\\u003C\\/a>.\\u003C\\/div>\\u003C\\/div>\\u003Cspan id=\\\"recaptcha_play_audio\\\">\\u003C\\/span>\\u003Cdiv class=\\\"audiocaptcha\\\">\\u003C\\/div>\\u003Cdiv id=\\\"recaptcha_image\\\" class=\\\"captcha_image\\\">\\u003C\\/div>\\u003Cdiv id=\\\"recaptcha_loading\\\">Loading... \\u003Cimg class=\\\"captcha_loading img\\\" src=\\\"https:\\/\\/s-static.ak.facebook.com\\/rsrc.php\\/v2\\/yb\\/r\\/GsNJNwuI-UM.gif\\\" alt=\\\"\\\" style=\\\"height:11px;width:16px;\\\" \\/>\\u003C\\/div>\\u003C\\/div>\\u003Cdiv class=\\\"captcha_input\\\">\\u003Clabel>Text in the box:\\u003C\\/label>\\u003Cdiv class=\\\"field_container\\\">\\u003Cinput type=\\\"text\\\" name=\\\"captcha_response\\\" id=\\\"captcha_response\\\" autocomplete=\\\"off\\\" aria-label=\\\"Captcha input. Type the words listed above to continue. Additionally you may also try the audio captcha by clicking the link above.\\\" \\/>\\u003C\\/div>\\u003Ca class=\\\"mlm\\\" href=\\\"#\\\" onclick=\\\"CSS.show($(&#039;captcha_whats_this&#039;)); return false;\\\">What&#039;s this?\\u003C\\/a>\\u003Cdiv id=\\\"captcha_whats_this\\\" class=\\\"hidden_elem\\\">\\u003Cdiv class=\\\"fsl fwb\\\">Security Check\\u003C\\/div>This is a standard security test that we use to prevent spammers from creating fake accounts and spamming users.\\u003C\\/div>\\u003C\\/div>\\u003C\\/div>\\u003C\\/div>\\u003C\\/div>\\u003Cdiv>\\u003Cdiv class=\\\"gridCol\\\">\\u003C\\/div>\\u003Cdiv class=\\\"gridCol\\\">\\u003C\\/div>\\u003C\\/div>\\u003Cdiv id=\\\"captcha_buttons\\\" class=\\\"clearfix\\\" style=\\\"display: none;\\\">\\u003Cdiv id=\\\"back_button\\\" class=\\\"gridCol\\\">\\u003Cdiv class=\\\"cancel_button_image\\\"> \\u00a0 \\u003C\\/div>\\u003Ca id=\\\"cancel_button\\\" href=\\\"#\\\" onclick=\\\"return run_with(this, [&quot;reg-util&quot;], function() &#123;RegUtil.getInstance().hide_captcha();RegUtil.getInstance().show_reg_form();&#125;);\\\">Back\\u003C\\/a>\\u003C\\/div>\\u003Cdiv id=\\\"A_btn_sign_up\\\" class=\\\"gridCol\\\">\\u003Cdiv class=\\\"reg_btn clearfix\\\">\\u003Clabel class=\\\"big_input uiButton uiButtonSpecial\\\" for=\\\"u4kzrn228\\\">\\u003Cinput value=\\\"Sign Up\\\" onclick=\\\"setFormTypeABTest($(&quot;reg&quot;)); return run_with(this, [&quot;reg-util&quot;], function() &#123;RegUtil.getInstance().ajax_validate_data(&quot;https:\\\\\\/\\\\\\/www.facebook.com\\\\\\/ajax\\\\\\/register.php&quot;, &#123;&#125;, &quot;&quot;, &quot;&quot;);&#125;);\\\" type=\\\"submit\\\" id=\\\"u4kzrn228\\\" \\/>\\u003C\\/label>\\u003Cspan id=\\\"captcha_async_status\\\" class=\\\"async_status\\\" style=\\\"display: none\\\">\\u003Cimg class=\\\"img\\\" src=\\\"https:\\/\\/s-static.ak.facebook.com\\/rsrc.php\\/v2\\/yb\\/r\\/GsNJNwuI-UM.gif\\\" alt=\\\"\\\" width=\\\"16\\\" height=\\\"11\\\" \\/>\\u003C\\/span>\\u003C\\/div>\\u003C\\/div>\\u003C\\/div>\\u003C\\/div>\\u003C\\/div>\\u003C\\/form>\\u003Cdiv id=\\\"reg_progress\\\" style=\\\"display: none\\\">\\u003Cdiv id=\\\"progress_wrap\\\">\\u003Cimg class=\\\"img\\\" src=\\\"https:\\/\\/s-static.ak.facebook.com\\/rsrc.php\\/v2\\/yb\\/r\\/GsNJNwuI-UM.gif\\\" alt=\\\"\\\" width=\\\"16\\\" height=\\\"11\\\" \\/>\\u003Cdiv id=\\\"progress_msg\\\">Registering\\u2026\\u003C\\/div>\\u003C\\/div>\\u003C\\/div>\\u003Cdiv id=\\\"reg_error\\\" class=\\\"hidden_elem\\\">\\u003Cdiv id=\\\"reg_error_inner\\\">An error occurred. Please try again.\\u003C\\/div>\\u003C\\/div>\\u003Cdiv id=\\\"tos_container\\\" class=\\\"tos_container\\\">\\u003Cp class=\\\"legal_tos hidden_elem\\\">By clicking Sign Up, you are indicating that you have read and agree to the \\u003Ca href=\\\"\\/legal\\/terms\\\" target=\\\"_blank\\\" rel=\\\"nofollow\\\">Terms of Use\\u003C\\/a> and \\u003Ca href=\\\"\\/policy.php\\\" target=\\\"_blank\\\" rel=\\\"nofollow\\\">Privacy Policy\\u003C\\/a>.\\u003C\\/p>\\u003C\\/div>\\u003C\\/div>\\u003C\\/div>\\u003C\\/div>\\u003C\\/div>\\u003C\\/div>\\u003C\\/div>\\u003Cdiv class=\\\"_14\\\">\\u003Cdiv class=\\\"pam uiOverlayFooter uiBoxGray topborder\\\">\\u003Ca class=\\\"layerCancel uiOverlayButton uiButton uiButtonConfirm uiButtonLarge\\\" href=\\\"#\\\" role=\\\"button\\\">\\u003Cspan class=\\\"uiButtonText\\\">Close\\u003C\\/span>\\u003C\\/a>\\u003C\\/div>\\u003C\\/div>\\u003C\\/div>\"},6]],\"elements\":[[\"m885243_32\",\"login_form\",2,\"m885243_31\"],[\"m885243_33\",\"loginbutton\",2,\"m885243_31\"],[\"m885243_34\",\"login_form\",2,\"m885243_31\"],[\"m885245_35\",\"reg\",2,\"m885243_31\"]],\"require\":[[\"m885243_30\"],[\"LoginFormController\",\"init\",[\"m885243_32\",\"m885243_33\"],[{\"__m\":\"m885243_32\"},{\"__m\":\"m885243_33\"}]],[\"TimelineSignUpDialog\",\"show\",[\"m885243_30\"],[{\"__m\":\"m885243_30\"},\"wait\",0]]]},\"css\":[\"BbtG6\",\"M7bAm\",\"zmU4p\",\"jg0IV\",\"VuNPD\"],\"bootloadable\":{\"editor\":{\"resources\":[\"JU6Q1\",\"MX2xp\",\"t9ps5\"]},\"reg-util\":{\"resources\":[\"JU6Q1\",\"MX2xp\",\"M7bAm\",\"A7lfH\",\"fsumT\",\"zmU4p\",\"T8gbI\",\"vB82p\",\"UJ8Bq\"]}},\"resource_map\":{\"A7lfH\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/yl\\/r\\/SgNfhSw_SOo.js\"},\"l7V35\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/yq\\/r\\/6eq9AE3_NWS.js\"},\"VuNPD\":{\"type\":\"css\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/yY\\/r\\/u8iA3kXb8Y1.css\"},\"t9ps5\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/yy\\/r\\/x6U-jMnCCkg.js\"},\"fsumT\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/yx\\/r\\/brsOAdl0jVK.js\"},\"vB82p\":{\"type\":\"css\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/y8\\/r\\/BuuYm8KfA4a.css\"},\"UJ8Bq\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/yH\\/r\\/ajLTK0Q3afC.js\"}},\"js\":[\"JU6Q1\",\"rKoQM\",\"MX2xp\",\"T8gbI\",\"A7lfH\",\"l7V35\"],\"onload\":[\"$(\\\"u4kzrn225\\\").value = tz_calculate(1352998852)\",\"try { $(\\\"email\\\").focus(); } catch (_ignore) { }\",\"\\n      if (typeof RegUtil == 'undefined') {\\n        RegUtil = function() {};\\n      }\\n      RegUtil.captcha_class=\\\"ReCaptchaCaptcha\\\";\\n      RegUtil.recaptcha_public_key=\\\"6LfDxsYSAAAAAGGLBGaRurawNnbvAGQw5UwRWYXL\\\";\\n      \",\"reg_bootload(\\\"\\\", false, \\\"reg\\\", \\\"form_focus\\\", \\\"https:\\\\\\/\\\\\\/www.facebook.com\\\\\\/ajax\\\\\\/register.php\\\", false);\",\"startFormTypeABTester($(\\\"reg\\\"))\"],\"id\":\"pagelet_loggedout_sign_up\",\"phase\":2})</script>\n", 
        "\n", 
        "<code class=\"hidden_elem\" id=\"u4kzrn2176\"><!-- <div class=\"fbTimelineSection fbTimelineCompactSection fbTimelineSectionTransparent\" id=\"u4kzrn231\"><div class=\"fbTimelineFilters fbTimelineCapsule\"><div class=\"uiSelector inlineBlock uiSelectorNormal uiSelectorDynamicLabel\"><div class=\"wrap\"><a class=\"uiSelectorButton uiButton\" href=\"#\" role=\"button\" id=\"fbTimelinePagesFilterSelectorButton\" aria-haspopup=\"1\" data-label=\"Highlights\" data-length=\"30\" rel=\"toggle\"><span class=\"uiButtonText\">Highlights</span></a><div class=\"uiSelectorMenuWrapper uiToggleFlyout\"><div role=\"menu\" class=\"uiMenu uiSelectorMenu\"><ul class=\"uiMenuInner\"><li class=\"uiMenuItem uiMenuItemRadio uiSelectorOption checked\" data-label=\"Highlights\"><a class=\"itemAnchor\" role=\"menuitemradio\" tabindex=\"0\" href=\"https://www.facebook.com/NorthCarolinaStateParks?filter=3\" aria-checked=\"true\" data-gt=\"&#123;&quot;pagestimeline&quot;:&quot;1&quot;,&quot;filter&quot;:&quot;highlights&quot;,&quot;pageid&quot;:&quot;143412869029&quot;,&quot;context&quot;:&quot;timeline&quot;&#125;\"><span class=\"itemLabel fsm\">Highlights</span></a></li><li class=\"uiMenuItem uiMenuItemRadio uiSelectorOption\" data-label=\"Posts by Page\"><a class=\"itemAnchor\" role=\"menuitemradio\" tabindex=\"-1\" href=\"https://www.facebook.com/NorthCarolinaStateParks?filter=1\" aria-checked=\"false\" data-gt=\"&#123;&quot;pagestimeline&quot;:&quot;1&quot;,&quot;filter&quot;:&quot;highlights&quot;,&quot;pageid&quot;:&quot;143412869029&quot;,&quot;context&quot;:&quot;timeline&quot;&#125;\"><span class=\"itemLabel fsm\">Posts by Page</span></a></li></ul></div></div></div><select><option value=\"\">Highlights</option><option value=\"3\" selected=\"1\">Highlights</option><option value=\"1\">Posts by Page</option></select></div></div><ol class=\"clearfix fbTimelineCapsule\" id=\"u4kzrn230\" data-referrer=\"pagelet_timeline_recent_tc\" data-start=\"1351254248\" data-end=\"1354348799\"><li class=\"fbTimelineUnit fbTimelineTwoColumn clearfix\" data-side=\"l\" data-fixed=\"1\" data-size=\"1\" id=\"tl_unit_7848281715889892654\"><div class=\"topBorder\"></div><div class=\"timelineUnitContainer\" id=\"u4kzrn237\" data-gt=\"&#123;&quot;action_type&quot;:&quot;1&quot;,&quot;aggregation_id&quot;:&quot;10151257845269030&quot;,&quot;aggregation_time&quot;:&quot;1352924021&quot;,&quot;actor&quot;:&quot;143412869029&quot;,&quot;creator&quot;:&quot;143412869029&quot;,&quot;pagestimeline&quot;:&quot;1&quot;,&quot;payload_type&quot;:&quot;7&quot;,&quot;eventtime&quot;:&quot;1352998853&quot;,&quot;viewerid&quot;:&quot;0&quot;,&quot;profileownerid&quot;:&quot;143412869029&quot;,&quot;unitimpressionid&quot;:&quot;f2d1d9d2&quot;,&quot;contentid&quot;:&quot;7848281715889892654&quot;,&quot;timeline_unit_type&quot;:&quot;AddSinglePhotoUnit&quot;,&quot;timewindowsize&quot;:&quot;2&quot;,&quot;contextwindowstart&quot;:&quot;1351753200&quot;,&quot;contextwindowend&quot;:&quot;1354348799&quot;&#125;\" data-time=\"1352924021\"><div class=\"\"><div role=\"article\"><div class=\"clearfix mbs pbs _1_m \"><a class=\"_29h _303\" href=\"http://www.facebook.com/NorthCarolinaStateParks\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-ash3/174880_143412869029_133040642_q.jpg\" alt=\"\" /></a><div class=\"_3dp _29k\"><h5 class=\"_1_s\"><span class=\"fcg\"><span class=\"fwb\"><a href=\"http://www.facebook.com/NorthCarolinaStateParks\">North Carolina State Parks and Recreation</a></span></span></h5><div class=\"_1_n fsm fwn fcg\"><a class=\"uiLinkSubtle\" href=\"http://www.facebook.com/photo.php?fbid=10151257845269030&amp;set=a.164247464029.118730.143412869029&amp;type=1\"><abbr title=\"Wednesday, November 14, 2012 at 12:13pm\" data-utime=\"1352924021\" class=\"timestamp livetimestamp\">20 hours ago</abbr></a><a data-hover=\"tooltip\" aria-label=\"Public\" class=\"uiStreamPrivacy inlineBlock fbStreamPrivacy fbPrivacyAudienceIndicator _1_o\" href=\"#\"><i class=\"lock img sp_2voup5 sx_4929d1\"></i></a></div></div></div><div class=\"aboveUnitContent\"><div class=\"userContentWrapper\"><div class=\"_wk\"><span class=\"userContent\">A cheerful splash of color taken 15 min. ago at Pilot Mountain State Park. We&#039;re already making plans to reopen as soon as possible. Keep up with progress on the park&#039;s webpage: <a href=\"http://www.facebook.com/l.php?u=http%3A%2F%2Fwww.ncparks.gov&amp;h=aAQENx0jQ&amp;s=1\" target=\"_blank\" rel=\"nofollow nofollow\" onmouseover=\"LinkshimAsyncLink.swap(this, &quot;http:\\\\/\\\\/www.ncparks.gov\\\\/&quot;);\" onclick=\"LinkshimAsyncLink.referrer_log(this, &quot;http:\\\\/\\\\/www.ncparks.gov\\\\/&quot;, &quot;http:\\\\/\\\\/www.facebook.com\\\\/si\\\\/ajax\\\\/l\\\\/render_linkshim_log\\\\/?u=http\\\\u00253A\\\\u00252F\\\\u00252Fwww.ncparks.gov&amp;h=aAQENx0jQ&amp;s=1&quot;);\">www.ncparks.gov</a></span></div></div></div><div class=\"photoUnit clearfix\"><div class=\"_53s uiScaledThumb photo photoWidth1\" data-cropped=\"1\"><a class=\"_6i9\" href=\"http://www.facebook.com/photo.php?fbid=10151257845269030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1\" rel=\"theater\" ajaxify=\"http://www.facebook.com/photo.php?fbid=10151257845269030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1&amp;src=https%3A%2F%2Fsphotos-a.xx.fbcdn.net%2Fhphotos-snc7%2F396736_10151257845269030_826810352_n.jpg&amp;size=800%2C630&amp;theater\"><div class=\"uiScaledImageContainer photoWrap\"><img class=\"scaledImageFitWidth img\" src=\"https://sphotos-a.xx.fbcdn.net/hphotos-snc7/c54.0.403.403/p403x403/396736_10151257845269030_826810352_n.jpg\" alt=\"Photo: A cheerful splash of color taken 15 min. ago at Pilot Mountain State Park. We&#039;re already making plans to reopen as soon as possible. Keep up with progress on the park&#039;s webpage: www.ncparks.gov\" width=\"403\" height=\"403\" /></div></a></div></div></div><div class=\"fbTimelineUFI uiCommentContainer\"><form rel=\"async\" class=\"commentable_item autoexpand_mode\" method=\"post\" action=\"/ajax/ufi/modify.php\" onsubmit=\"return window.Event &amp;&amp; Event.__inlineSubmit &amp;&amp; Event.__inlineSubmit(this,event)\" id=\"u4kzrn2142\"><input type=\"hidden\" name=\"charset_test\" value=\"&euro;,&acute;,\u20AC,\u00B4,\u6C34,\u0414,\u0404\" /><input type=\"hidden\" name=\"lsd\" value=\"AVqaEiN6\" autocomplete=\"off\" /><input type=\"hidden\" autocomplete=\"off\" name=\"feedback_params\" value=\"&#123;&quot;actor&quot;:&quot;143412869029&quot;,&quot;target_fbid&quot;:&quot;10151257845269030&quot;,&quot;target_profile_id&quot;:&quot;143412869029&quot;,&quot;type_id&quot;:&quot;7&quot;,&quot;assoc_obj_id&quot;:&quot;&quot;,&quot;source_app_id&quot;:&quot;0&quot;,&quot;extra_story_params&quot;:[],&quot;content_timestamp&quot;:&quot;1352924020&quot;,&quot;check_hash&quot;:&quot;AQBahYbGgHlIklzD&quot;,&quot;source&quot;:&quot;0&quot;&#125;\" /><input type=\"hidden\" autocomplete=\"off\" name=\"data_only_response\" value=\"1\" /><input type=\"hidden\" autocomplete=\"off\" name=\"timeline_ufi\" value=\"1\" /><input type=\"hidden\" name=\"timeline_log_data\" value=\"AQA3Ln_mmAUs9Fqv0SIoGvt5rxUdDULXnjSGlL3Ndq855bx3xOy421FPAe3qDYtTT943tQ0FwjZTcAvQSgSs1LLsPhZ7WKaKggKQcegYMJiNEiqfY89O7zzr1DBAfs3xgSyiqjwaNVa9HKDbZlvhprDl9m_NY8oAMfFOvW4aRhIauYC9ByI7_JWvOui2ZVg3P5QVeEmqlLVRtiPq4-s12NlQqBW846vmIwEbYov4PoomrNMZ657GOGWsBqCp7L3uZf_Zz8eo7dfnalesDdAUM0XtEt-gpRAF-abKHfUgB9xRxRUaN0TF2HzT9sJk1kzYoPP_wuxZ6J73WeVJJ6GjdAiVjP-LLPOtVa8KJOXll5BJz78t720BxT9Ri67Rbr5brz4\" /><div class=\"fbTimelineFeedbackHeader\"><div class=\"fbTimelineFeedbackActions clearfix\"><span class=\"UIActionLinks UIActionLinks_bottom\" data-ft=\"&#123;&quot;tn&quot;:&quot;=&quot;,&quot;type&quot;:20&#125;\"><a class=\"like_link\" title=\"Like this item\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151257845269030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Like</a> &#183; <a class=\"comment_link\" title=\"Leave a comment\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151257845269030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Comment</a> &#183; <a rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151257845269030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Share</a></span></div></div><div><div class=\"uiUfi UFIContainer utn\" id=\"u4kzrn286\"></div></div></form></div></div></div><i class=\"spinePointer\"></i><div class=\"bottomBorder\"></div></li><li class=\"fbTimelineUnit fbTimelineTwoColumn clearfix\" data-side=\"r\" data-fixed=\"1\" data-size=\"1\" id=\"u4kzrn233\"><div class=\"topBorder\"></div><div class=\"timelineUnitContainer\" id=\"u4kzrn234\" data-gt=\"&#123;&quot;eventtime&quot;:&quot;1352998853&quot;,&quot;viewerid&quot;:&quot;0&quot;,&quot;profileownerid&quot;:&quot;143412869029&quot;,&quot;unitimpressionid&quot;:&quot;f2d1d9d2&quot;,&quot;contentid&quot;:&quot;&quot;,&quot;timeline_unit_type&quot;:&quot;TimelinePagesLikedPagesUnit&quot;,&quot;timewindowsize&quot;:&quot;2&quot;,&quot;contextwindowstart&quot;:&quot;1351753200&quot;,&quot;contextwindowend&quot;:&quot;1354348799&quot;&#125;\" data-time=\"1351753200\"><div class=\"\"><div role=\"article\"><div class=\"timelinePageLikedPagesLabel\"><div class=\"clearfix\"><span class=\"lfloat fcg\">Likes</span><a class=\"rfloat\" href=\"/browse/fanned_pages/?id=143412869029\" ajaxify=\"/ajax/browser/dialog/fanned_pages/?id=143412869029\" rel=\"dialog\" role=\"button\">See All</a></div></div><ul class=\"uiList ptm _4kg _6-h _6-j _4ks\" id=\"liked_pages_timeline_unit_list\"><li class=\"uiListItem\"><div class=\"clearfix\"><a class=\"_8o _8s lfloat\" href=\"http://www.facebook.com/TNCNC\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-ash3/50265_170822448792_550479074_q.jpg\" alt=\"\" /></a><div class=\"clearfix _8m \"><div class=\"ptm rfloat\"><a class=\"uiIconText\" href=\"#\" style=\"padding-left: 20px;\" id=\"liked_pages_like_action_170822448792\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=170822448792&amp;next=http%3A%2F%2Fwww.facebook.com%2FTNCNC\" role=\"button\"><i class=\"img sp_1a6dlz sx_4ff1da\" style=\"top: 0px;\"></i>Like</a><a class=\"hidden_elem\" id=\"liked_pages_undo_action_170822448792\" href=\"#\" rel=\"async-post\" ajaxify=\"/ajax/pages/fan_status.php?fbpage_id=170822448792&amp;add=0&amp;reload=0&amp;preserve_tab=0&amp;use_primer=1&amp;fan_origin=page_timeline_liked_pages\">Undo</a></div><div><span class=\"fwb\"><a href=\"http://www.facebook.com/TNCNC\">The Nature Conservancy North Carolina Chapter</a></span><div class=\"fsm fwn fcg\"><div class=\"fsm fwn fcg\">Non-Profit Organization</div></div></div></div></div></li><li class=\"uiListItem\"><div class=\"clearfix\"><a class=\"_8o _8s lfloat\" href=\"http://www.facebook.com/ct4nc\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-snc6/276847_107684473037_1451772501_q.jpg\" alt=\"\" /></a><div class=\"clearfix _8m \"><div class=\"ptm rfloat\"><a class=\"uiIconText\" href=\"#\" style=\"padding-left: 20px;\" id=\"liked_pages_like_action_107684473037\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=107684473037&amp;next=http%3A%2F%2Fwww.facebook.com%2Fct4nc\" role=\"button\"><i class=\"img sp_1a6dlz sx_4ff1da\" style=\"top: 0px;\"></i>Like</a><a class=\"hidden_elem\" id=\"liked_pages_undo_action_107684473037\" href=\"#\" rel=\"async-post\" ajaxify=\"/ajax/pages/fan_status.php?fbpage_id=107684473037&amp;add=0&amp;reload=0&amp;preserve_tab=0&amp;use_primer=1&amp;fan_origin=page_timeline_liked_pages\">Undo</a></div><div><span class=\"fwb\"><a href=\"http://www.facebook.com/ct4nc\">Conservation Trust for North Carolina</a></span><div class=\"fsm fwn fcg\"><div class=\"fsm fwn fcg\">Environmental Conservation</div></div></div></div></div></li><li class=\"uiListItem\"><div class=\"clearfix\"><a class=\"_8o _8s lfloat\" href=\"http://www.facebook.com/NCRPA\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-prn1/50289_291163838873_1879852_q.jpg\" alt=\"\" /></a><div class=\"clearfix _8m \"><div class=\"ptm rfloat\"><a class=\"uiIconText\" href=\"#\" style=\"padding-left: 20px;\" id=\"liked_pages_like_action_291163838873\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=291163838873&amp;next=http%3A%2F%2Fwww.facebook.com%2FNCRPA\" role=\"button\"><i class=\"img sp_1a6dlz sx_4ff1da\" style=\"top: 0px;\"></i>Like</a><a class=\"hidden_elem\" id=\"liked_pages_undo_action_291163838873\" href=\"#\" rel=\"async-post\" ajaxify=\"/ajax/pages/fan_status.php?fbpage_id=291163838873&amp;add=0&amp;reload=0&amp;preserve_tab=0&amp;use_primer=1&amp;fan_origin=page_timeline_liked_pages\">Undo</a></div><div><span class=\"fwb\"><a href=\"http://www.facebook.com/NCRPA\">NC Recreation &amp; Park Association</a></span><div class=\"fsm fwn fcg\"><div class=\"fsm fwn fcg\">Non-Profit Organization</div></div></div></div></div></li><li class=\"uiListItem\"><div class=\"clearfix\"><a class=\"_8o _8s lfloat\" href=\"http://www.facebook.com/pages/Friends-of-the-Mountains-to-Sea-Trail/144643271890\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-snc7/592066_144643271890_845555063_q.jpg\" alt=\"\" /></a><div class=\"clearfix _8m \"><div class=\"ptm rfloat\"><a class=\"uiIconText\" href=\"#\" style=\"padding-left: 20px;\" id=\"liked_pages_like_action_144643271890\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=144643271890&amp;next=http%3A%2F%2Fwww.facebook.com%2Fpages%2FFriends-of-the-Mountains-to-Sea-Trail%2F144643271890\" role=\"button\"><i class=\"img sp_1a6dlz sx_4ff1da\" style=\"top: 0px;\"></i>Like</a><a class=\"hidden_elem\" id=\"liked_pages_undo_action_144643271890\" href=\"#\" rel=\"async-post\" ajaxify=\"/ajax/pages/fan_status.php?fbpage_id=144643271890&amp;add=0&amp;reload=0&amp;preserve_tab=0&amp;use_primer=1&amp;fan_origin=page_timeline_liked_pages\">Undo</a></div><div><span class=\"fwb\"><a href=\"http://www.facebook.com/pages/Friends-of-the-Mountains-to-Sea-Trail/144643271890\">Friends of the Mountains-to-Sea Trail</a></span><div class=\"fsm fwn fcg\"><div class=\"fsm fwn fcg\">Non-Profit Organization</div></div></div></div></div></li><li class=\"uiListItem\"><div class=\"clearfix\"><a class=\"_8o _8s lfloat\" href=\"http://www.facebook.com/pages/Friends-of-State-Parks/124338427526\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-ash4/373049_124338427526_1996898405_q.jpg\" alt=\"\" /></a><div class=\"clearfix _8m \"><div class=\"ptm rfloat\"><a class=\"uiIconText\" href=\"#\" style=\"padding-left: 20px;\" id=\"liked_pages_like_action_124338427526\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=124338427526&amp;next=http%3A%2F%2Fwww.facebook.com%2Fpages%2FFriends-of-State-Parks%2F124338427526\" role=\"button\"><i class=\"img sp_1a6dlz sx_4ff1da\" style=\"top: 0px;\"></i>Like</a><a class=\"hidden_elem\" id=\"liked_pages_undo_action_124338427526\" href=\"#\" rel=\"async-post\" ajaxify=\"/ajax/pages/fan_status.php?fbpage_id=124338427526&amp;add=0&amp;reload=0&amp;preserve_tab=0&amp;use_primer=1&amp;fan_origin=page_timeline_liked_pages\">Undo</a></div><div><span class=\"fwb\"><a href=\"http://www.facebook.com/pages/Friends-of-State-Parks/124338427526\">Friends of State Parks</a></span><div class=\"fsm fwn fcg\"><div class=\"fsm fwn fcg\">Non-Profit Organization</div></div></div></div></div></li></ul></div></div></div><div class=\"bottomBorder\"></div></li><li class=\"fbTimelineUnit fbTimelineTwoColumn clearfix\" data-side=\"l\" data-fixed=\"1\" data-size=\"1\" id=\"tl_unit_8165682813432251273\"><div class=\"topBorder\"></div><div class=\"timelineUnitContainer\" id=\"u4kzrn238\" data-gt=\"&#123;&quot;action_type&quot;:&quot;1&quot;,&quot;aggregation_id&quot;:&quot;10151257567514030&quot;,&quot;aggregation_time&quot;:&quot;1352909529&quot;,&quot;actor&quot;:&quot;143412869029&quot;,&quot;creator&quot;:&quot;143412869029&quot;,&quot;pagestimeline&quot;:&quot;1&quot;,&quot;payload_type&quot;:&quot;7&quot;,&quot;eventtime&quot;:&quot;1352998853&quot;,&quot;viewerid&quot;:&quot;0&quot;,&quot;profileownerid&quot;:&quot;143412869029&quot;,&quot;unitimpressionid&quot;:&quot;f2d1d9d2&quot;,&quot;contentid&quot;:&quot;8165682813432251273&quot;,&quot;timeline_unit_type&quot;:&quot;AddSinglePhotoUnit&quot;,&quot;timewindowsize&quot;:&quot;2&quot;,&quot;contextwindowstart&quot;:&quot;1351753200&quot;,&quot;contextwindowend&quot;:&quot;1354348799&quot;&#125;\" data-time=\"1352909529\"><div class=\"\"><div role=\"article\"><div class=\"clearfix mbs pbs _1_m \"><a class=\"_29h _303\" href=\"http://www.facebook.com/NorthCarolinaStateParks\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-ash3/174880_143412869029_133040642_q.jpg\" alt=\"\" /></a><div class=\"_3dp _29k\"><h5 class=\"_1_s\"><span class=\"fcg\"><span class=\"fwb\"><a href=\"http://www.facebook.com/NorthCarolinaStateParks\">North Carolina State Parks and Recreation</a></span></span></h5><div class=\"_1_n fsm fwn fcg\"><a class=\"uiLinkSubtle\" href=\"http://www.facebook.com/photo.php?fbid=10151257567514030&amp;set=a.164247464029.118730.143412869029&amp;type=1\"><abbr title=\"Wednesday, November 14, 2012 at 8:12am\" data-utime=\"1352909529\">Yesterday</abbr></a><a data-hover=\"tooltip\" aria-label=\"Public\" class=\"uiStreamPrivacy inlineBlock fbStreamPrivacy fbPrivacyAudienceIndicator _1_o\" href=\"#\"><i class=\"lock img sp_2voup5 sx_4929d1\"></i></a></div></div></div><div class=\"aboveUnitContent\"><div class=\"userContentWrapper\"><div class=\"_wk\"><span class=\"userContent\">Our popular Grand Camps at Haw River State Park that bring kids and their grandparents together to explore nature captured the Innovative Program Award this week from the N.C. Recreation and Parks Association. Congrats!</span></div></div></div><div class=\"photoUnit clearfix\"><div class=\"_53s uiScaledThumb photo photoWidth1\" data-cropped=\"1\"><a class=\"_6i9\" href=\"http://www.facebook.com/photo.php?fbid=10151257567514030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1\" rel=\"theater\" ajaxify=\"http://www.facebook.com/photo.php?fbid=10151257567514030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1&amp;src=https%3A%2F%2Fsphotos-b.xx.fbcdn.net%2Fhphotos-ash4%2F483052_10151257567514030_1833409390_n.jpg&amp;size=800%2C600&amp;theater\"><div class=\"uiScaledImageContainer photoWrap\"><img class=\"scaledImageFitWidth img\" src=\"https://sphotos-b.xx.fbcdn.net/hphotos-ash4/c67.0.403.403/p403x403/483052_10151257567514030_1833409390_n.jpg\" alt=\"Photo: Our popular Grand Camps at Haw River State Park that bring kids and their grandparents together to explore nature captured the Innovative Program Award this week from the N.C. Recreation and Parks Association. Congrats!\" width=\"403\" height=\"403\" /></div></a></div></div></div><div class=\"fbTimelineUFI uiCommentContainer\"><form rel=\"async\" class=\"commentable_item autoexpand_mode\" method=\"post\" action=\"/ajax/ufi/modify.php\" onsubmit=\"return window.Event &amp;&amp; Event.__inlineSubmit &amp;&amp; Event.__inlineSubmit(this,event)\" id=\"u4kzrn2143\"><input type=\"hidden\" name=\"charset_test\" value=\"&euro;,&acute;,\u20AC,\u00B4,\u6C34,\u0414,\u0404\" /><input type=\"hidden\" name=\"lsd\" value=\"AVqaEiN6\" autocomplete=\"off\" /><input type=\"hidden\" autocomplete=\"off\" name=\"feedback_params\" value=\"&#123;&quot;actor&quot;:&quot;143412869029&quot;,&quot;target_fbid&quot;:&quot;10151257567514030&quot;,&quot;target_profile_id&quot;:&quot;143412869029&quot;,&quot;type_id&quot;:&quot;7&quot;,&quot;assoc_obj_id&quot;:&quot;&quot;,&quot;source_app_id&quot;:&quot;0&quot;,&quot;extra_story_params&quot;:[],&quot;content_timestamp&quot;:&quot;1352909529&quot;,&quot;check_hash&quot;:&quot;AQA6o1p0o_WcQAuo&quot;,&quot;source&quot;:&quot;0&quot;&#125;\" /><input type=\"hidden\" autocomplete=\"off\" name=\"data_only_response\" value=\"1\" /><input type=\"hidden\" autocomplete=\"off\" name=\"timeline_ufi\" value=\"1\" /><input type=\"hidden\" name=\"timeline_log_data\" value=\"AQA8qzhPSztjIozKiQz1P95pq3-MAOpxEX-yUuSDLKE5IYjWaCwoRTNIoH_6qHYDVgcD6PwWAOMibLH3BdNQl7Lujcj4DF6qdAkeWiFnj0BlRp4CYGaIzjRUNWNvRQkX7UlhkKSPz2cZAdhJwZxj0OcSYbQxxMqDkJxGESKVCblw4CsWH3EfI1rQMs27QSxc8P92LcfHZ_A_xwW6gv4CP1o7lUzjq01Q4H5XJ5G0qRm3CPo67zMykVlpybwsJFWqABm2GqW1LiS1GKXtuML61PTCOzBCz6Aec40vQjBAK1EVGfT2qCX74XOG6dzxnzgdR7nlUNbBFckz_89PQKwK_6pwtlwEi2KtDZ2DcZsZVMBeYpi6BUvnlr-eDnsTlTSQ2xw\" /><div class=\"fbTimelineFeedbackHeader\"><div class=\"fbTimelineFeedbackActions clearfix\"><span class=\"fbTimelineFeedbackShares\"><a data-hover=\"tooltip\" aria-label=\"Show shares\" data-tooltip-alignh=\"right\" href=\"/shares/view?id=10151257567514030\" ajaxify=\"/ajax/shares/view?target_fbid=10151257567514030\" rel=\"async\"><i></i>1</a></span><span class=\"UIActionLinks UIActionLinks_bottom\" data-ft=\"&#123;&quot;tn&quot;:&quot;=&quot;,&quot;type&quot;:20&#125;\"><a class=\"like_link\" title=\"Like this item\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151257567514030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Like</a> &#183; <a class=\"comment_link\" title=\"Leave a comment\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151257567514030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Comment</a> &#183; <a rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151257567514030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Share</a></span></div></div><div><div class=\"uiUfi UFIContainer utn\" id=\"u4kzrn2121\"></div></div></form></div></div></div><i class=\"spinePointer\"></i><div class=\"bottomBorder\"></div></li><li class=\"fbTimelineUnit fbTimelineTwoColumn clearfix\" data-side=\"l\" data-fixed=\"1\" data-size=\"1\" id=\"tl_unit_1570148338748979515\"><div class=\"topBorder\"></div><div class=\"timelineUnitContainer\" id=\"u4kzrn239\" data-gt=\"&#123;&quot;action_type&quot;:&quot;1&quot;,&quot;aggregation_id&quot;:&quot;10151256220039030&quot;,&quot;aggregation_time&quot;:&quot;1352822105&quot;,&quot;actor&quot;:&quot;143412869029&quot;,&quot;creator&quot;:&quot;143412869029&quot;,&quot;pagestimeline&quot;:&quot;1&quot;,&quot;payload_type&quot;:&quot;7&quot;,&quot;eventtime&quot;:&quot;1352998853&quot;,&quot;viewerid&quot;:&quot;0&quot;,&quot;profileownerid&quot;:&quot;143412869029&quot;,&quot;unitimpressionid&quot;:&quot;f2d1d9d2&quot;,&quot;contentid&quot;:&quot;1570148338748979515&quot;,&quot;timeline_unit_type&quot;:&quot;AddSinglePhotoUnit&quot;,&quot;timewindowsize&quot;:&quot;2&quot;,&quot;contextwindowstart&quot;:&quot;1351753200&quot;,&quot;contextwindowend&quot;:&quot;1354348799&quot;&#125;\" data-time=\"1352822105\"><div class=\"\"><div role=\"article\"><div class=\"clearfix mbs pbs _1_m \"><a class=\"_29h _303\" href=\"http://www.facebook.com/NorthCarolinaStateParks\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-ash3/174880_143412869029_133040642_q.jpg\" alt=\"\" /></a><div class=\"_3dp _29k\"><h5 class=\"_1_s\"><span class=\"fcg\"><span class=\"fwb\"><a href=\"http://www.facebook.com/NorthCarolinaStateParks\">North Carolina State Parks and Recreation</a></span></span></h5><div class=\"_1_n fsm fwn fcg\"><a class=\"uiLinkSubtle\" href=\"http://www.facebook.com/photo.php?fbid=10151256220039030&amp;set=a.164247464029.118730.143412869029&amp;type=1\"><abbr title=\"Tuesday, November 13, 2012 at 7:55am\" data-utime=\"1352822105\">Tuesday</abbr></a><a data-hover=\"tooltip\" aria-label=\"Public\" class=\"uiStreamPrivacy inlineBlock fbStreamPrivacy fbPrivacyAudienceIndicator _1_o\" href=\"#\"><i class=\"lock img sp_2voup5 sx_4929d1\"></i></a></div></div></div><div class=\"aboveUnitContent\"><div class=\"userContentWrapper\"><div class=\"_wk\"><span class=\"userContent\">Guest starring in the popular TV series Revolution on Monday...the campground at Carolina Beach State Park, specifically sites 45 and 46.</span></div></div></div><div class=\"photoUnit clearfix\"><div class=\"_53s uiScaledThumb photo photoWidth1\" data-cropped=\"1\"><a class=\"_6i9\" href=\"http://www.facebook.com/photo.php?fbid=10151256220039030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1\" rel=\"theater\" ajaxify=\"http://www.facebook.com/photo.php?fbid=10151256220039030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1&amp;src=https%3A%2F%2Fsphotos-a.xx.fbcdn.net%2Fhphotos-snc6%2F602468_10151256220039030_1579313527_n.jpg&amp;size=800%2C719&amp;theater\"><div class=\"uiScaledImageContainer photoWrap\"><img class=\"scaledImageFitWidth img\" src=\"https://sphotos-a.xx.fbcdn.net/hphotos-snc6/c22.0.403.403/p403x403/602468_10151256220039030_1579313527_n.jpg\" alt=\"Photo: Guest starring in the popular TV series Revolution on Monday...the campground at Carolina Beach State Park, specifically sites 45 and 46.\" width=\"403\" height=\"403\" /></div></a></div></div></div><div class=\"fbTimelineUFI uiCommentContainer\"><form rel=\"async\" class=\"commentable_item autoexpand_mode\" method=\"post\" action=\"/ajax/ufi/modify.php\" onsubmit=\"return window.Event &amp;&amp; Event.__inlineSubmit &amp;&amp; Event.__inlineSubmit(this,event)\" id=\"u4kzrn2144\"><input type=\"hidden\" name=\"charset_test\" value=\"&euro;,&acute;,\u20AC,\u00B4,\u6C34,\u0414,\u0404\" /><input type=\"hidden\" name=\"lsd\" value=\"AVqaEiN6\" autocomplete=\"off\" /><input type=\"hidden\" autocomplete=\"off\" name=\"feedback_params\" value=\"&#123;&quot;actor&quot;:&quot;143412869029&quot;,&quot;target_fbid&quot;:&quot;10151256220039030&quot;,&quot;target_profile_id&quot;:&quot;143412869029&quot;,&quot;type_id&quot;:&quot;7&quot;,&quot;assoc_obj_id&quot;:&quot;&quot;,&quot;source_app_id&quot;:&quot;0&quot;,&quot;extra_story_params&quot;:[],&quot;content_timestamp&quot;:&quot;1352822105&quot;,&quot;check_hash&quot;:&quot;AQBVU6Kiwnj1MbGh&quot;,&quot;source&quot;:&quot;0&quot;&#125;\" /><input type=\"hidden\" autocomplete=\"off\" name=\"data_only_response\" value=\"1\" /><input type=\"hidden\" autocomplete=\"off\" name=\"timeline_ufi\" value=\"1\" /><input type=\"hidden\" name=\"timeline_log_data\" value=\"AQD3mjLrBfCc8XeIrxu6PIgpaYtd65n8VrHOC9sevvmFWC6E_diU2UU2pt0hkpzsBk2vSqOA4I0H1qyC4Oxf-Ig59QKnlh0SWvkdV2rrJv5QbG_9CW1xO_nfZ6Ygfhy8yqgbfIorRDYAVCdJ9TeM44YB_FhjhkKKoPrfe5fKQW-N7h0pUQ853FeIlZ2dbSOPfeXtV6l05zwBF2h5FG8E3WdFRqw5EfDc0V5oMqJZzMzHzg6RA0Awf2BqisSUFbAa9B88RlugZ0aR1fYMZltc3xSDvrYig-PzG9RtsAR7urM8QB_3n6DTypXgUDgEIYnSrS7VoSkiebKEhCo_Q-Ky-Z5YrXmVMV9bGI1ypXxC1weF125_x84IGH21WOoxcM3sSBc\" /><div class=\"fbTimelineFeedbackHeader\"><div class=\"fbTimelineFeedbackActions clearfix\"><span class=\"UIActionLinks UIActionLinks_bottom\" data-ft=\"&#123;&quot;tn&quot;:&quot;=&quot;,&quot;type&quot;:20&#125;\"><a class=\"like_link\" title=\"Like this item\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151256220039030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Like</a> &#183; <a class=\"comment_link\" title=\"Leave a comment\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151256220039030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Comment</a> &#183; <a rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151256220039030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Share</a></span></div></div><div><div class=\"uiUfi UFIContainer utn\" id=\"u4kzrn2122\"></div></div></form></div></div></div><i class=\"spinePointer\"></i><div class=\"bottomBorder\"></div></li><li class=\"fbTimelineUnit fbTimelineTwoColumn clearfix\" data-side=\"l\" data-fixed=\"1\" data-size=\"1\" id=\"tl_unit_-170373988893381171\"><div class=\"topBorder\"></div><div class=\"timelineUnitContainer\" id=\"u4kzrn240\" data-gt=\"&#123;&quot;action_type&quot;:&quot;1&quot;,&quot;aggregation_id&quot;:&quot;10151256146579030&quot;,&quot;aggregation_time&quot;:&quot;1352818576&quot;,&quot;actor&quot;:&quot;143412869029&quot;,&quot;creator&quot;:&quot;143412869029&quot;,&quot;pagestimeline&quot;:&quot;1&quot;,&quot;payload_type&quot;:&quot;7&quot;,&quot;eventtime&quot;:&quot;1352998853&quot;,&quot;viewerid&quot;:&quot;0&quot;,&quot;profileownerid&quot;:&quot;143412869029&quot;,&quot;unitimpressionid&quot;:&quot;f2d1d9d2&quot;,&quot;contentid&quot;:&quot;-170373988893381171&quot;,&quot;timeline_unit_type&quot;:&quot;AddSinglePhotoUnit&quot;,&quot;timewindowsize&quot;:&quot;2&quot;,&quot;contextwindowstart&quot;:&quot;1351753200&quot;,&quot;contextwindowend&quot;:&quot;1354348799&quot;&#125;\" data-time=\"1352818576\"><div class=\"\"><div role=\"article\"><div class=\"clearfix mbs pbs _1_m \"><a class=\"_29h _303\" href=\"http://www.facebook.com/NorthCarolinaStateParks\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-ash3/174880_143412869029_133040642_q.jpg\" alt=\"\" /></a><div class=\"_3dp _29k\"><h5 class=\"_1_s\"><span class=\"fcg\"><span class=\"fwb\"><a href=\"http://www.facebook.com/NorthCarolinaStateParks\">North Carolina State Parks and Recreation</a></span></span></h5><div class=\"_1_n fsm fwn fcg\"><a class=\"uiLinkSubtle\" href=\"http://www.facebook.com/photo.php?fbid=10151256146579030&amp;set=a.164247464029.118730.143412869029&amp;type=1\"><abbr title=\"Tuesday, November 13, 2012 at 6:56am\" data-utime=\"1352818576\">Tuesday</abbr></a><a data-hover=\"tooltip\" aria-label=\"Public\" class=\"uiStreamPrivacy inlineBlock fbStreamPrivacy fbPrivacyAudienceIndicator _1_o\" href=\"#\"><i class=\"lock img sp_2voup5 sx_4929d1\"></i></a></div></div></div><div class=\"aboveUnitContent\"><div class=\"userContentWrapper\"><div class=\"_wk\"><span class=\"userContent\">National Geographic recently named the Jocassee Gorges -\\-\\ the setting of Gorges State Park -\\-\\ as #9 among &quot;50 of the World&#039;s Last Great Places.&quot; Should be on your bucket list.</span></div></div></div><div class=\"photoUnit clearfix\"><div class=\"_53s uiScaledThumb photo photoWidth1\" data-cropped=\"1\"><a class=\"_6i9\" href=\"http://www.facebook.com/photo.php?fbid=10151256146579030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1\" rel=\"theater\" ajaxify=\"http://www.facebook.com/photo.php?fbid=10151256146579030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1&amp;src=https%3A%2F%2Fsphotos-a.xx.fbcdn.net%2Fhphotos-prn1%2F68251_10151256146579030_495096899_n.jpg&amp;size=600%2C787&amp;theater\"><div class=\"uiScaledImageContainer photoWrap\"><img class=\"scaledImageFitWidth img\" src=\"https://sphotos-a.xx.fbcdn.net/hphotos-prn1/c0.0.403.403/p403x403/68251_10151256146579030_495096899_n.jpg\" alt=\"Photo: National Geographic recently named the Jocassee Gorges -\\-\\ the setting of Gorges State Park -\\-\\ as #9 among &quot;50 of the World&#039;s Last Great Places.&quot; Should be on your bucket list.\" width=\"403\" height=\"403\" /></div></a></div></div></div><div class=\"fbTimelineUFI uiCommentContainer\"><form rel=\"async\" class=\"commentable_item autoexpand_mode\" method=\"post\" action=\"/ajax/ufi/modify.php\" onsubmit=\"return window.Event &amp;&amp; Event.__inlineSubmit &amp;&amp; Event.__inlineSubmit(this,event)\" id=\"u4kzrn2145\"><input type=\"hidden\" name=\"charset_test\" value=\"&euro;,&acute;,\u20AC,\u00B4,\u6C34,\u0414,\u0404\" /><input type=\"hidden\" name=\"lsd\" value=\"AVqaEiN6\" autocomplete=\"off\" /><input type=\"hidden\" autocomplete=\"off\" name=\"feedback_params\" value=\"&#123;&quot;actor&quot;:&quot;143412869029&quot;,&quot;target_fbid&quot;:&quot;10151256146579030&quot;,&quot;target_profile_id&quot;:&quot;143412869029&quot;,&quot;type_id&quot;:&quot;7&quot;,&quot;assoc_obj_id&quot;:&quot;&quot;,&quot;source_app_id&quot;:&quot;0&quot;,&quot;extra_story_params&quot;:[],&quot;content_timestamp&quot;:&quot;1352818576&quot;,&quot;check_hash&quot;:&quot;AQB_j_1Kc2tvI-z6&quot;,&quot;source&quot;:&quot;0&quot;&#125;\" /><input type=\"hidden\" autocomplete=\"off\" name=\"data_only_response\" value=\"1\" /><input type=\"hidden\" autocomplete=\"off\" name=\"timeline_ufi\" value=\"1\" /><input type=\"hidden\" name=\"timeline_log_data\" value=\"AQC7h0LfKotKK4FK-mSebPy1Fr2W-WJM39slH0FOKnDBDHdtikQTj8LIcDGoZaJJv6RCYsSdn0V1uoPZO-I590nrU5V8zi45uuKF5Yx5_0qezDkRGVFaanZsHET_PZpifzdp0MUlwEmO-Kw174APy-Bqk_sEerPe-FRvu7W91YT26nFYm0gkX5BriMCU0Ilh8h_cQwZiiij60sx21n8RG-lpbnt5Gdr0YlMkoj3-nGUHuCDiUBVrROcelaUVzPPAdSwEyp7KA518s0J_OQIIExahx-\\-\\BbCjt1Jxi-QyTEC16c3dW9Hy40Kc10vey5q-T0dRL6Rrgf0SPU2p_xneSVzoubMotOfOJq9cZnroE_tvA05YPR3psOtAxYBRaRzIC-DE\" /><div class=\"fbTimelineFeedbackHeader\"><div class=\"fbTimelineFeedbackActions clearfix\"><span class=\"fbTimelineFeedbackShares\"><a data-hover=\"tooltip\" aria-label=\"Show shares\" data-tooltip-alignh=\"right\" href=\"/shares/view?id=10151256146579030\" ajaxify=\"/ajax/shares/view?target_fbid=10151256146579030\" rel=\"async\"><i></i>41</a></span><span class=\"UIActionLinks UIActionLinks_bottom\" data-ft=\"&#123;&quot;tn&quot;:&quot;=&quot;,&quot;type&quot;:20&#125;\"><a class=\"like_link\" title=\"Like this item\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151256146579030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Like</a> &#183; <a class=\"comment_link\" title=\"Leave a comment\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151256146579030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Comment</a> &#183; <a rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151256146579030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Share</a></span></div></div><div><div class=\"uiUfi UFIContainer utn\" id=\"u4kzrn2123\"></div></div></form></div></div></div><i class=\"spinePointer\"></i><div class=\"bottomBorder\"></div></li><li class=\"fbTimelineUnit fbTimelineTwoColumn clearfix\" data-side=\"l\" data-fixed=\"1\" data-size=\"1\" id=\"tl_unit_3209293303011316676\"><div class=\"topBorder\"></div><div class=\"timelineUnitContainer\" id=\"u4kzrn241\" data-gt=\"&#123;&quot;action_type&quot;:&quot;1&quot;,&quot;aggregation_id&quot;:&quot;10151255074004030&quot;,&quot;aggregation_time&quot;:&quot;1352748410&quot;,&quot;actor&quot;:&quot;143412869029&quot;,&quot;creator&quot;:&quot;143412869029&quot;,&quot;pagestimeline&quot;:&quot;1&quot;,&quot;payload_type&quot;:&quot;7&quot;,&quot;eventtime&quot;:&quot;1352998853&quot;,&quot;viewerid&quot;:&quot;0&quot;,&quot;profileownerid&quot;:&quot;143412869029&quot;,&quot;unitimpressionid&quot;:&quot;f2d1d9d2&quot;,&quot;contentid&quot;:&quot;3209293303011316676&quot;,&quot;timeline_unit_type&quot;:&quot;AddSinglePhotoUnit&quot;,&quot;timewindowsize&quot;:&quot;2&quot;,&quot;contextwindowstart&quot;:&quot;1351753200&quot;,&quot;contextwindowend&quot;:&quot;1354348799&quot;&#125;\" data-time=\"1352748410\"><div class=\"\"><div role=\"article\"><div class=\"clearfix mbs pbs _1_m \"><a class=\"_29h _303\" href=\"http://www.facebook.com/NorthCarolinaStateParks\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-ash3/174880_143412869029_133040642_q.jpg\" alt=\"\" /></a><div class=\"_3dp _29k\"><h5 class=\"_1_s\"><span class=\"fcg\"><span class=\"fwb\"><a href=\"http://www.facebook.com/NorthCarolinaStateParks\">North Carolina State Parks and Recreation</a></span></span></h5><div class=\"_1_n fsm fwn fcg\"><a class=\"uiLinkSubtle\" href=\"http://www.facebook.com/photo.php?fbid=10151255074004030&amp;set=a.164247464029.118730.143412869029&amp;type=1\"><abbr title=\"Monday, November 12, 2012 at 11:26am\" data-utime=\"1352748410\">Monday</abbr></a><a data-hover=\"tooltip\" aria-label=\"Public\" class=\"uiStreamPrivacy inlineBlock fbStreamPrivacy fbPrivacyAudienceIndicator _1_o\" href=\"#\"><i class=\"lock img sp_2voup5 sx_4929d1\"></i></a></div></div></div><div class=\"aboveUnitContent\"><div class=\"userContentWrapper\"><div class=\"_wk\"><span class=\"userContent\">Our sincere thanks to all the rangers, NC Forest Service folks and volunteers who&#039;ve helped us manage the fire at Pilot Mountain State Park.</span></div></div></div><div class=\"photoUnit clearfix\"><div class=\"_53s uiScaledThumb photo photoWidth1\" data-cropped=\"1\"><a class=\"_6i9\" href=\"http://www.facebook.com/photo.php?fbid=10151255074004030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1\" rel=\"theater\" ajaxify=\"http://www.facebook.com/photo.php?fbid=10151255074004030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1&amp;src=https%3A%2F%2Fsphotos-b.xx.fbcdn.net%2Fhphotos-ash4%2F311341_10151255074004030_1780724402_n.jpg&amp;size=800%2C610&amp;theater\"><div class=\"uiScaledImageContainer photoWrap\"><img class=\"scaledImageFitWidth img\" src=\"https://sphotos-b.xx.fbcdn.net/hphotos-ash4/c62.0.403.403/p403x403/311341_10151255074004030_1780724402_n.jpg\" alt=\"Photo: Our sincere thanks to all the rangers, NC Forest Service folks and volunteers who&#039;ve helped us manage the fire at Pilot Mountain State Park.\" width=\"403\" height=\"403\" /></div></a></div></div></div><div class=\"fbTimelineUFI uiCommentContainer\"><form rel=\"async\" class=\"commentable_item autoexpand_mode\" method=\"post\" action=\"/ajax/ufi/modify.php\" onsubmit=\"return window.Event &amp;&amp; Event.__inlineSubmit &amp;&amp; Event.__inlineSubmit(this,event)\" id=\"u4kzrn2146\"><input type=\"hidden\" name=\"charset_test\" value=\"&euro;,&acute;,\u20AC,\u00B4,\u6C34,\u0414,\u0404\" /><input type=\"hidden\" name=\"lsd\" value=\"AVqaEiN6\" autocomplete=\"off\" /><input type=\"hidden\" autocomplete=\"off\" name=\"feedback_params\" value=\"&#123;&quot;actor&quot;:&quot;143412869029&quot;,&quot;target_fbid&quot;:&quot;10151255074004030&quot;,&quot;target_profile_id&quot;:&quot;143412869029&quot;,&quot;type_id&quot;:&quot;7&quot;,&quot;assoc_obj_id&quot;:&quot;&quot;,&quot;source_app_id&quot;:&quot;0&quot;,&quot;extra_story_params&quot;:[],&quot;content_timestamp&quot;:&quot;1352748409&quot;,&quot;check_hash&quot;:&quot;AQCo3HrAGuBfi0sq&quot;,&quot;source&quot;:&quot;0&quot;&#125;\" /><input type=\"hidden\" autocomplete=\"off\" name=\"data_only_response\" value=\"1\" /><input type=\"hidden\" autocomplete=\"off\" name=\"timeline_ufi\" value=\"1\" /><input type=\"hidden\" name=\"timeline_log_data\" value=\"AQBj4ApubcZLzM98wFDQz1UiEzCxbmSjVsz-eEBX_eKK3eI7_gvGe9QcwB-HIzK2SDdUslUH350OaOYH77BJy_4JhtUlkRhDXsGhFgHnG74pkhTj3jaICB9VbXJX_VUImo5ERbRGHWKbGt-lR2vozPt1JWLD_knd3IwGSe8_slL9uPyW4sTwYGfSNrGP-_Ceb2AtZCLTmOR921xhOQJFr1e_C1VsEvQwFczHTN8G1xIrhbnRxpI-0PyLo7vx2nweD4hnjLGe4NorZ9bJGbL5f2EQFN-zpfrRbay_d046li2e12WhBinzPROkbKntwccRhGdxA-QCCZRaQZI-ZgKf2VmZ4-ry_FljdF99bOqMZZxKn4VAzm-eqpMrIf6oJacOWEE\" /><div class=\"fbTimelineFeedbackHeader\"><div class=\"fbTimelineFeedbackActions clearfix\"><span class=\"fbTimelineFeedbackShares\"><a data-hover=\"tooltip\" aria-label=\"Show shares\" data-tooltip-alignh=\"right\" href=\"/shares/view?id=10151255074004030\" ajaxify=\"/ajax/shares/view?target_fbid=10151255074004030\" rel=\"async\"><i></i>13</a></span><span class=\"UIActionLinks UIActionLinks_bottom\" data-ft=\"&#123;&quot;tn&quot;:&quot;=&quot;,&quot;type&quot;:20&#125;\"><a class=\"like_link\" title=\"Like this item\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151255074004030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Like</a> &#183; <a class=\"comment_link\" title=\"Leave a comment\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151255074004030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Comment</a> &#183; <a rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151255074004030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Share</a></span></div></div><div><div class=\"uiUfi UFIContainer utn\" id=\"u4kzrn2124\"></div></div></form></div></div></div><i class=\"spinePointer\"></i><div class=\"bottomBorder\"></div></li><li class=\"fbTimelineUnit fbTimelineTwoColumn clearfix\" data-side=\"l\" data-fixed=\"1\" data-size=\"1\" id=\"tl_unit_3689851695770658638\"><div class=\"topBorder\"></div><div class=\"timelineUnitContainer\" id=\"u4kzrn242\" data-gt=\"&#123;&quot;action_type&quot;:&quot;1&quot;,&quot;aggregation_id&quot;:&quot;10151250347114030&quot;,&quot;aggregation_time&quot;:&quot;1352470430&quot;,&quot;actor&quot;:&quot;143412869029&quot;,&quot;creator&quot;:&quot;143412869029&quot;,&quot;pagestimeline&quot;:&quot;1&quot;,&quot;payload_type&quot;:&quot;7&quot;,&quot;eventtime&quot;:&quot;1352998853&quot;,&quot;viewerid&quot;:&quot;0&quot;,&quot;profileownerid&quot;:&quot;143412869029&quot;,&quot;unitimpressionid&quot;:&quot;f2d1d9d2&quot;,&quot;contentid&quot;:&quot;3689851695770658638&quot;,&quot;timeline_unit_type&quot;:&quot;AddSinglePhotoUnit&quot;,&quot;timewindowsize&quot;:&quot;2&quot;,&quot;contextwindowstart&quot;:&quot;1351753200&quot;,&quot;contextwindowend&quot;:&quot;1354348799&quot;&#125;\" data-time=\"1352470430\"><div class=\"\"><div role=\"article\"><div class=\"clearfix mbs pbs _1_m \"><a class=\"_29h _303\" href=\"http://www.facebook.com/NorthCarolinaStateParks\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-ash3/174880_143412869029_133040642_q.jpg\" alt=\"\" /></a><div class=\"_3dp _29k\"><h5 class=\"_1_s\"><span class=\"fcg\"><span class=\"fwb\"><a href=\"http://www.facebook.com/NorthCarolinaStateParks\">North Carolina State Parks and Recreation</a></span></span></h5><div class=\"_1_n fsm fwn fcg\"><a class=\"uiLinkSubtle\" href=\"http://www.facebook.com/photo.php?fbid=10151250347114030&amp;set=a.164247464029.118730.143412869029&amp;type=1\"><abbr title=\"Friday, November 9, 2012 at 6:13am\" data-utime=\"1352470430\">November 9</abbr></a><a data-hover=\"tooltip\" aria-label=\"Public\" class=\"uiStreamPrivacy inlineBlock fbStreamPrivacy fbPrivacyAudienceIndicator _1_o\" href=\"#\"><i class=\"lock img sp_2voup5 sx_4929d1\"></i></a></div></div></div><div class=\"aboveUnitContent\"><div class=\"userContentWrapper\"><div class=\"_wk\"><span class=\"userContent\">Autumn hues are reaching our eastern parks and it&#039;s a great time to visit. This pastoral scene was taken by one of our rangers visiting Lumber River State Park.</span></div></div></div><div class=\"photoUnit clearfix\"><div class=\"_53s uiScaledThumb photo photoWidth1\" data-cropped=\"1\"><a class=\"_6i9\" href=\"http://www.facebook.com/photo.php?fbid=10151250347114030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1\" rel=\"theater\" ajaxify=\"http://www.facebook.com/photo.php?fbid=10151250347114030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1&amp;src=https%3A%2F%2Fsphotos-a.xx.fbcdn.net%2Fhphotos-ash3%2F532256_10151250347114030_1579263044_n.jpg&amp;size=800%2C464&amp;theater\"><div class=\"uiScaledImageContainer photoWrap\"><img class=\"scaledImageFitWidth img\" src=\"https://sphotos-a.xx.fbcdn.net/hphotos-ash3/c145.0.403.403/p403x403/532256_10151250347114030_1579263044_n.jpg\" alt=\"Photo: Autumn hues are reaching our eastern parks and it&#039;s a great time to visit. This pastoral scene was taken by one of our rangers visiting Lumber River State Park.\" width=\"403\" height=\"403\" /></div></a></div></div></div><div class=\"fbTimelineUFI uiCommentContainer\"><form rel=\"async\" class=\"commentable_item autoexpand_mode\" method=\"post\" action=\"/ajax/ufi/modify.php\" onsubmit=\"return window.Event &amp;&amp; Event.__inlineSubmit &amp;&amp; Event.__inlineSubmit(this,event)\" id=\"u4kzrn2147\"><input type=\"hidden\" name=\"charset_test\" value=\"&euro;,&acute;,\u20AC,\u00B4,\u6C34,\u0414,\u0404\" /><input type=\"hidden\" name=\"lsd\" value=\"AVqaEiN6\" autocomplete=\"off\" /><input type=\"hidden\" autocomplete=\"off\" name=\"feedback_params\" value=\"&#123;&quot;actor&quot;:&quot;143412869029&quot;,&quot;target_fbid&quot;:&quot;10151250347114030&quot;,&quot;target_profile_id&quot;:&quot;143412869029&quot;,&quot;type_id&quot;:&quot;7&quot;,&quot;assoc_obj_id&quot;:&quot;&quot;,&quot;source_app_id&quot;:&quot;0&quot;,&quot;extra_story_params&quot;:[],&quot;content_timestamp&quot;:&quot;1352470430&quot;,&quot;check_hash&quot;:&quot;AQAS0KTwgcAnzbFK&quot;,&quot;source&quot;:&quot;0&quot;&#125;\" /><input type=\"hidden\" autocomplete=\"off\" name=\"data_only_response\" value=\"1\" /><input type=\"hidden\" autocomplete=\"off\" name=\"timeline_ufi\" value=\"1\" /><input type=\"hidden\" name=\"timeline_log_data\" value=\"AQAhy4cWXkO_7-k1hR2-Yc1U-hyQa0bMUOS8y6yPx2Qqwy0Wc0yN96Lzg8rs1eLL4GLIyEfDIjC_SVutuqRTNg_XUIth7aKK4rzkkLbQ7-Hhkvnd9rBSNIuNDyKOb4gc0KeFJWBil2qXpGF_zLAvbQvmtBhQQghQrLt3HBv7GBMQltcTih00p9pKXDCy-Wn7Rf7o8xKhe_61u7XlcflejlMQuPtSKGwVZHWsqCWGsvR5vSCvPPdTPFR_ZH7xyPIkgo5EdG7XvDEqL04u8-7Dg3p_lmk4nsBbrsH9-05ph_yxi2A22aggXxYXmWqhZ6y5UPG7pI6aRrQgC9Ivck7KtKOaJzxq2C2DpBnK6kww4fcFx1MHD2aqMDoVfXKLJtjA82s\" /><div class=\"fbTimelineFeedbackHeader\"><div class=\"fbTimelineFeedbackActions clearfix\"><span class=\"fbTimelineFeedbackShares\"><a data-hover=\"tooltip\" aria-label=\"Show shares\" data-tooltip-alignh=\"right\" href=\"/shares/view?id=10151250347114030\" ajaxify=\"/ajax/shares/view?target_fbid=10151250347114030\" rel=\"async\"><i></i>3</a></span><span class=\"UIActionLinks UIActionLinks_bottom\" data-ft=\"&#123;&quot;tn&quot;:&quot;=&quot;,&quot;type&quot;:20&#125;\"><a class=\"like_link\" title=\"Like this item\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151250347114030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Like</a> &#183; <a class=\"comment_link\" title=\"Leave a comment\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151250347114030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Comment</a> &#183; <a rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151250347114030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Share</a></span></div></div><div><div class=\"uiUfi UFIContainer utn\" id=\"u4kzrn287\"></div></div></form></div></div></div><i class=\"spinePointer\"></i><div class=\"bottomBorder\"></div></li><li class=\"fbTimelineUnit fbTimelineTwoColumn clearfix\" data-side=\"l\" data-fixed=\"1\" data-size=\"1\" id=\"tl_unit_-5394651607260822551\"><div class=\"topBorder\"></div><div class=\"timelineUnitContainer\" id=\"u4kzrn243\" data-gt=\"&#123;&quot;action_type&quot;:&quot;1&quot;,&quot;aggregation_id&quot;:&quot;10151249662709030&quot;,&quot;aggregation_time&quot;:&quot;1352419460&quot;,&quot;actor&quot;:&quot;143412869029&quot;,&quot;creator&quot;:&quot;143412869029&quot;,&quot;pagestimeline&quot;:&quot;1&quot;,&quot;payload_type&quot;:&quot;11&quot;,&quot;eventtime&quot;:&quot;1352998853&quot;,&quot;viewerid&quot;:&quot;0&quot;,&quot;profileownerid&quot;:&quot;143412869029&quot;,&quot;unitimpressionid&quot;:&quot;f2d1d9d2&quot;,&quot;contentid&quot;:&quot;-5394651607260822551&quot;,&quot;timeline_unit_type&quot;:&quot;StatusMessageUnit&quot;,&quot;timewindowsize&quot;:&quot;2&quot;,&quot;contextwindowstart&quot;:&quot;1351753200&quot;,&quot;contextwindowend&quot;:&quot;1354348799&quot;&#125;\" data-time=\"1352419460\"><div class=\"\"><div role=\"article\"><div class=\"clearfix mbs pbs _1_m \"><a class=\"_29h _303\" href=\"http://www.facebook.com/NorthCarolinaStateParks\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-ash3/174880_143412869029_133040642_q.jpg\" alt=\"\" /></a><div class=\"_3dp _29k\"><h5 class=\"_1_s\"><span class=\"fcg\"><span class=\"fwb\"><a href=\"http://www.facebook.com/NorthCarolinaStateParks\">North Carolina State Parks and Recreation</a></span></span></h5><div class=\"_1_n fsm fwn fcg\"><a class=\"uiLinkSubtle\" href=\"/NorthCarolinaStateParks/posts/10151249662709030\"><abbr title=\"Thursday, November 8, 2012 at 4:04pm\" data-utime=\"1352419460\">November 8</abbr></a> via <a class=\"uiLinkSubtle\" href=\"/mobile/?v=165907476854626\">mobile</a><a data-hover=\"tooltip\" aria-label=\"Public\" class=\"uiStreamPrivacy inlineBlock fbStreamPrivacy fbPrivacyAudienceIndicator _1_o\" href=\"#\"><i class=\"lock img sp_2voup5 sx_4929d1\"></i></a></div></div></div><div class=\"_1x1\"><div class=\"userContentWrapper\"><div class=\"_wk\"><span class=\"userContent\">Pilot Mountain State Park will be closed Friday due to the need to monitor a small woods fire near the summit visitor areas.</span></div></div></div></div><div class=\"fbTimelineUFI uiCommentContainer\"><form rel=\"async\" class=\"commentable_item autoexpand_mode\" method=\"post\" action=\"/ajax/ufi/modify.php\" onsubmit=\"return window.Event &amp;&amp; Event.__inlineSubmit &amp;&amp; Event.__inlineSubmit(this,event)\" id=\"u4kzrn2148\"><input type=\"hidden\" name=\"charset_test\" value=\"&euro;,&acute;,\u20AC,\u00B4,\u6C34,\u0414,\u0404\" /><input type=\"hidden\" name=\"lsd\" value=\"AVqaEiN6\" autocomplete=\"off\" /><input type=\"hidden\" autocomplete=\"off\" name=\"feedback_params\" value=\"&#123;&quot;actor&quot;:&quot;143412869029&quot;,&quot;target_fbid&quot;:&quot;10151249662709030&quot;,&quot;target_profile_id&quot;:&quot;143412869029&quot;,&quot;type_id&quot;:&quot;22&quot;,&quot;assoc_obj_id&quot;:&quot;&quot;,&quot;source_app_id&quot;:&quot;0&quot;,&quot;extra_story_params&quot;:[],&quot;content_timestamp&quot;:&quot;1352419460&quot;,&quot;check_hash&quot;:&quot;AQBRWde4KBeKkPrC&quot;,&quot;source&quot;:&quot;0&quot;&#125;\" /><input type=\"hidden\" autocomplete=\"off\" name=\"data_only_response\" value=\"1\" /><input type=\"hidden\" autocomplete=\"off\" name=\"timeline_ufi\" value=\"1\" /><input type=\"hidden\" name=\"timeline_log_data\" value=\"AQAEagyljGsWUQEX3FTGfzV3XfHi2wZyPPIxYTbNhG672WPUin0I8VEoO9xYf0mPGkMk6f_99IGNAORZfoxHo6AjPtJHSKffmVFfa0EFsjYVbfhH5dsAY2ZejXbf3jCd1IKRyqShHeZ6GanOvRZV1f4xH-xANVgPDZfClGifUMVPwA3P6gl1_u7b4OuKLDw0BzYqJD8v1a0YBCkorIs_25cG3m3AOeoco07FWRFJmQb_A223KuhUSMV14s0qTUwAOf3wtd7WJXxaFtNg99QRigMxdZGR_xEqgM3U3OTmKYDA2zz6s4M5QdwsE1L7zkLoJljI6lX1lNOjzO3lWkwBWQ14iQfGKSa7uRC4GwMRtCgOBa64tsjcQQ07ogdwx8DsRiE\" /><div class=\"fbTimelineFeedbackHeader\"><div class=\"fbTimelineFeedbackActions clearfix\"><span class=\"fbTimelineFeedbackShares\"><a data-hover=\"tooltip\" aria-label=\"Show shares\" data-tooltip-alignh=\"right\" href=\"/shares/view?id=10151249662709030\" ajaxify=\"/ajax/shares/view?target_fbid=10151249662709030\" rel=\"async\"><i></i>4</a></span><span class=\"UIActionLinks UIActionLinks_bottom\" data-ft=\"&#123;&quot;tn&quot;:&quot;=&quot;,&quot;type&quot;:20&#125;\"><a class=\"like_link\" title=\"Like this item\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2FNorthCarolinaStateParks%2Fposts%2F10151249662709030\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Like</a> &#183; <a class=\"comment_link\" title=\"Leave a comment\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2FNorthCarolinaStateParks%2Fposts%2F10151249662709030\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Comment</a> &#183; <a rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2FNorthCarolinaStateParks%2Fposts%2F10151249662709030\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Share</a></span></div></div><div><div class=\"uiUfi UFIContainer utn\" id=\"u4kzrn2125\"></div></div></form></div></div></div><i class=\"spinePointer\"></i><div class=\"bottomBorder\"></div></li><li class=\"fbTimelineUnit fbTimelineTwoColumn clearfix\" data-side=\"l\" data-fixed=\"1\" data-size=\"1\" id=\"tl_unit_8134846466346127333\"><div class=\"topBorder\"></div><div class=\"timelineUnitContainer\" id=\"u4kzrn245\" data-gt=\"&#123;&quot;action_type&quot;:&quot;1&quot;,&quot;aggregation_id&quot;:&quot;217369611728583&quot;,&quot;aggregation_time&quot;:&quot;1352312833&quot;,&quot;actor&quot;:&quot;143412869029&quot;,&quot;creator&quot;:&quot;143412869029&quot;,&quot;pagestimeline&quot;:&quot;1&quot;,&quot;payload_type&quot;:&quot;5&quot;,&quot;eventtime&quot;:&quot;1352998853&quot;,&quot;viewerid&quot;:&quot;0&quot;,&quot;profileownerid&quot;:&quot;143412869029&quot;,&quot;unitimpressionid&quot;:&quot;f2d1d9d2&quot;,&quot;contentid&quot;:&quot;8134846466346127333&quot;,&quot;timeline_unit_type&quot;:&quot;BestPostedLinkUnit&quot;,&quot;timewindowsize&quot;:&quot;2&quot;,&quot;contextwindowstart&quot;:&quot;1351753200&quot;,&quot;contextwindowend&quot;:&quot;1354348799&quot;&#125;\" data-time=\"1352312833\"><div class=\"\"><div role=\"article\"><div class=\"clearfix mbs pbs _1_m \"><a class=\"_29h _303\" href=\"http://www.facebook.com/NorthCarolinaStateParks\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-ash3/174880_143412869029_133040642_q.jpg\" alt=\"\" /></a><div class=\"_3dp _29k\"><h5 class=\"_1_s\"><span class=\"fcg\"><span class=\"fwb\"><a href=\"http://www.facebook.com/NorthCarolinaStateParks\">North Carolina State Parks and Recreation</a></span> shared a <a href=\"http://www.facebook.com/l.php?u=http%3A%2F%2Fwww.flickr.com%2Fphotos%2Finfobucket%2Fsets%2F72157631930079339%2F&amp;h=QAQHnf_bFAQEm68X7kQPHGZ9rw2IF6uoAGeAB0X6ukS809g&amp;enc=AZNX-wtavEM6swlA97BDcXnTc0cSpmABN4zVSPhc0z7Chww0sOzhe0au6AIhZ8zkx3LvJ_TFeejsI8onrZqi37c2&amp;s=1\" rel=\"nofollow\" onmouseover=\"LinkshimAsyncLink.swap(this, &quot;http:\\\\/\\\\/www.flickr.com\\\\/photos\\\\/infobucket\\\\/sets\\\\/72157631930079339\\\\/&quot;);\" onclick=\"LinkshimAsyncLink.referrer_log(this, &quot;http:\\\\/\\\\/www.flickr.com\\\\/photos\\\\/infobucket\\\\/sets\\\\/72157631930079339\\\\/&quot;, &quot;http:\\\\/\\\\/www.facebook.com\\\\/si\\\\/ajax\\\\/l\\\\/render_linkshim_log\\\\/?u=http\\\\u00253A\\\\u00252F\\\\u00252Fwww.flickr.com\\\\u00252Fphotos\\\\u00252Finfobucket\\\\u00252Fsets\\\\u00252F72157631930079339\\\\u00252F&amp;h=QAQHnf_bFAQEm68X7kQPHGZ9rw2IF6uoAGeAB0X6ukS809g&amp;enc=AZNX-wtavEM6swlA97BDcXnTc0cSpmABN4zVSPhc0z7Chww0sOzhe0au6AIhZ8zkx3LvJ_TFeejsI8onrZqi37c2&amp;s=1&quot;);\" target=\"_blank\">link</a>.</span></h5><div class=\"_1_n fsm fwn fcg\"><a class=\"uiLinkSubtle\" href=\"/NorthCarolinaStateParks/posts/217369611728583\"><abbr title=\"Wednesday, November 7, 2012 at 10:27am\" data-utime=\"1352312833\">November 7</abbr></a><a data-hover=\"tooltip\" aria-label=\"Public\" class=\"uiStreamPrivacy inlineBlock fbStreamPrivacy fbPrivacyAudienceIndicator _1_o\" href=\"#\"><i class=\"lock img sp_2voup5 sx_4929d1\"></i></a></div></div></div><div class=\"userContentWrapper aboveUnitContent\"><div class=\"_wk mbm\"><span class=\"userContent\">Old Fashion Day, a longtime tradition at Morrow Mountain State Park was held last week. Here are some great photos: <a href=\"http://www.facebook.com/l.php?u=http%3A%2F%2Fwww.flickr.com%2Fphotos%2Finfobucket%2Fsets%2F72157631930079339%2F&amp;h=pAQHUbf3M&amp;s=1\" target=\"_blank\" rel=\"nofollow nofollow\" onmouseover=\"LinkshimAsyncLink.swap(this, &quot;http:\\\\/\\\\/www.flickr.com\\\\/photos\\\\/infobucket\\\\/sets\\\\/72157631930079339\\\\/&quot;);\" onclick=\"LinkshimAsyncLink.referrer_log(this, &quot;http:\\\\/\\\\/www.flickr.com\\\\/photos\\\\/infobucket\\\\/sets\\\\/72157631930079339\\\\/&quot;, &quot;http:\\\\/\\\\/www.facebook.com\\\\/si\\\\/ajax\\\\/l\\\\/render_linkshim_log\\\\/?u=http\\\\u00253A\\\\u00252F\\\\u00252Fwww.flickr.com\\\\u00252Fphotos\\\\u00252Finfobucket\\\\u00252Fsets\\\\u00252F72157631930079339\\\\u00252F&amp;h=pAQHUbf3M&amp;s=1&quot;);\">http://www.flickr.com/photos/infobucket/sets/72157631930079339/</a></span></div></div><div class=\"shareUnit\"><a class=\"_1xw\" href=\"http://www.facebook.com/l.php?u=http%3A%2F%2Fwww.flickr.com%2Fphotos%2Finfobucket%2Fsets%2F72157631930079339%2F&amp;h=zAQGk0cyhAQGBvu9bC_i8XSsK28h9M0DrrVqlYXkuN9bkTw&amp;enc=AZNaHlyWvVU7BKtte5qJAljh4HEXBKDbiAVtkGwxP1rAOis4SoV6brCIt8KurslVBh_Kb1PC5SlHzazlZWoaB2BV&amp;s=1\" target=\"_blank\" rel=\"nofollow\" onmouseover=\"LinkshimAsyncLink.swap(this, &quot;http:\\\\/\\\\/www.flickr.com\\\\/photos\\\\/infobucket\\\\/sets\\\\/72157631930079339\\\\/&quot;);\" onclick=\"LinkshimAsyncLink.referrer_log(this, &quot;http:\\\\/\\\\/www.flickr.com\\\\/photos\\\\/infobucket\\\\/sets\\\\/72157631930079339\\\\/&quot;, &quot;http:\\\\/\\\\/www.facebook.com\\\\/si\\\\/ajax\\\\/l\\\\/render_linkshim_log\\\\/?u=http\\\\u00253A\\\\u00252F\\\\u00252Fwww.flickr.com\\\\u00252Fphotos\\\\u00252Finfobucket\\\\u00252Fsets\\\\u00252F72157631930079339\\\\u00252F&amp;h=zAQGk0cyhAQGBvu9bC_i8XSsK28h9M0DrrVqlYXkuN9bkTw&amp;enc=AZNaHlyWvVU7BKtte5qJAljh4HEXBKDbiAVtkGwxP1rAOis4SoV6brCIt8KurslVBh_Kb1PC5SlHzazlZWoaB2BV&amp;s=1&quot;);\"><div class=\"_1xx _1xz\"><div class=\"_1x-\"><div class=\"_1x_ fwb\">Old Fashion Day - 2012</div><div class=\"fsm fwn fcg\">www.flickr.com</div><div class=\"_1y1 fsm fwn fcg\">Old Fashion Day at Morrow Mountain State Park - November 4th, 2012</div></div></div></a></div></div><div class=\"fbTimelineUFI uiCommentContainer\"><form rel=\"async\" class=\"commentable_item autoexpand_mode\" method=\"post\" action=\"/ajax/ufi/modify.php\" onsubmit=\"return window.Event &amp;&amp; Event.__inlineSubmit &amp;&amp; Event.__inlineSubmit(this,event)\" id=\"u4kzrn2149\"><input type=\"hidden\" name=\"charset_test\" value=\"&euro;,&acute;,\u20AC,\u00B4,\u6C34,\u0414,\u0404\" /><input type=\"hidden\" name=\"lsd\" value=\"AVqaEiN6\" autocomplete=\"off\" /><input type=\"hidden\" autocomplete=\"off\" name=\"feedback_params\" value=\"&#123;&quot;actor&quot;:&quot;143412869029&quot;,&quot;target_fbid&quot;:&quot;217369611728583&quot;,&quot;target_profile_id&quot;:&quot;143412869029&quot;,&quot;type_id&quot;:&quot;17&quot;,&quot;assoc_obj_id&quot;:&quot;&quot;,&quot;source_app_id&quot;:&quot;0&quot;,&quot;extra_story_params&quot;:[],&quot;content_timestamp&quot;:&quot;1352312832&quot;,&quot;check_hash&quot;:&quot;AQAyjlMvpQQFmXBn&quot;,&quot;source&quot;:&quot;0&quot;&#125;\" /><input type=\"hidden\" autocomplete=\"off\" name=\"data_only_response\" value=\"1\" /><input type=\"hidden\" autocomplete=\"off\" name=\"timeline_ufi\" value=\"1\" />",
        "<input type=\"hidden\" name=\"timeline_log_data\" value=\"AQA31vqWBdLldeBpimgkNT0T6VctFQIu96PatdJI_iIJJTT8nzzQHbvMv6S4Bttg-xlhAu4V400PjnSf4zbHOQLxkChYuHzY4dlIszLUPCghARyGfHOtyL2AlTaRyyzG509joO14jIpt7fQo6G-fckSXmUsdsR1elLKlnDJl4YQSe3CTwsIjry5ZbClcY28Ad6qxRECpL7BShBPII7pG5DS3C6mtNUpBhBUouLtdzkEdLdzXCZo9BarvKOSjDqwFfeffzAiJy-SjfErBg6KZEknMq1YO_vGmapX5hitGpLzEYpraSQ5YX7YR0K9TQy7fkJvNO2e3DZQ5fihF-xg5mCtBtVeu1RJxIuXlYSdVLhR2jwID4F6fLF3sfUgQfNSz9ZI\" /><div class=\"fbTimelineFeedbackHeader\"><div class=\"fbTimelineFeedbackActions clearfix\"><span class=\"UIActionLinks UIActionLinks_bottom\" data-ft=\"&#123;&quot;tn&quot;:&quot;=&quot;,&quot;type&quot;:20&#125;\"><a class=\"like_link\" title=\"Like this item\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2FNorthCarolinaStateParks%2Fposts%2F217369611728583\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Like</a> &#183; <a class=\"comment_link\" title=\"Leave a comment\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2FNorthCarolinaStateParks%2Fposts%2F217369611728583\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Comment</a> &#183; <a rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2FNorthCarolinaStateParks%2Fposts%2F217369611728583\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Share</a></span></div></div><div><div class=\"uiUfi UFIContainer utn\" id=\"u4kzrn288\"></div></div></form></div></div></div><i class=\"spinePointer\"></i><div class=\"bottomBorder\"></div></li><li class=\"fbTimelineUnit fbTimelineTwoColumn clearfix\" data-side=\"l\" data-fixed=\"1\" data-size=\"1\" id=\"tl_unit_-6942550131812495592\"><div class=\"topBorder\"></div><div class=\"timelineUnitContainer\" id=\"u4kzrn246\" data-gt=\"&#123;&quot;action_type&quot;:&quot;1&quot;,&quot;aggregation_id&quot;:&quot;10151247636289030&quot;,&quot;aggregation_time&quot;:&quot;1352305539&quot;,&quot;actor&quot;:&quot;143412869029&quot;,&quot;creator&quot;:&quot;143412869029&quot;,&quot;pagestimeline&quot;:&quot;1&quot;,&quot;payload_type&quot;:&quot;7&quot;,&quot;eventtime&quot;:&quot;1352998853&quot;,&quot;viewerid&quot;:&quot;0&quot;,&quot;profileownerid&quot;:&quot;143412869029&quot;,&quot;unitimpressionid&quot;:&quot;f2d1d9d2&quot;,&quot;contentid&quot;:&quot;-6942550131812495592&quot;,&quot;timeline_unit_type&quot;:&quot;AddSinglePhotoUnit&quot;,&quot;timewindowsize&quot;:&quot;2&quot;,&quot;contextwindowstart&quot;:&quot;1351753200&quot;,&quot;contextwindowend&quot;:&quot;1354348799&quot;&#125;\" data-time=\"1352305539\"><div class=\"\"><div role=\"article\"><div class=\"clearfix mbs pbs _1_m \"><a class=\"_29h _303\" href=\"http://www.facebook.com/NorthCarolinaStateParks\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-ash3/174880_143412869029_133040642_q.jpg\" alt=\"\" /></a><div class=\"_3dp _29k\"><h5 class=\"_1_s\"><span class=\"fcg\"><span class=\"fwb\"><a href=\"http://www.facebook.com/NorthCarolinaStateParks\">North Carolina State Parks and Recreation</a></span></span></h5><div class=\"_1_n fsm fwn fcg\"><a class=\"uiLinkSubtle\" href=\"http://www.facebook.com/photo.php?fbid=10151247636289030&amp;set=a.164247464029.118730.143412869029&amp;type=1\"><abbr title=\"Wednesday, November 7, 2012 at 8:25am\" data-utime=\"1352305539\">November 7</abbr></a><a data-hover=\"tooltip\" aria-label=\"Public\" class=\"uiStreamPrivacy inlineBlock fbStreamPrivacy fbPrivacyAudienceIndicator _1_o\" href=\"#\"><i class=\"lock img sp_2voup5 sx_4929d1\"></i></a></div></div></div><div class=\"aboveUnitContent\"><div class=\"userContentWrapper\"><div class=\"_wk\"><span class=\"userContent\">Another great festival this weekend...this time in the Triad. Bring the kids and enjoy a day outside at Haw River State Park, one of our newest.</span></div></div></div><div class=\"photoUnit clearfix\"><div class=\"_53s uiScaledThumb photo photoWidth1\" data-cropped=\"1\"><a class=\"_6i9\" href=\"http://www.facebook.com/photo.php?fbid=10151247636289030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1\" rel=\"theater\" ajaxify=\"http://www.facebook.com/photo.php?fbid=10151247636289030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1&amp;src=https%3A%2F%2Fsphotos-a.xx.fbcdn.net%2Fhphotos-ash3%2F530716_10151247636289030_32005621_n.jpg&amp;size=630%2C810&amp;theater\"><div class=\"uiScaledImageContainer photoWrap\"><img class=\"scaledImageFitWidth img\" src=\"https://sphotos-a.xx.fbcdn.net/hphotos-ash3/c0.0.403.403/p403x403/530716_10151247636289030_32005621_n.jpg\" alt=\"Photo: Another great festival this weekend...this time in the Triad. Bring the kids and enjoy a day outside at Haw River State Park, one of our newest.\" width=\"403\" height=\"403\" /></div></a></div></div></div><div class=\"fbTimelineUFI uiCommentContainer\"><form rel=\"async\" class=\"commentable_item autoexpand_mode\" method=\"post\" action=\"/ajax/ufi/modify.php\" onsubmit=\"return window.Event &amp;&amp; Event.__inlineSubmit &amp;&amp; Event.__inlineSubmit(this,event)\" id=\"u4kzrn2150\"><input type=\"hidden\" name=\"charset_test\" value=\"&euro;,&acute;,\u20AC,\u00B4,\u6C34,\u0414,\u0404\" /><input type=\"hidden\" name=\"lsd\" value=\"AVqaEiN6\" autocomplete=\"off\" /><input type=\"hidden\" autocomplete=\"off\" name=\"feedback_params\" value=\"&#123;&quot;actor&quot;:&quot;143412869029&quot;,&quot;target_fbid&quot;:&quot;10151247636289030&quot;,&quot;target_profile_id&quot;:&quot;143412869029&quot;,&quot;type_id&quot;:&quot;7&quot;,&quot;assoc_obj_id&quot;:&quot;&quot;,&quot;source_app_id&quot;:&quot;0&quot;,&quot;extra_story_params&quot;:[],&quot;content_timestamp&quot;:&quot;1352305539&quot;,&quot;check_hash&quot;:&quot;AQC0dZdrazzMKjLG&quot;,&quot;source&quot;:&quot;0&quot;&#125;\" /><input type=\"hidden\" autocomplete=\"off\" name=\"data_only_response\" value=\"1\" /><input type=\"hidden\" autocomplete=\"off\" name=\"timeline_ufi\" value=\"1\" /><input type=\"hidden\" name=\"timeline_log_data\" value=\"AQCfrExipd8rvTCrVyqmjOtth9m8fmaQaujwrHbADpJLIuZdvGIabV2eazYPlSZY1e_JyuYdfjCrCPQm8BZZZ64CddxvkDr9bKnrCY5xK2BQOamt_XfgCPBcsju8lQmVgasplj3dfDxd1qJPG-swc7UtqNdaapVK_9Yo5wCKuOGl2f_kU1ZuorOtsxcWW23RuI0RZXIRFdWivAVcClNSRFfv6l8dsWMM-BF3FcCdw4Yg_fIUpHs7eoF2WBq_KP-WXJRenM-1N9FmNSk0ICoMDymzyFBwVa6WiZo9sOBQ88c6mmW_yrVLhsHz5RxCjeLAgcuseXi39rYKAB5IcH26sW9izXS1i0rNknZT6y7zKsFRZYLpPMIsl9-\\-\\4YWrSB0O8Gk\" /><div class=\"fbTimelineFeedbackHeader\"><div class=\"fbTimelineFeedbackActions clearfix\"><span class=\"fbTimelineFeedbackShares\"><a data-hover=\"tooltip\" aria-label=\"Show shares\" data-tooltip-alignh=\"right\" href=\"/shares/view?id=10151247636289030\" ajaxify=\"/ajax/shares/view?target_fbid=10151247636289030\" rel=\"async\"><i></i>4</a></span><span class=\"UIActionLinks UIActionLinks_bottom\" data-ft=\"&#123;&quot;tn&quot;:&quot;=&quot;,&quot;type&quot;:20&#125;\"><a class=\"like_link\" title=\"Like this item\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151247636289030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Like</a> &#183; <a class=\"comment_link\" title=\"Leave a comment\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151247636289030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Comment</a> &#183; <a rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151247636289030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Share</a></span></div></div><div><div class=\"uiUfi UFIContainer utn\" id=\"u4kzrn289\"></div></div></form></div></div></div><i class=\"spinePointer\"></i><div class=\"bottomBorder\"></div></li><li class=\"fbTimelineUnit fbTimelineTwoColumn clearfix\" data-side=\"l\" data-fixed=\"1\" data-size=\"1\" id=\"tl_unit_522684511461328539\"><div class=\"topBorder\"></div><div class=\"timelineUnitContainer\" id=\"u4kzrn247\" data-gt=\"&#123;&quot;action_type&quot;:&quot;1&quot;,&quot;aggregation_id&quot;:&quot;10151247406854030&quot;,&quot;aggregation_time&quot;:&quot;1352294493&quot;,&quot;actor&quot;:&quot;143412869029&quot;,&quot;creator&quot;:&quot;143412869029&quot;,&quot;pagestimeline&quot;:&quot;1&quot;,&quot;payload_type&quot;:&quot;7&quot;,&quot;eventtime&quot;:&quot;1352998853&quot;,&quot;viewerid&quot;:&quot;0&quot;,&quot;profileownerid&quot;:&quot;143412869029&quot;,&quot;unitimpressionid&quot;:&quot;f2d1d9d2&quot;,&quot;contentid&quot;:&quot;522684511461328539&quot;,&quot;timeline_unit_type&quot;:&quot;AddSinglePhotoUnit&quot;,&quot;timewindowsize&quot;:&quot;2&quot;,&quot;contextwindowstart&quot;:&quot;1351753200&quot;,&quot;contextwindowend&quot;:&quot;1354348799&quot;&#125;\" data-time=\"1352294493\"><div class=\"\"><div role=\"article\"><div class=\"clearfix mbs pbs _1_m \"><a class=\"_29h _303\" href=\"http://www.facebook.com/NorthCarolinaStateParks\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-ash3/174880_143412869029_133040642_q.jpg\" alt=\"\" /></a><div class=\"_3dp _29k\"><h5 class=\"_1_s\"><span class=\"fcg\"><span class=\"fwb\"><a href=\"http://www.facebook.com/NorthCarolinaStateParks\">North Carolina State Parks and Recreation</a></span></span></h5><div class=\"_1_n fsm fwn fcg\"><a class=\"uiLinkSubtle\" href=\"http://www.facebook.com/photo.php?fbid=10151247406854030&amp;set=a.164247464029.118730.143412869029&amp;type=1\"><abbr title=\"Wednesday, November 7, 2012 at 5:21am\" data-utime=\"1352294493\">November 7</abbr></a><a data-hover=\"tooltip\" aria-label=\"Public\" class=\"uiStreamPrivacy inlineBlock fbStreamPrivacy fbPrivacyAudienceIndicator _1_o\" href=\"#\"><i class=\"lock img sp_2voup5 sx_4929d1\"></i></a></div></div></div><div class=\"aboveUnitContent\"><div class=\"userContentWrapper\"><div class=\"_wk\"><span class=\"userContent\">Our blog post today is all about chasing traffic along the &#039;nectar superhighway&#039;: <a href=\"http://www.facebook.com/l.php?u=http%3A%2F%2Fncstateparks.wordpress.com%2F&amp;h=aAQENx0jQ&amp;s=1\" target=\"_blank\" rel=\"nofollow nofollow\" onmouseover=\"LinkshimAsyncLink.swap(this, &quot;http:\\\\/\\\\/ncstateparks.wordpress.com\\\\/&quot;);\" onclick=\"LinkshimAsyncLink.referrer_log(this, &quot;http:\\\\/\\\\/ncstateparks.wordpress.com\\\\/&quot;, &quot;http:\\\\/\\\\/www.facebook.com\\\\/si\\\\/ajax\\\\/l\\\\/render_linkshim_log\\\\/?u=http\\\\u00253A\\\\u00252F\\\\u00252Fncstateparks.wordpress.com\\\\u00252F&amp;h=aAQENx0jQ&amp;s=1&quot;);\">http://ncstateparks.wordpress.com/</a></span></div></div></div><div class=\"photoUnit clearfix\"><div class=\"_53s uiScaledThumb photo photoWidth1\" data-cropped=\"1\"><a class=\"_6i9\" href=\"http://www.facebook.com/photo.php?fbid=10151247406854030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1\" rel=\"theater\" ajaxify=\"http://www.facebook.com/photo.php?fbid=10151247406854030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1&amp;src=https%3A%2F%2Fsphotos-b.xx.fbcdn.net%2Fhphotos-ash3%2F30363_10151247406854030_895069397_n.jpg&amp;size=800%2C674&amp;theater\"><div class=\"uiScaledImageContainer photoWrap\"><img class=\"scaledImageFitWidth img\" src=\"https://sphotos-b.xx.fbcdn.net/hphotos-ash3/c37.0.403.403/p403x403/30363_10151247406854030_895069397_n.jpg\" alt=\"Photo: Our blog post today is all about chasing traffic along the &#039;nectar superhighway&#039;: http://ncstateparks.wordpress.com\" width=\"403\" height=\"403\" /></div></a></div></div></div><div class=\"fbTimelineUFI uiCommentContainer\"><form rel=\"async\" class=\"commentable_item autoexpand_mode\" method=\"post\" action=\"/ajax/ufi/modify.php\" onsubmit=\"return window.Event &amp;&amp; Event.__inlineSubmit &amp;&amp; Event.__inlineSubmit(this,event)\" id=\"u4kzrn2151\"><input type=\"hidden\" name=\"charset_test\" value=\"&euro;,&acute;,\u20AC,\u00B4,\u6C34,\u0414,\u0404\" /><input type=\"hidden\" name=\"lsd\" value=\"AVqaEiN6\" autocomplete=\"off\" /><input type=\"hidden\" autocomplete=\"off\" name=\"feedback_params\" value=\"&#123;&quot;actor&quot;:&quot;143412869029&quot;,&quot;target_fbid&quot;:&quot;10151247406854030&quot;,&quot;target_profile_id&quot;:&quot;143412869029&quot;,&quot;type_id&quot;:&quot;7&quot;,&quot;assoc_obj_id&quot;:&quot;&quot;,&quot;source_app_id&quot;:&quot;0&quot;,&quot;extra_story_params&quot;:[],&quot;content_timestamp&quot;:&quot;1352294493&quot;,&quot;check_hash&quot;:&quot;AQAiXWKzvEcT6mCl&quot;,&quot;source&quot;:&quot;0&quot;&#125;\" /><input type=\"hidden\" autocomplete=\"off\" name=\"data_only_response\" value=\"1\" /><input type=\"hidden\" autocomplete=\"off\" name=\"timeline_ufi\" value=\"1\" /><input type=\"hidden\" name=\"timeline_log_data\" value=\"AQCrd_m6uyiOcyebCsidrjjs6icVJ4rA5eAmGibXWYIWLq8e7RAGlF-uqFwPZNVhXUndlHOUSFUdlh6qQscPBqSRdN47Pd3sdGJfev1fKYteXKHqBEcsIIZPiMrqye3JX-vxx1EiT_48J67aVG3qBDadArG98EMFBu7OAVYUnx1w9hEmnDsgjCUzScl8L0AWOSAnUV15eIg3C7-tbm_6jFr3wQMh-TWYecpypV3296tRszLCs7Sz3JDHCErcXtq_y9VHyT7xEtu1lhWJyI5tBz-4Ii_1Y5skZGNg7C1yi_dcnRYkRb3dwbSMYbt5i_PASltO-SMCUaw_h7U4SPCHS2lLFRI68ePKwNalIRR1ZOW84-_NE-774hk8ifm7ZK1GyVg\" /><div class=\"fbTimelineFeedbackHeader\"><div class=\"fbTimelineFeedbackActions clearfix\"><span class=\"fbTimelineFeedbackShares\"><a data-hover=\"tooltip\" aria-label=\"Show shares\" data-tooltip-alignh=\"right\" href=\"/shares/view?id=10151247406854030\" ajaxify=\"/ajax/shares/view?target_fbid=10151247406854030\" rel=\"async\"><i></i>24</a></span><span class=\"UIActionLinks UIActionLinks_bottom\" data-ft=\"&#123;&quot;tn&quot;:&quot;=&quot;,&quot;type&quot;:20&#125;\"><a class=\"like_link\" title=\"Like this item\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151247406854030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Like</a> &#183; <a class=\"comment_link\" title=\"Leave a comment\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151247406854030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Comment</a> &#183; <a rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151247406854030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Share</a></span></div></div><div><div class=\"uiUfi UFIContainer utn\" id=\"u4kzrn2126\"></div></div></form></div></div></div><i class=\"spinePointer\"></i><div class=\"bottomBorder\"></div></li><li class=\"fbTimelineUnit fbTimelineTwoColumn clearfix\" data-side=\"l\" data-fixed=\"1\" data-size=\"1\" id=\"tl_unit_-1651662322930838557\"><div class=\"topBorder\"></div><div class=\"timelineUnitContainer\" id=\"u4kzrn248\" data-gt=\"&#123;&quot;action_type&quot;:&quot;1&quot;,&quot;aggregation_id&quot;:&quot;10151246283684030&quot;,&quot;aggregation_time&quot;:&quot;1352229052&quot;,&quot;actor&quot;:&quot;143412869029&quot;,&quot;creator&quot;:&quot;143412869029&quot;,&quot;pagestimeline&quot;:&quot;1&quot;,&quot;payload_type&quot;:&quot;7&quot;,&quot;eventtime&quot;:&quot;1352998853&quot;,&quot;viewerid&quot;:&quot;0&quot;,&quot;profileownerid&quot;:&quot;143412869029&quot;,&quot;unitimpressionid&quot;:&quot;f2d1d9d2&quot;,&quot;contentid&quot;:&quot;-1651662322930838557&quot;,&quot;timeline_unit_type&quot;:&quot;AddSinglePhotoUnit&quot;,&quot;timewindowsize&quot;:&quot;2&quot;,&quot;contextwindowstart&quot;:&quot;1351753200&quot;,&quot;contextwindowend&quot;:&quot;1354348799&quot;&#125;\" data-time=\"1352229052\"><div class=\"\"><div role=\"article\"><div class=\"clearfix mbs pbs _1_m \"><a class=\"_29h _303\" href=\"http://www.facebook.com/NorthCarolinaStateParks\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-ash3/174880_143412869029_133040642_q.jpg\" alt=\"\" /></a><div class=\"_3dp _29k\"><h5 class=\"_1_s\"><span class=\"fcg\"><span class=\"fwb\"><a href=\"http://www.facebook.com/NorthCarolinaStateParks\">North Carolina State Parks and Recreation</a></span></span></h5><div class=\"_1_n fsm fwn fcg\"><a class=\"uiLinkSubtle\" href=\"http://www.facebook.com/photo.php?fbid=10151246283684030&amp;set=a.164247464029.118730.143412869029&amp;type=1\"><abbr title=\"Tuesday, November 6, 2012 at 11:10am\" data-utime=\"1352229052\">November 6</abbr></a><a data-hover=\"tooltip\" aria-label=\"Public\" class=\"uiStreamPrivacy inlineBlock fbStreamPrivacy fbPrivacyAudienceIndicator _1_o\" href=\"#\"><i class=\"lock img sp_2voup5 sx_4929d1\"></i></a></div></div></div><div class=\"aboveUnitContent\"><div class=\"userContentWrapper\"><div class=\"_wk\"><span class=\"userContent\">Fur, Fangs and Feathers Fall Festival...try saying that 3 times fast. Crowders Mountain State Park&#039;s popular event is this Saturday.</span></div></div></div><div class=\"photoUnit clearfix\"><div class=\"_53s uiScaledThumb photo photoWidth1\" data-cropped=\"1\"><a class=\"_6i9\" href=\"http://www.facebook.com/photo.php?fbid=10151246283684030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1\" rel=\"theater\" ajaxify=\"http://www.facebook.com/photo.php?fbid=10151246283684030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1&amp;src=https%3A%2F%2Fsphotos-b.xx.fbcdn.net%2Fhphotos-prn1%2F12707_10151246283684030_1747962753_n.jpg&amp;size=600%2C785&amp;theater\"><div class=\"uiScaledImageContainer photoWrap\"><img class=\"scaledImageFitWidth img\" src=\"https://sphotos-b.xx.fbcdn.net/hphotos-prn1/c0.0.403.403/p403x403/12707_10151246283684030_1747962753_n.jpg\" alt=\"Photo: Fur, Fangs and Feathers Fall Festival...try saying that 3 times fast. Crowders Mountain State Park&#039;s popular event is this Saturday.\" width=\"403\" height=\"403\" /></div></a></div></div></div><div class=\"fbTimelineUFI uiCommentContainer\"><form rel=\"async\" class=\"commentable_item autoexpand_mode\" method=\"post\" action=\"/ajax/ufi/modify.php\" onsubmit=\"return window.Event &amp;&amp; Event.__inlineSubmit &amp;&amp; Event.__inlineSubmit(this,event)\" id=\"u4kzrn2152\"><input type=\"hidden\" name=\"charset_test\" value=\"&euro;,&acute;,\u20AC,\u00B4,\u6C34,\u0414,\u0404\" /><input type=\"hidden\" name=\"lsd\" value=\"AVqaEiN6\" autocomplete=\"off\" /><input type=\"hidden\" autocomplete=\"off\" name=\"feedback_params\" value=\"&#123;&quot;actor&quot;:&quot;143412869029&quot;,&quot;target_fbid&quot;:&quot;10151246283684030&quot;,&quot;target_profile_id&quot;:&quot;143412869029&quot;,&quot;type_id&quot;:&quot;7&quot;,&quot;assoc_obj_id&quot;:&quot;&quot;,&quot;source_app_id&quot;:&quot;0&quot;,&quot;extra_story_params&quot;:[],&quot;content_timestamp&quot;:&quot;1352229051&quot;,&quot;check_hash&quot;:&quot;AQDeF_tgEiCE6u9l&quot;,&quot;source&quot;:&quot;0&quot;&#125;\" /><input type=\"hidden\" autocomplete=\"off\" name=\"data_only_response\" value=\"1\" /><input type=\"hidden\" autocomplete=\"off\" name=\"timeline_ufi\" value=\"1\" /><input type=\"hidden\" name=\"timeline_log_data\" value=\"AQAF6nkB0mvVtiUcj1cWAXYQeRnr1EaeNDM1IRBQwSSidCcmAw5tBty4IdFDWDCIVC3mDrcvVEVdhAAl80HHAh8gdKN6cyx6KJRejQklsVVvo9PoBpra12auKI4ujRKwEycFNfE-MI9sL7Jp9In_9kam1sG4Wyr8_hkmaDy9jPipYRjE5YP-kMdvLvqNVTnMnxQjUCwGp8phNgGvsVLEL7Fn5MU6nvFI0HplV_DUrzDOyjR_AU-jv7PmkuOF4H-E7Jyba9ZUDQnIwLmHyDKQjThVyuDIp49B1WGWXKXp4XVKpsTzakR0XM7F7VpZyh6H7EQ0GPHcage09aIqAiAgTcX98ZEPgbjKQqOYgCZGSTdxnORTPcPVfJwKZtYBBIMZaa8\" /><div class=\"fbTimelineFeedbackHeader\"><div class=\"fbTimelineFeedbackActions clearfix\"><span class=\"fbTimelineFeedbackShares\"><a data-hover=\"tooltip\" aria-label=\"Show shares\" data-tooltip-alignh=\"right\" href=\"/shares/view?id=10151246283684030\" ajaxify=\"/ajax/shares/view?target_fbid=10151246283684030\" rel=\"async\"><i></i>1</a></span><span class=\"UIActionLinks UIActionLinks_bottom\" data-ft=\"&#123;&quot;tn&quot;:&quot;=&quot;,&quot;type&quot;:20&#125;\"><a class=\"like_link\" title=\"Like this item\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151246283684030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Like</a> &#183; <a class=\"comment_link\" title=\"Leave a comment\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151246283684030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Comment</a> &#183; <a rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151246283684030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Share</a></span></div></div><div><div class=\"uiUfi UFIContainer utn\" id=\"u4kzrn290\"></div></div></form></div></div></div><i class=\"spinePointer\"></i><div class=\"bottomBorder\"></div></li><li class=\"fbTimelineUnit fbTimelineTwoColumn clearfix\" data-side=\"l\" data-fixed=\"1\" data-size=\"1\" id=\"tl_unit_-7650629536826941552\"><div class=\"topBorder\"></div><div class=\"timelineUnitContainer\" id=\"u4kzrn249\" data-gt=\"&#123;&quot;action_type&quot;:&quot;1&quot;,&quot;aggregation_id&quot;:&quot;10151245839549030&quot;,&quot;aggregation_time&quot;:&quot;1352209557&quot;,&quot;actor&quot;:&quot;143412869029&quot;,&quot;creator&quot;:&quot;143412869029&quot;,&quot;pagestimeline&quot;:&quot;1&quot;,&quot;payload_type&quot;:&quot;7&quot;,&quot;eventtime&quot;:&quot;1352998853&quot;,&quot;viewerid&quot;:&quot;0&quot;,&quot;profileownerid&quot;:&quot;143412869029&quot;,&quot;unitimpressionid&quot;:&quot;f2d1d9d2&quot;,&quot;contentid&quot;:&quot;-7650629536826941552&quot;,&quot;timeline_unit_type&quot;:&quot;AddSinglePhotoUnit&quot;,&quot;timewindowsize&quot;:&quot;2&quot;,&quot;contextwindowstart&quot;:&quot;1351753200&quot;,&quot;contextwindowend&quot;:&quot;1354348799&quot;&#125;\" data-time=\"1352209557\"><div class=\"\"><div role=\"article\"><div class=\"clearfix mbs pbs _1_m \"><a class=\"_29h _303\" href=\"http://www.facebook.com/NorthCarolinaStateParks\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-ash3/174880_143412869029_133040642_q.jpg\" alt=\"\" /></a><div class=\"_3dp _29k\"><h5 class=\"_1_s\"><span class=\"fcg\"><span class=\"fwb\"><a href=\"http://www.facebook.com/NorthCarolinaStateParks\">North Carolina State Parks and Recreation</a></span></span></h5><div class=\"_1_n fsm fwn fcg\"><a class=\"uiLinkSubtle\" href=\"http://www.facebook.com/photo.php?fbid=10151245839549030&amp;set=a.164247464029.118730.143412869029&amp;type=1\"><abbr title=\"Tuesday, November 6, 2012 at 5:45am\" data-utime=\"1352209557\">November 6</abbr></a><a data-hover=\"tooltip\" aria-label=\"Public\" class=\"uiStreamPrivacy inlineBlock fbStreamPrivacy fbPrivacyAudienceIndicator _1_o\" href=\"#\"><i class=\"lock img sp_2voup5 sx_4929d1\"></i></a></div></div></div><div class=\"aboveUnitContent\"><div class=\"userContentWrapper\"><div class=\"_wk\"><span class=\"userContent\">Please vote today...and afterwards, if you&#039;ve got some time left, take a hike...and you can share our idea for a sticker.</span></div></div></div><div class=\"photoUnit clearfix\"><div class=\"_53s uiScaledThumb photo photoWidth1\" data-cropped=\"1\"><a class=\"_6i9\" href=\"http://www.facebook.com/photo.php?fbid=10151245839549030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1\" rel=\"theater\" ajaxify=\"http://www.facebook.com/photo.php?fbid=10151245839549030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1&amp;src=https%3A%2F%2Fsphotos-b.xx.fbcdn.net%2Fhphotos-ash3%2F530596_10151245839549030_960669189_n.jpg&amp;size=800%2C813&amp;theater\"><div class=\"uiScaledImageContainer photoWrap\"><img class=\"scaledImageFitWidth img\" src=\"https://sphotos-b.xx.fbcdn.net/hphotos-ash3/c0.0.403.403/p403x403/530596_10151245839549030_960669189_n.jpg\" alt=\"Photo: Please vote today...and afterwards, if you&#039;ve got some time left, take a hike...and you can share our idea for a sticker.\" width=\"403\" height=\"403\" /></div></a></div></div></div><div class=\"fbTimelineUFI uiCommentContainer\"><form rel=\"async\" class=\"commentable_item autoexpand_mode\" method=\"post\" action=\"/ajax/ufi/modify.php\" onsubmit=\"return window.Event &amp;&amp; Event.__inlineSubmit &amp;&amp; Event.__inlineSubmit(this,event)\" id=\"u4kzrn2153\"><input type=\"hidden\" name=\"charset_test\" value=\"&euro;,&acute;,\u20AC,\u00B4,\u6C34,\u0414,\u0404\" /><input type=\"hidden\" name=\"lsd\" value=\"AVqaEiN6\" autocomplete=\"off\" /><input type=\"hidden\" autocomplete=\"off\" name=\"feedback_params\" value=\"&#123;&quot;actor&quot;:&quot;143412869029&quot;,&quot;target_fbid&quot;:&quot;10151245839549030&quot;,&quot;target_profile_id&quot;:&quot;143412869029&quot;,&quot;type_id&quot;:&quot;7&quot;,&quot;assoc_obj_id&quot;:&quot;&quot;,&quot;source_app_id&quot;:&quot;0&quot;,&quot;extra_story_params&quot;:[],&quot;content_timestamp&quot;:&quot;1352209557&quot;,&quot;check_hash&quot;:&quot;AQAG9oynyt7wPR50&quot;,&quot;source&quot;:&quot;0&quot;&#125;\" /><input type=\"hidden\" autocomplete=\"off\" name=\"data_only_response\" value=\"1\" /><input type=\"hidden\" autocomplete=\"off\" name=\"timeline_ufi\" value=\"1\" /><input type=\"hidden\" name=\"timeline_log_data\" value=\"AQCLzqkRuHFMuIY518yuhXOwUckcZaVKnlZI-Yv_fk98gIX6o2gnFYK7IIthbqsbccmh2S0jRD63Vy28toZUVv2OjZ6XcNAFVowD14vUWstA9lw5TJTo9z4fXqQv0JH0P6CP_zbau14XY2uNCc14Z7U34QgCW3lnMISI3QUbRpgbpMAwHk6w0m5FFfL9LHKx7j7zeIIoUz_PEnre-M9XTytCOlnUvY0QaAbmfcUM4V5MQ9-aWBTWBtPE06HfxY-ioUOF8L5rBdHErJvKqIkX4fCcj82CF9OOQ6bY02TGuVNjnRsdJz-8hnaoDU-7AKui-ME6bw-_A7ls0RgspIAha7ZVtm_Hxy91SXCZ_uD0bdLtfFBKU2EPR0XqrltqzSbjSRs\" /><div class=\"fbTimelineFeedbackHeader\"><div class=\"fbTimelineFeedbackActions clearfix\"><span class=\"fbTimelineFeedbackShares\"><a data-hover=\"tooltip\" aria-label=\"Show shares\" data-tooltip-alignh=\"right\" href=\"/shares/view?id=10151245839549030\" ajaxify=\"/ajax/shares/view?target_fbid=10151245839549030\" rel=\"async\"><i></i>19</a></span><span class=\"UIActionLinks UIActionLinks_bottom\" data-ft=\"&#123;&quot;tn&quot;:&quot;=&quot;,&quot;type&quot;:20&#125;\"><a class=\"like_link\" title=\"Like this item\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151245839549030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Like</a> &#183; <a class=\"comment_link\" title=\"Leave a comment\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151245839549030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Comment</a> &#183; <a rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151245839549030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Share</a></span></div></div><div><div class=\"uiUfi UFIContainer utn\" id=\"u4kzrn2127\"></div></div></form></div></div></div><i class=\"spinePointer\"></i><div class=\"bottomBorder\"></div></li><li class=\"fbTimelineUnit fbTimelineTwoColumn clearfix\" data-side=\"l\" data-fixed=\"1\" data-size=\"1\" id=\"tl_unit_2149763130472413544\"><div class=\"topBorder\"></div><div class=\"timelineUnitContainer\" id=\"u4kzrn251\" data-gt=\"&#123;&quot;action_type&quot;:&quot;1&quot;,&quot;aggregation_id&quot;:&quot;130429140441330&quot;,&quot;aggregation_time&quot;:&quot;1352128614&quot;,&quot;actor&quot;:&quot;143412869029&quot;,&quot;creator&quot;:&quot;143412869029&quot;,&quot;pagestimeline&quot;:&quot;1&quot;,&quot;payload_type&quot;:&quot;5&quot;,&quot;eventtime&quot;:&quot;1352998853&quot;,&quot;viewerid&quot;:&quot;0&quot;,&quot;profileownerid&quot;:&quot;143412869029&quot;,&quot;unitimpressionid&quot;:&quot;f2d1d9d2&quot;,&quot;contentid&quot;:&quot;2149763130472413544&quot;,&quot;timeline_unit_type&quot;:&quot;BestPostedLinkUnit&quot;,&quot;timewindowsize&quot;:&quot;2&quot;,&quot;contextwindowstart&quot;:&quot;1351753200&quot;,&quot;contextwindowend&quot;:&quot;1354348799&quot;&#125;\" data-time=\"1352128614\"><div class=\"\"><div role=\"article\"><div class=\"clearfix mbs pbs _1_m \"><a class=\"_29h _303\" href=\"http://www.facebook.com/NorthCarolinaStateParks\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-ash3/174880_143412869029_133040642_q.jpg\" alt=\"\" /></a><div class=\"_3dp _29k\"><h5 class=\"_1_s\"><span class=\"fcg\"><span class=\"fwb\"><a href=\"http://www.facebook.com/NorthCarolinaStateParks\">North Carolina State Parks and Recreation</a></span> shared a <a href=\"http://www.facebook.com/l.php?u=http%3A%2F%2Fwww.techhive.com%2Farticle%2F2013463%2Fhow-to-ensure-facebook-posts-you-want-to-see-arent-hidden.html&amp;h=2AQFEdpc5AQEC0Ew5ru4QL2dHDVk7JAke7BI03g7prJpKqQ&amp;enc=AZOm3PLhWjjRt8_pZjnRRMgDskppuM29fLafAFncQT3LfShCBnwtFG1sf39pBrt9ByGPzVggyGdfEif3kDpN6cNa&amp;s=1\" rel=\"nofollow\" onmouseover=\"LinkshimAsyncLink.swap(this, &quot;http:\\\\/\\\\/www.techhive.com\\\\/article\\\\/2013463\\\\/how-to-ensure-facebook-posts-you-want-to-see-arent-hidden.html&quot;);\" onclick=\"LinkshimAsyncLink.referrer_log(this, &quot;http:\\\\/\\\\/www.techhive.com\\\\/article\\\\/2013463\\\\/how-to-ensure-facebook-posts-you-want-to-see-arent-hidden.html&quot;, &quot;http:\\\\/\\\\/www.facebook.com\\\\/si\\\\/ajax\\\\/l\\\\/render_linkshim_log\\\\/?u=http\\\\u00253A\\\\u00252F\\\\u00252Fwww.techhive.com\\\\u00252Farticle\\\\u00252F2013463\\\\u00252Fhow-to-ensure-facebook-posts-you-want-to-see-arent-hidden.html&amp;h=2AQFEdpc5AQEC0Ew5ru4QL2dHDVk7JAke7BI03g7prJpKqQ&amp;enc=AZOm3PLhWjjRt8_pZjnRRMgDskppuM29fLafAFncQT3LfShCBnwtFG1sf39pBrt9ByGPzVggyGdfEif3kDpN6cNa&amp;s=1&quot;);\" target=\"_blank\">link</a>.</span></h5><div class=\"_1_n fsm fwn fcg\"><a class=\"uiLinkSubtle\" href=\"/NorthCarolinaStateParks/posts/130429140441330\"><abbr title=\"Monday, November 5, 2012 at 7:16am\" data-utime=\"1352128614\">November 5</abbr></a><a data-hover=\"tooltip\" aria-label=\"Public\" class=\"uiStreamPrivacy inlineBlock fbStreamPrivacy fbPrivacyAudienceIndicator _1_o\" href=\"#\"><i class=\"lock img sp_2voup5 sx_4929d1\"></i></a></div></div></div><div class=\"userContentWrapper aboveUnitContent\"><div class=\"_wk mbm\"><span class=\"userContent\">Argggh. New Facebook rules may keep you from seeing all our news. Go to our FB page; hover over the &quot;Liked&quot; button below the cover photo; click &quot;Get Notifications&quot;:  <a href=\"http://www.facebook.com/l.php?u=http%3A%2F%2Fwww.techhive.com%2Farticle%2F2013463%2Fhow-to-ensure-facebook-posts-you-want-to-see-arent-hidden.html&amp;h=nAQESQtKi&amp;s=1\" target=\"_blank\" rel=\"nofollow nofollow\" onmouseover=\"LinkshimAsyncLink.swap(this, &quot;http:\\\\/\\\\/www.techhive.com\\\\/article\\\\/2013463\\\\/how-to-ensure-facebook-posts-you-want-to-see-arent-hidden.html&quot;);\" onclick=\"LinkshimAsyncLink.referrer_log(this, &quot;http:\\\\/\\\\/www.techhive.com\\\\/article\\\\/2013463\\\\/how-to-ensure-facebook-posts-you-want-to-see-arent-hidden.html&quot;, &quot;http:\\\\/\\\\/www.facebook.com\\\\/si\\\\/ajax\\\\/l\\\\/render_linkshim_log\\\\/?u=http\\\\u00253A\\\\u00252F\\\\u00252Fwww.techhive.com\\\\u00252Farticle\\\\u00252F2013463\\\\u00252Fhow-to-ensure-facebook-posts-you-want-to-see-arent-hidden.html&amp;h=nAQESQtKi&amp;s=1&quot;);\">http://www.techhive.com/article/2013463/how-to-ensure-facebook-posts-you-want-to-see-arent-hidden.html</a></span></div></div><div class=\"shareUnit\"><a class=\"_1xw _1y0\" href=\"http://www.facebook.com/l.php?u=http%3A%2F%2Fwww.techhive.com%2Farticle%2F2013463%2Fhow-to-ensure-facebook-posts-you-want-to-see-arent-hidden.html&amp;h=-AQG4vz5aAQHRJYb5ZfZyrYp2hjBRbPZ72YrDi7ZVHXWzEg&amp;enc=AZNxIqGQvgAwyiDJ27uOmFtuZUXbOt7DWz9rW9gaPerZJRVB2plyJWKtWA0AV0znEcK9izZUMGVCNZrUOybZ-zwK&amp;s=1\" target=\"_blank\" rel=\"nofollow\" onmouseover=\"LinkshimAsyncLink.swap(this, &quot;http:\\\\/\\\\/www.techhive.com\\\\/article\\\\/2013463\\\\/how-to-ensure-facebook-posts-you-want-to-see-arent-hidden.html&quot;);\" onclick=\"LinkshimAsyncLink.referrer_log(this, &quot;http:\\\\/\\\\/www.techhive.com\\\\/article\\\\/2013463\\\\/how-to-ensure-facebook-posts-you-want-to-see-arent-hidden.html&quot;, &quot;http:\\\\/\\\\/www.facebook.com\\\\/si\\\\/ajax\\\\/l\\\\/render_linkshim_log\\\\/?u=http\\\\u00253A\\\\u00252F\\\\u00252Fwww.techhive.com\\\\u00252Farticle\\\\u00252F2013463\\\\u00252Fhow-to-ensure-facebook-posts-you-want-to-see-arent-hidden.html&amp;h=-AQG4vz5aAQHRJYb5ZfZyrYp2hjBRbPZ72YrDi7ZVHXWzEg&amp;enc=AZNxIqGQvgAwyiDJ27uOmFtuZUXbOt7DWz9rW9gaPerZJRVB2plyJWKtWA0AV0znEcK9izZUMGVCNZrUOybZ-zwK&amp;s=1&quot;);\"><div class=\"_1xy _1xx\"><img class=\"img\" src=\"https://fbexternal-a.akamaihd.net/safe_image.php?d=AQBDU_k81DdZG1Md&amp;w=155&amp;h=114&amp;url=http%3A%2F%2Fcomputerworld.com.edgesuite.net%2Fcio%2Fog_image_logo%2FCIO_041612_002.gif\" alt=\"\" /></div><div class=\"_1xx _1xz\"><div class=\"_1x-\"><div class=\"_1x_ fwb\">How to Ensure Facebook Posts You Want to See Aren&#039;t Hidden</div><div class=\"fsm fwn fcg\">www.techhive.com</div><div class=\"_1y1 fsm fwn fcg\"></div></div></div></a></div></div><div class=\"fbTimelineUFI uiCommentContainer\"><form rel=\"async\" class=\"commentable_item autoexpand_mode\" method=\"post\" action=\"/ajax/ufi/modify.php\" onsubmit=\"return window.Event &amp;&amp; Event.__inlineSubmit &amp;&amp; Event.__inlineSubmit(this,event)\" id=\"u4kzrn2154\"><input type=\"hidden\" name=\"charset_test\" value=\"&euro;,&acute;,\u20AC,\u00B4,\u6C34,\u0414,\u0404\" /><input type=\"hidden\" name=\"lsd\" value=\"AVqaEiN6\" autocomplete=\"off\" /><input type=\"hidden\" autocomplete=\"off\" name=\"feedback_params\" value=\"&#123;&quot;actor&quot;:&quot;143412869029&quot;,&quot;target_fbid&quot;:&quot;130429140441330&quot;,&quot;target_profile_id&quot;:&quot;143412869029&quot;,&quot;type_id&quot;:&quot;17&quot;,&quot;assoc_obj_id&quot;:&quot;&quot;,&quot;source_app_id&quot;:&quot;0&quot;,&quot;extra_story_params&quot;:[],&quot;content_timestamp&quot;:&quot;1352128614&quot;,&quot;check_hash&quot;:&quot;AQAyoiqDvc3EI0Ju&quot;,&quot;source&quot;:&quot;0&quot;&#125;\" /><input type=\"hidden\" autocomplete=\"off\" name=\"data_only_response\" value=\"1\" /><input type=\"hidden\" autocomplete=\"off\" name=\"timeline_ufi\" value=\"1\" /><input type=\"hidden\" name=\"timeline_log_data\" value=\"AQD-Dpa30TK5BsBDqf3QPdXzpD6RMF0cyQEXQ_LIfkSsmTTlalqULRmJFWs_0dV0Jtb00T5q7Zp9kblolGbI515BjiaO5SKGJM5GoG7b2zTbPJ7YdcJv-\\-\\GrZ4bO0CrXR1pbFsURpF9QOgD-MwjUV5eeSYSsME690GHNBuHOaBukJK1X-_6tNsTfFoOj3JiCY63HcQIQ1KPZmH3zDPMpAQmatDS6fqdjjVLJwODGnVT53QTXmeuKvrPnAomabToVBYNckbdDwUYobowq8ROQYZooXIYdcG34qXLJXjEp-MxAG_Q-4pSakg1P4CosPlrnot4VC-a_lJqWqXlAZw-9PLPdZN2vHscRCzHUt27B4EIR5RwTsoolMCqcX8lPshxz6ns\" /><div class=\"fbTimelineFeedbackHeader\"><div class=\"fbTimelineFeedbackActions clearfix\"><span class=\"UIActionLinks UIActionLinks_bottom\" data-ft=\"&#123;&quot;tn&quot;:&quot;=&quot;,&quot;type&quot;:20&#125;\"><a class=\"like_link\" title=\"Like this item\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2FNorthCarolinaStateParks%2Fposts%2F130429140441330\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Like</a> &#183; <a class=\"comment_link\" title=\"Leave a comment\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2FNorthCarolinaStateParks%2Fposts%2F130429140441330\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Comment</a> &#183; <a rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2FNorthCarolinaStateParks%2Fposts%2F130429140441330\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Share</a></span></div></div><div><div class=\"uiUfi UFIContainer utn\" id=\"u4kzrn2128\"></div></div></form></div></div></div><i class=\"spinePointer\"></i><div class=\"bottomBorder\"></div></li><li class=\"fbTimelineUnit fbTimelineTwoColumn clearfix\" data-side=\"l\" data-fixed=\"1\" data-size=\"1\" id=\"tl_unit_1491461674822618306\"><div class=\"topBorder\"></div><div class=\"timelineUnitContainer\" id=\"u4kzrn252\" data-gt=\"&#123;&quot;action_type&quot;:&quot;1&quot;,&quot;aggregation_id&quot;:&quot;10151244259254030&quot;,&quot;aggregation_time&quot;:&quot;1352121462&quot;,&quot;actor&quot;:&quot;143412869029&quot;,&quot;creator&quot;:&quot;143412869029&quot;,&quot;pagestimeline&quot;:&quot;1&quot;,&quot;payload_type&quot;:&quot;7&quot;,&quot;eventtime&quot;:&quot;1352998853&quot;,&quot;viewerid&quot;:&quot;0&quot;,&quot;profileownerid&quot;:&quot;143412869029&quot;,&quot;unitimpressionid&quot;:&quot;f2d1d9d2&quot;,&quot;contentid&quot;:&quot;1491461674822618306&quot;,&quot;timeline_unit_type&quot;:&quot;AddSinglePhotoUnit&quot;,&quot;timewindowsize&quot;:&quot;2&quot;,&quot;contextwindowstart&quot;:&quot;1351753200&quot;,&quot;contextwindowend&quot;:&quot;1354348799&quot;&#125;\" data-time=\"1352121462\"><div class=\"\"><div role=\"article\"><div class=\"clearfix mbs pbs _1_m \"><a class=\"_29h _303\" href=\"http://www.facebook.com/NorthCarolinaStateParks\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-ash3/174880_143412869029_133040642_q.jpg\" alt=\"\" /></a><div class=\"_3dp _29k\"><h5 class=\"_1_s\"><span class=\"fcg\"><span class=\"fwb\"><a href=\"http://www.facebook.com/NorthCarolinaStateParks\">North Carolina State Parks and Recreation</a></span></span></h5><div class=\"_1_n fsm fwn fcg\"><a class=\"uiLinkSubtle\" href=\"http://www.facebook.com/photo.php?fbid=10151244259254030&amp;set=a.164247464029.118730.143412869029&amp;type=1\"><abbr title=\"Monday, November 5, 2012 at 5:17am\" data-utime=\"1352121462\">November 5</abbr></a><a data-hover=\"tooltip\" aria-label=\"Public\" class=\"uiStreamPrivacy inlineBlock fbStreamPrivacy fbPrivacyAudienceIndicator _1_o\" href=\"#\"><i class=\"lock img sp_2voup5 sx_4929d1\"></i></a></div></div></div><div class=\"aboveUnitContent\"><div class=\"userContentWrapper\"><div class=\"_wk\"><span class=\"userContent\">The longleaf pine forest was the topic when Superintendent Scott Hartley met with school kids last week during the Richmond County 4-H Science Adventure week.</span></div></div></div><div class=\"photoUnit clearfix\"><div class=\"_53s uiScaledThumb photo photoWidth1\" data-cropped=\"1\"><a class=\"_6i9\" href=\"http://www.facebook.com/photo.php?fbid=10151244259254030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1\" rel=\"theater\" ajaxify=\"http://www.facebook.com/photo.php?fbid=10151244259254030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1&amp;src=https%3A%2F%2Fsphotos-b.xx.fbcdn.net%2Fhphotos-ash3%2F550177_10151244259254030_1449134267_n.jpg&amp;size=800%2C695&amp;theater\"><div class=\"uiScaledImageContainer photoWrap\"><img class=\"scaledImageFitWidth img\" src=\"https://sphotos-b.xx.fbcdn.net/hphotos-ash3/c29.0.403.403/p403x403/550177_10151244259254030_1449134267_n.jpg\" alt=\"Photo: The longleaf pine forest was the topic when Superintendent Scott Hartley met with school kids last week during the Richmond County 4-H Science Adventure week.\" width=\"403\" height=\"403\" /></div></a></div></div></div><div class=\"fbTimelineUFI uiCommentContainer\"><form rel=\"async\" class=\"commentable_item autoexpand_mode\" method=\"post\" action=\"/ajax/ufi/modify.php\" onsubmit=\"return window.Event &amp;&amp; Event.__inlineSubmit &amp;&amp; Event.__inlineSubmit(this,event)\" id=\"u4kzrn2155\"><input type=\"hidden\" name=\"charset_test\" value=\"&euro;,&acute;,\u20AC,\u00B4,\u6C34,\u0414,\u0404\" /><input type=\"hidden\" name=\"lsd\" value=\"AVqaEiN6\" autocomplete=\"off\" /><input type=\"hidden\" autocomplete=\"off\" name=\"feedback_params\" value=\"&#123;&quot;actor&quot;:&quot;143412869029&quot;,&quot;target_fbid&quot;:&quot;10151244259254030&quot;,&quot;target_profile_id&quot;:&quot;143412869029&quot;,&quot;type_id&quot;:&quot;7&quot;,&quot;assoc_obj_id&quot;:&quot;&quot;,&quot;source_app_id&quot;:&quot;0&quot;,&quot;extra_story_params&quot;:[],&quot;content_timestamp&quot;:&quot;1352121462&quot;,&quot;check_hash&quot;:&quot;AQAyTVeWYZzqNRH0&quot;,&quot;source&quot;:&quot;0&quot;&#125;\" /><input type=\"hidden\" autocomplete=\"off\" name=\"data_only_response\" value=\"1\" /><input type=\"hidden\" autocomplete=\"off\" name=\"timeline_ufi\" value=\"1\" /><input type=\"hidden\" name=\"timeline_log_data\" value=\"AQAoVolFGpEfqWlwzbYXiaV8HOf_MAy3hUN_btt38_xz1Ik_gKHzXuQWuDkB9nNcFKQLJEmHUnKV-_00nujCxol1jChscWGCBpVvLlVTr0VI8e4JhMADTd-XSxv2vBKlwsqXEe1MGCoGYEhWAskcz2drecNjppAdwBQKiKFOEDp90CIXVTh6wyfdp9qefZFvdLSv2pV0RVJ1_w8lQtPkCi4RlAfpm6nPdBM67w027v2pL7n4RuqkGBTh08Au_b9F_mM3p4zxYqM33ftWn2E7Odhvp874qbVcnubfuNClEU0ynFNJLp2i-yFnJvlw8ocu-jL8xffqaw4unWbFwX7I01MwyCMBdNGkG3x5pXl1TwSGn4K1kHSB0OLJ1J8g800RE9U\" /><div class=\"fbTimelineFeedbackHeader\"><div class=\"fbTimelineFeedbackActions clearfix\"><span class=\"fbTimelineFeedbackShares\"><a data-hover=\"tooltip\" aria-label=\"Show shares\" data-tooltip-alignh=\"right\" href=\"/shares/view?id=10151244259254030\" ajaxify=\"/ajax/shares/view?target_fbid=10151244259254030\" rel=\"async\"><i></i>3</a></span><span class=\"UIActionLinks UIActionLinks_bottom\" data-ft=\"&#123;&quot;tn&quot;:&quot;=&quot;,&quot;type&quot;:20&#125;\"><a class=\"like_link\" title=\"Like this item\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151244259254030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Like</a> &#183; <a class=\"comment_link\" title=\"Leave a comment\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151244259254030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Comment</a> &#183; <a rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151244259254030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Share</a></span></div></div><div><div class=\"uiUfi UFIContainer utn\" id=\"u4kzrn2129\"></div></div></form></div></div></div><i class=\"spinePointer\"></i><div class=\"bottomBorder\"></div></li><li class=\"fbTimelineUnit fbTimelineTwoColumn clearfix\" data-side=\"l\" data-fixed=\"1\" data-size=\"1\" id=\"tl_unit_-4684120165449599717\"><div class=\"topBorder\"></div><div class=\"timelineUnitContainer\" id=\"u4kzrn253\" data-gt=\"&#123;&quot;action_type&quot;:&quot;1&quot;,&quot;aggregation_id&quot;:&quot;10151240100539030&quot;,&quot;aggregation_time&quot;:&quot;1351881754&quot;,&quot;actor&quot;:&quot;143412869029&quot;,&quot;creator&quot;:&quot;143412869029&quot;,&quot;pagestimeline&quot;:&quot;1&quot;,&quot;payload_type&quot;:&quot;7&quot;,&quot;eventtime&quot;:&quot;1352998853&quot;,&quot;viewerid&quot;:&quot;0&quot;,&quot;profileownerid&quot;:&quot;143412869029&quot;,&quot;unitimpressionid&quot;:&quot;f2d1d9d2&quot;,&quot;contentid&quot;:&quot;-4684120165449599717&quot;,&quot;timeline_unit_type&quot;:&quot;AddSinglePhotoUnit&quot;,&quot;timewindowsize&quot;:&quot;2&quot;,&quot;contextwindowstart&quot;:&quot;1351753200&quot;,&quot;contextwindowend&quot;:&quot;1354348799&quot;&#125;\" data-time=\"1351881754\"><div class=\"\"><div role=\"article\"><div class=\"clearfix mbs pbs _1_m \"><a class=\"_29h _303\" href=\"http://www.facebook.com/NorthCarolinaStateParks\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-ash3/174880_143412869029_133040642_q.jpg\" alt=\"\" /></a><div class=\"_3dp _29k\"><h5 class=\"_1_s\"><span class=\"fcg\"><span class=\"fwb\"><a href=\"http://www.facebook.com/NorthCarolinaStateParks\">North Carolina State Parks and Recreation</a></span></span></h5><div class=\"_1_n fsm fwn fcg\"><a class=\"uiLinkSubtle\" href=\"http://www.facebook.com/photo.php?fbid=10151240100539030&amp;set=a.164247464029.118730.143412869029&amp;type=1\"><abbr title=\"Friday, November 2, 2012 at 10:42am\" data-utime=\"1351878154\">November 2</abbr></a><a data-hover=\"tooltip\" aria-label=\"Public\" class=\"uiStreamPrivacy inlineBlock fbStreamPrivacy fbPrivacyAudienceIndicator _1_o\" href=\"#\"><i class=\"lock img sp_2voup5 sx_4929d1\"></i></a></div></div></div><div class=\"aboveUnitContent\"><div class=\"userContentWrapper\"><div class=\"_wk\"><span class=\"userContent\">Phriday Photo! Not all color in the fall is in the leaves. This is a pine barren gentian now in bloom at Weymouth Woods and elsewhere in the Sandhills. It thrives in areas that&#039;ve had prescribed burns.</span></div></div></div><div class=\"photoUnit clearfix\"><div class=\"_53s uiScaledThumb photo photoWidth1\" data-cropped=\"1\"><a class=\"_6i9\" href=\"http://www.facebook.com/photo.php?fbid=10151240100539030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1\" rel=\"theater\" ajaxify=\"http://www.facebook.com/photo.php?fbid=10151240100539030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1&amp;src=https%3A%2F%2Fsphotos-a.xx.fbcdn.net%2Fhphotos-snc6%2F665022_10151240100539030_1628734814_o.jpg&amp;smallsrc=https%3A%2F%2Fsphotos-a.xx.fbcdn.net%2Fhphotos-ash4%2F317145_10151240100539030_1628734814_n.jpg&amp;size=1200%2C1105&amp;theater\"><div class=\"uiScaledImageContainer photoWrap\"><img class=\"scaledImageFitWidth img\" src=\"https://sphotos-a.xx.fbcdn.net/hphotos-ash4/c17.0.403.403/p403x403/317145_10151240100539030_1628734814_n.jpg\" alt=\"Photo: Phriday Photo! Not all color in the fall is in the leaves. This is a pine barren gentian now in bloom at Weymouth Woods and elsewhere in the Sandhills. It thrives in areas that&#039;ve had prescribed burns.\" width=\"403\" height=\"403\" /></div></a></div></div></div><div class=\"fbTimelineUFI uiCommentContainer\"><form rel=\"async\" class=\"commentable_item autoexpand_mode\" method=\"post\" action=\"/ajax/ufi/modify.php\" onsubmit=\"return window.Event &amp;&amp; Event.__inlineSubmit &amp;&amp; Event.__inlineSubmit(this,event)\" id=\"u4kzrn2156\"><input type=\"hidden\" name=\"charset_test\" value=\"&euro;,&acute;,\u20AC,\u00B4,\u6C34,\u0414,\u0404\" /><input type=\"hidden\" name=\"lsd\" value=\"AVqaEiN6\" autocomplete=\"off\" /><input type=\"hidden\" autocomplete=\"off\" name=\"feedback_params\" value=\"&#123;&quot;actor&quot;:&quot;143412869029&quot;,&quot;target_fbid&quot;:&quot;10151240100539030&quot;,&quot;target_profile_id&quot;:&quot;143412869029&quot;,&quot;type_id&quot;:&quot;7&quot;,&quot;assoc_obj_id&quot;:&quot;&quot;,&quot;source_app_id&quot;:&quot;0&quot;,&quot;extra_story_params&quot;:[],&quot;content_timestamp&quot;:&quot;1351881754&quot;,&quot;check_hash&quot;:&quot;AQBjOl_N40DEHwPB&quot;,&quot;source&quot;:&quot;0&quot;&#125;\" /><input type=\"hidden\" autocomplete=\"off\" name=\"data_only_response\" value=\"1\" /><input type=\"hidden\" autocomplete=\"off\" name=\"timeline_ufi\" value=\"1\" /><input type=\"hidden\" name=\"timeline_log_data\" value=\"AQDThfqlyhON2N8_Gs3WD2OG0WIkl8SHjrupBknwoaqIwyonVKGriKb9oG6FCUOopWXxEyb5jCKhsb5p5doGXthIr0y_ZJ2HxC7q71RLsNNBI6X5KEjNZGIlBR9r7oapQc7Kh1_XS6O3mbsAsc9F3ja8b4B3kM19LifPLlu_1i2GzYVM55nixEhXHzB3PfqiBCznsJlZ_uFXItZo2HOF_Cls53hfRIUOxQ6V-fO55-vOsPYnwClOw3VDaElnW3Pn7Fn0ICBnY3Kav0FBVPKdt9QFOCJYADrh8JTkT-r07kgWpJzy7kFBN-FuAb-ZKqlfXfpp-UT5_-nsD19_E6YpHqF5Lvs9qxypGf5mdp4ThiVKCcHFsR7BgfP8Ed7YGQLME-I\" /><div class=\"fbTimelineFeedbackHeader\"><div class=\"fbTimelineFeedbackActions clearfix\"><span class=\"fbTimelineFeedbackShares\"><a data-hover=\"tooltip\" aria-label=\"Show shares\" data-tooltip-alignh=\"right\" href=\"/shares/view?id=10151240100539030\" ajaxify=\"/ajax/shares/view?target_fbid=10151240100539030\" rel=\"async\"><i></i>14</a></span><span class=\"UIActionLinks UIActionLinks_bottom\" data-ft=\"&#123;&quot;tn&quot;:&quot;=&quot;,&quot;type&quot;:20&#125;\"><a class=\"like_link\" title=\"Like this item\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151240100539030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Like</a> &#183; <a class=\"comment_link\" title=\"Leave a comment\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151240100539030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Comment</a> &#183; <a rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151240100539030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Share</a></span></div></div><div><div class=\"uiUfi UFIContainer utn\" id=\"u4kzrn2131\"></div></div></form></div></div></div><i class=\"spinePointer\"></i><div class=\"bottomBorder\"></div></li><li class=\"fbTimelineUnit fbTimelineTwoColumn clearfix\" data-side=\"l\" data-fixed=\"1\" data-size=\"1\" id=\"tl_unit_626691893126906489\"><div class=\"topBorder\"></div><div class=\"timelineUnitContainer\" id=\"u4kzrn255\" data-gt=\"&#123;&quot;action_type&quot;:&quot;1&quot;,&quot;aggregation_id&quot;:&quot;461599323892770&quot;,&quot;aggregation_time&quot;:&quot;1351860263&quot;,&quot;actor&quot;:&quot;143412869029&quot;,&quot;creator&quot;:&quot;143412869029&quot;,&quot;pagestimeline&quot;:&quot;1&quot;,&quot;payload_type&quot;:&quot;5&quot;,&quot;eventtime&quot;:&quot;1352998853&quot;,&quot;viewerid&quot;:&quot;0&quot;,&quot;profileownerid&quot;:&quot;143412869029&quot;,&quot;unitimpressionid&quot;:&quot;f2d1d9d2&quot;,&quot;contentid&quot;:&quot;626691893126906489&quot;,&quot;timeline_unit_type&quot;:&quot;BestPostedLinkUnit&quot;,&quot;timewindowsize&quot;:&quot;2&quot;,&quot;contextwindowstart&quot;:&quot;1351753200&quot;,&quot;contextwindowend&quot;:&quot;1354348799&quot;&#125;\" data-time=\"1351860263\"><div class=\"\"><div role=\"article\"><div class=\"clearfix mbs pbs _1_m \"><a class=\"_29h _303\" href=\"http://www.facebook.com/NorthCarolinaStateParks\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-ash3/174880_143412869029_133040642_q.jpg\" alt=\"\" /></a><div class=\"_3dp _29k\"><h5 class=\"_1_s\"><span class=\"fcg\"><span class=\"fwb\"><a href=\"http://www.facebook.com/NorthCarolinaStateParks\">North Carolina State Parks and Recreation</a></span> shared a <a href=\"http://www.facebook.com/l.php?u=http%3A%2F%2Fwww.ncparks.gov%2FNews%2Freleases%2Frelease.php%3Fid%3D270&amp;h=bAQEN3r6qAQGdsl9RmpnBtNEMbJmU069mM6J_RMdwuw46RA&amp;enc=AZPp-K2t29NgXIUCHq2ORz33dhddBgMuVykURbc-zUgb95SBU_eHLROs8fye0BuOnQMHINhFwS5BnC9RgsybJzDe&amp;s=1\" rel=\"nofollow\" onmouseover=\"LinkshimAsyncLink.swap(this, &quot;http:\\\\/\\\\/www.ncparks.gov\\\\/News\\\\/releases\\\\/release.php?id=270&quot;);\" onclick=\"LinkshimAsyncLink.referrer_log(this, &quot;http:\\\\/\\\\/www.ncparks.gov\\\\/News\\\\/releases\\\\/release.php?id=270&quot;, &quot;http:\\\\/\\\\/www.facebook.com\\\\/si\\\\/ajax\\\\/l\\\\/render_linkshim_log\\\\/?u=http\\\\u00253A\\\\u00252F\\\\u00252Fwww.ncparks.gov\\\\u00252FNews\\\\u00252Freleases\\\\u00252Frelease.php\\\\u00253Fid\\\\u00253D270&amp;h=bAQEN3r6qAQGdsl9RmpnBtNEMbJmU069mM6J_RMdwuw46RA&amp;enc=AZPp-K2t29NgXIUCHq2ORz33dhddBgMuVykURbc-zUgb95SBU_eHLROs8fye0BuOnQMHINhFwS5BnC9RgsybJzDe&amp;s=1&quot;);\" target=\"_blank\">link</a>.</span></h5><div class=\"_1_n fsm fwn fcg\"><a class=\"uiLinkSubtle\" href=\"/NorthCarolinaStateParks/posts/461599323892770\"><abbr title=\"Friday, November 2, 2012 at 4:44am\" data-utime=\"1351856663\">November 2</abbr></a><a data-hover=\"tooltip\" aria-label=\"Public\" class=\"uiStreamPrivacy inlineBlock fbStreamPrivacy fbPrivacyAudienceIndicator _1_o\" href=\"#\"><i class=\"lock img sp_2voup5 sx_4929d1\"></i></a></div></div></div><div class=\"userContentWrapper aboveUnitContent\"><div class=\"_wk mbm\"><span class=\"userContent\">If you love state parks, you&#039;re in good company. We&#039;ve counted 11.9 million visitors so far in 2012. Could be another record year for visitation:<a href=\"http://www.facebook.com/l.php?u=http%3A%2F%2Fwww.ncparks.gov%2FNews%2Freleases%2Frelease.php%3Fid%3D270&amp;h=0AQEvstfn&amp;s=1\" target=\"_blank\" rel=\"nofollow nofollow\" onmouseover=\"LinkshimAsyncLink.swap(this, &quot;http:\\\\/\\\\/www.ncparks.gov\\\\/News\\\\/releases\\\\/release.php?id=270&quot;);\" onclick=\"LinkshimAsyncLink.referrer_log(this, &quot;http:\\\\/\\\\/www.ncparks.gov\\\\/News\\\\/releases\\\\/release.php?id=270&quot;, &quot;http:\\\\/\\\\/www.facebook.com\\\\/si\\\\/ajax\\\\/l\\\\/render_linkshim_log\\\\/?u=http\\\\u00253A\\\\u00252F\\\\u00252Fwww.ncparks.gov\\\\u00252FNews\\\\u00252Freleases\\\\u00252Frelease.php\\\\u00253Fid\\\\u00253D270&amp;h=0AQEvstfn&amp;s=1&quot;);\">http://www.ncparks.gov/News/releases/release.php?id=270</a></span></div></div><div class=\"shareUnit\"><a class=\"_1xw _1y0\" href=\"http://www.facebook.com/l.php?u=http%3A%2F%2Fwww.ncparks.gov%2FNews%2Freleases%2Frelease.php%3Fid%3D270&amp;h=JAQEUzQYiAQHhE6LxfzJRaSOSMYBTr7mF-AIQimolKiZLXQ&amp;enc=AZN4Uo3eKUINuYIxxQVG3hUgGvR_HK_xIfcPDkXBvCk_XufvFNJMWpeNB3krwv4GgTdshzjFbEEt_-uXYYWmM4dC&amp;s=1\" target=\"_blank\" rel=\"nofollow\" onmouseover=\"LinkshimAsyncLink.swap(this, &quot;http:\\\\/\\\\/www.ncparks.gov\\\\/News\\\\/releases\\\\/release.php?id=270&quot;);\" onclick=\"LinkshimAsyncLink.referrer_log(this, &quot;http:\\\\/\\\\/www.ncparks.gov\\\\/News\\\\/releases\\\\/release.php?id=270&quot;, &quot;http:\\\\/\\\\/www.facebook.com\\\\/si\\\\/ajax\\\\/l\\\\/render_linkshim_log\\\\/?u=http\\\\u00253A\\\\u00252F\\\\u00252Fwww.ncparks.gov\\\\u00252FNews\\\\u00252Freleases\\\\u00252Frelease.php\\\\u00253Fid\\\\u00253D270&amp;h=JAQEUzQYiAQHhE6LxfzJRaSOSMYBTr7mF-AIQimolKiZLXQ&amp;enc=AZN4Uo3eKUINuYIxxQVG3hUgGvR_HK_xIfcPDkXBvCk_XufvFNJMWpeNB3krwv4GgTdshzjFbEEt_-uXYYWmM4dC&amp;s=1&quot;);\"><div class=\"_1xy _1xx\"><img class=\"img\" src=\"https://fbexternal-a.akamaihd.net/safe_image.php?d=AQAEwIcqjA3slAC0&amp;w=155&amp;h=114&amp;url=http%3A%2F%2Fwww.ncparks.gov%2FNews%2Freleases%2F..%2F..%2FNews%2Fpics%2Flicense_pl_unsat.jpg\" alt=\"\" /></div><div class=\"_1xx _1xz\"><div class=\"_1x-\"><div class=\"_1x_ fwb\">Official Website for the NC Division of Parks and Recreation</div><div class=\"fsm fwn fcg\">www.ncparks.gov</div><div class=\"_1y1 fsm fwn fcg\"></div></div></div></a></div></div><div class=\"fbTimelineUFI uiCommentContainer\"><form rel=\"async\" class=\"commentable_item autoexpand_mode\" method=\"post\" action=\"/ajax/ufi/modify.php\" onsubmit=\"return window.Event &amp;&amp; Event.__inlineSubmit &amp;&amp; Event.__inlineSubmit(this,event)\" id=\"u4kzrn2157\"><input type=\"hidden\" name=\"charset_test\" value=\"&euro;,&acute;,\u20AC,\u00B4,\u6C34,\u0414,\u0404\" /><input type=\"hidden\" name=\"lsd\" value=\"AVqaEiN6\" autocomplete=\"off\" /><input type=\"hidden\" autocomplete=\"off\" name=\"feedback_params\" value=\"&#123;&quot;actor&quot;:&quot;143412869029&quot;,&quot;target_fbid&quot;:&quot;461599323892770&quot;,&quot;target_profile_id&quot;:&quot;143412869029&quot;,&quot;type_id&quot;:&quot;17&quot;,&quot;assoc_obj_id&quot;:&quot;&quot;,&quot;source_app_id&quot;:&quot;0&quot;,&quot;extra_story_params&quot;:[],&quot;content_timestamp&quot;:&quot;1351860263&quot;,&quot;check_hash&quot;:&quot;AQDtgKQfeDWGV1fP&quot;,&quot;source&quot;:&quot;0&quot;&#125;\" /><input type=\"hidden\" autocomplete=\"off\" name=\"data_only_response\" value=\"1\" /><input type=\"hidden\" autocomplete=\"off\" name=\"timeline_ufi\" value=\"1\" /><input type=\"hidden\" name=\"timeline_log_data\" value=\"AQAbeWdr_N1EAr2HiUT94k7a1JZGO0EQ0DLF_0Zl6GioIDlOHVbrBteEroOFjwSXCjzHTRxO423i601CAulAHtLgisH6zeihRqyqOho14hNx8s-DfO21Dj0g04J0ircqqICLX2DK9epCFljRNlk0WS0SGuB2STEYDfqT7nG0g6XCx7qZUsAZhT0qAO7V8pGk6VyAWwzJYqT07YDMS5-T6X0MIMdYfItxPzqp9FwqysEN8FcfnrtQ1M_Cjbt4-SHwhEV5bs0yy1doUqghGJa65uvEPrzqyKvWyCkKU8hhkZlhB2QYbK1hFtUYsNNxzdOKlrYgHLYa_q1NeZWCSgVrxgT-5HV5gcnz4vt8437rij-\\-\\taH689bnXaWXiAOMuc6Ngow\" /><div class=\"fbTimelineFeedbackHeader\"><div class=\"fbTimelineFeedbackActions clearfix\"><span class=\"fbTimelineFeedbackShares\"><a data-hover=\"tooltip\" aria-label=\"Show shares\" data-tooltip-alignh=\"right\" href=\"/shares/view?id=461599323892770\" ajaxify=\"/ajax/shares/view?target_fbid=461599323892770\" rel=\"async\"><i></i>3</a></span><span class=\"UIActionLinks UIActionLinks_bottom\" data-ft=\"&#123;&quot;tn&quot;:&quot;=&quot;,&quot;type&quot;:20&#125;\"><a class=\"like_link\" title=\"Like this item\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2FNorthCarolinaStateParks%2Fposts%2F461599323892770\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Like</a> &#183; <a class=\"comment_link\" title=\"Leave a comment\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2FNorthCarolinaStateParks%2Fposts%2F461599323892770\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Comment</a> &#183; <a rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2FNorthCarolinaStateParks%2Fposts%2F461599323892770\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Share</a></span></div></div><div><div class=\"uiUfi UFIContainer utn\" id=\"u4kzrn2132\"></div></div></form></div></div></div><i class=\"spinePointer\"></i><div class=\"bottomBorder\"></div></li><li class=\"fbTimelineUnit fbTimelineTwoColumn clearfix\" data-side=\"l\" data-fixed=\"1\" data-size=\"1\" id=\"tl_unit_2230550431683067761\"><div class=\"topBorder\"></div><div class=\"timelineUnitContainer\" id=\"u4kzrn256\" data-gt=\"&#123;&quot;action_type&quot;:&quot;1&quot;,&quot;aggregation_id&quot;:&quot;10151239662119030&quot;,&quot;aggregation_time&quot;:&quot;1351859123&quot;,&quot;actor&quot;:&quot;143412869029&quot;,&quot;creator&quot;:&quot;143412869029&quot;,&quot;pagestimeline&quot;:&quot;1&quot;,&quot;payload_type&quot;:&quot;7&quot;,&quot;eventtime&quot;:&quot;1352998853&quot;,&quot;viewerid&quot;:&quot;0&quot;,&quot;profileownerid&quot;:&quot;143412869029&quot;,&quot;unitimpressionid&quot;:&quot;f2d1d9d2&quot;,&quot;contentid&quot;:&quot;2230550431683067761&quot;,&quot;timeline_unit_type&quot;:&quot;AddSinglePhotoUnit&quot;,&quot;timewindowsize&quot;:&quot;2&quot;,&quot;contextwindowstart&quot;:&quot;1351753200&quot;,&quot;contextwindowend&quot;:&quot;1354348799&quot;&#125;\" data-time=\"1351859123\"><div class=\"\"><div role=\"article\"><div class=\"clearfix mbs pbs _1_m \"><a class=\"_29h _303\" href=\"http://www.facebook.com/NorthCarolinaStateParks\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-ash3/174880_143412869029_133040642_q.jpg\" alt=\"\" /></a><div class=\"_3dp _29k\"><h5 class=\"_1_s\"><span class=\"fcg\"><span class=\"fwb\"><a href=\"http://www.facebook.com/NorthCarolinaStateParks\">North Carolina State Parks and Recreation</a></span></span></h5>",
        "<div class=\"_1_n fsm fwn fcg\"><a class=\"uiLinkSubtle\" href=\"http://www.facebook.com/photo.php?fbid=10151239662119030&amp;set=a.164247464029.118730.143412869029&amp;type=1\"><abbr title=\"Friday, November 2, 2012 at 4:25am\" data-utime=\"1351855523\">November 2</abbr></a><a data-hover=\"tooltip\" aria-label=\"Public\" class=\"uiStreamPrivacy inlineBlock fbStreamPrivacy fbPrivacyAudienceIndicator _1_o\" href=\"#\"><i class=\"lock img sp_2voup5 sx_4929d1\"></i></a></div></div></div><div class=\"aboveUnitContent\"><div class=\"userContentWrapper\"><div class=\"_wk\"><span class=\"userContent\">Mount Mitchell after the storm...this was taken late yesterday when skies finally cleared. The park remains closed due to power problems and the Blue Ridge Parkway approaches being closed.</span></div></div></div><div class=\"photoUnit clearfix\"><div class=\"_53s uiScaledThumb photo photoWidth1\" data-cropped=\"1\"><a class=\"_6i9\" href=\"http://www.facebook.com/photo.php?fbid=10151239662119030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1\" rel=\"theater\" ajaxify=\"http://www.facebook.com/photo.php?fbid=10151239662119030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1&amp;src=https%3A%2F%2Fsphotos-b.xx.fbcdn.net%2Fhphotos-prn1%2F134665_10151239662119030_745996597_o.jpg&amp;smallsrc=https%3A%2F%2Fsphotos-b.xx.fbcdn.net%2Fhphotos-ash3%2F522771_10151239662119030_745996597_n.jpg&amp;size=1200%2C814&amp;theater\"><div class=\"uiScaledImageContainer photoWrap\"><img class=\"scaledImageFitWidth img\" src=\"https://sphotos-b.xx.fbcdn.net/hphotos-ash3/c95.0.403.403/p403x403/522771_10151239662119030_745996597_n.jpg\" alt=\"Photo: Mount Mitchell after the storm...this was taken late yesterday when skies finally cleared. The park remains closed due to power problems and the Blue Ridge Parkway approaches being closed.\" width=\"403\" height=\"403\" /></div></a></div></div></div><div class=\"fbTimelineUFI uiCommentContainer\"><form rel=\"async\" class=\"commentable_item autoexpand_mode\" method=\"post\" action=\"/ajax/ufi/modify.php\" onsubmit=\"return window.Event &amp;&amp; Event.__inlineSubmit &amp;&amp; Event.__inlineSubmit(this,event)\" id=\"u4kzrn2158\"><input type=\"hidden\" name=\"charset_test\" value=\"&euro;,&acute;,\u20AC,\u00B4,\u6C34,\u0414,\u0404\" /><input type=\"hidden\" name=\"lsd\" value=\"AVqaEiN6\" autocomplete=\"off\" /><input type=\"hidden\" autocomplete=\"off\" name=\"feedback_params\" value=\"&#123;&quot;actor&quot;:&quot;143412869029&quot;,&quot;target_fbid&quot;:&quot;10151239662119030&quot;,&quot;target_profile_id&quot;:&quot;143412869029&quot;,&quot;type_id&quot;:&quot;7&quot;,&quot;assoc_obj_id&quot;:&quot;&quot;,&quot;source_app_id&quot;:&quot;0&quot;,&quot;extra_story_params&quot;:[],&quot;content_timestamp&quot;:&quot;1351859123&quot;,&quot;check_hash&quot;:&quot;AQA_J_zshr0DSA6_&quot;,&quot;source&quot;:&quot;0&quot;&#125;\" /><input type=\"hidden\" autocomplete=\"off\" name=\"data_only_response\" value=\"1\" /><input type=\"hidden\" autocomplete=\"off\" name=\"timeline_ufi\" value=\"1\" /><input type=\"hidden\" name=\"timeline_log_data\" value=\"AQDggTONWQyiumOGxWo2dZ2txSjn9xfZAF3BYV4YKgevjA8kjwh0EY1QCVlyTnPn3RpPJqObQCY6CP_IoGSGtUtaJHrEuLDT4Kwl8bMfdpW92qeDqwHzX5VRCPNS_4K_SujORjPVKDMBKO71-PHI5HnfuSNlHSgH8qI4lkOJenKPJdsZu8OU6b2At_DJkF_uD73pbciBmr0uOhxjFV_YdTr2uFsDn1r6YBsd4j7GAXX_ucpEQ2nBQFwDuq0iIv7ZLaDQfVc3_es3jnftTMRk-SK1UobOa9Sfle4Aq51sZXqsX3OLzj4DZgYPYxWTU9WxBf4kM55O_gEMo3U2KUQtyA3dv9YKiemHQ2IwqHfi8FK21eWAcryGL-cHNHaK9687AXk\" /><div class=\"fbTimelineFeedbackHeader\"><div class=\"fbTimelineFeedbackActions clearfix\"><span class=\"fbTimelineFeedbackShares\"><a data-hover=\"tooltip\" aria-label=\"Show shares\" data-tooltip-alignh=\"right\" href=\"/shares/view?id=10151239662119030\" ajaxify=\"/ajax/shares/view?target_fbid=10151239662119030\" rel=\"async\"><i></i>133</a></span><span class=\"UIActionLinks UIActionLinks_bottom\" data-ft=\"&#123;&quot;tn&quot;:&quot;=&quot;,&quot;type&quot;:20&#125;\"><a class=\"like_link\" title=\"Like this item\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151239662119030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Like</a> &#183; <a class=\"comment_link\" title=\"Leave a comment\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151239662119030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Comment</a> &#183; <a rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151239662119030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Share</a></span></div></div><div><div class=\"uiUfi UFIContainer utn\" id=\"u4kzrn2133\"></div></div></form></div></div></div><i class=\"spinePointer\"></i><div class=\"bottomBorder\"></div></li><li class=\"fbTimelineUnit fbTimelineTwoColumn clearfix\" data-side=\"l\" data-fixed=\"1\" data-size=\"1\" id=\"tl_unit_-425691253677834257\"><div class=\"topBorder\"></div><div class=\"timelineUnitContainer\" id=\"u4kzrn257\" data-gt=\"&#123;&quot;action_type&quot;:&quot;1&quot;,&quot;aggregation_id&quot;:&quot;10151238196354030&quot;,&quot;aggregation_time&quot;:&quot;1351772447&quot;,&quot;actor&quot;:&quot;143412869029&quot;,&quot;creator&quot;:&quot;143412869029&quot;,&quot;pagestimeline&quot;:&quot;1&quot;,&quot;payload_type&quot;:&quot;7&quot;,&quot;eventtime&quot;:&quot;1352998853&quot;,&quot;viewerid&quot;:&quot;0&quot;,&quot;profileownerid&quot;:&quot;143412869029&quot;,&quot;unitimpressionid&quot;:&quot;f2d1d9d2&quot;,&quot;contentid&quot;:&quot;-425691253677834257&quot;,&quot;timeline_unit_type&quot;:&quot;AddSinglePhotoUnit&quot;,&quot;timewindowsize&quot;:&quot;2&quot;,&quot;contextwindowstart&quot;:&quot;1351753200&quot;,&quot;contextwindowend&quot;:&quot;1354348799&quot;&#125;\" data-time=\"1351772447\"><div class=\"\"><div role=\"article\"><div class=\"clearfix mbs pbs _1_m \"><a class=\"_29h _303\" href=\"http://www.facebook.com/NorthCarolinaStateParks\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-ash3/174880_143412869029_133040642_q.jpg\" alt=\"\" /></a><div class=\"_3dp _29k\"><h5 class=\"_1_s\"><span class=\"fcg\"><span class=\"fwb\"><a href=\"http://www.facebook.com/NorthCarolinaStateParks\">North Carolina State Parks and Recreation</a></span></span></h5><div class=\"_1_n fsm fwn fcg\"><a class=\"uiLinkSubtle\" href=\"http://www.facebook.com/photo.php?fbid=10151238196354030&amp;set=a.164247464029.118730.143412869029&amp;type=1\"><abbr title=\"Thursday, November 1, 2012 at 4:20am\" data-utime=\"1351768847\">November 1</abbr></a><a data-hover=\"tooltip\" aria-label=\"Public\" class=\"uiStreamPrivacy inlineBlock fbStreamPrivacy fbPrivacyAudienceIndicator _1_o\" href=\"#\"><i class=\"lock img sp_2voup5 sx_4929d1\"></i></a></div></div></div><div class=\"aboveUnitContent\"><div class=\"userContentWrapper\"><div class=\"_wk\"><span class=\"userContent\">Another new friends support group has been organized for Cliffs of the Neuse State Park. Congrats. Most of these organizations are now affiliated with the statewide Friends of State Parks. Nice to have such good friends.</span></div></div></div><div class=\"photoUnit clearfix\"><div class=\"_53s uiScaledThumb photo photoWidth1\" data-cropped=\"1\"><a class=\"_6i9\" href=\"http://www.facebook.com/photo.php?fbid=10151238196354030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1\" rel=\"theater\" ajaxify=\"http://www.facebook.com/photo.php?fbid=10151238196354030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1&amp;src=https%3A%2F%2Fsphotos-b.xx.fbcdn.net%2Fhphotos-prn1%2F135806_10151238196354030_29953425_o.jpg&amp;smallsrc=https%3A%2F%2Fsphotos-b.xx.fbcdn.net%2Fhphotos-prn1%2F29371_10151238196354030_29953425_n.jpg&amp;size=1200%2C803&amp;theater\"><div class=\"uiScaledImageContainer photoWrap\"><img class=\"scaledImageFitWidth img\" src=\"https://sphotos-b.xx.fbcdn.net/hphotos-prn1/c99.0.403.403/p403x403/29371_10151238196354030_29953425_n.jpg\" alt=\"Photo: Another new friends support group has been organized for Cliffs of the Neuse State Park. Congrats. Most of these organizations are now affiliated with the statewide Friends of State Parks. Nice to have such good friends.\" width=\"403\" height=\"403\" /></div></a></div></div></div><div class=\"fbTimelineUFI uiCommentContainer\"><form rel=\"async\" class=\"commentable_item autoexpand_mode\" method=\"post\" action=\"/ajax/ufi/modify.php\" onsubmit=\"return window.Event &amp;&amp; Event.__inlineSubmit &amp;&amp; Event.__inlineSubmit(this,event)\" id=\"u4kzrn2159\"><input type=\"hidden\" name=\"charset_test\" value=\"&euro;,&acute;,\u20AC,\u00B4,\u6C34,\u0414,\u0404\" /><input type=\"hidden\" name=\"lsd\" value=\"AVqaEiN6\" autocomplete=\"off\" /><input type=\"hidden\" autocomplete=\"off\" name=\"feedback_params\" value=\"&#123;&quot;actor&quot;:&quot;143412869029&quot;,&quot;target_fbid&quot;:&quot;10151238196354030&quot;,&quot;target_profile_id&quot;:&quot;143412869029&quot;,&quot;type_id&quot;:&quot;7&quot;,&quot;assoc_obj_id&quot;:&quot;&quot;,&quot;source_app_id&quot;:&quot;0&quot;,&quot;extra_story_params&quot;:[],&quot;content_timestamp&quot;:&quot;1351772447&quot;,&quot;check_hash&quot;:&quot;AQBtMr23XNBJTtk_&quot;,&quot;source&quot;:&quot;0&quot;&#125;\" /><input type=\"hidden\" autocomplete=\"off\" name=\"data_only_response\" value=\"1\" /><input type=\"hidden\" autocomplete=\"off\" name=\"timeline_ufi\" value=\"1\" /><input type=\"hidden\" name=\"timeline_log_data\" value=\"AQDuRVfg_8dXBhIsIw-zeGtZPbdf-zz8-bLQzsP6ye-ibq60RrHOgBae6EDtd7RE7wsQOBU1exBD7U2LnkP6nm9sUFaXey87FvZKHMtYV_nxoT0ODFDWec4scQ_mdTgle8znVNp4Q72dMjt40ONUx7SCCzSSINjJmM6l0dEiQVlmCIyvUcEiXgIuY_e_Yye6273stvLgW2rR7mg1sRhzXAsISrQAJFk3VKoAmHMYvUnRwFkBjrveoFzgtNQFqNSEMK5P0-s_4Z2o242zUrbER8as2JNAXeUxwF5-loVDolpA-RpEppBaLNf4sfG_VgIT5TPnDzN1_KED3SH4oce5beGH1-Lw4L6AlE_E8jgFGIsTzyx-OzWBKMzI0lFNxgBfb48\" /><div class=\"fbTimelineFeedbackHeader\"><div class=\"fbTimelineFeedbackActions clearfix\"><span class=\"fbTimelineFeedbackShares\"><a data-hover=\"tooltip\" aria-label=\"Show shares\" data-tooltip-alignh=\"right\" href=\"/shares/view?id=10151238196354030\" ajaxify=\"/ajax/shares/view?target_fbid=10151238196354030\" rel=\"async\"><i></i>3</a></span><span class=\"UIActionLinks UIActionLinks_bottom\" data-ft=\"&#123;&quot;tn&quot;:&quot;=&quot;,&quot;type&quot;:20&#125;\"><a class=\"like_link\" title=\"Like this item\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151238196354030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Like</a> &#183; <a class=\"comment_link\" title=\"Leave a comment\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151238196354030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Comment</a> &#183; <a rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151238196354030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Share</a></span></div></div><div><div class=\"uiUfi UFIContainer utn\" id=\"u4kzrn2141\"></div></div></form></div></div></div><i class=\"spinePointer\"></i><div class=\"bottomBorder\"></div></li><li class=\"fbTimelineUnit fbTimelineTwoColumn clearfix\" data-side=\"l\" data-fixed=\"1\" data-size=\"1\" id=\"tl_unit_815093699853453383\"><div class=\"topBorder\"></div><div class=\"timelineUnitContainer\" id=\"u4kzrn258\" data-gt=\"&#123;&quot;action_type&quot;:&quot;1&quot;,&quot;aggregation_id&quot;:&quot;10151237038709030&quot;,&quot;aggregation_time&quot;:&quot;1351710557&quot;,&quot;actor&quot;:&quot;143412869029&quot;,&quot;creator&quot;:&quot;143412869029&quot;,&quot;pagestimeline&quot;:&quot;1&quot;,&quot;payload_type&quot;:&quot;7&quot;,&quot;eventtime&quot;:&quot;1352998853&quot;,&quot;viewerid&quot;:&quot;0&quot;,&quot;profileownerid&quot;:&quot;143412869029&quot;,&quot;unitimpressionid&quot;:&quot;f2d1d9d2&quot;,&quot;contentid&quot;:&quot;815093699853453383&quot;,&quot;timeline_unit_type&quot;:&quot;AddSinglePhotoUnit&quot;,&quot;timewindowsize&quot;:&quot;2&quot;,&quot;contextwindowstart&quot;:&quot;1349074800&quot;,&quot;contextwindowend&quot;:&quot;1351753199&quot;&#125;\" data-time=\"1351710557\"><div class=\"\"><div role=\"article\"><div class=\"clearfix mbs pbs _1_m \"><a class=\"_29h _303\" href=\"http://www.facebook.com/NorthCarolinaStateParks\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-ash3/174880_143412869029_133040642_q.jpg\" alt=\"\" /></a><div class=\"_3dp _29k\"><h5 class=\"_1_s\"><span class=\"fcg\"><span class=\"fwb\"><a href=\"http://www.facebook.com/NorthCarolinaStateParks\">North Carolina State Parks and Recreation</a></span></span></h5><div class=\"_1_n fsm fwn fcg\"><a class=\"uiLinkSubtle\" href=\"http://www.facebook.com/photo.php?fbid=10151237038709030&amp;set=a.164247464029.118730.143412869029&amp;type=1\"><abbr title=\"Wednesday, October 31, 2012 at 11:09am\" data-utime=\"1351706957\">October 31</abbr></a><a data-hover=\"tooltip\" aria-label=\"Public\" class=\"uiStreamPrivacy inlineBlock fbStreamPrivacy fbPrivacyAudienceIndicator _1_o\" href=\"#\"><i class=\"lock img sp_2voup5 sx_4929d1\"></i></a></div></div></div><div class=\"aboveUnitContent\"><div class=\"userContentWrapper\"><div class=\"_wk\"><span class=\"userContent\">Have a safe, happy Halloween. Remember, it&#039;s still the Year of the Bat in state parks. So if you see any tonight, be nice to &#039;em.</span></div></div></div><div class=\"photoUnit clearfix\"><div class=\"_53s uiScaledThumb photo photoWidth1\" data-cropped=\"1\"><a class=\"_6i9\" href=\"http://www.facebook.com/photo.php?fbid=10151237038709030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1\" rel=\"theater\" ajaxify=\"http://www.facebook.com/photo.php?fbid=10151237038709030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1&amp;src=https%3A%2F%2Fsphotos-b.xx.fbcdn.net%2Fhphotos-snc6%2F247670_10151237038709030_1604193137_n.jpg&amp;size=500%2C500&amp;theater\"><div class=\"uiScaledImageContainer photoWrap\"><img class=\"scaledImageFitWidth img\" src=\"https://sphotos-b.xx.fbcdn.net/hphotos-snc6/c0.0.403.403/p403x403/247670_10151237038709030_1604193137_n.jpg\" alt=\"Photo: Have a safe, happy Halloween. Remember, it&#039;s still the Year of the Bat in state parks. So if you see any tonight, be nice to &#039;em.\" width=\"403\" height=\"403\" /></div></a></div></div></div><div class=\"fbTimelineUFI uiCommentContainer\"><form rel=\"async\" class=\"commentable_item autoexpand_mode\" method=\"post\" action=\"/ajax/ufi/modify.php\" onsubmit=\"return window.Event &amp;&amp; Event.__inlineSubmit &amp;&amp; Event.__inlineSubmit(this,event)\" id=\"u4kzrn2160\"><input type=\"hidden\" name=\"charset_test\" value=\"&euro;,&acute;,\u20AC,\u00B4,\u6C34,\u0414,\u0404\" /><input type=\"hidden\" name=\"lsd\" value=\"AVqaEiN6\" autocomplete=\"off\" /><input type=\"hidden\" autocomplete=\"off\" name=\"feedback_params\" value=\"&#123;&quot;actor&quot;:&quot;143412869029&quot;,&quot;target_fbid&quot;:&quot;10151237038709030&quot;,&quot;target_profile_id&quot;:&quot;143412869029&quot;,&quot;type_id&quot;:&quot;7&quot;,&quot;assoc_obj_id&quot;:&quot;&quot;,&quot;source_app_id&quot;:&quot;0&quot;,&quot;extra_story_params&quot;:[],&quot;content_timestamp&quot;:&quot;1351710556&quot;,&quot;check_hash&quot;:&quot;AQBsqMLK8zZElYDd&quot;,&quot;source&quot;:&quot;0&quot;&#125;\" /><input type=\"hidden\" autocomplete=\"off\" name=\"data_only_response\" value=\"1\" /><input type=\"hidden\" autocomplete=\"off\" name=\"timeline_ufi\" value=\"1\" /><input type=\"hidden\" name=\"timeline_log_data\" value=\"AQDBN6Pc9A7jk_U40N_sIxtjjCaqUP-6n6L2edehhoOf-9mvN8CwvRZQGTlVL_C-AiDRb96EXR4uI9isoPL-sitY-5i3RwbHLdPwwfJVV2kJdkGJLSi5FnRIoGPbK8K7HA4l-c_RV0_wVe3RpAbhBPJssv3HwCxzYX4F9j8ZR-JVtuMl7fjwdHLBj9nRfu3pPhsubOvOwNY87-zg0i-Ai56o0FjUoUphWI1SkUcyZE20nUuYj6Vd7FXV4M0ovD6bm9c7fESPcbff59My0FoTiXhbFJqSKq66_KgWzrxKOy6oj-AFKvfQRq1G9cQ1u6FZg7BDjc5ovW3TTvjNxapb4fsuNAblRrJ8JPwvaXlxPxAzNH4pZrGj0v1IaTomHYGEx8c\" /><div class=\"fbTimelineFeedbackHeader\"><div class=\"fbTimelineFeedbackActions clearfix\"><span class=\"fbTimelineFeedbackShares\"><a data-hover=\"tooltip\" aria-label=\"Show shares\" data-tooltip-alignh=\"right\" href=\"/shares/view?id=10151237038709030\" ajaxify=\"/ajax/shares/view?target_fbid=10151237038709030\" rel=\"async\"><i></i>1</a></span><span class=\"UIActionLinks UIActionLinks_bottom\" data-ft=\"&#123;&quot;tn&quot;:&quot;=&quot;,&quot;type&quot;:20&#125;\"><a class=\"like_link\" title=\"Like this item\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151237038709030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Like</a> &#183; <a class=\"comment_link\" title=\"Leave a comment\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151237038709030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Comment</a> &#183; <a rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151237038709030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Share</a></span></div></div><div><div class=\"uiUfi UFIContainer utn\" id=\"u4kzrn291\"></div></div></form></div></div></div><i class=\"spinePointer\"></i><div class=\"bottomBorder\"></div></li><li class=\"fbTimelineUnit fbTimelineTwoColumn clearfix\" data-side=\"l\" data-fixed=\"1\" data-size=\"1\" id=\"tl_unit_5181084083100967931\"><div class=\"topBorder\"></div><div class=\"timelineUnitContainer\" id=\"u4kzrn259\" data-gt=\"&#123;&quot;action_type&quot;:&quot;1&quot;,&quot;aggregation_id&quot;:&quot;10151236651299030&quot;,&quot;aggregation_time&quot;:&quot;1351693645&quot;,&quot;actor&quot;:&quot;143412869029&quot;,&quot;creator&quot;:&quot;143412869029&quot;,&quot;pagestimeline&quot;:&quot;1&quot;,&quot;payload_type&quot;:&quot;7&quot;,&quot;eventtime&quot;:&quot;1352998853&quot;,&quot;viewerid&quot;:&quot;0&quot;,&quot;profileownerid&quot;:&quot;143412869029&quot;,&quot;unitimpressionid&quot;:&quot;f2d1d9d2&quot;,&quot;contentid&quot;:&quot;5181084083100967931&quot;,&quot;timeline_unit_type&quot;:&quot;AddSinglePhotoUnit&quot;,&quot;timewindowsize&quot;:&quot;2&quot;,&quot;contextwindowstart&quot;:&quot;1349074800&quot;,&quot;contextwindowend&quot;:&quot;1351753199&quot;&#125;\" data-time=\"1351693645\"><div class=\"\"><div role=\"article\"><div class=\"clearfix mbs pbs _1_m \"><a class=\"_29h _303\" href=\"http://www.facebook.com/NorthCarolinaStateParks\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-ash3/174880_143412869029_133040642_q.jpg\" alt=\"\" /></a><div class=\"_3dp _29k\"><h5 class=\"_1_s\"><span class=\"fcg\"><span class=\"fwb\"><a href=\"http://www.facebook.com/NorthCarolinaStateParks\">North Carolina State Parks and Recreation</a></span></span></h5><div class=\"_1_n fsm fwn fcg\"><a class=\"uiLinkSubtle\" href=\"http://www.facebook.com/photo.php?fbid=10151236651299030&amp;set=a.164247464029.118730.143412869029&amp;type=1\"><abbr title=\"Wednesday, October 31, 2012 at 6:27am\" data-utime=\"1351690045\">October 31</abbr></a><a data-hover=\"tooltip\" aria-label=\"Public\" class=\"uiStreamPrivacy inlineBlock fbStreamPrivacy fbPrivacyAudienceIndicator _1_o\" href=\"#\"><i class=\"lock img sp_2voup5 sx_4929d1\"></i></a></div></div></div><div class=\"aboveUnitContent\"><div class=\"userContentWrapper\"><div class=\"_wk\"><span class=\"userContent\">A weird week and likely the first time state parks were closed on the coast for hurricane winds and in the mountains for snow. Find out more about state parks and nasty storms in our blog today: <a href=\"http://www.facebook.com/l.php?u=http%3A%2F%2Fncstateparks.wordpress.com%2F&amp;h=EAQG7ZeET&amp;s=1\" target=\"_blank\" rel=\"nofollow nofollow\" onmouseover=\"LinkshimAsyncLink.swap(this, &quot;http:\\\\/\\\\/ncstateparks.wordpress.com\\\\/&quot;);\" onclick=\"LinkshimAsyncLink.referrer_log(this, &quot;http:\\\\/\\\\/ncstateparks.wordpress.com\\\\/&quot;, &quot;http:\\\\/\\\\/www.facebook.com\\\\/si\\\\/ajax\\\\/l\\\\/render_linkshim_log\\\\/?u=http\\\\u00253A\\\\u00252F\\\\u00252Fncstateparks.wordpress.com\\\\u00252F&amp;h=EAQG7ZeET&amp;s=1&quot;);\">http://ncstateparks.wordpress.com/</a></span></div></div></div><div class=\"photoUnit clearfix\"><div class=\"_53s uiScaledThumb photo photoWidth1\" data-cropped=\"1\"><a class=\"_6i9\" href=\"http://www.facebook.com/photo.php?fbid=10151236651299030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1\" rel=\"theater\" ajaxify=\"http://www.facebook.com/photo.php?fbid=10151236651299030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1&amp;src=https%3A%2F%2Fsphotos-a.xx.fbcdn.net%2Fhphotos-prn1%2F33842_10151236651299030_1601138943_n.jpg&amp;size=800%2C553&amp;theater\"><div class=\"uiScaledImageContainer photoWrap\"><img class=\"scaledImageFitWidth img\" src=\"https://sphotos-a.xx.fbcdn.net/hphotos-prn1/c89.0.403.403/p403x403/33842_10151236651299030_1601138943_n.jpg\" alt=\"Photo: A weird week and likely the first time state parks were closed on the coast for hurricane winds and in the mountains for snow. Find out more about state parks and nasty storms in our blog today: http://ncstateparks.wordpress.com\" width=\"403\" height=\"403\" /></div></a></div></div></div><div class=\"fbTimelineUFI uiCommentContainer\"><form rel=\"async\" class=\"commentable_item autoexpand_mode\" method=\"post\" action=\"/ajax/ufi/modify.php\" onsubmit=\"return window.Event &amp;&amp; Event.__inlineSubmit &amp;&amp; Event.__inlineSubmit(this,event)\" id=\"u4kzrn2161\"><input type=\"hidden\" name=\"charset_test\" value=\"&euro;,&acute;,\u20AC,\u00B4,\u6C34,\u0414,\u0404\" /><input type=\"hidden\" name=\"lsd\" value=\"AVqaEiN6\" autocomplete=\"off\" /><input type=\"hidden\" autocomplete=\"off\" name=\"feedback_params\" value=\"&#123;&quot;actor&quot;:&quot;143412869029&quot;,&quot;target_fbid&quot;:&quot;10151236651299030&quot;,&quot;target_profile_id&quot;:&quot;143412869029&quot;,&quot;type_id&quot;:&quot;7&quot;,&quot;assoc_obj_id&quot;:&quot;&quot;,&quot;source_app_id&quot;:&quot;0&quot;,&quot;extra_story_params&quot;:[],&quot;content_timestamp&quot;:&quot;1351693645&quot;,&quot;check_hash&quot;:&quot;AQBJOgV8xsHXD83J&quot;,&quot;source&quot;:&quot;0&quot;&#125;\" /><input type=\"hidden\" autocomplete=\"off\" name=\"data_only_response\" value=\"1\" /><input type=\"hidden\" autocomplete=\"off\" name=\"timeline_ufi\" value=\"1\" /><input type=\"hidden\" name=\"timeline_log_data\" value=\"AQCgytFRPm3iSAqNqVO2XZs-ZSK9sJnkbuF3cHrCtBE09hMzZmSMVfSO7FHdov7EznFPj4jEG-i9_wCAOyoXty4W88h77Hwj2EZD0zEdkkFgIOt5uIVCuj3cpqnGIuZbWv9fpFmWA_vrdGnRQK2W-\\-\\Oc4ZgEbJXP5DyksKxxykyk0azvOda6HRpY1daAk8mGz1sKpcXgECGma0PX-6HXasotySQfwlOplnoIRdnlJX_KyQxvB_SuzmL94pcYMSmgOQB0JERm8bEpCICV6EL3A8qQkFFbRvMDtc22dx1ZxNxJGfdLWA2YkUKNbo3Sv8W91EGQI2aOeSJOzNYVovxES8NNiQ8Xh-suuFBUEtTyWUvZ4vhXsxShb4ZIi8Ov0tY_PiM\" /><div class=\"fbTimelineFeedbackHeader\"><div class=\"fbTimelineFeedbackActions clearfix\"><span class=\"fbTimelineFeedbackShares\"><a data-hover=\"tooltip\" aria-label=\"Show shares\" data-tooltip-alignh=\"right\" href=\"/shares/view?id=10151236651299030\" ajaxify=\"/ajax/shares/view?target_fbid=10151236651299030\" rel=\"async\"><i></i>15</a></span><span class=\"UIActionLinks UIActionLinks_bottom\" data-ft=\"&#123;&quot;tn&quot;:&quot;=&quot;,&quot;type&quot;:20&#125;\"><a class=\"like_link\" title=\"Like this item\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151236651299030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Like</a> &#183; <a class=\"comment_link\" title=\"Leave a comment\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151236651299030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Comment</a> &#183; <a rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151236651299030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Share</a></span></div></div><div><div class=\"uiUfi UFIContainer utn\" id=\"u4kzrn2134\"></div></div></form></div></div></div><i class=\"spinePointer\"></i><div class=\"bottomBorder\"></div></li><li class=\"fbTimelineUnit fbTimelineTwoColumn clearfix\" data-side=\"l\" data-fixed=\"1\" data-size=\"1\" id=\"tl_unit_7505335592811485177\"><div class=\"topBorder\"></div><div class=\"timelineUnitContainer\" id=\"u4kzrn260\" data-gt=\"&#123;&quot;action_type&quot;:&quot;1&quot;,&quot;aggregation_id&quot;:&quot;10151236567709030&quot;,&quot;aggregation_time&quot;:&quot;1351689977&quot;,&quot;actor&quot;:&quot;143412869029&quot;,&quot;creator&quot;:&quot;143412869029&quot;,&quot;pagestimeline&quot;:&quot;1&quot;,&quot;payload_type&quot;:&quot;7&quot;,&quot;eventtime&quot;:&quot;1352998853&quot;,&quot;viewerid&quot;:&quot;0&quot;,&quot;profileownerid&quot;:&quot;143412869029&quot;,&quot;unitimpressionid&quot;:&quot;f2d1d9d2&quot;,&quot;contentid&quot;:&quot;7505335592811485177&quot;,&quot;timeline_unit_type&quot;:&quot;AddSinglePhotoUnit&quot;,&quot;timewindowsize&quot;:&quot;2&quot;,&quot;contextwindowstart&quot;:&quot;1349074800&quot;,&quot;contextwindowend&quot;:&quot;1351753199&quot;&#125;\" data-time=\"1351689977\"><div class=\"\"><div role=\"article\"><div class=\"clearfix mbs pbs _1_m \"><a class=\"_29h _303\" href=\"http://www.facebook.com/NorthCarolinaStateParks\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-ash3/174880_143412869029_133040642_q.jpg\" alt=\"\" /></a><div class=\"_3dp _29k\"><h5 class=\"_1_s\"><span class=\"fcg\"><span class=\"fwb\"><a href=\"http://www.facebook.com/NorthCarolinaStateParks\">North Carolina State Parks and Recreation</a></span></span></h5><div class=\"_1_n fsm fwn fcg\"><a class=\"uiLinkSubtle\" href=\"http://www.facebook.com/photo.php?fbid=10151236567709030&amp;set=a.164247464029.118730.143412869029&amp;type=1\"><abbr title=\"Wednesday, October 31, 2012 at 5:26am\" data-utime=\"1351686377\">October 31</abbr></a><a data-hover=\"tooltip\" aria-label=\"Public\" class=\"uiStreamPrivacy inlineBlock fbStreamPrivacy fbPrivacyAudienceIndicator _1_o\" href=\"#\"><i class=\"lock img sp_2voup5 sx_4929d1\"></i></a></div></div></div><div class=\"aboveUnitContent\"><div class=\"userContentWrapper\"><div class=\"_wk\"><span class=\"userContent\">Nice editorial on the state&#039;s foresight in setting aside the land for Gorges State Park where we opened a new visitor center this month: <a href=\"http://www.facebook.com/l.php?u=http%3A%2F%2Fwww.blueridgenow.com%2Farticle%2F20121031%2FARTICLES%2F210311000%2F1016%2FOPINION02%3Fp%3D2%26tc%3Dar&amp;h=xAQFWHACK&amp;s=1\" target=\"_blank\" rel=\"nofollow nofollow\" onmouseover=\"LinkshimAsyncLink.swap(this, &quot;http:\\\\/\\\\/www.blueridgenow.com\\\\/article\\\\/20121031\\\\/ARTICLES\\\\/210311000\\\\/1016\\\\/OPINION02?p=2&amp;tc=ar&quot;);\" onclick=\"LinkshimAsyncLink.referrer_log(this, &quot;http:\\\\/\\\\/www.blueridgenow.com\\\\/article\\\\/20121031\\\\/ARTICLES\\\\/210311000\\\\/1016\\\\/OPINION02?p=2&amp;tc=ar&quot;, &quot;http:\\\\/\\\\/www.facebook.com\\\\/si\\\\/ajax\\\\/l\\\\/render_linkshim_log\\\\/?u=http\\\\u00253A\\\\u00252F\\\\u00252Fwww.blueridgenow.com\\\\u00252Farticle\\\\u00252F20121031\\\\u00252FARTICLES\\\\u00252F210311000\\\\u00252F1016\\\\u00252FOPINION02\\\\u00253Fp\\\\u00253D2\\\\u002526tc\\\\u00253Dar&amp;h=xAQFWHACK&amp;s=1&quot;);\">http://www.blueridgenow.com/article/20121031/ARTICLES/210311000/1016/OPINION02?p=2&amp;tc=ar</a></span></div></div></div><div class=\"photoUnit clearfix\"><div class=\"_53s uiScaledThumb photo photoWidth1\" data-cropped=\"1\"><a class=\"_6i9\" href=\"http://www.facebook.com/photo.php?fbid=10151236567709030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1\" rel=\"theater\" ajaxify=\"http://www.facebook.com/photo.php?fbid=10151236567709030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1&amp;src=https%3A%2F%2Fsphotos-a.xx.fbcdn.net%2Fhphotos-ash4%2F374875_10151236567709030_284189565_n.jpg&amp;size=800%2C566&amp;theater\"><div class=\"uiScaledImageContainer photoWrap\"><img class=\"scaledImageFitWidth img\" src=\"https://sphotos-a.xx.fbcdn.net/hphotos-ash4/c83.0.403.403/p403x403/374875_10151236567709030_284189565_n.jpg\" alt=\"Photo: Nice editorial on the state&#039;s foresight in setting aside the land for Gorges State Park where we opened a new visitor center this month: http://www.blueridgenow.com/article/20121031/ARTICLES/210311000/1016/OPINION02?p=2&amp;tc=pg&amp;tc=ar\" width=\"403\" height=\"403\" /></div></a></div></div></div><div class=\"fbTimelineUFI uiCommentContainer\"><form rel=\"async\" class=\"commentable_item autoexpand_mode\" method=\"post\" action=\"/ajax/ufi/modify.php\" onsubmit=\"return window.Event &amp;&amp; Event.__inlineSubmit &amp;&amp; Event.__inlineSubmit(this,event)\" id=\"u4kzrn2162\"><input type=\"hidden\" name=\"charset_test\" value=\"&euro;,&acute;,\u20AC,\u00B4,\u6C34,\u0414,\u0404\" /><input type=\"hidden\" name=\"lsd\" value=\"AVqaEiN6\" autocomplete=\"off\" /><input type=\"hidden\" autocomplete=\"off\" name=\"feedback_params\" value=\"&#123;&quot;actor&quot;:&quot;143412869029&quot;,&quot;target_fbid&quot;:&quot;10151236567709030&quot;,&quot;target_profile_id&quot;:&quot;143412869029&quot;,&quot;type_id&quot;:&quot;7&quot;,&quot;assoc_obj_id&quot;:&quot;&quot;,&quot;source_app_id&quot;:&quot;0&quot;,&quot;extra_story_params&quot;:[],&quot;content_timestamp&quot;:&quot;1351689977&quot;,&quot;check_hash&quot;:&quot;AQA3frztTEzmWABd&quot;,&quot;source&quot;:&quot;0&quot;&#125;\" /><input type=\"hidden\" autocomplete=\"off\" name=\"data_only_response\" value=\"1\" /><input type=\"hidden\" autocomplete=\"off\" name=\"timeline_ufi\" value=\"1\" /><input type=\"hidden\" name=\"timeline_log_data\" value=\"AQDOFO2Unh_vPBOf5UBafz9CF_02TR9Igjb6vEG8hmDGFbOqXG8Ostae2EHBvl0j7T9XVOBrLOqOLoOwlzt8urAUNuNBwUgwL6lkOA8Ycq3XRS-jJvpmD3px6LeJMNDkQYYD0eNiVQq2dvcFg09dD4OfODjntJ2h4SWLKNRW3WRFTYnU1bhckkKRjyrvhNBG_LeCUir-kngtemBwMVG3XTtva9CyEE4sCejNOefFlLCpHrAo5eiwgM1rSIjrkzSfeo0sHbEXwipiDlplASzUMHPwwZDmUhVz2-K_elnfoeJ6kW-bIqslZcjsB_PDaDjxVY-bT79ZjhBjpev0D1UJG-u2NKzXFJ9CB7WSI2qQZ885Ai3weKYpnQqqSNWVecFNwgQ\" /><div class=\"fbTimelineFeedbackHeader\"><div class=\"fbTimelineFeedbackActions clearfix\"><span class=\"fbTimelineFeedbackShares\"><a data-hover=\"tooltip\" aria-label=\"Show shares\" data-tooltip-alignh=\"right\" href=\"/shares/view?id=10151236567709030\" ajaxify=\"/ajax/shares/view?target_fbid=10151236567709030\" rel=\"async\"><i></i>1</a></span><span class=\"UIActionLinks UIActionLinks_bottom\" data-ft=\"&#123;&quot;tn&quot;:&quot;=&quot;,&quot;type&quot;:20&#125;\"><a class=\"like_link\" title=\"Like this item\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151236567709030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Like</a> &#183; <a class=\"comment_link\" title=\"Leave a comment\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151236567709030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Comment</a> &#183; <a rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151236567709030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Share</a></span></div></div><div><div class=\"uiUfi UFIContainer utn\" id=\"u4kzrn292\"></div></div></form></div></div></div><i class=\"spinePointer\"></i><div class=\"bottomBorder\"></div></li><li class=\"fbTimelineUnit fbTimelineTwoColumn clearfix\" data-side=\"l\" data-fixed=\"1\" data-size=\"1\" id=\"tl_unit_-351409199215131782\"><div class=\"topBorder\"></div><div class=\"timelineUnitContainer\" id=\"u4kzrn261\" data-gt=\"&#123;&quot;action_type&quot;:&quot;1&quot;,&quot;aggregation_id&quot;:&quot;10151236478529030&quot;,&quot;aggregation_time&quot;:&quot;1351686544&quot;,&quot;actor&quot;:&quot;143412869029&quot;,&quot;creator&quot;:&quot;143412869029&quot;,&quot;pagestimeline&quot;:&quot;1&quot;,&quot;payload_type&quot;:&quot;7&quot;,&quot;eventtime&quot;:&quot;1352998853&quot;,&quot;viewerid&quot;:&quot;0&quot;,&quot;profileownerid&quot;:&quot;143412869029&quot;,&quot;unitimpressionid&quot;:&quot;f2d1d9d2&quot;,&quot;contentid&quot;:&quot;-351409199215131782&quot;,&quot;timeline_unit_type&quot;:&quot;AddSinglePhotoUnit&quot;,&quot;timewindowsize&quot;:&quot;2&quot;,&quot;contextwindowstart&quot;:&quot;1349074800&quot;,&quot;contextwindowend&quot;:&quot;1351753199&quot;&#125;\" data-time=\"1351686544\"><div class=\"\"><div role=\"article\"><div class=\"clearfix mbs pbs _1_m \"><a class=\"_29h _303\" href=\"http://www.facebook.com/NorthCarolinaStateParks\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-ash3/174880_143412869029_133040642_q.jpg\" alt=\"\" /></a><div class=\"_3dp _29k\"><h5 class=\"_1_s\"><span class=\"fcg\"><span class=\"fwb\"><a href=\"http://www.facebook.com/NorthCarolinaStateParks\">North Carolina State Parks and Recreation</a></span></span></h5><div class=\"_1_n fsm fwn fcg\"><a class=\"uiLinkSubtle\" href=\"http://www.facebook.com/photo.php?fbid=10151236478529030&amp;set=a.164247464029.118730.143412869029&amp;type=1\"><abbr title=\"Wednesday, October 31, 2012 at 4:29am\" data-utime=\"1351682944\">October 31</abbr></a><a data-hover=\"tooltip\" aria-label=\"Public\" class=\"uiStreamPrivacy inlineBlock fbStreamPrivacy fbPrivacyAudienceIndicator _1_o\" href=\"#\"><i class=\"lock img sp_2voup5 sx_4929d1\"></i></a></div></div></div><div class=\"aboveUnitContent\"><div class=\"userContentWrapper\"><div class=\"_wk\"><span class=\"userContent\">Here&#039;s a great chance to have some fun Saturday and also benefit Raven Rock State Park through its new friends group.</span></div></div></div><div class=\"photoUnit clearfix\"><div class=\"_53s uiScaledThumb photo photoWidth1\" data-cropped=\"1\"><a class=\"_6i9\" href=\"http://www.facebook.com/photo.php?fbid=10151236478529030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1\" rel=\"theater\" ajaxify=\"http://www.facebook.com/photo.php?fbid=10151236478529030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1&amp;src=https%3A%2F%2Fsphotos-b.xx.fbcdn.net%2Fhphotos-snc6%2F169098_10151236478529030_1251981654_o.jpg&amp;smallsrc=https%3A%2F%2Fsphotos-b.xx.fbcdn.net%2Fhphotos-ash4%2F409059_10151236478529030_1251981654_n.jpg&amp;size=800%2C1015&amp;theater\"><div class=\"uiScaledImageContainer photoWrap\"><img class=\"scaledImageFitWidth img\" src=\"https://sphotos-b.xx.fbcdn.net/hphotos-ash4/c0.0.403.403/p403x403/409059_10151236478529030_1251981654_n.jpg\" alt=\"Photo: Here&#039;s a great chance to have some fun Saturday and also benefit Raven Rock State Park through its new friends group.\" width=\"403\" height=\"403\" /></div></a></div></div></div><div class=\"fbTimelineUFI uiCommentContainer\"><form rel=\"async\" class=\"commentable_item autoexpand_mode\" method=\"post\" action=\"/ajax/ufi/modify.php\" onsubmit=\"return window.Event &amp;&amp; Event.__inlineSubmit &amp;&amp; Event.__inlineSubmit(this,event)\" id=\"u4kzrn2163\"><input type=\"hidden\" name=\"charset_test\" value=\"&euro;,&acute;,\u20AC,\u00B4,\u6C34,\u0414,\u0404\" /><input type=\"hidden\" name=\"lsd\" value=\"AVqaEiN6\" autocomplete=\"off\" /><input type=\"hidden\" autocomplete=\"off\" name=\"feedback_params\" value=\"&#123;&quot;actor&quot;:&quot;143412869029&quot;,&quot;target_fbid&quot;:&quot;10151236478529030&quot;,&quot;target_profile_id&quot;:&quot;143412869029&quot;,&quot;type_id&quot;:&quot;7&quot;,&quot;assoc_obj_id&quot;:&quot;&quot;,&quot;source_app_id&quot;:&quot;0&quot;,&quot;extra_story_params&quot;:[],&quot;content_timestamp&quot;:&quot;1351686544&quot;,&quot;check_hash&quot;:&quot;AQDVkb0pab3Icy__&quot;,&quot;source&quot;:&quot;0&quot;&#125;\" /><input type=\"hidden\" autocomplete=\"off\" name=\"data_only_response\" value=\"1\" /><input type=\"hidden\" autocomplete=\"off\" name=\"timeline_ufi\" value=\"1\" /><input type=\"hidden\" name=\"timeline_log_data\" value=\"AQCBY2G4YST4zmje2BkF-ippVRMlnGAQtV967SyjLgSm6-u-vdYySHMdVT8c4FBmfx_pIEWFihyNy47bGJht2tskm0JCSFRF6VIOAbuLid5LHrDamVUT3KG6GFA2mW8xTHriAyi_B3K1Nr3LGr4JmQ5_fTg93wiS-NI0pAnBpjBZVnJKf86Wz1nLyMsew9t5wqVlvOeX0RmtEN9U0uA7tJ7oXFLjDq4wBJleSw3_iYs1_EBRyR_q7PKSXSKYu7YWvBezs7xZ_qPYt__8pziaiNtL4HGFe1oldPgR73deMIparL3hBRI8R-3H-EOc9NAIWVTbOB5hzD-c4Z90bxrDhsJTPbsCSFgIbjKqqBI2d9w5qhynvZeP9Zdyyk8VI8Q-Pio\" /><div class=\"fbTimelineFeedbackHeader\"><div class=\"fbTimelineFeedbackActions clearfix\"><span class=\"fbTimelineFeedbackShares\"><a data-hover=\"tooltip\" aria-label=\"Show shares\" data-tooltip-alignh=\"right\" href=\"/shares/view?id=10151236478529030\" ajaxify=\"/ajax/shares/view?target_fbid=10151236478529030\" rel=\"async\"><i></i>25</a></span><span class=\"UIActionLinks UIActionLinks_bottom\" data-ft=\"&#123;&quot;tn&quot;:&quot;=&quot;,&quot;type&quot;:20&#125;\"><a class=\"like_link\" title=\"Like this item\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151236478529030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Like</a> &#183; <a class=\"comment_link\" title=\"Leave a comment\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151236478529030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Comment</a> &#183; <a rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151236478529030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Share</a></span></div></div><div><div class=\"uiUfi UFIContainer utn\" id=\"u4kzrn293\"></div></div></form></div></div></div><i class=\"spinePointer\"></i><div class=\"bottomBorder\"></div></li><li class=\"fbTimelineUnit fbTimelineTwoColumn clearfix\" data-side=\"l\" data-fixed=\"1\" data-size=\"1\" id=\"tl_unit_-6319555045563901530\"><div class=\"topBorder\"></div><div class=\"timelineUnitContainer\" id=\"u4kzrn262\" data-gt=\"&#123;&quot;action_type&quot;:&quot;1&quot;,&quot;aggregation_id&quot;:&quot;10151235033604030&quot;,&quot;aggregation_time&quot;:&quot;1351605958&quot;,&quot;actor&quot;:&quot;143412869029&quot;,&quot;creator&quot;:&quot;143412869029&quot;,&quot;pagestimeline&quot;:&quot;1&quot;,&quot;payload_type&quot;:&quot;11&quot;,&quot;eventtime&quot;:&quot;1352998853&quot;,&quot;viewerid&quot;:&quot;0&quot;,&quot;profileownerid&quot;:&quot;143412869029&quot;,&quot;unitimpressionid&quot;:&quot;f2d1d9d2&quot;,&quot;contentid&quot;:&quot;-6319555045563901530&quot;,&quot;timeline_unit_type&quot;:&quot;StatusMessageUnit&quot;,&quot;timewindowsize&quot;:&quot;2&quot;,&quot;contextwindowstart&quot;:&quot;1349074800&quot;,&quot;contextwindowend&quot;:&quot;1351753199&quot;&#125;\" data-time=\"1351605959\"><div class=\"\"><div role=\"article\"><div class=\"clearfix mbs pbs _1_m \"><a class=\"_29h _303\" href=\"http://www.facebook.com/NorthCarolinaStateParks\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-ash3/174880_143412869029_133040642_q.jpg\" alt=\"\" /></a><div class=\"_3dp _29k\"><h5 class=\"_1_s\"><span class=\"fcg\"><span class=\"fwb\"><a href=\"http://www.facebook.com/NorthCarolinaStateParks\">North Carolina State Parks and Recreation</a></span></span></h5><div class=\"_1_n fsm fwn fcg\"><a class=\"uiLinkSubtle\" href=\"/NorthCarolinaStateParks/posts/10151235033604030\"><abbr title=\"Tuesday, October 30, 2012 at 6:05am\" data-utime=\"1351602359\">October 30</abbr></a><a data-hover=\"tooltip\" aria-label=\"Public\" class=\"uiStreamPrivacy inlineBlock fbStreamPrivacy fbPrivacyAudienceIndicator _1_o\" href=\"#\"><i class=\"lock img sp_2voup5 sx_4929d1\"></i></a></div></div></div><div class=\"_1x1\"><div class=\"userContentWrapper\"><div class=\"_wk\"><span class=\"userContent\">Still snowing -\\-\\ Grandfather Mountain and Mount Mitchell state parks and Mount Jefferson State Natural Area are closed. Other mountain parks open to this point but road conditions are hazardous.</span></div></div></div></div><div class=\"fbTimelineUFI uiCommentContainer\"><form rel=\"async\" class=\"commentable_item autoexpand_mode\" method=\"post\" action=\"/ajax/ufi/modify.php\" onsubmit=\"return window.Event &amp;&amp; Event.__inlineSubmit &amp;&amp; Event.__inlineSubmit(this,event)\" id=\"u4kzrn2164\"><input type=\"hidden\" name=\"charset_test\" value=\"&euro;,&acute;,\u20AC,\u00B4,\u6C34,\u0414,\u0404\" /><input type=\"hidden\" name=\"lsd\" value=\"AVqaEiN6\" autocomplete=\"off\" /><input type=\"hidden\" autocomplete=\"off\" name=\"feedback_params\" value=\"&#123;&quot;actor&quot;:&quot;143412869029&quot;,&quot;target_fbid&quot;:&quot;10151235033604030&quot;,&quot;target_profile_id&quot;:&quot;143412869029&quot;,&quot;type_id&quot;:&quot;22&quot;,&quot;assoc_obj_id&quot;:&quot;&quot;,&quot;source_app_id&quot;:&quot;0&quot;,&quot;extra_story_params&quot;:[],&quot;content_timestamp&quot;:&quot;1351605958&quot;,&quot;check_hash&quot;:&quot;AQBRLUUeCPOzzbQH&quot;,&quot;source&quot;:&quot;0&quot;&#125;\" /><input type=\"hidden\" autocomplete=\"off\" name=\"data_only_response\" value=\"1\" /><input type=\"hidden\" autocomplete=\"off\" name=\"timeline_ufi\" value=\"1\" /><input type=\"hidden\" name=\"timeline_log_data\" value=\"AQCaJQnHWDIIq7D4fMnT2vWyOQ5gD-E6J0k48gB_mRWFlnwmu7qlgrEGnU20gQW606JDgeFn2NGKhMndFnJfaxcJsroxMiOHrnmw7oPOt1ZoBpIZCDbMQ2Wb6Fg-eYqgw8QwUHNipbs2sXypICI8IrJ44Rt4HzqnYxqgklsMfVdql5hpDrX0QWSzh4NAElUXSKD_tiN18fIEdX2LyDruqCgTbnGgQfdUI-T8AVra7KCGt24F9mkehN5rUItBQ1yVt_WQHUzdupYsBebEcGqOv4kFyKrVSr5nX6Z53Ao0laXuyFWtWouJICvx1ZCyIBsAo9JWJB6XpUam_UUC60GYi018SNX-RhP_SzznrVXv97x1o575LnjamB5ZSVZM4BCSdHE\" /><div class=\"fbTimelineFeedbackHeader\"><div class=\"fbTimelineFeedbackActions clearfix\"><span class=\"fbTimelineFeedbackShares\"><a data-hover=\"tooltip\" aria-label=\"Show shares\" data-tooltip-alignh=\"right\" href=\"/shares/view?id=10151235033604030\" ajaxify=\"/ajax/shares/view?target_fbid=10151235033604030\" rel=\"async\"><i></i>1</a></span><span class=\"UIActionLinks UIActionLinks_bottom\" data-ft=\"&#123;&quot;tn&quot;:&quot;=&quot;,&quot;type&quot;:20&#125;\"><a class=\"like_link\" title=\"Like this item\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2FNorthCarolinaStateParks%2Fposts%2F10151235033604030\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Like</a> &#183; <a class=\"comment_link\" title=\"Leave a comment\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2FNorthCarolinaStateParks%2Fposts%2F10151235033604030\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Comment</a> &#183; <a rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2FNorthCarolinaStateParks%2Fposts%2F10151235033604030\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Share</a></span></div></div><div><div class=\"uiUfi UFIContainer utn\" id=\"u4kzrn294\"></div></div></form></div></div></div><i class=\"spinePointer\"></i><div class=\"bottomBorder\"></div></li><li class=\"fbTimelineUnit fbTimelineTwoColumn clearfix\" data-side=\"l\" data-fixed=\"1\" data-size=\"1\" id=\"tl_unit_5060400664960079647\"><div class=\"topBorder\"></div><div class=\"timelineUnitContainer\" id=\"u4kzrn264\" data-gt=\"&#123;&quot;action_type&quot;:&quot;1&quot;,&quot;aggregation_id&quot;:&quot;399953756744985&quot;,&quot;aggregation_time&quot;:&quot;1351602897&quot;,&quot;actor&quot;:&quot;143412869029&quot;,&quot;creator&quot;:&quot;143412869029&quot;,&quot;pagestimeline&quot;:&quot;1&quot;,&quot;payload_type&quot;:&quot;5&quot;,&quot;eventtime&quot;:&quot;1352998853&quot;,&quot;viewerid&quot;:&quot;0&quot;,&quot;profileownerid&quot;:&quot;143412869029&quot;,&quot;unitimpressionid&quot;:&quot;f2d1d9d2&quot;,&quot;contentid&quot;:&quot;5060400664960079647&quot;,&quot;timeline_unit_type&quot;:&quot;BestPostedLinkUnit&quot;,&quot;timewindowsize&quot;:&quot;2&quot;,&quot;contextwindowstart&quot;:&quot;1349074800&quot;,&quot;contextwindowend&quot;:&quot;1351753199&quot;&#125;\" data-time=\"1351602897\"><div class=\"\"><div role=\"article\"><div class=\"clearfix mbs pbs _1_m \"><a class=\"_29h _303\" href=\"http://www.facebook.com/NorthCarolinaStateParks\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-ash3/174880_143412869029_133040642_q.jpg\" alt=\"\" /></a><div class=\"_3dp _29k\"><h5 class=\"_1_s\"><span class=\"fcg\"><span class=\"fwb\"><a href=\"http://www.facebook.com/NorthCarolinaStateParks\">North Carolina State Parks and Recreation</a></span> shared a <a href=\"http://www.facebook.com/l.php?u=http%3A%2F%2Fwww2.wataugademocrat.com%2FNews%2Fstory%2FUPDATE-4-8-inches-of-snow-forecasted-id-009466&amp;h=bAQEN3r6qAQGNFIbzZURfhtnuv1tYZmNqlsF8DP9WIbWvqg&amp;enc=AZOltUz084kUAXq2ol-NYOuLIEYFZhlssYQUxb2H27f2S4Vj7EETJo2gGOe-K1CvmNM0KRlg4ZguhjDhvmtZoJ66&amp;s=1\" rel=\"nofollow\" onmouseover=\"LinkshimAsyncLink.swap(this, &quot;http:\\\\/\\\\/www2.wataugademocrat.com\\\\/News\\\\/story\\\\/UPDATE-4-8-inches-of-snow-forecasted-id-009466&quot;);\" onclick=\"LinkshimAsyncLink.referrer_log(this, &quot;http:\\\\/\\\\/www2.wataugademocrat.com\\\\/News\\\\/story\\\\/UPDATE-4-8-inches-of-snow-forecasted-id-009466&quot;, &quot;http:\\\\/\\\\/www.facebook.com\\\\/si\\\\/ajax\\\\/l\\\\/render_linkshim_log\\\\/?u=http\\\\u00253A\\\\u00252F\\\\u00252Fwww2.wataugademocrat.com\\\\u00252FNews\\\\u00252Fstory\\\\u00252FUPDATE-4-8-inches-of-snow-forecasted-id-009466&amp;h=bAQEN3r6qAQGNFIbzZURfhtnuv1tYZmNqlsF8DP9WIbWvqg&amp;enc=AZOltUz084kUAXq2ol-NYOuLIEYFZhlssYQUxb2H27f2S4Vj7EETJo2gGOe-K1CvmNM0KRlg4ZguhjDhvmtZoJ66&amp;s=1&quot;);\" target=\"_blank\">link</a>.</span></h5><div class=\"_1_n fsm fwn fcg\"><a class=\"uiLinkSubtle\" href=\"/NorthCarolinaStateParks/posts/399953756744985\"><abbr title=\"Tuesday, October 30, 2012 at 5:14am\" data-utime=\"1351599297\">October 30</abbr></a><a data-hover=\"tooltip\" aria-label=\"Public\" class=\"uiStreamPrivacy inlineBlock fbStreamPrivacy fbPrivacyAudienceIndicator _1_o\" href=\"#\"><i class=\"lock img sp_2voup5 sx_4929d1\"></i></a></div></div></div><div class=\"userContentWrapper aboveUnitContent\"><div class=\"_wk mbm\"><span class=\"userContent\">Report from Mount Mitchell - 19 degrees, wind gusting to 40 mph, 6 inches of snow so far: <a href=\"http://www.facebook.com/l.php?u=http%3A%2F%2Fwww2.wataugademocrat.com%2FNews%2Fstory%2FUPDATE-4-8-inches-of-snow-forecasted-id-009466&amp;h=ZAQHJsopg&amp;s=1\" target=\"_blank\" rel=\"nofollow nofollow\" onmouseover=\"LinkshimAsyncLink.swap(this, &quot;http:\\\\/\\\\/www2.wataugademocrat.com\\\\/News\\\\/story\\\\/UPDATE-4-8-inches-of-snow-forecasted-id-009466&quot;);\" onclick=\"LinkshimAsyncLink.referrer_log(this, &quot;http:\\\\/\\\\/www2.wataugademocrat.com\\\\/News\\\\/story\\\\/UPDATE-4-8-inches-of-snow-forecasted-id-009466&quot;, &quot;http:\\\\/\\\\/www.facebook.com\\\\/si\\\\/ajax\\\\/l\\\\/render_linkshim_log\\\\/?u=http\\\\u00253A\\\\u00252F\\\\u00252Fwww2.wataugademocrat.com\\\\u00252FNews\\\\u00252Fstory\\\\u00252FUPDATE-4-8-inches-of-snow-forecasted-id-009466&amp;h=ZAQHJsopg&amp;s=1&quot;);\">http://www2.wataugademocrat.com/News/story/UPDATE-4-8-inches-of-snow-forecasted-id-009466</a></span></div></div><div class=\"shareUnit\"><a class=\"_1xw _1y0\" href=\"http://www.facebook.com/l.php?u=http%3A%2F%2Fwww2.wataugademocrat.com%2FNews%2Fstory%2FUPDATE-4-8-inches-of-snow-forecasted-id-009466&amp;h=uAQHr8pT-AQEMbKRQw3NOi7E8Rrr5cijNVUVvd3jWdGhLUg&amp;enc=AZPxVKnuQaELsakE5LyOh1Dc3PwX4dxUM_rf6PdUyrjAqAeQjv1DDjSsEEFSUik88ue6zOkifK87miqec-Y5HuvV&amp;s=1\" target=\"_blank\" rel=\"nofollow\" onmouseover=\"LinkshimAsyncLink.swap(this, &quot;http:\\\\/\\\\/www2.wataugademocrat.com\\\\/News\\\\/story\\\\/UPDATE-4-8-inches-of-snow-forecasted-id-009466&quot;);\" onclick=\"LinkshimAsyncLink.referrer_log(this, &quot;http:\\\\/\\\\/www2.wataugademocrat.com\\\\/News\\\\/story\\\\/UPDATE-4-8-inches-of-snow-forecasted-id-009466&quot;, &quot;http:\\\\/\\\\/www.facebook.com\\\\/si\\\\/ajax\\\\/l\\\\/render_linkshim_log\\\\/?u=http\\\\u00253A\\\\u00252F\\\\u00252Fwww2.wataugademocrat.com\\\\u00252FNews\\\\u00252Fstory\\\\u00252FUPDATE-4-8-inches-of-snow-forecasted-id-009466&amp;h=uAQHr8pT-AQEMbKRQw3NOi7E8Rrr5cijNVUVvd3jWdGhLUg&amp;enc=AZPxVKnuQaELsakE5LyOh1Dc3PwX4dxUM_rf6PdUyrjAqAeQjv1DDjSsEEFSUik88ue6zOkifK87miqec-Y5HuvV&amp;s=1&quot;);\"><div class=\"_1xy _1xx\"><img class=\"img\" src=\"https://fbexternal-a.akamaihd.net/safe_image.php?d=AQAs-rjUWSscJRW8&amp;w=155&amp;h=114&amp;url=http%3A%2F%2Fwww2.wataugademocrat.com%2Fattachments%2F2012%2F10%2FjaeqjSandySnow__AOweb.jpg\" alt=\"\" /></div><div class=\"_1xx _1xz\"><div class=\"_1x-\"><div class=\"_1x_ fwb\">UPDATE: 4-8 Inches Of Snow Forecasted - The Watauga Democrat</div><div class=\"fsm fwn fcg\">www2.wataugademocrat.com</div><div class=\"_1y1 fsm fwn fcg\"></div></div></div></a></div></div><div class=\"fbTimelineUFI uiCommentContainer\"><form rel=\"async\" class=\"commentable_item autoexpand_mode\" method=\"post\" action=\"/ajax/ufi/modify.php\" onsubmit=\"return window.Event &amp;&amp; Event.__inlineSubmit &amp;&amp; Event.__inlineSubmit(this,event)\" id=\"u4kzrn2165\"><input type=\"hidden\" name=\"charset_test\" value=\"&euro;,&acute;,\u20AC,\u00B4,\u6C34,\u0414,\u0404\" /><input type=\"hidden\" name=\"lsd\" value=\"AVqaEiN6\" autocomplete=\"off\" /><input type=\"hidden\" autocomplete=\"off\" name=\"feedback_params\" value=\"&#123;&quot;actor&quot;:&quot;143412869029&quot;,&quot;target_fbid&quot;:&quot;399953756744985&quot;,&quot;target_profile_id&quot;:&quot;143412869029&quot;,&quot;type_id&quot;:&quot;17&quot;,&quot;assoc_obj_id&quot;:&quot;&quot;,&quot;source_app_id&quot;:&quot;0&quot;,&quot;extra_story_params&quot;:[],&quot;content_timestamp&quot;:&quot;1351602897&quot;,&quot;check_hash&quot;:&quot;AQDi4gXwq6dJfNNV&quot;,&quot;source&quot;:&quot;0&quot;&#125;\" /><input type=\"hidden\" autocomplete=\"off\" name=\"data_only_response\" value=\"1\" /><input type=\"hidden\" autocomplete=\"off\" name=\"timeline_ufi\" value=\"1\" /><input type=\"hidden\" name=\"timeline_log_data\" value=\"AQC_Ya7lJvnfPhVvhl6YdJray5UhLUUqnC5IE0B4IqdpvecV8OMXVktoAlg4cD-V9yCVZO6DFY7fJScmm-7BRmEmq9IJjAA3DQruDS8F8pt_fw8y0IWX5xMjBFkfLy2O2h7xAs0yXoTkgkYXpwcHGknOGL-24hK0wxAR-TFhhJe9t0vI87r2kxVI0JGMNrINvL_cQZpS_xf6nS67yQcMWekpyc7squtzJPxOi8DJ7caBPpC3T13Sb1KlNwQzkwFyMr_9bGP4Odn4AYqY_NJI7XQpklcbvxvckw4q2lsqVwO1T_qDykHwkHhVvIwR3tc6IGQW8FxYPzry2v4d6OFlkYIQOXJzA09PboAW0e-Ff8kmFvaV5xT6Wb79W0cmJFxin-c\" /><div class=\"fbTimelineFeedbackHeader\"><div class=\"fbTimelineFeedbackActions clearfix\"><span class=\"fbTimelineFeedbackShares\"><a data-hover=\"tooltip\" aria-label=\"Show shares\" data-tooltip-alignh=\"right\" href=\"/shares/view?id=399953756744985\" ajaxify=\"/ajax/shares/view?target_fbid=399953756744985\" rel=\"async\"><i></i>3</a></span><span class=\"UIActionLinks UIActionLinks_bottom\" data-ft=\"&#123;&quot;tn&quot;:&quot;=&quot;,&quot;type&quot;:20&#125;\"><a class=\"like_link\" title=\"Like this item\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2FNorthCarolinaStateParks%2Fposts%2F399953756744985\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Like</a> &#183; <a class=\"comment_link\" title=\"Leave a comment\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2FNorthCarolinaStateParks%2Fposts%2F399953756744985\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Comment</a> &#183; <a rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2FNorthCarolinaStateParks%2Fposts%2F399953756744985\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Share</a></span></div></div><div><div class=\"uiUfi UFIContainer utn\" id=\"u4kzrn2135\"></div></div></form></div></div></div><i class=\"spinePointer\"></i><div class=\"bottomBorder\"></div></li><li class=\"fbTimelineUnit fbTimelineTwoColumn clearfix\" data-side=\"l\" data-fixed=\"1\" data-size=\"1\" id=\"tl_unit_-7624703520450932554\"><div class=\"topBorder\"></div><div class=\"timelineUnitContainer\" id=\"u4kzrn265\" data-gt=\"&#123;&quot;action_type&quot;:&quot;1&quot;,&quot;aggregation_id&quot;:&quot;10151234924589030&quot;,&quot;aggregation_time&quot;:&quot;1351600566&quot;,&quot;actor&quot;:&quot;143412869029&quot;,&quot;creator&quot;:&quot;143412869029&quot;,&quot;pagestimeline&quot;:&quot;1&quot;,&quot;payload_type&quot;:&quot;7&quot;,&quot;eventtime&quot;:&quot;1352998853&quot;,&quot;viewerid&quot;:&quot;0&quot;,&quot;profileownerid&quot;:&quot;143412869029&quot;,&quot;unitimpressionid&quot;:&quot;f2d1d9d2&quot;,&quot;contentid&quot;:&quot;-7624703520450932554&quot;,&quot;timeline_unit_type&quot;:&quot;AddSinglePhotoUnit&quot;,&quot;timewindowsize&quot;:&quot;2&quot;,&quot;contextwindowstart&quot;:&quot;1349074800&quot;,&quot;contextwindowend&quot;:&quot;1351753199&quot;&#125;\" data-time=\"1351600566\"><div class=\"\"><div role=\"article\"><div class=\"clearfix mbs pbs _1_m \"><a class=\"_29h _303\" href=\"http://www.facebook.com/NorthCarolinaStateParks\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-ash3/174880_143412869029_133040642_q.jpg\" alt=\"\" /></a><div class=\"_3dp _29k\"><h5 class=\"_1_s\"><span class=\"fcg\"><span class=\"fwb\"><a href=\"http://www.facebook.com/NorthCarolinaStateParks\">North Carolina State Parks and Recreation</a></span></span></h5><div class=\"_1_n fsm fwn fcg\"><a class=\"uiLinkSubtle\" href=\"http://www.facebook.com/photo.php?fbid=10151234924589030&amp;set=a.164247464029.118730.143412869029&amp;type=1\"><abbr title=\"Tuesday, October 30, 2012 at 4:36am\" data-utime=\"1351596966\">October 30</abbr></a><a data-hover=\"tooltip\" aria-label=\"Public\" class=\"uiStreamPrivacy inlineBlock fbStreamPrivacy fbPrivacyAudienceIndicator _1_o\" href=\"#\"><i class=\"lock img sp_2voup5 sx_4929d1\"></i></a></div></div></div><div class=\"aboveUnitContent\"><div class=\"userContentWrapper\"><div class=\"_wk\"><span class=\"userContent\">The hike with a ranger was the best part after a kid-friendly Track Trail was dedicated at Stone Mountain State Park Saturday. It&#039;s one of a network of trails in western NC to entice youngsters into nature.</span></div></div></div><div class=\"photoUnit clearfix\"><div class=\"_53s uiScaledThumb photo photoWidth1\" data-cropped=\"1\"><a class=\"_6i9\" href=\"http://www.facebook.com/photo.php?fbid=10151234924589030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1\" rel=\"theater\" ajaxify=\"http://www.facebook.com/photo.php?fbid=10151234924589030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1&amp;src=https%3A%2F%2Fsphotos-b.xx.fbcdn.net%2Fhphotos-prn1%2F76656_10151234924589030_897374255_n.jpg&amp;size=798%2C914&amp;theater\"><div class=\"uiScaledImageContainer photoWrap\"><img class=\"scaledImageFitWidth img\" src=\"https://sphotos-b.xx.fbcdn.net/hphotos-prn1/c0.0.403.403/p403x403/76656_10151234924589030_897374255_n.jpg\" alt=\"Photo: The hike with a ranger was the best part after a kid-friendly Track Trail was dedicated at Stone Mountain State Park Saturday. It&#039;s one of a network of trails in western NC to entice youngsters into nature.\" width=\"403\" height=\"403\" /></div></a></div></div></div><div class=\"fbTimelineUFI uiCommentContainer\"><form rel=\"async\" class=\"commentable_item autoexpand_mode\" method=\"post\" action=\"/ajax/ufi/modify.php\" onsubmit=\"return window.Event &amp;&amp; Event.__inlineSubmit &amp;&amp; Event.__inlineSubmit(this,event)\" id=\"u4kzrn2166\"><input type=\"hidden\" name=\"charset_test\" value=\"&euro;,&acute;,\u20AC,\u00B4,\u6C34,\u0414,\u0404\" /><input type=\"hidden\" name=\"lsd\" value=\"AVqaEiN6\" autocomplete=\"off\" /><input type=\"hidden\" autocomplete=\"off\" name=\"feedback_params\" value=\"&#123;&quot;actor&quot;:&quot;143412869029&quot;,&quot;target_fbid&quot;:&quot;10151234924589030&quot;,&quot;target_profile_id&quot;:&quot;143412869029&quot;,&quot;type_id&quot;:&quot;7&quot;,&quot;assoc_obj_id&quot;:&quot;&quot;,&quot;source_app_id&quot;:&quot;0&quot;,&quot;extra_story_params&quot;:[],&quot;content_timestamp&quot;:&quot;1351600566&quot;,&quot;check_hash&quot;:&quot;AQASAujje4BnIDs9&quot;,&quot;source&quot;:&quot;0&quot;&#125;\" /><input type=\"hidden\" autocomplete=\"off\" name=\"data_only_response\" value=\"1\" /><input type=\"hidden\" autocomplete=\"off\" name=\"timeline_ufi\" value=\"1\" /><input type=\"hidden\" name=\"timeline_log_data\" value=\"AQDGs6WfBBsxWWPB21-mGcg-2nqscDD3NDO4Z4uJMrPuyIEQnsKJp1unJaFr21DAvCNLnOMCKbIojSWG3fkkbWiHJXeR3dHR0ROHVVBRb_OSjvBBanrNBi5u56frQvXcxUFEBpaSXjMi2bOiEB2RhAcQ93UPPIAGaZXUPFroemGfOTxYLxEl8h9TRUPwX1_OHtxp070v9N7PozLEpxl34kfJMtcCxOuaIXeRc328qAaSGJRDOBld8nCCF2MeV8tEWz5YBiXQaBivoSDR51TFZvd22v9-_ysaGYodbf3sifaJozfhd-GVBHsMa-VyVNB_d9_84YQ2uCpWceeaZlw5DHBF-FnPmQgKQa7TD1iJVWiHWp1CtopkbOaH7y4EM1T01Xw\" /><div class=\"fbTimelineFeedbackHeader\"><div class=\"fbTimelineFeedbackActions clearfix\"><span class=\"UIActionLinks UIActionLinks_bottom\" data-ft=\"&#123;&quot;tn&quot;:&quot;=&quot;,&quot;type&quot;:20&#125;\"><a class=\"like_link\" title=\"Like this item\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151234924589030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Like</a> &#183; <a class=\"comment_link\" title=\"Leave a comment\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151234924589030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Comment</a> &#183; <a rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151234924589030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Share</a></span></div></div><div><div class=\"uiUfi UFIContainer utn\" id=\"u4kzrn2136\"></div></div></form></div></div></div><i class=\"spinePointer\"></i><div class=\"bottomBorder\"></div></li>",
        "<li class=\"fbTimelineUnit fbTimelineTwoColumn clearfix\" data-side=\"l\" data-fixed=\"1\" data-size=\"1\" id=\"tl_unit_8160881010165548025\"><div class=\"topBorder\"></div><div class=\"timelineUnitContainer\" id=\"u4kzrn266\" data-gt=\"&#123;&quot;action_type&quot;:&quot;1&quot;,&quot;aggregation_id&quot;:&quot;10151233722359030&quot;,&quot;aggregation_time&quot;:&quot;1351532717&quot;,&quot;actor&quot;:&quot;143412869029&quot;,&quot;creator&quot;:&quot;143412869029&quot;,&quot;pagestimeline&quot;:&quot;1&quot;,&quot;payload_type&quot;:&quot;7&quot;,&quot;eventtime&quot;:&quot;1352998853&quot;,&quot;viewerid&quot;:&quot;0&quot;,&quot;profileownerid&quot;:&quot;143412869029&quot;,&quot;unitimpressionid&quot;:&quot;f2d1d9d2&quot;,&quot;contentid&quot;:&quot;8160881010165548025&quot;,&quot;timeline_unit_type&quot;:&quot;AddSinglePhotoUnit&quot;,&quot;timewindowsize&quot;:&quot;2&quot;,&quot;contextwindowstart&quot;:&quot;1349074800&quot;,&quot;contextwindowend&quot;:&quot;1351753199&quot;&#125;\" data-time=\"1351532717\"><div class=\"\"><div role=\"article\"><div class=\"clearfix mbs pbs _1_m \"><a class=\"_29h _303\" href=\"http://www.facebook.com/NorthCarolinaStateParks\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-ash3/174880_143412869029_133040642_q.jpg\" alt=\"\" /></a><div class=\"_3dp _29k\"><h5 class=\"_1_s\"><span class=\"fcg\"><span class=\"fwb\"><a href=\"http://www.facebook.com/NorthCarolinaStateParks\">North Carolina State Parks and Recreation</a></span></span></h5><div class=\"_1_n fsm fwn fcg\"><a class=\"uiLinkSubtle\" href=\"http://www.facebook.com/photo.php?fbid=10151233722359030&amp;set=a.164247464029.118730.143412869029&amp;type=1\"><abbr title=\"Monday, October 29, 2012 at 9:45am\" data-utime=\"1351529117\">October 29</abbr></a><a data-hover=\"tooltip\" aria-label=\"Public\" class=\"uiStreamPrivacy inlineBlock fbStreamPrivacy fbPrivacyAudienceIndicator _1_o\" href=\"#\"><i class=\"lock img sp_2voup5 sx_4929d1\"></i></a></div></div></div><div class=\"aboveUnitContent\"><div class=\"userContentWrapper\"><div class=\"_wk\"><span class=\"userContent\">Our congratulations to the new Friends of Morrow Mountain State Park which organized last week. If you love the park and/or live in the area, consider joining.</span></div></div></div><div class=\"photoUnit clearfix\"><div class=\"_53s uiScaledThumb photo photoWidth1\" data-cropped=\"1\"><a class=\"_6i9\" href=\"http://www.facebook.com/photo.php?fbid=10151233722359030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1\" rel=\"theater\" ajaxify=\"http://www.facebook.com/photo.php?fbid=10151233722359030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1&amp;src=https%3A%2F%2Fsphotos-a.xx.fbcdn.net%2Fhphotos-prn1%2F22630_10151233722359030_1415002181_n.jpg&amp;size=799%2C571&amp;theater\"><div class=\"uiScaledImageContainer photoWrap\"><img class=\"scaledImageFitWidth img\" src=\"https://sphotos-a.xx.fbcdn.net/hphotos-prn1/c79.0.403.403/p403x403/22630_10151233722359030_1415002181_n.jpg\" alt=\"Photo: Our congratulations to the new Friends of Morrow Mountain State Park which organized last week. If you love the park and/or live in the area, consider joining.\" width=\"403\" height=\"403\" /></div></a></div></div></div><div class=\"fbTimelineUFI uiCommentContainer\"><form rel=\"async\" class=\"commentable_item autoexpand_mode\" method=\"post\" action=\"/ajax/ufi/modify.php\" onsubmit=\"return window.Event &amp;&amp; Event.__inlineSubmit &amp;&amp; Event.__inlineSubmit(this,event)\" id=\"u4kzrn2167\"><input type=\"hidden\" name=\"charset_test\" value=\"&euro;,&acute;,\u20AC,\u00B4,\u6C34,\u0414,\u0404\" /><input type=\"hidden\" name=\"lsd\" value=\"AVqaEiN6\" autocomplete=\"off\" /><input type=\"hidden\" autocomplete=\"off\" name=\"feedback_params\" value=\"&#123;&quot;actor&quot;:&quot;143412869029&quot;,&quot;target_fbid&quot;:&quot;10151233722359030&quot;,&quot;target_profile_id&quot;:&quot;143412869029&quot;,&quot;type_id&quot;:&quot;7&quot;,&quot;assoc_obj_id&quot;:&quot;&quot;,&quot;source_app_id&quot;:&quot;0&quot;,&quot;extra_story_params&quot;:[],&quot;content_timestamp&quot;:&quot;1351532717&quot;,&quot;check_hash&quot;:&quot;AQBXfvHJLfTOEP8Y&quot;,&quot;source&quot;:&quot;0&quot;&#125;\" /><input type=\"hidden\" autocomplete=\"off\" name=\"data_only_response\" value=\"1\" /><input type=\"hidden\" autocomplete=\"off\" name=\"timeline_ufi\" value=\"1\" /><input type=\"hidden\" name=\"timeline_log_data\" value=\"AQDnFwuyVIiJwfeZUy9stHYfda8iYrNEtChJ_hL-YHn0T7Mu0BAr2SIAn31-PAmHH4HGlmXRuhDOM1iLmGnW8qBUoFWvS6fGFkgGqq8srwmmJ13n_2rnITxsn9tMXJD_0OnV6uNu157OWAa9CoB_iBvYHaZy2i1KKjZ5Iovf2Fs813S3JFPz16ZvAQLu_nRHxCP93Q8yadQXKTTjOnBEfvqvZ_l4tKoTj99NMS3zyzDO8uLrO_u5uKtmm5Lkqn_y0-sfm-0qq_-C16xmrHUAQ5AOXNgi-rZcm81HQW02jkoSKLbgPT8FM0u8HNL_66MpwsXVH9WlJqHUux5Xd0kE-7V2v5Y41aKjnLY9eD6I6F6cnSepQPDyC0jl0K5OyRHxt2c\" /><div class=\"fbTimelineFeedbackHeader\"><div class=\"fbTimelineFeedbackActions clearfix\"><span class=\"fbTimelineFeedbackShares\"><a data-hover=\"tooltip\" aria-label=\"Show shares\" data-tooltip-alignh=\"right\" href=\"/shares/view?id=10151233722359030\" ajaxify=\"/ajax/shares/view?target_fbid=10151233722359030\" rel=\"async\"><i></i>4</a></span><span class=\"UIActionLinks UIActionLinks_bottom\" data-ft=\"&#123;&quot;tn&quot;:&quot;=&quot;,&quot;type&quot;:20&#125;\"><a class=\"like_link\" title=\"Like this item\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151233722359030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Like</a> &#183; <a class=\"comment_link\" title=\"Leave a comment\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151233722359030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Comment</a> &#183; <a rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151233722359030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Share</a></span></div></div><div><div class=\"uiUfi UFIContainer utn\" id=\"u4kzrn299\"></div></div></form></div></div></div><i class=\"spinePointer\"></i><div class=\"bottomBorder\"></div></li><li class=\"fbTimelineUnit fbTimelineTwoColumn clearfix\" data-side=\"l\" data-fixed=\"1\" data-size=\"1\" id=\"tl_unit_-5274868237973642515\"><div class=\"topBorder\"></div><div class=\"timelineUnitContainer\" id=\"u4kzrn268\" data-gt=\"&#123;&quot;action_type&quot;:&quot;1&quot;,&quot;aggregation_id&quot;:&quot;196718357130532&quot;,&quot;aggregation_time&quot;:&quot;1351519437&quot;,&quot;actor&quot;:&quot;143412869029&quot;,&quot;creator&quot;:&quot;143412869029&quot;,&quot;pagestimeline&quot;:&quot;1&quot;,&quot;payload_type&quot;:&quot;5&quot;,&quot;eventtime&quot;:&quot;1352998853&quot;,&quot;viewerid&quot;:&quot;0&quot;,&quot;profileownerid&quot;:&quot;143412869029&quot;,&quot;unitimpressionid&quot;:&quot;f2d1d9d2&quot;,&quot;contentid&quot;:&quot;-5274868237973642515&quot;,&quot;timeline_unit_type&quot;:&quot;BestPostedLinkUnit&quot;,&quot;timewindowsize&quot;:&quot;2&quot;,&quot;contextwindowstart&quot;:&quot;1349074800&quot;,&quot;contextwindowend&quot;:&quot;1351753199&quot;&#125;\" data-time=\"1351519437\"><div class=\"\"><div role=\"article\"><div class=\"clearfix mbs pbs _1_m \"><a class=\"_29h _303\" href=\"http://www.facebook.com/NorthCarolinaStateParks\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-ash3/174880_143412869029_133040642_q.jpg\" alt=\"\" /></a><div class=\"_3dp _29k\"><h5 class=\"_1_s\"><span class=\"fcg\"><span class=\"fwb\"><a href=\"http://www.facebook.com/NorthCarolinaStateParks\">North Carolina State Parks and Recreation</a></span> shared a <a href=\"http://www.facebook.com/l.php?u=https%3A%2F%2Ftwitter.com%2FNCparks&amp;h=oAQHrW15DAQEPrqnPf8P3YzZGTwgNjf0qWUu6GRnqREtxdw&amp;enc=AZPZt8g-c7Fft-dHYjNrXeYzFjl12eA4YEMsPd1RT9b_M9VeiDc98reMlN6X1R93gO2c4oEL1SIz1UpgAJ80ouW0&amp;s=1\" rel=\"nofollow\" onmouseover=\"LinkshimAsyncLink.swap(this, &quot;https:\\\\/\\\\/twitter.com\\\\/NCparks&quot;);\" onclick=\"LinkshimAsyncLink.referrer_log(this, &quot;https:\\\\/\\\\/twitter.com\\\\/NCparks&quot;, &quot;http:\\\\/\\\\/www.facebook.com\\\\/si\\\\/ajax\\\\/l\\\\/render_linkshim_log\\\\/?u=https\\\\u00253A\\\\u00252F\\\\u00252Ftwitter.com\\\\u00252FNCparks&amp;h=oAQHrW15DAQEPrqnPf8P3YzZGTwgNjf0qWUu6GRnqREtxdw&amp;enc=AZPZt8g-c7Fft-dHYjNrXeYzFjl12eA4YEMsPd1RT9b_M9VeiDc98reMlN6X1R93gO2c4oEL1SIz1UpgAJ80ouW0&amp;s=1&quot;);\" target=\"_blank\">link</a>.</span></h5><div class=\"_1_n fsm fwn fcg\"><a class=\"uiLinkSubtle\" href=\"/NorthCarolinaStateParks/posts/196718357130532\"><abbr title=\"Monday, October 29, 2012 at 6:03am\" data-utime=\"1351515837\">October 29</abbr></a><a data-hover=\"tooltip\" aria-label=\"Public\" class=\"uiStreamPrivacy inlineBlock fbStreamPrivacy fbPrivacyAudienceIndicator _1_o\" href=\"#\"><i class=\"lock img sp_2voup5 sx_4929d1\"></i></a></div></div></div><div class=\"userContentWrapper aboveUnitContent\"><div class=\"_wk mbm\"><span class=\"userContent\">Three parks - Dismal Swamp, Jockey&#039;s Ridge, Merchants Millpond - remain closed due to Hurricane Sandy. In the west, high winds have closed Chimney Rock and snow may affect other western parks. Stay tuned for the latest with our Twitter feed: <a href=\"http://www.facebook.com/l.php?u=https%3A%2F%2Ftwitter.com%2FNCparks&amp;h=8AQGAIFeF&amp;s=1\" target=\"_blank\" rel=\"nofollow nofollow\" onmouseover=\"LinkshimAsyncLink.swap(this, &quot;https:\\\\/\\\\/twitter.com\\\\/NCparks&quot;);\" onclick=\"LinkshimAsyncLink.referrer_log(this, &quot;https:\\\\/\\\\/twitter.com\\\\/NCparks&quot;, &quot;http:\\\\/\\\\/www.facebook.com\\\\/si\\\\/ajax\\\\/l\\\\/render_linkshim_log\\\\/?u=https\\\\u00253A\\\\u00252F\\\\u00252Ftwitter.com\\\\u00252FNCparks&amp;h=8AQGAIFeF&amp;s=1&quot;);\">https://twitter.com/NCparks</a></span></div></div><div class=\"shareUnit\"><a class=\"_1xw _1y0\" href=\"http://www.facebook.com/l.php?u=https%3A%2F%2Ftwitter.com%2FNCparks&amp;h=-AQG4vz5aAQH-Egqbfkb4xaG0WPVDPp0nTxtP3I2HAcJSaQ&amp;enc=AZPCrNx9GpWxR3xIJw8ZrcyRVeMPnTU_lFfLdVVRsQ8UHA_lPBsovbwEAcfHKk0dMRU3FdxQTC8A7x-mvy_9q1FA&amp;s=1\" target=\"_blank\" rel=\"nofollow\" onmouseover=\"LinkshimAsyncLink.swap(this, &quot;https:\\\\/\\\\/twitter.com\\\\/NCparks&quot;);\" onclick=\"LinkshimAsyncLink.referrer_log(this, &quot;https:\\\\/\\\\/twitter.com\\\\/NCparks&quot;, &quot;http:\\\\/\\\\/www.facebook.com\\\\/si\\\\/ajax\\\\/l\\\\/render_linkshim_log\\\\/?u=https\\\\u00253A\\\\u00252F\\\\u00252Ftwitter.com\\\\u00252FNCparks&amp;h=-AQG4vz5aAQH-Egqbfkb4xaG0WPVDPp0nTxtP3I2HAcJSaQ&amp;enc=AZPCrNx9GpWxR3xIJw8ZrcyRVeMPnTU_lFfLdVVRsQ8UHA_lPBsovbwEAcfHKk0dMRU3FdxQTC8A7x-mvy_9q1FA&amp;s=1&quot;);\"><div class=\"_1xy _1xx\"><img class=\"img\" src=\"https://fbexternal-a.akamaihd.net/safe_image.php?d=AQDDqkjdbGg0u4Pr&amp;w=155&amp;h=114&amp;url=https%3A%2F%2Fsi0.twimg.com%2Fprofile_images%2F479569785%2Fdprcolor1_reasonably_small.png\" alt=\"\" /></div><div class=\"_1xx _1xz\"><div class=\"_1x-\"><div class=\"_1x_ fwb\">NC State Parks (NCparks) on Twitter</div><div class=\"fsm fwn fcg\">twitter.com</div><div class=\"_1y1 fsm fwn fcg\"></div></div></div></a></div></div><div class=\"fbTimelineUFI uiCommentContainer\"><form rel=\"async\" class=\"commentable_item autoexpand_mode\" method=\"post\" action=\"/ajax/ufi/modify.php\" onsubmit=\"return window.Event &amp;&amp; Event.__inlineSubmit &amp;&amp; Event.__inlineSubmit(this,event)\" id=\"u4kzrn2168\"><input type=\"hidden\" name=\"charset_test\" value=\"&euro;,&acute;,\u20AC,\u00B4,\u6C34,\u0414,\u0404\" /><input type=\"hidden\" name=\"lsd\" value=\"AVqaEiN6\" autocomplete=\"off\" /><input type=\"hidden\" autocomplete=\"off\" name=\"feedback_params\" value=\"&#123;&quot;actor&quot;:&quot;143412869029&quot;,&quot;target_fbid&quot;:&quot;196718357130532&quot;,&quot;target_profile_id&quot;:&quot;143412869029&quot;,&quot;type_id&quot;:&quot;17&quot;,&quot;assoc_obj_id&quot;:&quot;&quot;,&quot;source_app_id&quot;:&quot;0&quot;,&quot;extra_story_params&quot;:[],&quot;content_timestamp&quot;:&quot;1351519437&quot;,&quot;check_hash&quot;:&quot;AQDtnLTn29vO0wo2&quot;,&quot;source&quot;:&quot;0&quot;&#125;\" /><input type=\"hidden\" autocomplete=\"off\" name=\"data_only_response\" value=\"1\" /><input type=\"hidden\" autocomplete=\"off\" name=\"timeline_ufi\" value=\"1\" /><input type=\"hidden\" name=\"timeline_log_data\" value=\"AQA5qb40PSKFdE55h_BmyEKwZcs8vPHEt5Gak0aObjV1eOvsbo-vOUhG7tsOaEWeptNifZhDq8l4kSd2HMSvZUyjfq3JhbwRewPmzFEv5fW5hDAdbnGrDq_XTxI5NjjdyegXpMWWBllFXushcO6xb1SowFMg1aDV_I1mpTW6RL4dQJFu73GfKVib4wiQ-DtX0cgfXKLQL4nMAjS9rar7CMj4bcjt4M4CLKXnVdrR9d5fvRibSwFKYZAvCyrziaJ5DpygaWwUqMSIyRMDC64b6ADrZmBEbKdm-gfxci8uC5XqTOgccc9ZQQFk-1VX_GwS4D3rv2t1A0xCTGeHMMRCv-zmA7L90ictdYKeMK09LWzdU-n3tA7xh1kWLrePCT-rKjI\" /><div class=\"fbTimelineFeedbackHeader\"><div class=\"fbTimelineFeedbackActions clearfix\"><span class=\"fbTimelineFeedbackShares\"><a data-hover=\"tooltip\" aria-label=\"Show shares\" data-tooltip-alignh=\"right\" href=\"/shares/view?id=196718357130532\" ajaxify=\"/ajax/shares/view?target_fbid=196718357130532\" rel=\"async\"><i></i>2</a></span><span class=\"UIActionLinks UIActionLinks_bottom\" data-ft=\"&#123;&quot;tn&quot;:&quot;=&quot;,&quot;type&quot;:20&#125;\"><a class=\"like_link\" title=\"Like this item\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2FNorthCarolinaStateParks%2Fposts%2F196718357130532\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Like</a> &#183; <a class=\"comment_link\" title=\"Leave a comment\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2FNorthCarolinaStateParks%2Fposts%2F196718357130532\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Comment</a> &#183; <a rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2FNorthCarolinaStateParks%2Fposts%2F196718357130532\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Share</a></span></div></div><div><div class=\"uiUfi UFIContainer utn\" id=\"u4kzrn2137\"></div></div></form></div></div></div><i class=\"spinePointer\"></i><div class=\"bottomBorder\"></div></li><li class=\"fbTimelineUnit fbTimelineTwoColumn clearfix\" data-side=\"l\" data-fixed=\"1\" data-size=\"1\" id=\"tl_unit_-1161764595458730036\"><div class=\"topBorder\"></div><div class=\"timelineUnitContainer\" id=\"u4kzrn270\" data-gt=\"&#123;&quot;action_type&quot;:&quot;1&quot;,&quot;aggregation_id&quot;:&quot;455084064537202&quot;,&quot;aggregation_time&quot;:&quot;1351514075&quot;,&quot;actor&quot;:&quot;143412869029&quot;,&quot;creator&quot;:&quot;143412869029&quot;,&quot;pagestimeline&quot;:&quot;1&quot;,&quot;payload_type&quot;:&quot;5&quot;,&quot;eventtime&quot;:&quot;1352998853&quot;,&quot;viewerid&quot;:&quot;0&quot;,&quot;profileownerid&quot;:&quot;143412869029&quot;,&quot;unitimpressionid&quot;:&quot;f2d1d9d2&quot;,&quot;contentid&quot;:&quot;-1161764595458730036&quot;,&quot;timeline_unit_type&quot;:&quot;BestPostedLinkUnit&quot;,&quot;timewindowsize&quot;:&quot;2&quot;,&quot;contextwindowstart&quot;:&quot;1349074800&quot;,&quot;contextwindowend&quot;:&quot;1351753199&quot;&#125;\" data-time=\"1351514075\"><div class=\"\"><div role=\"article\"><div class=\"clearfix mbs pbs _1_m \"><a class=\"_29h _303\" href=\"http://www.facebook.com/NorthCarolinaStateParks\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-ash3/174880_143412869029_133040642_q.jpg\" alt=\"\" /></a><div class=\"_3dp _29k\"><h5 class=\"_1_s\"><span class=\"fcg\"><span class=\"fwb\"><a href=\"http://www.facebook.com/NorthCarolinaStateParks\">North Carolina State Parks and Recreation</a></span> shared a <a href=\"http://www.facebook.com/l.php?u=http%3A%2F%2Fderscott.blogspot.com%2F2012%2F10%2Fson-of-uwharrie-eno-river-run-11-miler.html&amp;h=CAQF8Bvp8AQEjAkd6cqYTV0bwHUF81GsQ6zdTogrAitZsgg&amp;enc=AZPdNFzWtZGz3EgeqrlV08bIW5LHMaPwBxVufhZwkGOU7ctKSmDqAIoVe32FHg3uXGBG9n6prwdj8TQ6oouoTNch&amp;s=1\" rel=\"nofollow\" onmouseover=\"LinkshimAsyncLink.swap(this, &quot;http:\\\\/\\\\/derscott.blogspot.com\\\\/2012\\\\/10\\\\/son-of-uwharrie-eno-river-run-11-miler.html&quot;);\" onclick=\"LinkshimAsyncLink.referrer_log(this, &quot;http:\\\\/\\\\/derscott.blogspot.com\\\\/2012\\\\/10\\\\/son-of-uwharrie-eno-river-run-11-miler.html&quot;, &quot;http:\\\\/\\\\/www.facebook.com\\\\/si\\\\/ajax\\\\/l\\\\/render_linkshim_log\\\\/?u=http\\\\u00253A\\\\u00252F\\\\u00252Fderscott.blogspot.com\\\\u00252F2012\\\\u00252F10\\\\u00252Fson-of-uwharrie-eno-river-run-11-miler.html&amp;h=CAQF8Bvp8AQEjAkd6cqYTV0bwHUF81GsQ6zdTogrAitZsgg&amp;enc=AZPdNFzWtZGz3EgeqrlV08bIW5LHMaPwBxVufhZwkGOU7ctKSmDqAIoVe32FHg3uXGBG9n6prwdj8TQ6oouoTNch&amp;s=1&quot;);\" target=\"_blank\">link</a>.</span></h5><div class=\"_1_n fsm fwn fcg\"><a class=\"uiLinkSubtle\" href=\"/NorthCarolinaStateParks/posts/455084064537202\"><abbr title=\"Monday, October 29, 2012 at 4:34am\" data-utime=\"1351510475\">October 29</abbr></a><a data-hover=\"tooltip\" aria-label=\"Public\" class=\"uiStreamPrivacy inlineBlock fbStreamPrivacy fbPrivacyAudienceIndicator _1_o\" href=\"#\"><i class=\"lock img sp_2voup5 sx_4929d1\"></i></a></div></div></div><div class=\"userContentWrapper aboveUnitContent\"><div class=\"_wk mbm\"><span class=\"userContent\">The first-ever Eno River Run was Saturday...short but a bit brutal according to this blogger: <a href=\"http://www.facebook.com/l.php?u=http%3A%2F%2Fderscott.blogspot.com%2F2012%2F10%2Fson-of-uwharrie-eno-river-run-11-miler.html&amp;h=HAQGKveDe&amp;s=1\" target=\"_blank\" rel=\"nofollow nofollow\" onmouseover=\"LinkshimAsyncLink.swap(this, &quot;http:\\\\/\\\\/derscott.blogspot.com\\\\/2012\\\\/10\\\\/son-of-uwharrie-eno-river-run-11-miler.html&quot;);\" onclick=\"LinkshimAsyncLink.referrer_log(this, &quot;http:\\\\/\\\\/derscott.blogspot.com\\\\/2012\\\\/10\\\\/son-of-uwharrie-eno-river-run-11-miler.html&quot;, &quot;http:\\\\/\\\\/www.facebook.com\\\\/si\\\\/ajax\\\\/l\\\\/render_linkshim_log\\\\/?u=http\\\\u00253A\\\\u00252F\\\\u00252Fderscott.blogspot.com\\\\u00252F2012\\\\u00252F10\\\\u00252Fson-of-uwharrie-eno-river-run-11-miler.html&amp;h=HAQGKveDe&amp;s=1&quot;);\">http://derscott.blogspot.com/2012/10/son-of-uwharrie-eno-river-run-11-miler.html</a></span></div></div><div class=\"shareUnit\"><a class=\"_1xw _1y0\" href=\"http://www.facebook.com/l.php?u=http%3A%2F%2Fderscott.blogspot.com%2F2012%2F10%2Fson-of-uwharrie-eno-river-run-11-miler.html&amp;h=MAQEyZKBuAQEItDA-XeBtfn9vchVILMBhU3qg_vDzXwzOGw&amp;enc=AZOG9quv6DGMmbJYcvwXW-tV2XlvVLry2ugPXlh8AruE0bPLFANrFt1Dj_m7iIKrlMi8v-u1dy4LysBEjGf0FRQ0&amp;s=1\" target=\"_blank\" rel=\"nofollow\" onmouseover=\"LinkshimAsyncLink.swap(this, &quot;http:\\\\/\\\\/derscott.blogspot.com\\\\/2012\\\\/10\\\\/son-of-uwharrie-eno-river-run-11-miler.html&quot;);\" onclick=\"LinkshimAsyncLink.referrer_log(this, &quot;http:\\\\/\\\\/derscott.blogspot.com\\\\/2012\\\\/10\\\\/son-of-uwharrie-eno-river-run-11-miler.html&quot;, &quot;http:\\\\/\\\\/www.facebook.com\\\\/si\\\\/ajax\\\\/l\\\\/render_linkshim_log\\\\/?u=http\\\\u00253A\\\\u00252F\\\\u00252Fderscott.blogspot.com\\\\u00252F2012\\\\u00252F10\\\\u00252Fson-of-uwharrie-eno-river-run-11-miler.html&amp;h=MAQEyZKBuAQEItDA-XeBtfn9vchVILMBhU3qg_vDzXwzOGw&amp;enc=AZOG9quv6DGMmbJYcvwXW-tV2XlvVLry2ugPXlh8AruE0bPLFANrFt1Dj_m7iIKrlMi8v-u1dy4LysBEjGf0FRQ0&amp;s=1&quot;);\"><div class=\"_1xy _1xx\"><img class=\"img\" src=\"https://fbexternal-a.akamaihd.net/safe_image.php?d=AQDQ3aNncnHM3Vih&amp;w=155&amp;h=114&amp;url=http%3A%2F%2F1.bp.blogspot.com%2F-JODQ9Z2HDH4%2FUIwzUzgu2uI%2FAAAAAAAAM18%2F77-ur0jPLdE%2Fs72-c%2FEno_River_Run.png\" alt=\"\" /></div><div class=\"_1xx _1xz\"><div class=\"_1x-\"><div class=\"_1x_ fwb\">Der Scott: Son of Uwharrie - Eno River Run 11 Miler</div><div class=\"fsm fwn fcg\">derscott.blogspot.com</div><div class=\"_1y1 fsm fwn fcg\"></div></div></div></a></div></div><div class=\"fbTimelineUFI uiCommentContainer\"><form rel=\"async\" class=\"commentable_item autoexpand_mode\" method=\"post\" action=\"/ajax/ufi/modify.php\" onsubmit=\"return window.Event &amp;&amp; Event.__inlineSubmit &amp;&amp; Event.__inlineSubmit(this,event)\" id=\"u4kzrn2169\"><input type=\"hidden\" name=\"charset_test\" value=\"&euro;,&acute;,\u20AC,\u00B4,\u6C34,\u0414,\u0404\" /><input type=\"hidden\" name=\"lsd\" value=\"AVqaEiN6\" autocomplete=\"off\" /><input type=\"hidden\" autocomplete=\"off\" name=\"feedback_params\" value=\"&#123;&quot;actor&quot;:&quot;143412869029&quot;,&quot;target_fbid&quot;:&quot;455084064537202&quot;,&quot;target_profile_id&quot;:&quot;143412869029&quot;,&quot;type_id&quot;:&quot;17&quot;,&quot;assoc_obj_id&quot;:&quot;&quot;,&quot;source_app_id&quot;:&quot;0&quot;,&quot;extra_story_params&quot;:[],&quot;content_timestamp&quot;:&quot;1351514075&quot;,&quot;check_hash&quot;:&quot;AQBVG4YTJe3xvDxz&quot;,&quot;source&quot;:&quot;0&quot;&#125;\" /><input type=\"hidden\" autocomplete=\"off\" name=\"data_only_response\" value=\"1\" /><input type=\"hidden\" autocomplete=\"off\" name=\"timeline_ufi\" value=\"1\" /><input type=\"hidden\" name=\"timeline_log_data\" value=\"AQCD4RV4MW1YWLwLMGiavl4UAfheCCNXlW77qmI3PjoQwlsQEjrvG2G1zaEME4ARqXuutKRMIUnJIkhRTgsvShqSoL04BsMTbFN19iyQdVGf_KM5PrQDm_t8Rujybz5hmsZ03wkwCjRAc-yQorowYoNiWL4NdqvGMg_MLx6vPvHoLqGlaNn7Nt_o0x-7HLATGD55M9i93uSJ_lSYvbjwU6jlKYk2FsZuY_TtXzPo88__1276RYZJsFwBdAxohRSDIhGGMcYDzXKxTv3Pgo8yfMaCFS7nB9LPL-elPfVacKl6ZoSW0fC5Ur4k0_pam3gW1DV1AAk5kXc_2Q60GqUo_dH11hS4RJwzP_wSdOOadfzm1Hj9DIxjj8wGpVCKzFbe_Qk\" /><div class=\"fbTimelineFeedbackHeader\"><div class=\"fbTimelineFeedbackActions clearfix\"><span class=\"UIActionLinks UIActionLinks_bottom\" data-ft=\"&#123;&quot;tn&quot;:&quot;=&quot;,&quot;type&quot;:20&#125;\"><a class=\"like_link\" title=\"Like this item\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2FNorthCarolinaStateParks%2Fposts%2F455084064537202\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Like</a> &#183; <a class=\"comment_link\" title=\"Leave a comment\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2FNorthCarolinaStateParks%2Fposts%2F455084064537202\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Comment</a> &#183; <a rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2FNorthCarolinaStateParks%2Fposts%2F455084064537202\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Share</a></span></div></div><div><div class=\"uiUfi UFIContainer utn\" id=\"u4kzrn295\"></div></div></form></div></div></div><i class=\"spinePointer\"></i><div class=\"bottomBorder\"></div></li><li class=\"fbTimelineUnit fbTimelineTwoColumn clearfix\" data-side=\"l\" data-fixed=\"1\" data-size=\"1\" id=\"tl_unit_118442752173505026\"><div class=\"topBorder\"></div><div class=\"timelineUnitContainer\" id=\"u4kzrn271\" data-gt=\"&#123;&quot;action_type&quot;:&quot;1&quot;,&quot;aggregation_id&quot;:&quot;10151231518534030&quot;,&quot;aggregation_time&quot;:&quot;1351429095&quot;,&quot;actor&quot;:&quot;143412869029&quot;,&quot;creator&quot;:&quot;143412869029&quot;,&quot;pagestimeline&quot;:&quot;1&quot;,&quot;payload_type&quot;:&quot;7&quot;,&quot;eventtime&quot;:&quot;1352998853&quot;,&quot;viewerid&quot;:&quot;0&quot;,&quot;profileownerid&quot;:&quot;143412869029&quot;,&quot;unitimpressionid&quot;:&quot;f2d1d9d2&quot;,&quot;contentid&quot;:&quot;118442752173505026&quot;,&quot;timeline_unit_type&quot;:&quot;AddSinglePhotoUnit&quot;,&quot;timewindowsize&quot;:&quot;2&quot;,&quot;contextwindowstart&quot;:&quot;1349074800&quot;,&quot;contextwindowend&quot;:&quot;1351753199&quot;&#125;\" data-time=\"1351429095\"><div class=\"\"><div role=\"article\"><div class=\"clearfix mbs pbs _1_m \"><a class=\"_29h _303\" href=\"http://www.facebook.com/NorthCarolinaStateParks\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-ash3/174880_143412869029_133040642_q.jpg\" alt=\"\" /></a><div class=\"_3dp _29k\"><h5 class=\"_1_s\"><span class=\"fcg\"><span class=\"fwb\"><a href=\"http://www.facebook.com/NorthCarolinaStateParks\">North Carolina State Parks and Recreation</a></span></span></h5><div class=\"_1_n fsm fwn fcg\"><a class=\"uiLinkSubtle\" href=\"http://www.facebook.com/photo.php?fbid=10151231518534030&amp;set=a.164247464029.118730.143412869029&amp;type=1\"><abbr title=\"Sunday, October 28, 2012 at 4:58am\" data-utime=\"1351425495\">October 28</abbr></a><a data-hover=\"tooltip\" aria-label=\"Public\" class=\"uiStreamPrivacy inlineBlock fbStreamPrivacy fbPrivacyAudienceIndicator _1_o\" href=\"#\"><i class=\"lock img sp_2voup5 sx_4929d1\"></i></a></div></div></div><div class=\"aboveUnitContent\"><div class=\"userContentWrapper\"><div class=\"_wk\"><span class=\"userContent\">It&#039;s getting rough on the coast. Jockey&#039;s Ridge and Dismal Swamp state parks are closed until further notice. Before heading to a state park on or near the coast, call the park office or monitor our website: <a href=\"http://www.facebook.com/l.php?u=http%3A%2F%2Fwww.ncparks.gov&amp;h=qAQGVDcGZ&amp;s=1\" target=\"_blank\" rel=\"nofollow nofollow\" onmouseover=\"LinkshimAsyncLink.swap(this, &quot;http:\\\\/\\\\/www.ncparks.gov\\\\/&quot;);\" onclick=\"LinkshimAsyncLink.referrer_log(this, &quot;http:\\\\/\\\\/www.ncparks.gov\\\\/&quot;, &quot;http:\\\\/\\\\/www.facebook.com\\\\/si\\\\/ajax\\\\/l\\\\/render_linkshim_log\\\\/?u=http\\\\u00253A\\\\u00252F\\\\u00252Fwww.ncparks.gov&amp;h=qAQGVDcGZ&amp;s=1&quot;);\">www.ncparks.gov</a></span></div></div></div><div class=\"photoUnit clearfix\"><div class=\"_53s uiScaledThumb photo photoWidth1\" data-cropped=\"1\"><a class=\"_6i9\" href=\"http://www.facebook.com/photo.php?fbid=10151231518534030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1\" rel=\"theater\" ajaxify=\"http://www.facebook.com/photo.php?fbid=10151231518534030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1&amp;src=https%3A%2F%2Fsphotos-a.xx.fbcdn.net%2Fhphotos-snc7%2F395079_10151231518534030_773270012_n.jpg&amp;size=600%2C390&amp;theater\"><div class=\"uiScaledImageContainer photoWrap\" style=\"height:390px;\"><img class=\"img\" src=\"https://sphotos-a.xx.fbcdn.net/hphotos-snc7/c98.0.403.403/p403x403/395079_10151231518534030_773270012_n.jpg\" alt=\"Photo: It&#039;s getting rough on the coast. Jockey&#039;s Ridge and Dismal Swamp state parks are closed until further notice. Before heading to a state park on or near the coast, call the park office or monitor our website: www.ncparks.gov\" width=\"403\" height=\"390\" /></div></a></div></div></div><div class=\"fbTimelineUFI uiCommentContainer\"><form rel=\"async\" class=\"commentable_item autoexpand_mode\" method=\"post\" action=\"/ajax/ufi/modify.php\" onsubmit=\"return window.Event &amp;&amp; Event.__inlineSubmit &amp;&amp; Event.__inlineSubmit(this,event)\" id=\"u4kzrn2170\"><input type=\"hidden\" name=\"charset_test\" value=\"&euro;,&acute;,\u20AC,\u00B4,\u6C34,\u0414,\u0404\" /><input type=\"hidden\" name=\"lsd\" value=\"AVqaEiN6\" autocomplete=\"off\" /><input type=\"hidden\" autocomplete=\"off\" name=\"feedback_params\" value=\"&#123;&quot;actor&quot;:&quot;143412869029&quot;,&quot;target_fbid&quot;:&quot;10151231518534030&quot;,&quot;target_profile_id&quot;:&quot;143412869029&quot;,&quot;type_id&quot;:&quot;7&quot;,&quot;assoc_obj_id&quot;:&quot;&quot;,&quot;source_app_id&quot;:&quot;0&quot;,&quot;extra_story_params&quot;:[],&quot;content_timestamp&quot;:&quot;1351429095&quot;,&quot;check_hash&quot;:&quot;AQBQJNxoJZ1HChN3&quot;,&quot;source&quot;:&quot;0&quot;&#125;\" /><input type=\"hidden\" autocomplete=\"off\" name=\"data_only_response\" value=\"1\" /><input type=\"hidden\" autocomplete=\"off\" name=\"timeline_ufi\" value=\"1\" /><input type=\"hidden\" name=\"timeline_log_data\" value=\"AQBuRyXbEs55soMQM5MufbdFMCnG0iDrSP9x34wPVBd_IUpEsO02S_gp9H_NCDrRefjSu74M9uG1PHeDf9EAkwzTyE3Q2mMjjIuM9HFykZVCRmOFITZzlJ2gVDCcpQ4m2bGAyDJLdwqT8JjXFLa-ZWvOzmkwxFaHroDTXHDfAm1BGOEnBZ2szOhVBrdZXQ1N4gLZXAMLzfNOkvk9vg7wcz7304hG8b65GeeNRXxxS8VH5SgoTBzovViwGCEYfX0AtX76w0Vb4ju2FvC2BxMHguDtPRCs6a7D4c33JOSxI3PpdDuFHQ8N1r5IaN21YRXqQsIn9j1i-5wJxbu4cIw0j8wfzAEByp4lMBBNwrBc4Ht4oujc33DOJJwBSyy1BJtCAb0\" /><div class=\"fbTimelineFeedbackHeader\"><div class=\"fbTimelineFeedbackActions clearfix\"><span class=\"fbTimelineFeedbackShares\"><a data-hover=\"tooltip\" aria-label=\"Show shares\" data-tooltip-alignh=\"right\" href=\"/shares/view?id=10151231518534030\" ajaxify=\"/ajax/shares/view?target_fbid=10151231518534030\" rel=\"async\"><i></i>5</a></span><span class=\"UIActionLinks UIActionLinks_bottom\" data-ft=\"&#123;&quot;tn&quot;:&quot;=&quot;,&quot;type&quot;:20&#125;\"><a class=\"like_link\" title=\"Like this item\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151231518534030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Like</a> &#183; <a class=\"comment_link\" title=\"Leave a comment\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151231518534030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Comment</a> &#183; <a rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151231518534030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Share</a></span></div></div><div><div class=\"uiUfi UFIContainer utn\" id=\"u4kzrn297\"></div></div></form></div></div></div><i class=\"spinePointer\"></i><div class=\"bottomBorder\"></div></li><li class=\"fbTimelineUnit fbTimelineTwoColumn clearfix\" data-side=\"l\" data-fixed=\"1\" data-size=\"1\" id=\"tl_unit_-8822072180524338871\"><div class=\"topBorder\"></div><div class=\"timelineUnitContainer\" id=\"u4kzrn272\" data-gt=\"&#123;&quot;action_type&quot;:&quot;1&quot;,&quot;aggregation_id&quot;:&quot;10151228750629030&quot;,&quot;aggregation_time&quot;:&quot;1351272809&quot;,&quot;actor&quot;:&quot;143412869029&quot;,&quot;creator&quot;:&quot;143412869029&quot;,&quot;pagestimeline&quot;:&quot;1&quot;,&quot;payload_type&quot;:&quot;7&quot;,&quot;eventtime&quot;:&quot;1352998853&quot;,&quot;viewerid&quot;:&quot;0&quot;,&quot;profileownerid&quot;:&quot;143412869029&quot;,&quot;unitimpressionid&quot;:&quot;f2d1d9d2&quot;,&quot;contentid&quot;:&quot;-8822072180524338871&quot;,&quot;timeline_unit_type&quot;:&quot;AddSinglePhotoUnit&quot;,&quot;timewindowsize&quot;:&quot;2&quot;,&quot;contextwindowstart&quot;:&quot;1349074800&quot;,&quot;contextwindowend&quot;:&quot;1351753199&quot;&#125;\" data-time=\"1351272809\"><div class=\"\"><div role=\"article\"><div class=\"clearfix mbs pbs _1_m \"><a class=\"_29h _303\" href=\"http://www.facebook.com/NorthCarolinaStateParks\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-ash3/174880_143412869029_133040642_q.jpg\" alt=\"\" /></a><div class=\"_3dp _29k\"><h5 class=\"_1_s\"><span class=\"fcg\"><span class=\"fwb\"><a href=\"http://www.facebook.com/NorthCarolinaStateParks\">North Carolina State Parks and Recreation</a></span></span></h5><div class=\"_1_n fsm fwn fcg\"><a class=\"uiLinkSubtle\" href=\"http://www.facebook.com/photo.php?fbid=10151228750629030&amp;set=a.164247464029.118730.143412869029&amp;type=1\"><abbr title=\"Friday, October 26, 2012 at 9:33am\" data-utime=\"1351269209\">October 26</abbr></a><a data-hover=\"tooltip\" aria-label=\"Public\" class=\"uiStreamPrivacy inlineBlock fbStreamPrivacy fbPrivacyAudienceIndicator _1_o\" href=\"#\"><i class=\"lock img sp_2voup5 sx_4929d1\"></i></a></div></div></div><div class=\"aboveUnitContent\"><div class=\"userContentWrapper\"><div class=\"_wk\"><span class=\"userContent\">Phriday Photo! Just a reminder that not all the great fall color appears in the mountains...these ferns were spotted yesterday at Weymouth Woods in the Sandhills.</span></div></div></div><div class=\"photoUnit clearfix\"><div class=\"_53s uiScaledThumb photo photoWidth1\" data-cropped=\"1\"><a class=\"_6i9\" href=\"http://www.facebook.com/photo.php?fbid=10151228750629030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1\" rel=\"theater\" ajaxify=\"http://www.facebook.com/photo.php?fbid=10151228750629030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1&amp;src=https%3A%2F%2Fsphotos-b.xx.fbcdn.net%2Fhphotos-ash3%2F560758_10151228750629030_1999705611_n.jpg&amp;size=800%2C628&amp;theater\"><div class=\"uiScaledImageContainer photoWrap\"><img class=\"scaledImageFitWidth img\" src=\"https://sphotos-b.xx.fbcdn.net/hphotos-ash3/c55.0.403.403/p403x403/560758_10151228750629030_1999705611_n.jpg\" alt=\"Photo: Phriday Photo! Just a reminder that not all the great fall color appears in the mountains...these ferns were spotted yesterday at Weymouth Woods in the Sandhills.\" width=\"403\" height=\"403\" /></div></a></div></div></div><div class=\"fbTimelineUFI uiCommentContainer\"><form rel=\"async\" class=\"commentable_item autoexpand_mode\" method=\"post\" action=\"/ajax/ufi/modify.php\" onsubmit=\"return window.Event &amp;&amp; Event.__inlineSubmit &amp;&amp; Event.__inlineSubmit(this,event)\" id=\"u4kzrn2171\"><input type=\"hidden\" name=\"charset_test\" value=\"&euro;,&acute;,\u20AC,\u00B4,\u6C34,\u0414,\u0404\" /><input type=\"hidden\" name=\"lsd\" value=\"AVqaEiN6\" autocomplete=\"off\" /><input type=\"hidden\" autocomplete=\"off\" name=\"feedback_params\" value=\"&#123;&quot;actor&quot;:&quot;143412869029&quot;,&quot;target_fbid&quot;:&quot;10151228750629030&quot;,&quot;target_profile_id&quot;:&quot;143412869029&quot;,&quot;type_id&quot;:&quot;7&quot;,&quot;assoc_obj_id&quot;:&quot;&quot;,&quot;source_app_id&quot;:&quot;0&quot;,&quot;extra_story_params&quot;:[],&quot;content_timestamp&quot;:&quot;1351272808&quot;,&quot;check_hash&quot;:&quot;AQCTfIrIBwUbnjIr&quot;,&quot;source&quot;:&quot;0&quot;&#125;\" /><input type=\"hidden\" autocomplete=\"off\" name=\"data_only_response\" value=\"1\" /><input type=\"hidden\" autocomplete=\"off\" name=\"timeline_ufi\" value=\"1\" /><input type=\"hidden\" name=\"timeline_log_data\" value=\"AQDy0RyWEAASiujUkEJO2LK37rsfJ-NYK9j4NTcO80ox0xyCNawOUezRE_SN4QxiKz79Ysf5rsx7WHElt0RczM0NgE6c3IRMVocuQxzYTb5JFDLvDxpv4qyZ9ozv-WeJ_SeyzKbU2BO70EgOg0CrxyqDhwjPAMkRa6elfilpq6Ny_RPgGJTLdniMorR1l5vW0W8YLukNeG-bMM5_oS4Z49P5G9zLnscF-cwLl_vqiqs5oWKbTeRQW8hLxRlieCP32w1O3e5sy8npIP-mkVApOeVANx8BIda1sQVijbwMm_gbhTzTIk5xn7Ch7sTIdU3KkGaJyB0tkac_B4yoCUGFEkmK7L4BnJH-1gfQC0Ms34LXR6KNwuLKeZedl2V49vLxsgI\" /><div class=\"fbTimelineFeedbackHeader\"><div class=\"fbTimelineFeedbackActions clearfix\"><span class=\"fbTimelineFeedbackShares\"><a data-hover=\"tooltip\" aria-label=\"Show shares\" data-tooltip-alignh=\"right\" href=\"/shares/view?id=10151228750629030\" ajaxify=\"/ajax/shares/view?target_fbid=10151228750629030\" rel=\"async\"><i></i>2</a></span><span class=\"UIActionLinks UIActionLinks_bottom\" data-ft=\"&#123;&quot;tn&quot;:&quot;=&quot;,&quot;type&quot;:20&#125;\"><a class=\"like_link\" title=\"Like this item\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151228750629030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Like</a> &#183; <a class=\"comment_link\" title=\"Leave a comment\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151228750629030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Comment</a> &#183; <a rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151228750629030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Share</a></span></div></div><div><div class=\"uiUfi UFIContainer utn\" id=\"u4kzrn2138\"></div></div></form></div></div></div><i class=\"spinePointer\"></i><div class=\"bottomBorder\"></div></li><li class=\"fbTimelineUnit fbTimelineTwoColumn clearfix\" data-side=\"l\" data-fixed=\"1\" data-size=\"1\" id=\"tl_unit_6745945291601565463\"><div class=\"topBorder\"></div><div class=\"timelineUnitContainer\" id=\"u4kzrn273\" data-gt=\"&#123;&quot;action_type&quot;:&quot;1&quot;,&quot;aggregation_id&quot;:&quot;10151228504104030&quot;,&quot;aggregation_time&quot;:&quot;1351261059&quot;,&quot;actor&quot;:&quot;143412869029&quot;,&quot;creator&quot;:&quot;143412869029&quot;,&quot;pagestimeline&quot;:&quot;1&quot;,&quot;payload_type&quot;:&quot;11&quot;,&quot;eventtime&quot;:&quot;1352998853&quot;,&quot;viewerid&quot;:&quot;0&quot;,&quot;profileownerid&quot;:&quot;143412869029&quot;,&quot;unitimpressionid&quot;:&quot;f2d1d9d2&quot;,&quot;contentid&quot;:&quot;6745945291601565463&quot;,&quot;timeline_unit_type&quot;:&quot;StatusMessageUnit&quot;,&quot;timewindowsize&quot;:&quot;2&quot;,&quot;contextwindowstart&quot;:&quot;1349074800&quot;,&quot;contextwindowend&quot;:&quot;1351753199&quot;&#125;\" data-time=\"1351261059\"><div class=\"\"><div role=\"article\"><div class=\"clearfix mbs pbs _1_m \"><a class=\"_29h _303\" href=\"http://www.facebook.com/NorthCarolinaStateParks\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-ash3/174880_143412869029_133040642_q.jpg\" alt=\"\" /></a><div class=\"_3dp _29k\"><h5 class=\"_1_s\"><span class=\"fcg\"><span class=\"fwb\"><a href=\"http://www.facebook.com/NorthCarolinaStateParks\">North Carolina State Parks and Recreation</a></span></span></h5><div class=\"_1_n fsm fwn fcg\"><a class=\"uiLinkSubtle\" href=\"/NorthCarolinaStateParks/posts/10151228504104030\"><abbr title=\"Friday, October 26, 2012 at 6:17am\" data-utime=\"1351257459\">October 26</abbr></a><a data-hover=\"tooltip\" aria-label=\"Public\" class=\"uiStreamPrivacy inlineBlock fbStreamPrivacy fbPrivacyAudienceIndicator _1_o\" href=\"#\"><i class=\"lock img sp_2voup5 sx_4929d1\"></i></a></div></div></div><div class=\"_1x1\"><div class=\"userContentWrapper\"><div class=\"_wk\"><span class=\"userContent\">Dang, the haunted hayride and trunk-or-treat at Cliffs of the Neuse State Park were cancelled, too, due to weather concerns.</span></div></div></div></div><div class=\"fbTimelineUFI uiCommentContainer\"><form rel=\"async\" class=\"commentable_item autoexpand_mode\" method=\"post\" action=\"/ajax/ufi/modify.php\" onsubmit=\"return window.Event &amp;&amp; Event.__inlineSubmit &amp;&amp; Event.__inlineSubmit(this,event)\" id=\"u4kzrn2172\"><input type=\"hidden\" name=\"charset_test\" value=\"&euro;,&acute;,\u20AC,\u00B4,\u6C34,\u0414,\u0404\" /><input type=\"hidden\" name=\"lsd\" value=\"AVqaEiN6\" autocomplete=\"off\" /><input type=\"hidden\" autocomplete=\"off\" name=\"feedback_params\" value=\"&#123;&quot;actor&quot;:&quot;143412869029&quot;,&quot;target_fbid&quot;:&quot;10151228504104030&quot;,&quot;target_profile_id&quot;:&quot;143412869029&quot;,&quot;type_id&quot;:&quot;22&quot;,&quot;assoc_obj_id&quot;:&quot;&quot;,&quot;source_app_id&quot;:&quot;0&quot;,&quot;extra_story_params&quot;:[],&quot;content_timestamp&quot;:&quot;1351261059&quot;,&quot;check_hash&quot;:&quot;AQAJWtj28-dJleQJ&quot;,&quot;source&quot;:&quot;0&quot;&#125;\" /><input type=\"hidden\" autocomplete=\"off\" name=\"data_only_response\" value=\"1\" /><input type=\"hidden\" autocomplete=\"off\" name=\"timeline_ufi\" value=\"1\" /><input type=\"hidden\" name=\"timeline_log_data\" value=\"AQAnqduipAgoOm_9IjjltRCBJyhIBv2-\\-\\eLrjr4l_RRflt2UWyH0EPF2kVqzF4Q12S5pfMKmua5i3Ih93QBatDveVbELZ65qOpzYxXZzGwFtgzfa0fu173jC-rcygo27udXQeX_lMB4R4fV9zUYUpfhC_cdrY4PE-9BLE8Gp5Idd-UGSHP50CUJyyh7gFYJEXsWOfxh7a5iCh6VkqaavWrwXgESYx04gW7S6zKfVC4up7JuXIE6E8YTsPgg3WPB1Rz-g6Y6coNu8nnk7p3McPFiocJiOwjt4Wh7CB7bDVAtAe7E1168oJ7B3R-9lbX53nxlibQl-L6gx0v3C3w-9bqKnhcanSMX6LXUFb-2zwSbivYmm2R3blGuA9ihYamQy7zA\" /><div class=\"fbTimelineFeedbackHeader\"><div class=\"fbTimelineFeedbackActions clearfix\"><span class=\"UIActionLinks UIActionLinks_bottom\" data-ft=\"&#123;&quot;tn&quot;:&quot;=&quot;,&quot;type&quot;:20&#125;\"><a class=\"like_link\" title=\"Like this item\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2FNorthCarolinaStateParks%2Fposts%2F10151228504104030\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Like</a> &#183; <a class=\"comment_link\" title=\"Leave a comment\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2FNorthCarolinaStateParks%2Fposts%2F10151228504104030\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Comment</a> &#183; <a rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2FNorthCarolinaStateParks%2Fposts%2F10151228504104030\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Share</a></span></div></div><div><div class=\"uiUfi UFIContainer utn\" id=\"u4kzrn2139\"></div></div></form></div></div></div><i class=\"spinePointer\"></i><div class=\"bottomBorder\"></div></li><li class=\"fbTimelineUnit fbTimelineTwoColumn clearfix\" data-side=\"l\" data-fixed=\"1\" data-size=\"1\" id=\"tl_unit_7377182109995536428\"><div class=\"topBorder\"></div><div class=\"timelineUnitContainer\" id=\"u4kzrn274\" data-gt=\"&#123;&quot;action_type&quot;:&quot;1&quot;,&quot;aggregation_id&quot;:&quot;10151228392704030&quot;,&quot;aggregation_time&quot;:&quot;1351255612&quot;,&quot;actor&quot;:&quot;143412869029&quot;,&quot;creator&quot;:&quot;143412869029&quot;,&quot;pagestimeline&quot;:&quot;1&quot;,&quot;payload_type&quot;:&quot;11&quot;,&quot;eventtime&quot;:&quot;1352998853&quot;,&quot;viewerid&quot;:&quot;0&quot;,&quot;profileownerid&quot;:&quot;143412869029&quot;,&quot;unitimpressionid&quot;:&quot;f2d1d9d2&quot;,&quot;contentid&quot;:&quot;7377182109995536428&quot;,&quot;timeline_unit_type&quot;:&quot;StatusMessageUnit&quot;,&quot;timewindowsize&quot;:&quot;2&quot;,&quot;contextwindowstart&quot;:&quot;1349074800&quot;,&quot;contextwindowend&quot;:&quot;1351753199&quot;&#125;\" data-time=\"1351255612\"><div class=\"\"><div role=\"article\"><div class=\"clearfix mbs pbs _1_m \"><a class=\"_29h _303\" href=\"http://www.facebook.com/NorthCarolinaStateParks\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-ash3/174880_143412869029_133040642_q.jpg\" alt=\"\" /></a><div class=\"_3dp _29k\"><h5 class=\"_1_s\"><span class=\"fcg\"><span class=\"fwb\"><a href=\"http://www.facebook.com/NorthCarolinaStateParks\">North Carolina State Parks and Recreation</a></span></span></h5><div class=\"_1_n fsm fwn fcg\"><a class=\"uiLinkSubtle\" href=\"/NorthCarolinaStateParks/posts/10151228392704030\"><abbr title=\"Friday, October 26, 2012 at 4:46am\" data-utime=\"1351252012\">October 26</abbr></a><a data-hover=\"tooltip\" aria-label=\"Public\" class=\"uiStreamPrivacy inlineBlock fbStreamPrivacy fbPrivacyAudienceIndicator _1_o\" href=\"#\"><i class=\"lock img sp_2voup5 sx_4929d1\"></i></a></div></div></div><div class=\"_1x1\"><div class=\"userContentWrapper\"><div class=\"_wk\"><span class=\"userContent\">Hurricanes can be scary, too. Fort Fisher&#039;s fall festival and haunted trail walk have been cancelled for Saturday due to the promise of lousy weather.</span></div></div></div></div><div class=\"fbTimelineUFI uiCommentContainer\"><form rel=\"async\" class=\"commentable_item autoexpand_mode\" method=\"post\" action=\"/ajax/ufi/modify.php\" onsubmit=\"return window.Event &amp;&amp; Event.__inlineSubmit &amp;&amp; Event.__inlineSubmit(this,event)\" id=\"u4kzrn2173\"><input type=\"hidden\" name=\"charset_test\" value=\"&euro;,&acute;,\u20AC,\u00B4,\u6C34,\u0414,\u0404\" /><input type=\"hidden\" name=\"lsd\" value=\"AVqaEiN6\" autocomplete=\"off\" /><input type=\"hidden\" autocomplete=\"off\" name=\"feedback_params\" value=\"&#123;&quot;actor&quot;:&quot;143412869029&quot;,&quot;target_fbid&quot;:&quot;10151228392704030&quot;,&quot;target_profile_id&quot;:&quot;143412869029&quot;,&quot;type_id&quot;:&quot;22&quot;,&quot;assoc_obj_id&quot;:&quot;&quot;,&quot;source_app_id&quot;:&quot;0&quot;,&quot;extra_story_params&quot;:[],&quot;content_timestamp&quot;:&quot;1351255612&quot;,&quot;check_hash&quot;:&quot;AQB9zHN8_d5UJupb&quot;,&quot;source&quot;:&quot;0&quot;&#125;\" /><input type=\"hidden\" autocomplete=\"off\" name=\"data_only_response\" value=\"1\" /><input type=\"hidden\" autocomplete=\"off\" name=\"timeline_ufi\" value=\"1\" /><input type=\"hidden\" name=\"timeline_log_data\" value=\"AQDkfiAFcMFmC4_ZDIx9bQS1fakojF9Xy_eFn8wrPOXfkINtK4qKeQeu7r2Ao072L_43OkiIzj7jBLEXJufthAR9yH_tEpV6o5KWyFSuIl_qwb8NXyTXCUwi-9t4SAkco3BJa-xRphunCC16bmZyNdNaCoFyK7n767709BoyS7iL4-HldWOzzz6ZVJSeZxhARCS0kDqKYJiNA5P-VUAdmBnGz15lhOR8U3VTj62mZf1TTsRulRWXSk9NB9Ob3502-cARFB-mZuoLZaZshBhfmYHvLxJa9-J1J-YRfDQWi3i70QPxEavoW-uCmsLwkSUjtH4sZcAM9B8yshm2Alg401-3otztTlacqDXV4rt7Kxc2CRzCVwA1_fE68zRBASaotko\" /><div class=\"fbTimelineFeedbackHeader\"><div class=\"fbTimelineFeedbackActions clearfix\"><span class=\"fbTimelineFeedbackShares\"><a data-hover=\"tooltip\" aria-label=\"Show shares\" data-tooltip-alignh=\"right\" href=\"/shares/view?id=10151228392704030\" ajaxify=\"/ajax/shares/view?target_fbid=10151228392704030\" rel=\"async\"><i></i>2</a></span><span class=\"UIActionLinks UIActionLinks_bottom\" data-ft=\"&#123;&quot;tn&quot;:&quot;=&quot;,&quot;type&quot;:20&#125;\"><a class=\"like_link\" title=\"Like this item\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2FNorthCarolinaStateParks%2Fposts%2F10151228392704030\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Like</a> &#183; <a class=\"comment_link\" title=\"Leave a comment\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2FNorthCarolinaStateParks%2Fposts%2F10151228392704030\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Comment</a> &#183; <a rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2FNorthCarolinaStateParks%2Fposts%2F10151228392704030\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Share</a></span></div></div><div><div class=\"uiUfi UFIContainer utn\" id=\"u4kzrn2140\"></div></div></form></div></div></div><i class=\"spinePointer\"></i><div class=\"bottomBorder\"></div></li><li class=\"fbTimelineUnit lastCapsule fbTimelineTwoColumn clearfix\" data-side=\"l\" data-fixed=\"1\" data-size=\"1\" id=\"tl_unit_-5935640137999532876\"><div class=\"topBorder\"></div><div class=\"timelineUnitContainer\" id=\"u4kzrn275\" data-gt=\"&#123;&quot;action_type&quot;:&quot;1&quot;,&quot;aggregation_id&quot;:&quot;10151228371144030&quot;,&quot;aggregation_time&quot;:&quot;1351254248&quot;,&quot;actor&quot;:&quot;143412869029&quot;,&quot;creator&quot;:&quot;143412869029&quot;,&quot;pagestimeline&quot;:&quot;1&quot;,&quot;payload_type&quot;:&quot;7&quot;,&quot;eventtime&quot;:&quot;1352998853&quot;,&quot;viewerid&quot;:&quot;0&quot;,&quot;profileownerid&quot;:&quot;143412869029&quot;,&quot;unitimpressionid&quot;:&quot;f2d1d9d2&quot;,&quot;contentid&quot;:&quot;-5935640137999532876&quot;,&quot;timeline_unit_type&quot;:&quot;AddSinglePhotoUnit&quot;,&quot;timewindowsize&quot;:&quot;2&quot;,&quot;contextwindowstart&quot;:&quot;1349074800&quot;,&quot;contextwindowend&quot;:&quot;1351753199&quot;&#125;\" data-time=\"1351254248\"><div class=\"\"><div role=\"article\"><div class=\"clearfix mbs pbs _1_m \"><a class=\"_29h _303\" href=\"http://www.facebook.com/NorthCarolinaStateParks\" tabindex=\"-1\" aria-hidden=\"true\"><img class=\"_s0 _rx img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-ash3/174880_143412869029_133040642_q.jpg\" alt=\"\" /></a><div class=\"_3dp _29k\"><h5 class=\"_1_s\"><span class=\"fcg\"><span class=\"fwb\"><a href=\"http://www.facebook.com/NorthCarolinaStateParks\">North Carolina State Parks and Recreation</a></span></span></h5><div class=\"_1_n fsm fwn fcg\"><a class=\"uiLinkSubtle\" href=\"http://www.facebook.com/photo.php?fbid=10151228371144030&amp;set=a.164247464029.118730.143412869029&amp;type=1\"><abbr title=\"Friday, October 26, 2012 at 4:24am\" data-utime=\"1351250648\">October 26</abbr></a><a data-hover=\"tooltip\" aria-label=\"Public\" class=\"uiStreamPrivacy inlineBlock fbStreamPrivacy fbPrivacyAudienceIndicator _1_o\" href=\"#\"><i class=\"lock img sp_2voup5 sx_4929d1\"></i></a></div></div></div><div class=\"aboveUnitContent\"><div class=\"userContentWrapper\"><div class=\"_wk\"><span class=\"userContent\">Raven Rock State Park&#039;s north side bridle trails were augmented with a new vault toilet this week (replacing an aged wooden outhouse). It has solar lights and will help protect the environment, too.</span></div></div></div><div class=\"photoUnit clearfix\"><div class=\"_53s uiScaledThumb photo photoWidth1\" data-cropped=\"1\"><a class=\"_6i9\" href=\"http://www.facebook.com/photo.php?fbid=10151228371144030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1\" rel=\"theater\" ajaxify=\"http://www.facebook.com/photo.php?fbid=10151228371144030&amp;set=a.164247464029.118730.143412869029&amp;type=1&amp;relevant_count=1&amp;src=https%3A%2F%2Fsphotos-b.xx.fbcdn.net%2Fhphotos-ash4%2F404330_10151228371144030_1232066847_n.jpg&amp;size=800%2C655&amp;theater\"><div class=\"uiScaledImageContainer photoWrap\"><img class=\"scaledImageFitWidth img\" src=\"https://sphotos-b.xx.fbcdn.net/hphotos-ash4/c44.0.403.403/p403x403/404330_10151228371144030_1232066847_n.jpg\" alt=\"Photo: Raven Rock State Park&#039;s north side bridle trails were augmented with a new vault toilet this week (replacing an aged wooden outhouse). It has solar lights and will help protect the environment, too.\" width=\"403\" height=\"403\" /></div></a></div></div></div><div class=\"fbTimelineUFI uiCommentContainer\"><form rel=\"async\" class=\"commentable_item autoexpand_mode\" method=\"post\" action=\"/ajax/ufi/modify.php\" onsubmit=\"return window.Event &amp;&amp; Event.__inlineSubmit &amp;&amp; Event.__inlineSubmit(this,event)\" id=\"u4kzrn2174\"><input type=\"hidden\" name=\"charset_test\" value=\"&euro;,&acute;,\u20AC,\u00B4,\u6C34,\u0414,\u0404\" /><input type=\"hidden\" name=\"lsd\" value=\"AVqaEiN6\" autocomplete=\"off\" /><input type=\"hidden\" autocomplete=\"off\" name=\"feedback_params\" value=\"&#123;&quot;actor&quot;:&quot;143412869029&quot;,&quot;target_fbid&quot;:&quot;10151228371144030&quot;,&quot;target_profile_id&quot;:&quot;143412869029&quot;,&quot;type_id&quot;:&quot;7&quot;,&quot;assoc_obj_id&quot;:&quot;&quot;,&quot;source_app_id&quot;:&quot;0&quot;,&quot;extra_story_params&quot;:[],&quot;content_timestamp&quot;:&quot;1351254247&quot;,&quot;check_hash&quot;:&quot;AQDCI5i9VM4H4-8g&quot;,&quot;source&quot;:&quot;0&quot;&#125;\" /><input type=\"hidden\" autocomplete=\"off\" name=\"data_only_response\" value=\"1\" /><input type=\"hidden\" autocomplete=\"off\" name=\"timeline_ufi\" value=\"1\" /><input type=\"hidden\" name=\"timeline_log_data\" value=\"AQDxmaeRXL9NZF77OaCWqNbVsV8D3gBFiDCaZ-D1RpMeROpUEi_dej4jyA_WcWnKuvZ9H8UTc31jNFRntInqRzGUldzqVOJ0FigdKiY0je3mmpKh4MkKvX7mIVbjF-HOF72CDOTcknfmSWbDxF3_oK6uXG36vulv2yzJvOGjfKBXKbwb6x8eVToPBPG9fl6oBlIZDShbdIZ7dj5alqWawK5_Y-_vydPAPPrRwU_EFYupjP6bfltCQJFycjoubzzgX38e0B01jzZNm1DbdKZLMCBZxo4I3z7GF4q0xzBFHqVDAOmi-uIHevLM7ma7EMa6mZ5bnqMzDkwDxsHsmVolXTE8qkOVnt9FCzWv5VRoOGZJRFS3UiB7XxnviiGJ8pGgH0c\" /><div class=\"fbTimelineFeedbackHeader\"><div class=\"fbTimelineFeedbackActions clearfix\"><span class=\"fbTimelineFeedbackShares\"><a data-hover=\"tooltip\" aria-label=\"Show shares\" data-tooltip-alignh=\"right\" href=\"/shares/view?id=10151228371144030\" ajaxify=\"/ajax/shares/view?target_fbid=10151228371144030\" rel=\"async\"><i></i>2</a></span><span class=\"UIActionLinks UIActionLinks_bottom\" data-ft=\"&#123;&quot;tn&quot;:&quot;=&quot;,&quot;type&quot;:20&#125;\"><a class=\"like_link\" title=\"Like this item\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151228371144030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Like</a> &#183; <a class=\"comment_link\" title=\"Leave a comment\" rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151228371144030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Comment</a> &#183; <a rel=\"dialog\" ajaxify=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2Fphoto.php%3Ffbid%3D10151228371144030%26set%3Da.164247464029.118730.143412869029%26type%3D1\" href=\"http://www.facebook.com/r.php?fbpage_id=143412869029&amp;r=111\">Share</a></span></div></div><div><div class=\"uiUfi UFIContainer utn\" id=\"u4kzrn2130\"></div></div></form></div></div></div><i class=\"spinePointer\"></i><div class=\"bottomBorder\"></div></li><li class=\"fbTimelineUnit lastCapsule fbTimelineTwoColumn clearfix\" data-side=\"l\" data-fixed=\"1\" data-size=\"1\" id=\"tl_unit_2530096808_recent\"><div class=\"topBorder\"></div><div class=\"timelineReportContainer\" id=\"u4kzrn235\" data-gt=\"&#123;&quot;eventtime&quot;:&quot;1352998853&quot;,&quot;viewerid&quot;:&quot;0&quot;,&quot;profileownerid&quot;:&quot;143412869029&quot;,&quot;unitimpressionid&quot;:&quot;f2d1d9d2&quot;,&quot;contentid&quot;:&quot;&quot;,&quot;timeline_unit_type&quot;:&quot;TimelineReportUnit&quot;,&quot;timewindowsize&quot;:&quot;2&quot;,&quot;contextwindowstart&quot;:&quot;1351753200&quot;,&quot;contextwindowend&quot;:&quot;1354348799&quot;&#125;\" role=\"complementary\"><div class=\"\"><div role=\"article\"><div class=\"timelineReportHeader\"><span class=\"timelineReportCol logo\"><i style=\"width:32px;height:32px\" class=\"img sp_ep77xh sx_f7344f\"></i></span><div class=\"timelineReportCol titleWithSubtitle\"><div class=\"title fsl fwn\">Activity</div><div class=\"date fwn fcg\">November</div></div></div><div class=\"timelineReportContent timelineNoSubheaderReport\"><div class=\"timelineUnitContainer ogProfileLastUnit\" id=\"u4kzrn236\" data-gt=\"&#123;&quot;eventtime&quot;:&quot;1352998853&quot;,&quot;viewerid&quot;:&quot;0&quot;,&quot;profileownerid&quot;:&quot;143412869029&quot;,&quot;unitimpressionid&quot;:&quot;f2d1d9d2&quot;,&quot;contentid&quot;:&quot;-1355571233344405583&quot;,&quot;timeline_unit_type&quot;:&quot;TimelinePagesInsightsReportUnit&quot;,&quot;timewindowsize&quot;:&quot;2&quot;,&quot;contextwindowstart&quot;:&quot;1351753200&quot;,&quot;contextwindowend&quot;:&quot;1354348799&quot;&#125;\" data-time=\"1353050999\"><div class=\"\"><div role=\"article\"><div class=\"timelinePagesInsightsReport\"><div class=\"mbm timelinePagesInsightsSingle\"><div class=\"aboveUnitContent\"><span class=\"fwb\">People Who Like This</span></div><div class=\"bigNumber\">85</div></div></div></div></div></div></div></div></div></div><i class=\"spinePointer\"></i><div class=\"bottomBorder\"></div></li><li aria-hidden=\"true\" class=\"fbTimelineBalancer\" id=\"u4kzrn232\" data-nonunits=\"1\"><ol><li class=\"hidden_elem\" data-endmarker=\"recent\" data-pageindex=\"0\"></li></ol></li></ol><div id=\"pagelet_timeline_recent_more_pager\"></div><div class=\"fbTimelineSubSections\"></div><div id=\"u4kzrn231_scroll_trigger\"></div></div> --></code>\n", 
        "<script>bigPipe.onPageletArrive({\"is_last\":true,\"is_second_to_last_phase\":true,\"content\":{\"pagelet_timeline_recent\":{\"container_id\":\"u4kzrn2176\"}},\"jsmods\":{\"instances\":[[\"m885382_72\",[\"UFIController\",\"m885382_71\",\"m885382_91\"],[{\"__m\":\"m885382_71\"},{\"ftentidentifier\":\"10151257845269030\",\"instanceid\":\"u4kzrn276\",\"source\":0,\"showaddcomment\":true,\"collapseaddcomment\":false,\"markedcomment\":null,\"scrollcomments\":false,\"scrollwidth\":null,\"showshares\":false,\"shownub\":false,\"numberdelimiter\":\",\",\"showtyping\":0,\"logtyping\":0,\"entstream\":false,\"timelinelogdata\":\"AQA3Ln_mmAUs9Fqv0SIoGvt5rxUdDULXnjSGlL3Ndq855bx3xOy421FPAe3qDYtTT943tQ0FwjZTcAvQSgSs1LLsPhZ7WKaKggKQcegYMJiNEiqfY89O7zzr1DBAfs3xgSyiqjwaNVa9HKDbZlvhprDl9m_NY8oAMfFOvW4aRhIauYC9ByI7_JWvOui2ZVg3P5QVeEmqlLVRtiPq4-s12NlQqBW846vmIwEbYov4PoomrNMZ657GOGWsBqCp7L3uZf_Zz8eo7dfnalesDdAUM0XtEt-gpRAF-abKHfUgB9xRxRUaN0TF2HzT9sJk1kzYoPP_wuxZ6J73WeVJJ6GjdAiVjP-LLPOtVa8KJOXll5BJz78t720BxT9Ri67Rbr5brz4\"},{\"feedbacktargets\":[{\"targetfbid\":\"10151257845269030\",\"entidentifier\":\"10151257845269030\",\"permalink\":\"http:\\/\\/www.facebook.com\\/photo.php?fbid=10151257845269030&set=a.164247464029.118730.143412869029&type=1\",\"commentcount\":0,\"likecount\":26,\"viewercanlike\":false,\"hasviewerliked\":false,\"cancomment\":false,\"showremovemenu\":false,\"actorforpost\":\"\",\"viewerid\":\"0\",\"likesentences\":{\"current\":{\"text\":\"Scott Fister, Michael Moore, Bert Hayes and 23 others like this.\",\"ranges\":[{\"offset\":0,\"length\":12,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/scott.fister.7\",\"id\":1000114791}]},{\"offset\":14,\"length\":13,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/people\\/Michael-Moore\\/1572484594\",\"id\":1572484594}]},{\"offset\":29,\"length\":10,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/bert.hayes.7\",\"id\":1217252943}]}],\"aggregatedranges\":[{\"offset\":44,\"length\":9,\"count\":23}]},\"alternate\":{\"text\":\"You, Scott Fister, Michael Moore, Bert Hayes and 23 others like this.\",\"ranges\":[{\"offset\":5,\"length\":12,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/scott.fister.7\",\"id\":1000114791}]},{\"offset\":19,\"length\":13,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/people\\/Michael-Moore\\/1572484594\",\"id\":1572484594}]},{\"offset\":34,\"length\":10,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/bert.hayes.7\",\"id\":1217252943}]}],\"aggregatedranges\":[{\"offset\":49,\"length\":9,\"count\":23}]}},\"sharecount\":0,\"ownerid\":\"143412869029\",\"lastseentime\":null,\"canremoveall\":false,\"seencount\":0,\"seenbyall\":false,\"hasunseencollapsed\":false,\"mentionsdatasource\":{\"__m\":\"m885382_91\"},\"isranked\":false,\"isthreaded\":false,\"isownerpage\":true}],\"comments\":[],\"profiles\":[],\"actions\":[],\"commentlists\":{\"10151257845269030\":{\"range\":{\"offset\":0,\"length\":0},\"values\":[]}},\"servertime\":1352998853}],1],[\"m885382_74\",[\"UFIController\",\"m885382_73\",\"m885382_92\"],[{\"__m\":\"m885382_73\"},{\"ftentidentifier\":\"10151250347114030\",\"instanceid\":\"u4kzrn277\",\"source\":0,\"showaddcomment\":true,\"collapseaddcomment\":false,\"markedcomment\":null,\"scrollcomments\":false,\"scrollwidth\":null,\"showshares\":false,\"shownub\":false,\"numberdelimiter\":\",\",\"showtyping\":0,\"logtyping\":0,\"entstream\":false,\"timelinelogdata\":\"AQAhy4cWXkO_7-k1hR2-Yc1U-hyQa0bMUOS8y6yPx2Qqwy0Wc0yN96Lzg8rs1eLL4GLIyEfDIjC_SVutuqRTNg_XUIth7aKK4rzkkLbQ7-Hhkvnd9rBSNIuNDyKOb4gc0KeFJWBil2qXpGF_zLAvbQvmtBhQQghQrLt3HBv7GBMQltcTih00p9pKXDCy-Wn7Rf7o8xKhe_61u7XlcflejlMQuPtSKGwVZHWsqCWGsvR5vSCvPPdTPFR_ZH7xyPIkgo5EdG7XvDEqL04u8-7Dg3p_lmk4nsBbrsH9-05ph_yxi2A22aggXxYXmWqhZ6y5UPG7pI6aRrQgC9Ivck7KtKOaJzxq2C2DpBnK6kww4fcFx1MHD2aqMDoVfXKLJtjA82s\"},{\"feedbacktargets\":[{\"targetfbid\":\"10151250347114030\",\"entidentifier\":\"10151250347114030\",\"permalink\":\"http:\\/\\/www.facebook.com\\/photo.php?fbid=10151250347114030&set=a.164247464029.118730.143412869029&type=1\",\"commentcount\":0,\"likecount\":57,\"viewercanlike\":false,\"hasviewerliked\":false,\"cancomment\":false,\"showremovemenu\":false,\"actorforpost\":\"\",\"viewerid\":\"0\",\"likesentences\":{\"current\":{\"text\":\"Chris Helms, Donna Henley, Barbara Jones Ward and 54 others like this.\",\"ranges\":[{\"offset\":0,\"length\":11,\"entities\":[{\"url\":\"\",\"shimhash\":\"KAQFhCPEx\",\"id\":1540322263}]},{\"offset\":13,\"length\":12,\"entities\":[{\"url\":\"\",\"shimhash\":\"RAQEwajm0\",\"id\":100004084513069}]},{\"offset\":27,\"length\":18,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/barbara.j.ward3\",\"id\":1517953328}]}],\"aggregatedranges\":[{\"offset\":50,\"length\":9,\"count\":54}]},\"alternate\":{\"text\":\"You, Chris Helms, Donna Henley, Barbara Jones Ward and 54 others like this.\",\"ranges\":[{\"offset\":5,\"length\":11,\"entities\":[{\"url\":\"\",\"shimhash\":\"aAQENx0jQ\",\"id\":1540322263}]},{\"offset\":18,\"length\":12,\"entities\":[{\"url\":\"\",\"shimhash\":\"uAQHr8pT-\",\"id\":100004084513069}]},{\"offset\":32,\"length\":18,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/barbara.j.ward3\",\"id\":1517953328}]}],\"aggregatedranges\":[{\"offset\":55,\"length\":9,\"count\":54}]}},\"sharecount\":3,\"ownerid\":\"143412869029\",\"lastseentime\":null,\"canremoveall\":false,\"seencount\":0,\"seenbyall\":false,\"hasunseencollapsed\":false,\"mentionsdatasource\":{\"__m\":\"m885382_92\"},\"isranked\":false,\"isthreaded\":false,\"isownerpage\":true}],\"comments\":[],\"profiles\":[],\"actions\":[],\"commentlists\":{\"10151250347114030\":{\"range\":{\"offset\":0,\"length\":0},\"values\":[]}},\"servertime\":1352998853}],1],[\"m885382_76\",[\"UFIController\",\"m885382_75\",\"m885382_93\"],[{\"__m\":\"m885382_75\"},{\"ftentidentifier\":\"217369611728583\",\"instanceid\":\"u4kzrn278\",\"source\":0,\"showaddcomment\":true,\"collapseaddcomment\":false,\"markedcomment\":null,\"scrollcomments\":false,\"scrollwidth\":null,\"showshares\":false,\"shownub\":false,\"numberdelimiter\":\",\",\"showtyping\":0,\"logtyping\":0,\"entstream\":false,\"timelinelogdata\":\"AQA31vqWBdLldeBpimgkNT0T6VctFQIu96PatdJI_iIJJTT8nzzQHbvMv6S4Bttg-xlhAu4V400PjnSf4zbHOQLxkChYuHzY4dlIszLUPCghARyGfHOtyL2AlTaRyyzG509joO14jIpt7fQo6G-fckSXmUsdsR1elLKlnDJl4YQSe3CTwsIjry5ZbClcY28Ad6qxRECpL7BShBPII7pG5DS3C6mtNUpBhBUouLtdzkEdLdzXCZo9BarvKOSjDqwFfeffzAiJy-SjfErBg6KZEknMq1YO_vGmapX5hitGpLzEYpraSQ5YX7YR0K9TQy7fkJvNO2e3DZQ5fihF-xg5mCtBtVeu1RJxIuXlYSdVLhR2jwID4F6fLF3sfUgQfNSz9ZI\"},{\"feedbacktargets\":[{\"targetfbid\":\"217369611728583\",\"entidentifier\":\"217369611728583\",\"permalink\":\"\\/NorthCarolinaStateParks\\/posts\\/217369611728583\",\"commentcount\":0,\"likecount\":11,\"viewercanlike\":false,\"hasviewerliked\":false,\"cancomment\":false,\"showremovemenu\":false,\"actorforpost\":\"\",\"viewerid\":\"0\",\"likesentences\":{\"current\":{\"text\":\"Catawba County Parks, Richard Carder, Taffy Wild and 8 others like this.\",\"ranges\":[{\"offset\":0,\"length\":20,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/catawbacountyparks\",\"id\":94474264671}]},{\"offset\":22,\"length\":14,\"entities\":[{\"url\":\"\",\"shimhash\":\"FAQHPmOF9\",\"id\":1843933820}]},{\"offset\":38,\"length\":10,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/taffy.wild\",\"id\":100001711258648}]}],\"aggregatedranges\":[{\"offset\":53,\"length\":8,\"count\":8}]},\"alternate\":{\"text\":\"You, Catawba County Parks, Richard Carder, Taffy Wild and 8 others like this.\",\"ranges\":[{\"offset\":5,\"length\":20,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/catawbacountyparks\",\"id\":94474264671}]},{\"offset\":27,\"length\":14,\"entities\":[{\"url\":\"\",\"shimhash\":\"NAQEyoCZQ\",\"id\":1843933820}]},{\"offset\":43,\"length\":10,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/taffy.wild\",\"id\":100001711258648}]}],\"aggregatedranges\":[{\"offset\":58,\"length\":8,\"count\":8}]}},\"sharecount\":0,\"ownerid\":\"143412869029\",\"lastseentime\":null,\"canremoveall\":false,\"seencount\":0,\"seenbyall\":false,\"hasunseencollapsed\":false,\"mentionsdatasource\":{\"__m\":\"m885382_93\"},\"isranked\":false,\"isthreaded\":false,\"isownerpage\":true}],\"comments\":[],\"profiles\":[],\"actions\":[],\"commentlists\":{\"217369611728583\":{\"range\":{\"offset\":0,\"length\":0},\"values\":[]}},\"servertime\":1352998853}],1],[\"m885382_78\",[\"UFIController\",\"m885382_77\",\"m885382_94\"],[{\"__m\":\"m885382_77\"},{\"ftentidentifier\":\"10151247636289030\",\"instanceid\":\"u4kzrn279\",\"source\":0,\"showaddcomment\":true,\"collapseaddcomment\":false,\"markedcomment\":null,\"scrollcomments\":false,\"scrollwidth\":null,\"showshares\":false,\"shownub\":false,\"numberdelimiter\":\",\",\"showtyping\":0,\"logtyping\":0,\"entstream\":false,\"timelinelogdata\":\"AQCfrExipd8rvTCrVyqmjOtth9m8fmaQaujwrHbADpJLIuZdvGIabV2eazYPlSZY1e_JyuYdfjCrCPQm8BZZZ64CddxvkDr9bKnrCY5xK2BQOamt_XfgCPBcsju8lQmVgasplj3dfDxd1qJPG-swc7UtqNdaapVK_9Yo5wCKuOGl2f_kU1ZuorOtsxcWW23RuI0RZXIRFdWivAVcClNSRFfv6l8dsWMM-BF3FcCdw4Yg_fIUpHs7eoF2WBq_KP-WXJRenM-1N9FmNSk0ICoMDymzyFBwVa6WiZo9sOBQ88c6mmW_yrVLhsHz5RxCjeLAgcuseXi39rYKAB5IcH26sW9izXS1i0rNknZT6y7zKsFRZYLpPMIsl9--4YWrSB0O8Gk\"},{\"feedbacktargets\":[{\"targetfbid\":\"10151247636289030\",\"entidentifier\":\"10151247636289030\",\"permalink\":\"http:\\/\\/www.facebook.com\\/photo.php?fbid=10151247636289030&set=a.164247464029.118730.143412869029&type=1\",\"commentcount\":0,\"likecount\":18,\"viewercanlike\":false,\"hasviewerliked\":false,\"cancomment\":false,\"showremovemenu\":false,\"actorforpost\":\"\",\"viewerid\":\"0\",\"likesentences\":{\"current\":{\"text\":\"Chris Helms, Janet White Pearson, Catawba County Parks and 15 others like this.\",\"ranges\":[{\"offset\":0,\"length\":11,\"entities\":[{\"url\":\"\",\"shimhash\":\"4AQEws-m0\",\"id\":1540322263}]},{\"offset\":13,\"length\":19,\"entities\":[{\"url\":\"\",\"shimhash\":\"UAQHMbuxY\",\"id\":1579249248}]},{\"offset\":34,\"length\":20,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/catawbacountyparks\",\"id\":94474264671}]}],\"aggregatedranges\":[{\"offset\":59,\"length\":9,\"count\":15}]},\"alternate\":{\"text\":\"You, Chris Helms, Janet White Pearson, Catawba County Parks and 15 others like this.\",\"ranges\":[{\"offset\":5,\"length\":11,\"entities\":[{\"url\":\"\",\"shimhash\":\"KAQFhCPEx\",\"id\":1540322263}]},{\"offset\":18,\"length\":19,\"entities\":[{\"url\":\"\",\"shimhash\":\"SAQEjLIXR\",\"id\":1579249248}]},{\"offset\":39,\"length\":20,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/catawbacountyparks\",\"id\":94474264671}]}],\"aggregatedranges\":[{\"offset\":64,\"length\":9,\"count\":15}]}},\"sharecount\":4,\"ownerid\":\"143412869029\",\"lastseentime\":null,\"canremoveall\":false,\"seencount\":0,\"seenbyall\":false,\"hasunseencollapsed\":false,\"mentionsdatasource\":{\"__m\":\"m885382_94\"},\"isranked\":false,\"isthreaded\":false,\"isownerpage\":true}],\"comments\":[],\"profiles\":[],\"actions\":[],\"commentlists\":{\"10151247636289030\":{\"range\":{\"offset\":0,\"length\":0},\"values\":[]}},\"servertime\":1352998853}],1],[\"m885382_80\",[\"UFIController\",\"m885382_79\",\"m885383_95\"],[{\"__m\":\"m885382_79\"},{\"ftentidentifier\":\"10151246283684030\",\"instanceid\":\"u4kzrn280\",\"source\":0,\"showaddcomment\":true,\"collapseaddcomment\":false,\"markedcomment\":null,\"scrollcomments\":false,\"scrollwidth\":null,\"showshares\":false,\"shownub\":false,\"numberdelimiter\":\",\",\"showtyping\":0,\"logtyping\":0,\"entstream\":false,\"timelinelogdata\":\"AQAF6nkB0mvVtiUcj1cWAXYQeRnr1EaeNDM1IRBQwSSidCcmAw5tBty4IdFDWDCIVC3mDrcvVEVdhAAl80HHAh8gdKN6cyx6KJRejQklsVVvo9PoBpra12auKI4ujRKwEycFNfE-MI9sL7Jp9In_9kam1sG4Wyr8_hkmaDy9jPipYRjE5YP-kMdvLvqNVTnMnxQjUCwGp8phNgGvsVLEL7Fn5MU6nvFI0HplV_DUrzDOyjR_AU-jv7PmkuOF4H-E7Jyba9ZUDQnIwLmHyDKQjThVyuDIp49B1WGWXKXp4XVKpsTzakR0XM7F7VpZyh6H7EQ0GPHcage09aIqAiAgTcX98ZEPgbjKQqOYgCZGSTdxnORTPcPVfJwKZtYBBIMZaa8\"},{\"feedbacktargets\":[{\"targetfbid\":\"10151246283684030\",\"entidentifier\":\"10151246283684030\",\"permalink\":\"http:\\/\\/www.facebook.com\\/photo.php?fbid=10151246283684030&set=a.164247464029.118730.143412869029&type=1\",\"commentcount\":0,\"likecount\":15,\"viewercanlike\":false,\"hasviewerliked\":false,\"cancomment\":false,\"showremovemenu\":false,\"actorforpost\":\"\",\"viewerid\":\"0\",\"likesentences\":{\"current\":{\"text\":\"Joey Pyron, Mike Kerkado, Paul T. Brown II and 12 others like this.\",\"ranges\":[{\"offset\":0,\"length\":10,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/joey.pyron\",\"id\":100002897607424}]},{\"offset\":12,\"length\":12,\"entities\":[{\"url\":\"\",\"shimhash\":\"FAQHPmOF9\",\"id\":645854192}]},{\"offset\":26,\"length\":16,\"entities\":[{\"url\":\"\",\"shimhash\":\"rAQEKPnyM\",\"id\":100003240249568}]}],\"aggregatedranges\":[{\"offset\":47,\"length\":9,\"count\":12}]},\"alternate\":{\"text\":\"You, Joey Pyron, Mike Kerkado, Paul T. Brown II and 12 others like this.\",\"ranges\":[{\"offset\":5,\"length\":10,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/joey.pyron\",\"id\":100002897607424}]},{\"offset\":17,\"length\":12,\"entities\":[{\"url\":\"\",\"shimhash\":\"nAQESQtKi\",\"id\":645854192}]},{\"offset\":31,\"length\":16,\"entities\":[{\"url\":\"\",\"shimhash\":\"zAQGk0cyh\",\"id\":100003240249568}]}],\"aggregatedranges\":[{\"offset\":52,\"length\":9,\"count\":12}]}},\"sharecount\":1,\"ownerid\":\"143412869029\",\"lastseentime\":null,\"canremoveall\":false,\"seencount\":0,\"seenbyall\":false,\"hasunseencollapsed\":false,\"mentionsdatasource\":{\"__m\":\"m885383_95\"},\"isranked\":false,\"isthreaded\":false,\"isownerpage\":true}],\"comments\":[],\"profiles\":[],\"actions\":[],\"commentlists\":{\"10151246283684030\":{\"range\":{\"offset\":0,\"length\":0},\"values\":[]}},\"servertime\":1352998853}],1],[\"m885382_82\",[\"UFIController\",\"m885382_81\",\"m885383_96\"],[{\"__m\":\"m885382_81\"},{\"ftentidentifier\":\"10151237038709030\",\"instanceid\":\"u4kzrn281\",\"source\":0,\"showaddcomment\":true,\"collapseaddcomment\":false,\"markedcomment\":null,\"scrollcomments\":false,\"scrollwidth\":null,\"showshares\":false,\"shownub\":false,\"numberdelimiter\":\",\",\"showtyping\":0,\"logtyping\":0,\"entstream\":false,\"timelinelogdata\":\"AQDBN6Pc9A7jk_U40N_sIxtjjCaqUP-6n6L2edehhoOf-9mvN8CwvRZQGTlVL_C-AiDRb96EXR4uI9isoPL-sitY-5i3RwbHLdPwwfJVV2kJdkGJLSi5FnRIoGPbK8K7HA4l-c_RV0_wVe3RpAbhBPJssv3HwCxzYX4F9j8ZR-JVtuMl7fjwdHLBj9nRfu3pPhsubOvOwNY87-zg0i-Ai56o0FjUoUphWI1SkUcyZE20nUuYj6Vd7FXV4M0ovD6bm9c7fESPcbff59My0FoTiXhbFJqSKq66_KgWzrxKOy6oj-AFKvfQRq1G9cQ1u6FZg7BDjc5ovW3TTvjNxapb4fsuNAblRrJ8JPwvaXlxPxAzNH4pZrGj0v1IaTomHYGEx8c\"},{\"feedbacktargets\":[{\"targetfbid\":\"10151237038709030\",\"entidentifier\":\"10151237038709030\",\"permalink\":\"http:\\/\\/www.facebook.com\\/photo.php?fbid=10151237038709030&set=a.164247464029.118730.143412869029&type=1\",\"commentcount\":0,\"likecount\":39,\"viewercanlike\":false,\"hasviewerliked\":false,\"cancomment\":false,\"showremovemenu\":false,\"actorforpost\":\"\",\"viewerid\":\"0\",\"likesentences\":{\"current\":{\"text\":\"Laura McGuinn, Betty Anderson, Dan Schlosser and 36 others like this.\",\"ranges\":[{\"offset\":0,\"length\":13,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/foolofatook\",\"id\":845930579}]},{\"offset\":15,\"length\":14,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/betty.anderson.75\",\"id\":100003251197199}]},{\"offset\":31,\"length\":13,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/dan.schlosser.39\",\"id\":755647228}]}],\"aggregatedranges\":[{\"offset\":49,\"length\":9,\"count\":36}]},\"alternate\":{\"text\":\"You, Laura McGuinn, Betty Anderson, Dan Schlosser and 36 others like this.\",\"ranges\":[{\"offset\":5,\"length\":13,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/foolofatook\",\"id\":845930579}]},{\"offset\":20,\"length\":14,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/betty.anderson.75\",\"id\":100003251197199}]},{\"offset\":36,\"length\":13,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/dan.schlosser.39\",\"id\":755647228}]}],\"aggregatedranges\":[{\"offset\":54,\"length\":9,\"count\":36}]}},\"sharecount\":1,\"ownerid\":\"143412869029\",\"lastseentime\":null,\"canremoveall\":false,\"seencount\":0,\"seenbyall\":false,\"hasunseencollapsed\":false,\"mentionsdatasource\":{\"__m\":\"m885383_96\"},\"isranked\":false,\"isthreaded\":false,\"isownerpage\":true}],\"comments\":[],\"profiles\":[],\"actions\":[],\"commentlists\":{\"10151237038709030\":{\"range\":{\"offset\":0,\"length\":0},\"values\":[]}},\"servertime\":1352998853}],1],[\"m885382_84\",[\"UFIController\",\"m885382_83\",\"m885383_97\"],[{\"__m\":\"m885382_83\"},{\"ftentidentifier\":\"10151236567709030\",\"instanceid\":\"u4kzrn282\",\"source\":0,\"showaddcomment\":true,\"collapseaddcomment\":false,\"markedcomment\":null,\"scrollcomments\":false,\"scrollwidth\":null,\"showshares\":false,\"shownub\":false,\"numberdelimiter\":\",\",\"showtyping\":0,\"logtyping\":0,\"entstream\":false,\"timelinelogdata\":\"AQDOFO2Unh_vPBOf5UBafz9CF_02TR9Igjb6vEG8hmDGFbOqXG8Ostae2EHBvl0j7T9XVOBrLOqOLoOwlzt8urAUNuNBwUgwL6lkOA8Ycq3XRS-jJvpmD3px6LeJMNDkQYYD0eNiVQq2dvcFg09dD4OfODjntJ2h4SWLKNRW3WRFTYnU1bhckkKRjyrvhNBG_LeCUir-kngtemBwMVG3XTtva9CyEE4sCejNOefFlLCpHrAo5eiwgM1rSIjrkzSfeo0sHbEXwipiDlplASzUMHPwwZDmUhVz2-K_elnfoeJ6kW-bIqslZcjsB_PDaDjxVY-bT79ZjhBjpev0D1UJG-u2NKzXFJ9CB7WSI2qQZ885Ai3weKYpnQqqSNWVecFNwgQ\"},{\"feedbacktargets\":[{\"targetfbid\":\"10151236567709030\",\"entidentifier\":\"10151236567709030\",\"permalink\":\"http:\\/\\/www.facebook.com\\/photo.php?fbid=10151236567709030&set=a.164247464029.118730.143412869029&type=1\",\"commentcount\":0,\"likecount\":23,\"viewercanlike\":false,\"hasviewerliked\":false,\"cancomment\":false,\"showremovemenu\":false,\"actorforpost\":\"\",\"viewerid\":\"0\",\"likesentences\":{\"current\":{\"text\":\"Denise Antonacci Kenney, Kimberly Jackson, Cliff McCall and 20 others like this.\",\"ranges\":[{\"offset\":0,\"length\":23,\"entities\":[{\"url\":\"\",\"shimhash\":\"0AQEvstfn\",\"id\":1319934126}]},{\"offset\":25,\"length\":16,\"entities\":[{\"url\":\"\",\"shimhash\":\"xAQFWHACK\",\"id\":100003035312349}]},{\"offset\":43,\"length\":12,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/micronautp\",\"id\":1274303760}]}],\"aggregatedranges\":[{\"offset\":60,\"length\":9,\"count\":20}]},\"alternate\":{\"text\":\"You, Denise Antonacci Kenney, Kimberly Jackson, Cliff McCall and 20 others like this.\",\"ranges\":[{\"offset\":5,\"length\":23,\"entities\":[{\"url\":\"\",\"shimhash\":\"HAQGKveDe\",\"id\":1319934126}]},{\"offset\":30,\"length\":16,\"entities\":[{\"url\":\"\",\"shimhash\":\"XAQGPMp6b\",\"id\":100003035312349}]},{\"offset\":48,\"length\":12,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/micronautp\",\"id\":1274303760}]}],\"aggregatedranges\":[{\"offset\":65,\"length\":9,\"count\":20}]}},\"sharecount\":1,\"ownerid\":\"143412869029\",\"lastseentime\":null,\"canremoveall\":false,\"seencount\":0,\"seenbyall\":false,\"hasunseencollapsed\":false,\"mentionsdatasource\":{\"__m\":\"m885383_97\"},\"isranked\":false,\"isthreaded\":false,\"isownerpage\":true}],\"comments\":[],\"profiles\":[],\"actions\":[],\"commentlists\":{\"10151236567709030\":{\"range\":{\"offset\":0,\"length\":0},\"values\":[]}},\"servertime\":1352998853}],1],[\"m885382_86\",[\"UFIController\",\"m885382_85\",\"m885383_98\"],[{\"__m\":\"m885382_85\"},{\"ftentidentifier\":\"10151236478529030\",\"instanceid\":\"u4kzrn283\",\"source\":0,\"showaddcomment\":true,\"collapseaddcomment\":false,\"markedcomment\":null,\"scrollcomments\":false,\"scrollwidth\":null,\"showshares\":false,\"shownub\":false,\"numberdelimiter\":\",\",\"showtyping\":0,\"logtyping\":0,\"entstream\":false,\"timelinelogdata\":\"AQCBY2G4YST4zmje2BkF-ippVRMlnGAQtV967SyjLgSm6-u-vdYySHMdVT8c4FBmfx_pIEWFihyNy47bGJht2tskm0JCSFRF6VIOAbuLid5LHrDamVUT3KG6GFA2mW8xTHriAyi_B3K1Nr3LGr4JmQ5_fTg93wiS-NI0pAnBpjBZVnJKf86Wz1nLyMsew9t5wqVlvOeX0RmtEN9U0uA7tJ7oXFLjDq4wBJleSw3_iYs1_EBRyR_q7PKSXSKYu7YWvBezs7xZ_qPYt__8pziaiNtL4HGFe1oldPgR73deMIparL3hBRI8R-3H-EOc9NAIWVTbOB5hzD-c4Z90bxrDhsJTPbsCSFgIbjKqqBI2d9w5qhynvZeP9Zdyyk8VI8Q-Pio\"},{\"feedbacktargets\":[{\"targetfbid\":\"10151236478529030\",\"entidentifier\":\"10151236478529030\",\"permalink\":\"http:\\/\\/www.facebook.com\\/photo.php?fbid=10151236478529030&set=a.164247464029.118730.143412869029&type=1\",\"commentcount\":0,\"likecount\":26,\"viewercanlike\":false,\"hasviewerliked\":false,\"cancomment\":false,\"showremovemenu\":false,\"actorforpost\":\"\",\"viewerid\":\"0\",\"likesentences\":{\"current\":{\"text\":\"Janet White Pearson, Theresa O'Brien Thompson, Kimberly Jackson and 23 others like this.\",\"ranges\":[{\"offset\":0,\"length\":19,\"entities\":[{\"url\":\"\",\"shimhash\":\"xAQFWHACK\",\"id\":1579249248}]},{\"offset\":21,\"length\":24,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/tothompson1\",\"id\":29706483}]},{\"offset\":47,\"length\":16,\"entities\":[{\"url\":\"\",\"shimhash\":\"6AQFK7RpU\",\"id\":100003035312349}]}],\"aggregatedranges\":[{\"offset\":68,\"length\":9,\"count\":23}]},\"alternate\":{\"text\":\"You, Janet White Pearson, Theresa O'Brien Thompson, Kimberly Jackson and 23 others like this.\",\"ranges\":[{\"offset\":5,\"length\":19,\"entities\":[{\"url\":\"\",\"shimhash\":\"UAQHMbuxY\",\"id\":1579249248}]},{\"offset\":26,\"length\":24,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/tothompson1\",\"id\":29706483}]},{\"offset\":52,\"length\":16,\"entities\":[{\"url\":\"\",\"shimhash\":\"nAQESQtKi\",\"id\":100003035312349}]}],\"aggregatedranges\":[{\"offset\":73,\"length\":9,\"count\":23}]}},\"sharecount\":25,\"ownerid\":\"143412869029\",\"lastseentime\":null,\"canremoveall\":false,\"seencount\":0,\"seenbyall\":false,\"hasunseencollapsed\":false,\"mentionsdatasource\":{\"__m\":\"m885383_98\"},\"isranked\":false,\"isthreaded\":false,\"isownerpage\":true}],\"comments\":[],\"profiles\":[],\"actions\":[],\"commentlists\":{\"10151236478529030\":{\"range\":{\"offset\":0,\"length\":0},\"values\":[]}},\"servertime\":1352998853}],1],[\"m885382_88\",[\"UFIController\",\"m885382_87\",\"m885383_99\"],[{\"__m\":\"m885382_87\"},{\"ftentidentifier\":\"10151235033604030\",\"instanceid\":\"u4kzrn284\",\"source\":0,\"showaddcomment\":true,\"collapseaddcomment\":false,\"markedcomment\":null,\"scrollcomments\":false,\"scrollwidth\":null,\"showshares\":false,\"shownub\":false,\"numberdelimiter\":\",\",\"showtyping\":0,\"logtyping\":0,\"entstream\":false,\"timelinelogdata\":\"AQCaJQnHWDIIq7D4fMnT2vWyOQ5gD-E6J0k48gB_mRWFlnwmu7qlgrEGnU20gQW606JDgeFn2NGKhMndFnJfaxcJsroxMiOHrnmw7oPOt1ZoBpIZCDbMQ2Wb6Fg-eYqgw8QwUHNipbs2sXypICI8IrJ44Rt4HzqnYxqgklsMfVdql5hpDrX0QWSzh4NAElUXSKD_tiN18fIEdX2LyDruqCgTbnGgQfdUI-T8AVra7KCGt24F9mkehN5rUItBQ1yVt_WQHUzdupYsBebEcGqOv4kFyKrVSr5nX6Z53Ao0laXuyFWtWouJICvx1ZCyIBsAo9JWJB6XpUam_UUC60GYi018SNX-RhP_SzznrVXv97x1o575LnjamB5ZSVZM4BCSdHE\"},{\"feedbacktargets\":[{\"targetfbid\":\"10151235033604030\",\"entidentifier\":\"10151235033604030\",\"permalink\":\"\\/NorthCarolinaStateParks\\/posts\\/10151235033604030\",\"commentcount\":0,\"likecount\":33,\"viewercanlike\":false,\"hasviewerliked\":false,\"cancomment\":false,\"showremovemenu\":false,\"actorforpost\":\"\",\"viewerid\":\"0\",\"likesentences\":{\"current\":{\"text\":\"Paul T. Brown II, Kevin Stubbs, Keith Green and 30 others like this.\",\"ranges\":[{\"offset\":0,\"length\":16,\"entities\":[{\"url\":\"\",\"shimhash\":\"4AQEws-m0\",\"id\":100003240249568}]},{\"offset\":18,\"length\":12,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/kevin.stubbs.14\",\"id\":1691186129}]},{\"offset\":32,\"length\":11,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/keith.green.948\",\"id\":100001184660063}]}],\"aggregatedranges\":[{\"offset\":48,\"length\":9,\"count\":30}]},\"alternate\":{\"text\":\"You, Paul T. Brown II, Kevin Stubbs, Keith Green and 30 others like this.\",\"ranges\":[{\"offset\":5,\"length\":16,\"entities\":[{\"url\":\"\",\"shimhash\":\"nAQESQtKi\",\"id\":100003240249568}]},{\"offset\":23,\"length\":12,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/kevin.stubbs.14\",\"id\":1691186129}]},{\"offset\":37,\"length\":11,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/keith.green.948\",\"id\":100001184660063}]}],\"aggregatedranges\":[{\"offset\":53,\"length\":9,\"count\":30}]}},\"sharecount\":1,\"ownerid\":\"143412869029\",\"lastseentime\":null,\"canremoveall\":false,\"seencount\":0,\"seenbyall\":false,\"hasunseencollapsed\":false,\"mentionsdatasource\":{\"__m\":\"m885383_99\"},\"isranked\":false,\"isthreaded\":false,\"isownerpage\":true}],\"comments\":[],\"profiles\":[],\"actions\":[],\"commentlists\":{\"10151235033604030\":{\"range\":{\"offset\":0,\"length\":0},\"values\":[]}},\"servertime\":1352998853}],1],[\"m885382_90\",[\"UFIController\",\"m885382_89\",\"m885383_100\"],[{\"__m\":\"m885382_89\"},{\"ftentidentifier\":\"455084064537202\",\"instanceid\":\"u4kzrn285\",\"source\":0,\"showaddcomment\":true,\"collapseaddcomment\":false,\"markedcomment\":null,\"scrollcomments\":false,\"scrollwidth\":null,\"showshares\":false,\"shownub\":false,\"numberdelimiter\":\",\",\"showtyping\":0,\"logtyping\":0,\"entstream\":false,\"timelinelogdata\":\"AQCD4RV4MW1YWLwLMGiavl4UAfheCCNXlW77qmI3PjoQwlsQEjrvG2G1zaEME4ARqXuutKRMIUnJIkhRTgsvShqSoL04BsMTbFN19iyQdVGf_KM5PrQDm_t8Rujybz5hmsZ03wkwCjRAc-yQorowYoNiWL4NdqvGMg_MLx6vPvHoLqGlaNn7Nt_o0x-7HLATGD55M9i93uSJ_lSYvbjwU6jlKYk2FsZuY_TtXzPo88__1276RYZJsFwBdAxohRSDIhGGMcYDzXKxTv3Pgo8yfMaCFS7nB9LPL-elPfVacKl6ZoSW0fC5Ur4k0_pam3gW1DV1AAk5kXc_2Q60GqUo_dH11hS4RJwzP_wSdOOadfzm1Hj9DIxjj8wGpVCKzFbe_Qk\"},{\"feedbacktargets\":[{\"targetfbid\":\"455084064537202\",\"entidentifier\":\"455084064537202\",\"permalink\":\"\\/NorthCarolinaStateParks\\/posts\\/455084064537202\",\"commentcount\":0,\"likecount\":4,\"viewercanlike\":false,\"hasviewerliked\":false,\"cancomment\":false,\"showremovemenu\":false,\"actorforpost\":\"\",\"viewerid\":\"0\",\"likesentences\":{\"current\":{\"text\":\"Sam Bland, Paul T. Brown II and 2 others like this.\",\"ranges\":[{\"offset\":0,\"length\":9,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/sam.bland.7\",\"id\":100001451230238}]},{\"offset\":11,\"length\":16,\"entities\":[{\"url\":\"\",\"shimhash\":\"KAQFhCPEx\",\"id\":100003240249568}]}],\"aggregatedranges\":[{\"offset\":32,\"length\":8,\"count\":2}]},\"alternate\":{\"text\":\"You, Sam Bland, Paul T. Brown II and 2 others like this.\",\"ranges\":[{\"offset\":5,\"length\":9,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/sam.bland.7\",\"id\":100001451230238}]},{\"offset\":16,\"length\":16,\"entities\":[{\"url\":\"\",\"shimhash\":\"bAQEN3r6q\",\"id\":100003240249568}]}],\"aggregatedranges\":[{\"offset\":37,\"length\":8,\"count\":2}]}},\"sharecount\":0,\"ownerid\":\"143412869029\",\"lastseentime\":null,\"canremoveall\":false,\"seencount\":0,\"seenbyall\":false,\"hasunseencollapsed\":false,\"mentionsdatasource\":{\"__m\":\"m885383_100\"},\"isranked\":false,\"isthreaded\":false,\"isownerpage\":true}],\"comments\":[],\"profiles\":[],\"actions\":[],\"commentlists\":{\"455084064537202\":{\"range\":{\"offset\":0,\"length\":0},\"values\":[]}},\"servertime\":1352998853}],1],[\"m885388_102\",[\"UFIController\",\"m885388_101\",\"m885388_103\"],[{\"__m\":\"m885388_101\"},{\"ftentidentifier\":\"10151231518534030\",\"instanceid\":\"u4kzrn296\",\"source\":0,\"showaddcomment\":true,\"collapseaddcomment\":false,\"markedcomment\":null,\"scrollcomments\":false,\"scrollwidth\":null,\"showshares\":false,\"shownub\":false,\"numberdelimiter\":\",\",\"showtyping\":0,\"logtyping\":0,\"entstream\":false,\"timelinelogdata\":\"AQBuRyXbEs55soMQM5MufbdFMCnG0iDrSP9x34wPVBd_IUpEsO02S_gp9H_NCDrRefjSu74M9uG1PHeDf9EAkwzTyE3Q2mMjjIuM9HFykZVCRmOFITZzlJ2gVDCcpQ4m2bGAyDJLdwqT8JjXFLa-ZWvOzmkwxFaHroDTXHDfAm1BGOEnBZ2szOhVBrdZXQ1N4gLZXAMLzfNOkvk9vg7wcz7304hG8b65GeeNRXxxS8VH5SgoTBzovViwGCEYfX0AtX76w0Vb4ju2FvC2BxMHguDtPRCs6a7D4c33JOSxI3PpdDuFHQ8N1r5IaN21YRXqQsIn9j1i-5wJxbu4cIw0j8wfzAEByp4lMBBNwrBc4Ht4oujc33DOJJwBSyy1BJtCAb0\"},{\"feedbacktargets\":[{\"targetfbid\":\"10151231518534030\",\"entidentifier\":\"10151231518534030\",\"permalink\":\"http:\\/\\/www.facebook.com\\/photo.php?fbid=10151231518534030&set=a.164247464029.118730.143412869029&type=1\",\"commentcount\":1,\"likecount\":12,\"viewercanlike\":false,\"hasviewerliked\":false,\"cancomment\":false,\"showremovemenu\":false,\"actorforpost\":\"\",\"viewerid\":\"0\",\"likesentences\":{\"current\":{\"text\":\"Paul T. Brown II, Joey Pyron, Tim Jeffrey and 9 others like this.\",\"ranges\":[{\"offset\":0,\"length\":16,\"entities\":[{\"url\":\"\",\"shimhash\":\"SAQEjLIXR\",\"id\":100003240249568}]},{\"offset\":18,\"length\":10,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/joey.pyron\",\"id\":100002897607424}]},{\"offset\":30,\"length\":11,\"entities\":[{\"url\":\"\",\"shimhash\":\"EAQG7ZeET\",\"id\":100004131669089}]}],\"aggregatedranges\":[{\"offset\":46,\"length\":8,\"count\":9}]},\"alternate\":{\"text\":\"You, Paul T. Brown II, Joey Pyron, Tim Jeffrey and 9 others like this.\",\"ranges\":[{\"offset\":5,\"length\":16,\"entities\":[{\"url\":\"\",\"shimhash\":\"1AQFaCi3m\",\"id\":100003240249568}]},{\"offset\":23,\"length\":10,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/joey.pyron\",\"id\":100002897607424}]},{\"offset\":35,\"length\":11,\"entities\":[{\"url\":\"\",\"shimhash\":\"7AQEijtPg\",\"id\":100004131669089}]}],\"aggregatedranges\":[{\"offset\":51,\"length\":8,\"count\":9}]}},\"sharecount\":5,\"ownerid\":\"143412869029\",\"lastseentime\":null,\"canremoveall\":false,\"seencount\":0,\"seenbyall\":false,\"hasunseencollapsed\":false,\"mentionsdatasource\":{\"__m\":\"m885388_103\"},\"isranked\":false,\"isthreaded\":false,\"isownerpage\":true}],\"comments\":[{\"id\":\"10151231518534030_8249595\",\"fbid\":\"10151231520519030\",\"legacyid\":\"8249595\",\"body\":{\"text\":\"Merchants Millpond State Park has also just closed. And, all trails at Pettigrew State Park are closed.\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"143412869029\",\"ftentidentifier\":\"10151231518534030\",\"timestamp\":{\"time\":1351429253,\"text\":\"October 28 at 6:00am\",\"verbose\":\"Sunday, October 28, 2012 at 6:00am\",\"original\":1351429205},\"likecount\":0,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":null,\"istranslatable\":false}],\"profiles\":[{\"id\":\"143412869029\",\"name\":\"North Carolina State Parks and Recreation\",\"firstName\":\"North Carolina State Parks and Recreation\",\"vanity\":\"NorthCarolinaStateParks\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-ash3\\/174880_143412869029_133040642_q.jpg\",\"uri\":\"http:\\/\\/www.facebook.com\\/NorthCarolinaStateParks\",\"type\":\"page\"}],\"actions\":[],\"commentlists\":{\"10151231518534030\":{\"range\":{\"offset\":0,\"length\":1},\"values\":[\"10151231518534030_8249595\"]}},\"servertime\":1352998853}],1],[\"m885388_105\",[\"UFIController\",\"m885388_104\",\"m885388_106\"],[{\"__m\":\"m885388_104\"},{\"ftentidentifier\":\"10151233722359030\",\"instanceid\":\"u4kzrn298\",\"source\":0,\"showaddcomment\":true,\"collapseaddcomment\":false,\"markedcomment\":null,\"scrollcomments\":false,\"scrollwidth\":null,\"showshares\":false,\"shownub\":false,\"numberdelimiter\":\",\",\"showtyping\":0,\"logtyping\":0,\"entstream\":false,\"timelinelogdata\":\"AQDnFwuyVIiJwfeZUy9stHYfda8iYrNEtChJ_hL-YHn0T7Mu0BAr2SIAn31-PAmHH4HGlmXRuhDOM1iLmGnW8qBUoFWvS6fGFkgGqq8srwmmJ13n_2rnITxsn9tMXJD_0OnV6uNu157OWAa9CoB_iBvYHaZy2i1KKjZ5Iovf2Fs813S3JFPz16ZvAQLu_nRHxCP93Q8yadQXKTTjOnBEfvqvZ_l4tKoTj99NMS3zyzDO8uLrO_u5uKtmm5Lkqn_y0-sfm-0qq_-C16xmrHUAQ5AOXNgi-rZcm81HQW02jkoSKLbgPT8FM0u8HNL_66MpwsXVH9WlJqHUux5Xd0kE-7V2v5Y41aKjnLY9eD6I6F6cnSepQPDyC0jl0K5OyRHxt2c\"},{\"feedbacktargets\":[{\"targetfbid\":\"10151233722359030\",\"entidentifier\":\"10151233722359030\",\"permalink\":\"http:\\/\\/www.facebook.com\\/photo.php?fbid=10151233722359030&set=a.164247464029.118730.143412869029&type=1\",\"commentcount\":2,\"likecount\":40,\"viewercanlike\":false,\"hasviewerliked\":false,\"cancomment\":false,\"showremovemenu\":false,\"actorforpost\":\"\",\"viewerid\":\"0\",\"likesentences\":{\"current\":{\"text\":\"Tammy Lynn Walser, Michael Moore, Laura McGuinn and 37 others like this.\",\"ranges\":[{\"offset\":0,\"length\":17,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/TLWalser\",\"id\":532302702}]},{\"offset\":19,\"length\":13,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/people\\/Michael-Moore\\/1572484594\",\"id\":1572484594}]},{\"offset\":34,\"length\":13,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/foolofatook\",\"id\":845930579}]}],\"aggregatedranges\":[{\"offset\":52,\"length\":9,\"count\":37}]},\"alternate\":{\"text\":\"You, Tammy Lynn Walser, Michael Moore, Laura McGuinn and 37 others like this.\",\"ranges\":[{\"offset\":5,\"length\":17,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/TLWalser\",\"id\":532302702}]},{\"offset\":24,\"length\":13,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/people\\/Michael-Moore\\/1572484594\",\"id\":1572484594}]},{\"offset\":39,\"length\":13,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/foolofatook\",\"id\":845930579}]}],\"aggregatedranges\":[{\"offset\":57,\"length\":9,\"count\":37}]}},\"sharecount\":4,\"ownerid\":\"143412869029\",\"lastseentime\":null,\"canremoveall\":false,\"seencount\":0,\"seenbyall\":false,\"hasunseencollapsed\":false,\"mentionsdatasource\":{\"__m\":\"m885388_106\"},\"isranked\":false,\"isthreaded\":false,\"isownerpage\":true}],\"comments\":[{\"id\":\"10151233722359030_8263093\",\"fbid\":\"10151234382589030\",\"legacyid\":\"8263093\",\"body\":{\"text\":\"Love this Park, I'm headed there next weekend for a day hike and autumn colors.. :)\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"1044978470\",\"ftentidentifier\":\"10151233722359030\",\"timestamp\":{\"time\":1351561503,\"text\":\"October 29 at 6:45pm\",\"verbose\":\"Monday, October 29, 2012 at 6:45pm\"},\"likecount\":0,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":null,\"istranslatable\":false},{\"id\":\"10151233722359030_8284783\",\"fbid\":\"10151238791639030\",\"legacyid\":\"8284783\",\"body\":{\"text\":\"I like this park, went there for the first time a few weeks ago.  http:\\/\\/www.mmooredesign.com\\/\",\"ranges\":[{\"offset\":66,\"length\":28,\"entities\":[{\"url\":\"http:\\/\\/www.mmooredesign.com\\/\",\"shimhash\":\"2AQFEdpc5\",\"id\":\"8d47530eb8de30eebd72070eb14a28d0\",\"external\":true}]}],\"aggregatedranges\":[]},\"author\":\"1572484594\",\"ftentidentifier\":\"10151233722359030\",\"timestamp\":{\"time\":1351799770,\"text\":\"November 1 at 12:56pm\",\"verbose\":\"Thursday, November 1, 2012 at 12:56pm\",\"original\":1351799752},\"likecount\":0,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":null,\"istranslatable\":false}],\"profiles\":[{\"id\":\"1044978470\",\"name\":\"Brad Carter\",\"firstName\":\"Brad\",\"vanity\":\"tbcarter2\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-snc6\\/187207_1044978470_87905912_q.jpg\",\"uri\":null,\"gender\":2,\"type\":\"user\"},{\"id\":\"1572484594\",\"name\":\"Michael Moore\",\"firstName\":\"Michael\",\"vanity\":\"\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-ash4\\/276016_1572484594_1920018_q.jpg\",\"uri\":\"http:\\/\\/www.facebook.com\\/people\\/Michael-Moore\\/1572484594\",\"gender\":2,\"type\":\"user\"}],\"actions\":[],\"commentlists\":{\"10151233722359030\":{\"range\":{\"offset\":0,\"length\":2},\"values\":[\"10151233722359030_8263093\",\"10151233722359030_8284783\"]}},\"servertime\":1352998853}],1],[\"m885388_108\",[\"UFIController\",\"m885388_107\",\"m885388_149\"],[{\"__m\":\"m885388_107\"},{\"ftentidentifier\":\"10151257567514030\",\"instanceid\":\"u4kzrn2100\",\"source\":0,\"showaddcomment\":true,\"collapseaddcomment\":false,\"markedcomment\":null,\"scrollcomments\":false,\"scrollwidth\":null,\"showshares\":false,\"shownub\":false,\"numberdelimiter\":\",\",\"showtyping\":0,\"logtyping\":0,\"entstream\":false,\"timelinelogdata\":\"AQA8qzhPSztjIozKiQz1P95pq3-MAOpxEX-yUuSDLKE5IYjWaCwoRTNIoH_6qHYDVgcD6PwWAOMibLH3BdNQl7Lujcj4DF6qdAkeWiFnj0BlRp4CYGaIzjRUNWNvRQkX7UlhkKSPz2cZAdhJwZxj0OcSYbQxxMqDkJxGESKVCblw4CsWH3EfI1rQMs27QSxc8P92LcfHZ_A_xwW6gv4CP1o7lUzjq01Q4H5XJ5G0qRm3CPo67zMykVlpybwsJFWqABm2GqW1LiS1GKXtuML61PTCOzBCz6Aec40vQjBAK1EVGfT2qCX74XOG6dzxnzgdR7nlUNbBFckz_89PQKwK_6pwtlwEi2KtDZ2DcZsZVMBeYpi6BUvnlr-eDnsTlTSQ2xw\"},{\"feedbacktargets\":[{\"targetfbid\":\"10151257567514030\",\"entidentifier\":\"10151257567514030\",\"permalink\":\"http:\\/\\/www.facebook.com\\/photo.php?fbid=10151257567514030&set=a.164247464029.118730.143412869029&type=1\",\"commentcount\":2,\"likecount\":16,\"viewercanlike\":false,\"hasviewerliked\":false,\"cancomment\":false,\"showremovemenu\":false,\"actorforpost\":\"\",\"viewerid\":\"0\",\"likesentences\":{\"current\":{\"text\":\"Barbara Edlen, Londa Strong, LuAnn Bryan and 13 others like this.\",\"ranges\":[{\"offset\":0,\"length\":13,\"entities\":[{\"url\":\"\",\"shimhash\":\"FAQHPmOF9\",\"id\":1074294432}]},{\"offset\":15,\"length\":12,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/londa.strong\",\"id\":638264840}]},{\"offset\":29,\"length\":11,\"entities\":[{\"url\":\"\",\"shimhash\":\"nAQESQtKi\",\"id\":1052836190}]}],\"aggregatedranges\":[{\"offset\":45,\"length\":9,\"count\":13}]},\"alternate\":{\"text\":\"You, Barbara Edlen, Londa Strong, LuAnn Bryan and 13 others like this.\",\"ranges\":[{\"offset\":5,\"length\":13,\"entities\":[{\"url\":\"\",\"shimhash\":\"1AQFaCi3m\",\"id\":1074294432}]},{\"offset\":20,\"length\":12,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/londa.strong\",\"id\":638264840}]},{\"offset\":34,\"length\":11,\"entities\":[{\"url\":\"\",\"shimhash\":\"mAQF5-WQL\",\"id\":1052836190}]}],\"aggregatedranges\":[{\"offset\":50,\"length\":9,\"count\":13}]}},\"sharecount\":1,\"ownerid\":\"143412869029\",\"lastseentime\":null,\"canremoveall\":false,\"seencount\":0,\"seenbyall\":false,\"hasunseencollapsed\":false,\"mentionsdatasource\":{\"__m\":\"m885388_149\"},\"isranked\":false,\"isthreaded\":false,\"isownerpage\":true}],\"comments\":[{\"id\":\"10151257567514030_8378000\",\"fbid\":\"10151257762534030\",\"legacyid\":\"8378000\",\"body\":{\"text\":\"Sounds good to me! Let me know where to sign up!!\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"1624951186\",\"ftentidentifier\":\"10151257567514030\",\"timestamp\":{\"time\":1352919876,\"text\":\"21 hours ago\",\"verbose\":\"Wednesday, November 14, 2012 at 11:04am\"},\"likecount\":0,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":null,\"istranslatable\":false},{\"id\":\"10151257567514030_8378425\",\"fbid\":\"10151257838044030\",\"legacyid\":\"8378425\",\"body\":{\"text\":\"Call Haw River State Park at (336) 342-6163 to inquire about next summer's camps.\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"143412869029\",\"ftentidentifier\":\"10151257567514030\",\"timestamp\":{\"time\":1352923682,\"text\":\"20 hours ago\",\"verbose\":\"Wednesday, November 14, 2012 at 12:08pm\"},\"likecount\":0,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":null,\"istranslatable\":false}],\"profiles\":[{\"id\":\"1624951186\",\"name\":\"Linda Davis DuBose\",\"firstName\":\"Linda\",\"vanity\":\"linda.d.dubose\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-snc6\\/187129_1624951186_1835875686_q.jpg\",\"uri\":\"http:\\/\\/www.facebook.com\\/linda.d.dubose\",\"gender\":1,\"type\":\"user\"},{\"id\":\"143412869029\",\"name\":\"North Carolina State Parks and Recreation\",\"firstName\":\"North Carolina State Parks and Recreation\",\"vanity\":\"NorthCarolinaStateParks\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-ash3\\/174880_143412869029_133040642_q.jpg\",\"uri\":\"http:\\/\\/www.facebook.com\\/NorthCarolinaStateParks\",\"type\":\"page\"}],\"actions\":[],\"commentlists\":{\"10151257567514030\":{\"range\":{\"offset\":0,\"length\":2},\"values\":[\"10151257567514030_8378000\",\"10151257567514030_8378425\"]}},\"servertime\":1352998853}],1],[\"m885388_110\",[\"UFIController\",\"m885388_109\",\"m885388_150\"],[{\"__m\":\"m885388_109\"},{\"ftentidentifier\":\"10151256220039030\",\"instanceid\":\"u4kzrn2101\",\"source\":0,\"showaddcomment\":true,\"collapseaddcomment\":false,\"markedcomment\":null,\"scrollcomments\":false,\"scrollwidth\":null,\"showshares\":false,\"shownub\":false,\"numberdelimiter\":\",\",\"showtyping\":0,\"logtyping\":0,\"entstream\":false,\"timelinelogdata\":\"AQD3mjLrBfCc8XeIrxu6PIgpaYtd65n8VrHOC9sevvmFWC6E_diU2UU2pt0hkpzsBk2vSqOA4I0H1qyC4Oxf-Ig59QKnlh0SWvkdV2rrJv5QbG_9CW1xO_nfZ6Ygfhy8yqgbfIorRDYAVCdJ9TeM44YB_FhjhkKKoPrfe5fKQW-N7h0pUQ853FeIlZ2dbSOPfeXtV6l05zwBF2h5FG8E3WdFRqw5EfDc0V5oMqJZzMzHzg6RA0Awf2BqisSUFbAa9B88RlugZ0aR1fYMZltc3xSDvrYig-PzG9RtsAR7urM8QB_3n6DTypXgUDgEIYnSrS7VoSkiebKEhCo_Q-Ky-Z5YrXmVMV9bGI1ypXxC1weF125_x84IGH21WOoxcM3sSBc\"},{\"feedbacktargets\":[{\"targetfbid\":\"10151256220039030\",\"entidentifier\":\"10151256220039030\",\"permalink\":\"http:\\/\\/www.facebook.com\\/photo.php?fbid=10151256220039030&set=a.164247464029.118730.143412869029&type=1\",\"commentcount\":4,\"likecount\":32,\"viewercanlike\":false,\"hasviewerliked\":false,\"cancomment\":false,\"showremovemenu\":false,\"actorforpost\":\"\",\"viewerid\":\"0\",\"likesentences\":{\"current\":{\"text\":\"Lynn West, Pstty Shirley, Donna Engler LaPlante and 29 others like this.\",\"ranges\":[{\"offset\":0,\"length\":9,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/people\\/Lynn-West\\/100004289917948\",\"id\":100004289917948}]},{\"offset\":11,\"length\":13,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/pstty.shirley\",\"id\":100002750821855}]},{\"offset\":26,\"length\":21,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/donna.laplante1\",\"id\":100001892394365}]}],\"aggregatedranges\":[{\"offset\":52,\"length\":9,\"count\":29}]},\"alternate\":{\"text\":\"You, Lynn West, Pstty Shirley, Donna Engler LaPlante and 29 others like this.\",\"ranges\":[{\"offset\":5,\"length\":9,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/people\\/Lynn-West\\/100004289917948\",\"id\":100004289917948}]},{\"offset\":16,\"length\":13,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/pstty.shirley\",\"id\":100002750821855}]},{\"offset\":31,\"length\":21,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/donna.laplante1\",\"id\":100001892394365}]}],\"aggregatedranges\":[{\"offset\":57,\"length\":9,\"count\":29}]}},\"sharecount\":0,\"ownerid\":\"143412869029\",\"lastseentime\":null,\"canremoveall\":false,\"seencount\":0,\"seenbyall\":false,\"hasunseencollapsed\":false,\"mentionsdatasource\":{\"__m\":\"m885388_150\"},\"isranked\":false,\"isthreaded\":false,\"isownerpage\":true}],\"comments\":[{\"id\":\"10151256220039030_8369651\",\"fbid\":\"10151256250769030\",\"legacyid\":\"8369651\",\"body\":{\"text\":\"One of my fav NC Parks.\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"1379114029\",\"ftentidentifier\":\"10151256220039030\",\"timestamp\":{\"time\":1352823740,\"text\":\"Tuesday at 8:22am\",\"verbose\":\"Tuesday, November 13, 2012 at 8:22am\"},\"likecount\":1,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":null,\"istranslatable\":false},{\"id\":\"10151256220039030_8376282\",\"fbid\":\"10151257449434030\",\"legacyid\":\"8376282\",\"body\":{\"text\":\"Eeerrie!!!\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"100002750821855\",\"ftentidentifier\":\"10151256220039030\",\"timestamp\":{\"time\":1352902793,\"text\":\"Yesterday at 6:19am\",\"verbose\":\"Wednesday, November 14, 2012 at 6:19am\"},\"likecount\":0,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":null,\"istranslatable\":false}],\"profiles\":[{\"id\":\"1379114029\",\"name\":\"Trey McLamb\",\"firstName\":\"Trey\",\"vanity\":\"trey.mclamb.3\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-ash4\\/368893_1379114029_909009035_q.jpg\",\"uri\":\"http:\\/\\/www.facebook.com\\/trey.mclamb.3\",\"gender\":2,\"type\":\"user\"},{\"id\":\"100002750821855\",\"name\":\"Pstty Shirley\",\"firstName\":\"Pstty\",\"vanity\":\"pstty.shirley\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-snc6\\/273541_100002750821855_4422200_q.jpg\",\"uri\":\"http:\\/\\/www.facebook.com\\/pstty.shirley\",\"gender\":1,\"type\":\"user\"}],\"actions\":[],\"commentlists\":{\"10151256220039030\":{\"range\":{\"offset\":2,\"length\":2},\"values\":[\"10151256220039030_8369651\",\"10151256220039030_8376282\"]}},\"servertime\":1352998853}],1],[\"m885388_112\",[\"UFIController\",\"m885388_111\",\"m885388_151\"],[{\"__m\":\"m885388_111\"},{\"ftentidentifier\":\"10151256146579030\",\"instanceid\":\"u4kzrn2102\",\"source\":0,\"showaddcomment\":true,\"collapseaddcomment\":false,\"markedcomment\":null,\"scrollcomments\":false,\"scrollwidth\":null,\"showshares\":false,\"shownub\":false,\"numberdelimiter\":\",\",\"showtyping\":0,\"logtyping\":0,\"entstream\":false,\"timelinelogdata\":\"AQC7h0LfKotKK4FK-mSebPy1Fr2W-WJM39slH0FOKnDBDHdtikQTj8LIcDGoZaJJv6RCYsSdn0V1uoPZO-I590nrU5V8zi45uuKF5Yx5_0qezDkRGVFaanZsHET_PZpifzdp0MUlwEmO-Kw174APy-Bqk_sEerPe-FRvu7W91YT26nFYm0gkX5BriMCU0Ilh8h_cQwZiiij60sx21n8RG-lpbnt5Gdr0YlMkoj3-nGUHuCDiUBVrROcelaUVzPPAdSwEyp7KA518s0J_OQIIExahx--BbCjt1Jxi-QyTEC16c3dW9Hy40Kc10vey5q-T0dRL6Rrgf0SPU2p_xneSVzoubMotOfOJq9cZnroE_tvA05YPR3psOtAxYBRaRzIC-DE\"},{\"feedbacktargets\":[{\"targetfbid\":\"10151256146579030\",\"entidentifier\":\"10151256146579030\",\"permalink\":\"http:\\/\\/www.facebook.com\\/photo.php?fbid=10151256146579030&set=a.164247464029.118730.143412869029&type=1\",\"commentcount\":5,\"likecount\":96,\"viewercanlike\":false,\"hasviewerliked\":false,\"cancomment\":false,\"showremovemenu\":false,\"actorforpost\":\"\",\"viewerid\":\"0\",\"likesentences\":{\"current\":{\"text\":\"Yvonne Lee, NatureFind Charlotte, NatureFind Raleigh-Durham and 93 others like this.\",\"ranges\":[{\"offset\":0,\"length\":10,\"entities\":[{\"url\":\"\",\"shimhash\":\"vAQF9o7cr\",\"id\":1531397905}]},{\"offset\":12,\"length\":20,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/NatureFindCharlotte\",\"id\":102003176554068}]},{\"offset\":34,\"length\":25,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/NatureFindRaleighDurham\",\"id\":171854356200534}]}],\"aggregatedranges\":[{\"offset\":64,\"length\":9,\"count\":93}]},\"alternate\":{\"text\":\"You, Yvonne Lee, NatureFind Charlotte, NatureFind Raleigh-Durham and 93 others like this.\",\"ranges\":[{\"offset\":5,\"length\":10,\"entities\":[{\"url\":\"\",\"shimhash\":\"cAQHGPsjp\",\"id\":1531397905}]},{\"offset\":17,\"length\":20,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/NatureFindCharlotte\",\"id\":102003176554068}]},{\"offset\":39,\"length\":25,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/NatureFindRaleighDurham\",\"id\":171854356200534}]}],\"aggregatedranges\":[{\"offset\":69,\"length\":9,\"count\":93}]}},\"sharecount\":41,\"ownerid\":\"143412869029\",\"lastseentime\":null,\"canremoveall\":false,\"seencount\":0,\"seenbyall\":false,\"hasunseencollapsed\":false,\"mentionsdatasource\":{\"__m\":\"m885388_151\"},\"isranked\":false,\"isthreaded\":false,\"isownerpage\":true}],\"comments\":[{\"id\":\"10151256146579030_8371954\",\"fbid\":\"10151256648179030\",\"legacyid\":\"8371954\",\"body\":{\"text\":\"I think it truly desrves this  ... it is beautiful.\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"100000573931493\",\"ftentidentifier\":\"10151256146579030\",\"timestamp\":{\"time\":1352844511,\"text\":\"Tuesday at 2:08pm\",\"verbose\":\"Tuesday, November 13, 2012 at 2:08pm\"},\"likecount\":0,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":null,\"istranslatable\":false},{\"id\":\"10151256146579030_8372941\",\"fbid\":\"10151256815104030\",\"legacyid\":\"8372941\",\"body\":{\"text\":\"where is it?\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"528102547\",\"ftentidentifier\":\"10151256146579030\",\"timestamp\":{\"time\":1352854686,\"text\":\"Tuesday at 4:58pm\",\"verbose\":\"Tuesday, November 13, 2012 at 4:58pm\"},\"likecount\":0,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":null,\"istranslatable\":false}],\"profiles\":[{\"id\":\"100000573931493\",\"name\":\"Anne Runyon\",\"firstName\":\"Anne\",\"vanity\":\"anne.runyon.1\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-ash4\\/211617_100000573931493_1691646344_q.jpg\",\"uri\":\"http:\\/\\/www.facebook.com\\/anne.runyon.1\",\"gender\":1,\"type\":\"user\"},{\"id\":\"528102547\",\"name\":\"Tim Bollinger\",\"firstName\":\"Tim\",\"vanity\":\"tim.bollinger.16\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-ash4\\/195354_528102547_1908087663_q.jpg\",\"uri\":null,\"gender\":2,\"type\":\"user\"}],\"actions\":[],\"commentlists\":{\"10151256146579030\":{\"range\":{\"offset\":3,\"length\":2},\"values\":[\"10151256146579030_8371954\",\"10151256146579030_8372941\"]}},\"servertime\":1352998853}],1],[\"m885388_114\",[\"UFIController\",\"m885388_113\",\"m885388_152\"],[{\"__m\":\"m885388_113\"},{\"ftentidentifier\":\"10151255074004030\",\"instanceid\":\"u4kzrn2103\",\"source\":0,\"showaddcomment\":true,\"collapseaddcomment\":false,\"markedcomment\":null,\"scrollcomments\":false,\"scrollwidth\":null,\"showshares\":false,\"shownub\":false,\"numberdelimiter\":\",\",\"showtyping\":0,\"logtyping\":0,\"entstream\":false,\"timelinelogdata\":\"AQBj4ApubcZLzM98wFDQz1UiEzCxbmSjVsz-eEBX_eKK3eI7_gvGe9QcwB-HIzK2SDdUslUH350OaOYH77BJy_4JhtUlkRhDXsGhFgHnG74pkhTj3jaICB9VbXJX_VUImo5ERbRGHWKbGt-lR2vozPt1JWLD_knd3IwGSe8_slL9uPyW4sTwYGfSNrGP-_Ceb2AtZCLTmOR921xhOQJFr1e_C1VsEvQwFczHTN8G1xIrhbnRxpI-0PyLo7vx2nweD4hnjLGe4NorZ9bJGbL5f2EQFN-zpfrRbay_d046li2e12WhBinzPROkbKntwccRhGdxA-QCCZRaQZI-ZgKf2VmZ4-ry_FljdF99bOqMZZxKn4VAzm-eqpMrIf6oJacOWEE\"},{\"feedbacktargets\":[{\"targetfbid\":\"10151255074004030\",\"entidentifier\":\"10151255074004030\",\"permalink\":\"http:\\/\\/www.facebook.com\\/photo.php?fbid=10151255074004030&set=a.164247464029.118730.143412869029&type=1\",\"commentcount\":8,\"likecount\":143,\"viewercanlike\":false,\"hasviewerliked\":false,\"cancomment\":false,\"showremovemenu\":false,\"actorforpost\":\"\",\"viewerid\":\"0\",\"likesentences\":{\"current\":{\"text\":\"Teresa Bowman, Ultimate Office Assistants, Deborah Spence and 140 others like this.\",\"ranges\":[{\"offset\":0,\"length\":13,\"entities\":[{\"url\":\"\",\"shimhash\":\"bAQEN3r6q\",\"id\":1253575449}]},{\"offset\":15,\"length\":26,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/ultimateofficeassistants\",\"id\":192095360810344}]},{\"offset\":43,\"length\":14,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/deborah.spence2\",\"id\":1835920678}]}],\"aggregatedranges\":[{\"offset\":62,\"length\":10,\"count\":140}]},\"alternate\":{\"text\":\"You, Teresa Bowman, Ultimate Office Assistants, Deborah Spence and 140 others like this.\",\"ranges\":[{\"offset\":5,\"length\":13,\"entities\":[{\"url\":\"\",\"shimhash\":\"wAQGAB9qs\",\"id\":1253575449}]},{\"offset\":20,\"length\":26,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/ultimateofficeassistants\",\"id\":192095360810344}]},{\"offset\":48,\"length\":14,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/deborah.spence2\",\"id\":1835920678}]}],\"aggregatedranges\":[{\"offset\":67,\"length\":10,\"count\":140}]}},\"sharecount\":13,\"ownerid\":\"143412869029\",\"lastseentime\":null,\"canremoveall\":false,\"seencount\":0,\"seenbyall\":false,\"hasunseencollapsed\":false,\"mentionsdatasource\":{\"__m\":\"m885388_152\"},\"isranked\":false,\"isthreaded\":false,\"isownerpage\":true}],\"comments\":[{\"id\":\"10151255074004030_8369412\",\"fbid\":\"10151256221539030\",\"legacyid\":\"8369412\",\"body\":{\"text\":\"Indeed, a well done job by all, including State Park office, maintenance, and staff drawn from other parks.\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"100000929967603\",\"ftentidentifier\":\"10151255074004030\",\"timestamp\":{\"time\":1352822188,\"text\":\"Tuesday at 7:56am\",\"verbose\":\"Tuesday, November 13, 2012 at 7:56am\"},\"likecount\":0,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":null,\"istranslatable\":false},{\"id\":\"10151255074004030_8373877\",\"fbid\":\"10151256978329030\",\"legacyid\":\"8373877\",\"body\":{\"text\":\"Thank so very much!\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"1835920678\",\"ftentidentifier\":\"10151255074004030\",\"timestamp\":{\"time\":1352865005,\"text\":\"Tuesday at 7:50pm\",\"verbose\":\"Tuesday, November 13, 2012 at 7:50pm\"},\"likecount\":0,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":null,\"istranslatable\":false}],\"profiles\":[{\"id\":\"100000929967603\",\"name\":\"Joseph Williams\",\"firstName\":\"Joseph\",\"vanity\":\"JosephKellyWilliams\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-ash4\\/370434_100000929967603_1340823992_q.jpg\",\"uri\":null,\"gender\":2,\"type\":\"user\"},{\"id\":\"1835920678\",\"name\":\"Deborah Spence\",\"firstName\":\"Deborah\",\"vanity\":\"deborah.spence2\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/static-ak\\/rsrc.php\\/v2\\/y9\\/r\\/IB7NOFmPw2a.gif\",\"uri\":\"http:\\/\\/www.facebook.com\\/deborah.spence2\",\"gender\":1,\"type\":\"user\"}],\"actions\":[],\"commentlists\":{\"10151255074004030\":{\"range\":{\"offset\":6,\"length\":2},\"values\":[\"10151255074004030_8369412\",\"10151255074004030_8373877\"]}},\"servertime\":1352998853}],1],[\"m885388_116\",[\"UFIController\",\"m885388_115\",\"m885388_153\"],[{\"__m\":\"m885388_115\"},{\"ftentidentifier\":\"10151249662709030\",\"instanceid\":\"u4kzrn2104\",\"source\":0,\"showaddcomment\":true,\"collapseaddcomment\":false,\"markedcomment\":null,\"scrollcomments\":false,\"scrollwidth\":null,\"showshares\":false,\"shownub\":false,\"numberdelimiter\":\",\",\"showtyping\":0,\"logtyping\":0,\"entstream\":false,\"timelinelogdata\":\"AQAEagyljGsWUQEX3FTGfzV3XfHi2wZyPPIxYTbNhG672WPUin0I8VEoO9xYf0mPGkMk6f_99IGNAORZfoxHo6AjPtJHSKffmVFfa0EFsjYVbfhH5dsAY2ZejXbf3jCd1IKRyqShHeZ6GanOvRZV1f4xH-xANVgPDZfClGifUMVPwA3P6gl1_u7b4OuKLDw0BzYqJD8v1a0YBCkorIs_25cG3m3AOeoco07FWRFJmQb_A223KuhUSMV14s0qTUwAOf3wtd7WJXxaFtNg99QRigMxdZGR_xEqgM3U3OTmKYDA2zz6s4M5QdwsE1L7zkLoJljI6lX1lNOjzO3lWkwBWQ14iQfGKSa7uRC4GwMRtCgOBa64tsjcQQ07ogdwx8DsRiE\"},{\"feedbacktargets\":[{\"targetfbid\":\"10151249662709030\",\"entidentifier\":\"10151249662709030\",\"permalink\":\"\\/NorthCarolinaStateParks\\/posts\\/10151249662709030\",\"commentcount\":1,\"likecount\":12,\"viewercanlike\":false,\"hasviewerliked\":false,\"cancomment\":false,\"showremovemenu\":false,\"actorforpost\":\"\",\"viewerid\":\"0\",\"likesentences\":{\"current\":{\"text\":\"Four Seasons Cabins & Cottages, Visit Gaston, Donald Barnett and 9 others like this.\",\"ranges\":[{\"offset\":0,\"length\":30,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/fourseasonscottages\",\"id\":106968456044405}]},{\"offset\":32,\"length\":12,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/visitgaston\",\"id\":136945563049549}]},{\"offset\":46,\"length\":14,\"entities\":[{\"url\":\"\",\"shimhash\":\"FAQHPmOF9\",\"id\":100000400766417}]}],\"aggregatedranges\":[{\"offset\":65,\"length\":8,\"count\":9}]},\"alternate\":{\"text\":\"You, Four Seasons Cabins & Cottages, Visit Gaston, Donald Barnett and 9 others like this.\",\"ranges\":[{\"offset\":5,\"length\":30,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/fourseasonscottages\",\"id\":106968456044405}]},{\"offset\":37,\"length\":12,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/visitgaston\",\"id\":136945563049549}]},{\"offset\":51,\"length\":14,\"entities\":[{\"url\":\"\",\"shimhash\":\"wAQGAB9qs\",\"id\":100000400766417}]}],\"aggregatedranges\":[{\"offset\":70,\"length\":8,\"count\":9}]}},\"sharecount\":4,\"ownerid\":\"143412869029\",\"lastseentime\":null,\"canremoveall\":false,\"seencount\":0,\"seenbyall\":false,\"hasunseencollapsed\":false,\"mentionsdatasource\":{\"__m\":\"m885388_153\"},\"isranked\":false,\"isthreaded\":false,\"isownerpage\":true}],\"comments\":[{\"id\":\"10151249662709030_25062266\",\"fbid\":\"10151249683489030\",\"legacyid\":\"25062266\",\"body\":{\"text\":\"Controlled burn got outta control?\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"819247968\",\"ftentidentifier\":\"10151249662709030\",\"timestamp\":{\"time\":1352420088,\"text\":\"November 8 at 4:14pm\",\"verbose\":\"Thursday, November 8, 2012 at 4:14pm\"},\"likecount\":1,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":1,\"istranslatable\":false}],\"profiles\":[{\"id\":\"819247968\",\"name\":\"Michelle Leigh\",\"firstName\":\"Michelle\",\"vanity\":\"shelleybean11\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-snc6\\/274998_819247968_1267305762_q.jpg\",\"uri\":null,\"gender\":1,\"type\":\"user\"}],\"actions\":[],\"commentlists\":{\"10151249662709030\":{\"range\":{\"offset\":0,\"length\":1},\"values\":[\"10151249662709030_25062266\"]}},\"servertime\":1352998853}],1],[\"m885388_118\",[\"UFIController\",\"m885388_117\",\"m885388_154\"],[{\"__m\":\"m885388_117\"},{\"ftentidentifier\":\"10151247406854030\",\"instanceid\":\"u4kzrn2105\",\"source\":0,\"showaddcomment\":true,\"collapseaddcomment\":false,\"markedcomment\":null,\"scrollcomments\":false,\"scrollwidth\":null,\"showshares\":false,\"shownub\":false,\"numberdelimiter\":\",\",\"showtyping\":0,\"logtyping\":0,\"entstream\":false,\"timelinelogdata\":\"AQCrd_m6uyiOcyebCsidrjjs6icVJ4rA5eAmGibXWYIWLq8e7RAGlF-uqFwPZNVhXUndlHOUSFUdlh6qQscPBqSRdN47Pd3sdGJfev1fKYteXKHqBEcsIIZPiMrqye3JX-vxx1EiT_48J67aVG3qBDadArG98EMFBu7OAVYUnx1w9hEmnDsgjCUzScl8L0AWOSAnUV15eIg3C7-tbm_6jFr3wQMh-TWYecpypV3296tRszLCs7Sz3JDHCErcXtq_y9VHyT7xEtu1lhWJyI5tBz-4Ii_1Y5skZGNg7C1yi_dcnRYkRb3dwbSMYbt5i_PASltO-SMCUaw_h7U4SPCHS2lLFRI68ePKwNalIRR1ZOW84-_NE-774hk8ifm7ZK1GyVg\"},{\"feedbacktargets\":[{\"targetfbid\":\"10151247406854030\",\"entidentifier\":\"10151247406854030\",\"permalink\":\"http:\\/\\/www.facebook.com\\/photo.php?fbid=10151247406854030&set=a.164247464029.118730.143412869029&type=1\",\"commentcount\":2,\"likecount\":64,\"viewercanlike\":false,\"hasviewerliked\":false,\"cancomment\":false,\"showremovemenu\":false,\"actorforpost\":\"\",\"viewerid\":\"0\",\"likesentences\":{\"current\":{\"text\":\"Denise Antonacci Kenney, Philip Woodward, LaNan Justice and 61 others like this.\",\"ranges\":[{\"offset\":0,\"length\":23,\"entities\":[{\"url\":\"\",\"shimhash\":\"bAQEN3r6q\",\"id\":1319934126}]},{\"offset\":25,\"length\":15,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/philip.woodward.33\",\"id\":100001424529980}]},{\"offset\":42,\"length\":13,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/lanan.justice\",\"id\":1336167930}]}],\"aggregatedranges\":[{\"offset\":60,\"length\":9,\"count\":61}]},\"alternate\":{\"text\":\"You, Denise Antonacci Kenney, Philip Woodward, LaNan Justice and 61 others like this.\",\"ranges\":[{\"offset\":5,\"length\":23,\"entities\":[{\"url\":\"\",\"shimhash\":\"6AQFK7RpU\",\"id\":1319934126}]},{\"offset\":30,\"length\":15,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/philip.woodward.33\",\"id\":100001424529980}]},{\"offset\":47,\"length\":13,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/lanan.justice\",\"id\":1336167930}]}],\"aggregatedranges\":[{\"offset\":65,\"length\":9,\"count\":61}]}},\"sharecount\":24,\"ownerid\":\"143412869029\",\"lastseentime\":null,\"canremoveall\":false,\"seencount\":0,\"seenbyall\":false,\"hasunseencollapsed\":false,\"mentionsdatasource\":{\"__m\":\"m885388_154\"},\"isranked\":false,\"isthreaded\":false,\"isownerpage\":true}],\"comments\":[{\"id\":\"10151247406854030_8326113\",\"fbid\":\"10151247492029030\",\"legacyid\":\"8326113\",\"body\":{\"text\":\"Here is another story on Monarchs posted on The NCCF on Line Review.  http:\\/\\/www.nccoast.org\\/Blog-Post.aspx?k=bb953419-4b4b-432f-acbc-8686489cbec9\",\"ranges\":[{\"offset\":70,\"length\":76,\"entities\":[{\"url\":\"http:\\/\\/www.nccoast.org\\/Blog-Post.aspx?k=bb953419-4b4b-432f-acbc-8686489cbec9\",\"shimhash\":\"UAQHMbuxY\",\"id\":\"2e530020f422da28eb65a1610f6e6e1b\",\"external\":true}]}],\"aggregatedranges\":[]},\"author\":\"100001451230238\",\"ftentidentifier\":\"10151247406854030\",\"timestamp\":{\"time\":1352298837,\"text\":\"November 7 at 6:33am\",\"verbose\":\"Wednesday, November 7, 2012 at 6:33am\"},\"likecount\":0,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":null,\"istranslatable\":false},{\"id\":\"10151247406854030_8329225\",\"fbid\":\"10151248209999030\",\"legacyid\":\"8329225\",\"body\":{\"text\":\"A few years ago, we were in Bethany Beach, Delaware.  It was early October.  One afternoon, while enjoying an indoor pool, with glass all around.  A friend took a walk and found hundreds of Monarch Butterflies, in a bush, by this pool, on the beach.  They were there for that day and the next",
        " morning, then they left.  We were in the same place, this past Sept. and there were Monarchs, again, but not nearly as many as before.  They are beautiful and fascinating.\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"100000084052035\",\"ftentidentifier\":\"10151247406854030\",\"timestamp\":{\"time\":1352328943,\"text\":\"November 7 at 2:55pm\",\"verbose\":\"Wednesday, November 7, 2012 at 2:55pm\"},\"likecount\":0,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":null,\"istranslatable\":false}],\"profiles\":[{\"id\":\"100001451230238\",\"name\":\"Sam Bland\",\"firstName\":\"Sam\",\"vanity\":\"sam.bland.7\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-prn1\\/41383_100001451230238_4728_q.jpg\",\"uri\":\"http:\\/\\/www.facebook.com\\/sam.bland.7\",\"gender\":2,\"type\":\"user\"},{\"id\":\"100000084052035\",\"name\":\"Jane Miller\",\"firstName\":\"Jane\",\"vanity\":\"jane.miller.5817\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-prn1\\/161355_100000084052035_8034453_q.jpg\",\"uri\":\"http:\\/\\/www.facebook.com\\/jane.miller.5817\",\"gender\":1,\"type\":\"user\"}],\"actions\":[],\"commentlists\":{\"10151247406854030\":{\"range\":{\"offset\":0,\"length\":2},\"values\":[\"10151247406854030_8326113\",\"10151247406854030_8329225\"]}},\"servertime\":1352998853}],1],[\"m885388_120\",[\"UFIController\",\"m885388_119\",\"m885388_155\"],[{\"__m\":\"m885388_119\"},{\"ftentidentifier\":\"10151245839549030\",\"instanceid\":\"u4kzrn2106\",\"source\":0,\"showaddcomment\":true,\"collapseaddcomment\":false,\"markedcomment\":null,\"scrollcomments\":false,\"scrollwidth\":null,\"showshares\":false,\"shownub\":false,\"numberdelimiter\":\",\",\"showtyping\":0,\"logtyping\":0,\"entstream\":false,\"timelinelogdata\":\"AQCLzqkRuHFMuIY518yuhXOwUckcZaVKnlZI-Yv_fk98gIX6o2gnFYK7IIthbqsbccmh2S0jRD63Vy28toZUVv2OjZ6XcNAFVowD14vUWstA9lw5TJTo9z4fXqQv0JH0P6CP_zbau14XY2uNCc14Z7U34QgCW3lnMISI3QUbRpgbpMAwHk6w0m5FFfL9LHKx7j7zeIIoUz_PEnre-M9XTytCOlnUvY0QaAbmfcUM4V5MQ9-aWBTWBtPE06HfxY-ioUOF8L5rBdHErJvKqIkX4fCcj82CF9OOQ6bY02TGuVNjnRsdJz-8hnaoDU-7AKui-ME6bw-_A7ls0RgspIAha7ZVtm_Hxy91SXCZ_uD0bdLtfFBKU2EPR0XqrltqzSbjSRs\"},{\"feedbacktargets\":[{\"targetfbid\":\"10151245839549030\",\"entidentifier\":\"10151245839549030\",\"permalink\":\"http:\\/\\/www.facebook.com\\/photo.php?fbid=10151245839549030&set=a.164247464029.118730.143412869029&type=1\",\"commentcount\":3,\"likecount\":115,\"viewercanlike\":false,\"hasviewerliked\":false,\"cancomment\":false,\"showremovemenu\":false,\"actorforpost\":\"\",\"viewerid\":\"0\",\"likesentences\":{\"current\":{\"text\":\"Greensboro Gal, Kim Haigler, Kellie DuBose Duke and 112 others like this.\",\"ranges\":[{\"offset\":0,\"length\":14,\"entities\":[{\"url\":\"\",\"shimhash\":\"CAQF8Bvp8\",\"id\":100001359534375}]},{\"offset\":16,\"length\":11,\"entities\":[{\"url\":\"\",\"shimhash\":\"ZAQHJsopg\",\"id\":100002631863425}]},{\"offset\":29,\"length\":18,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/kellie.d.duke\",\"id\":1541076476}]}],\"aggregatedranges\":[{\"offset\":52,\"length\":10,\"count\":112}]},\"alternate\":{\"text\":\"You, Greensboro Gal, Kim Haigler, Kellie DuBose Duke and 112 others like this.\",\"ranges\":[{\"offset\":5,\"length\":14,\"entities\":[{\"url\":\"\",\"shimhash\":\"UAQHMbuxY\",\"id\":100001359534375}]},{\"offset\":21,\"length\":11,\"entities\":[{\"url\":\"\",\"shimhash\":\"vAQF9o7cr\",\"id\":100002631863425}]},{\"offset\":34,\"length\":18,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/kellie.d.duke\",\"id\":1541076476}]}],\"aggregatedranges\":[{\"offset\":57,\"length\":10,\"count\":112}]}},\"sharecount\":19,\"ownerid\":\"143412869029\",\"lastseentime\":null,\"canremoveall\":false,\"seencount\":0,\"seenbyall\":false,\"hasunseencollapsed\":false,\"mentionsdatasource\":{\"__m\":\"m885388_155\"},\"isranked\":false,\"isthreaded\":false,\"isownerpage\":true}],\"comments\":[{\"id\":\"10151245839549030_8318215\",\"fbid\":\"10151245873584030\",\"legacyid\":\"8318215\",\"body\":{\"text\":\"'love this!!!!\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"674380708\",\"ftentidentifier\":\"10151245839549030\",\"timestamp\":{\"time\":1352211379,\"text\":\"November 6 at 6:16am\",\"verbose\":\"Tuesday, November 6, 2012 at 6:16am\"},\"likecount\":1,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":null,\"istranslatable\":false},{\"id\":\"10151245839549030_8318594\",\"fbid\":\"10151245961949030\",\"legacyid\":\"8318594\",\"body\":{\"text\":\"Please consider supporting candidates who recognize the value of our great park system.  Keep them open and accessible!\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"576324451\",\"ftentidentifier\":\"10151245839549030\",\"timestamp\":{\"time\":1352215841,\"text\":\"November 6 at 7:30am\",\"verbose\":\"Tuesday, November 6, 2012 at 7:30am\"},\"likecount\":2,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":null,\"istranslatable\":false},{\"id\":\"10151245839549030_8320209\",\"fbid\":\"10151246312634030\",\"legacyid\":\"8320209\",\"body\":{\"text\":\"Exactly what I did. Florence Nature Preserve to Little Pisgah.\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"501001443\",\"ftentidentifier\":\"10151245839549030\",\"timestamp\":{\"time\":1352230515,\"text\":\"November 6 at 11:35am\",\"verbose\":\"Tuesday, November 6, 2012 at 11:35am\"},\"likecount\":1,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":null,\"istranslatable\":false}],\"profiles\":[{\"id\":\"674380708\",\"name\":\"Beverly Clifton Murrelle\",\"firstName\":\"Beverly\",\"vanity\":\"bevmurrelle\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-snc6\\/203329_674380708_1730813702_q.jpg\",\"uri\":\"http:\\/\\/www.facebook.com\\/bevmurrelle\",\"gender\":1,\"type\":\"user\"},{\"id\":\"576324451\",\"name\":\"Mary Mcnaught\",\"firstName\":\"Mary\",\"vanity\":\"mary.mcnaught.142\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-snc6\\/276359_576324451_2009470_q.jpg\",\"uri\":null,\"gender\":7,\"type\":\"user\"},{\"id\":\"501001443\",\"name\":\"Kevin Scott Simpson\",\"firstName\":\"Kevin\",\"vanity\":\"kevin.s.simpson\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-snc6\\/275414_501001443_645524956_q.jpg\",\"uri\":\"http:\\/\\/www.facebook.com\\/kevin.s.simpson\",\"gender\":7,\"type\":\"user\"}],\"actions\":[],\"commentlists\":{\"10151245839549030\":{\"range\":{\"offset\":0,\"length\":3},\"values\":[\"10151245839549030_8318215\",\"10151245839549030_8318594\",\"10151245839549030_8320209\"]}},\"servertime\":1352998853}],1],[\"m885388_122\",[\"UFIController\",\"m885388_121\",\"m885388_156\"],[{\"__m\":\"m885388_121\"},{\"ftentidentifier\":\"130429140441330\",\"instanceid\":\"u4kzrn2107\",\"source\":0,\"showaddcomment\":true,\"collapseaddcomment\":false,\"markedcomment\":null,\"scrollcomments\":false,\"scrollwidth\":null,\"showshares\":false,\"shownub\":false,\"numberdelimiter\":\",\",\"showtyping\":0,\"logtyping\":0,\"entstream\":false,\"timelinelogdata\":\"AQD-Dpa30TK5BsBDqf3QPdXzpD6RMF0cyQEXQ_LIfkSsmTTlalqULRmJFWs_0dV0Jtb00T5q7Zp9kblolGbI515BjiaO5SKGJM5GoG7b2zTbPJ7YdcJv--GrZ4bO0CrXR1pbFsURpF9QOgD-MwjUV5eeSYSsME690GHNBuHOaBukJK1X-_6tNsTfFoOj3JiCY63HcQIQ1KPZmH3zDPMpAQmatDS6fqdjjVLJwODGnVT53QTXmeuKvrPnAomabToVBYNckbdDwUYobowq8ROQYZooXIYdcG34qXLJXjEp-MxAG_Q-4pSakg1P4CosPlrnot4VC-a_lJqWqXlAZw-9PLPdZN2vHscRCzHUt27B4EIR5RwTsoolMCqcX8lPshxz6ns\"},{\"feedbacktargets\":[{\"targetfbid\":\"130429140441330\",\"entidentifier\":\"130429140441330\",\"permalink\":\"\\/NorthCarolinaStateParks\\/posts\\/130429140441330\",\"commentcount\":2,\"likecount\":4,\"viewercanlike\":false,\"hasviewerliked\":false,\"cancomment\":false,\"showremovemenu\":false,\"actorforpost\":\"\",\"viewerid\":\"0\",\"likesentences\":{\"current\":{\"text\":\"Paul T. Brown II, Elaine Williams and 2 others like this.\",\"ranges\":[{\"offset\":0,\"length\":16,\"entities\":[{\"url\":\"\",\"shimhash\":\"xAQFWHACK\",\"id\":100003240249568}]},{\"offset\":18,\"length\":15,\"entities\":[{\"url\":\"\",\"shimhash\":\"xAQFWHACK\",\"id\":100000966877835}]}],\"aggregatedranges\":[{\"offset\":38,\"length\":8,\"count\":2}]},\"alternate\":{\"text\":\"You, Paul T. Brown II, Elaine Williams and 2 others like this.\",\"ranges\":[{\"offset\":5,\"length\":16,\"entities\":[{\"url\":\"\",\"shimhash\":\"sAQEQPezu\",\"id\":100003240249568}]},{\"offset\":23,\"length\":15,\"entities\":[{\"url\":\"\",\"shimhash\":\"-AQG4vz5a\",\"id\":100000966877835}]}],\"aggregatedranges\":[{\"offset\":43,\"length\":8,\"count\":2}]}},\"sharecount\":0,\"ownerid\":\"143412869029\",\"lastseentime\":null,\"canremoveall\":false,\"seencount\":0,\"seenbyall\":false,\"hasunseencollapsed\":false,\"mentionsdatasource\":{\"__m\":\"m885388_156\"},\"isranked\":false,\"isthreaded\":false,\"isownerpage\":true}],\"comments\":[{\"id\":\"130429140441330_190307\",\"fbid\":\"130433100440934\",\"legacyid\":\"190307\",\"body\":{\"text\":\"This is not the solution...please read the comments on the link you provide...this issue has been a headache for both users and businesses for a long time!!!...\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"645854192\",\"ftentidentifier\":\"130429140441330\",\"timestamp\":{\"time\":1352129414,\"text\":\"November 5 at 7:30am\",\"verbose\":\"Monday, November 5, 2012 at 7:30am\"},\"likecount\":0,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":0,\"istranslatable\":false},{\"id\":\"130429140441330_190442\",\"fbid\":\"130448403772737\",\"legacyid\":\"190442\",\"body\":{\"text\":\"I think you actually want people to click on \\\"Show in News Feed.\\\" Last line of the CIO article - \\\"Click the second option, 'Show in News Feed,' to make sure all the page's posts appear in your news feed.\\\"\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"100002714150505\",\"ftentidentifier\":\"130429140441330\",\"timestamp\":{\"time\":1352132955,\"text\":\"November 5 at 8:29am\",\"verbose\":\"Monday, November 5, 2012 at 8:29am\"},\"likecount\":0,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":0,\"istranslatable\":false}],\"profiles\":[{\"id\":\"645854192\",\"name\":\"Mike Kerkado\",\"firstName\":\"Mike\",\"vanity\":\"kerkado\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-snc6\\/273893_645854192_1467633529_q.jpg\",\"uri\":null,\"gender\":2,\"type\":\"user\"},{\"id\":\"100002714150505\",\"name\":\"Chris Nelson\",\"firstName\":\"Chris\",\"vanity\":\"\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-ash4\\/186082_100002714150505_3824686_q.jpg\",\"uri\":\"http:\\/\\/www.facebook.com\\/people\\/Chris-Nelson\\/100002714150505\",\"gender\":2,\"type\":\"user\"}],\"actions\":[],\"commentlists\":{\"130429140441330\":{\"range\":{\"offset\":0,\"length\":2},\"values\":[\"130429140441330_190307\",\"130429140441330_190442\"]}},\"servertime\":1352998853}],1],[\"m885388_124\",[\"UFIController\",\"m885388_123\",\"m885388_157\"],[{\"__m\":\"m885388_123\"},{\"ftentidentifier\":\"10151244259254030\",\"instanceid\":\"u4kzrn2108\",\"source\":0,\"showaddcomment\":true,\"collapseaddcomment\":false,\"markedcomment\":null,\"scrollcomments\":false,\"scrollwidth\":null,\"showshares\":false,\"shownub\":false,\"numberdelimiter\":\",\",\"showtyping\":0,\"logtyping\":0,\"entstream\":false,\"timelinelogdata\":\"AQAoVolFGpEfqWlwzbYXiaV8HOf_MAy3hUN_btt38_xz1Ik_gKHzXuQWuDkB9nNcFKQLJEmHUnKV-_00nujCxol1jChscWGCBpVvLlVTr0VI8e4JhMADTd-XSxv2vBKlwsqXEe1MGCoGYEhWAskcz2drecNjppAdwBQKiKFOEDp90CIXVTh6wyfdp9qefZFvdLSv2pV0RVJ1_w8lQtPkCi4RlAfpm6nPdBM67w027v2pL7n4RuqkGBTh08Au_b9F_mM3p4zxYqM33ftWn2E7Odhvp874qbVcnubfuNClEU0ynFNJLp2i-yFnJvlw8ocu-jL8xffqaw4unWbFwX7I01MwyCMBdNGkG3x5pXl1TwSGn4K1kHSB0OLJ1J8g800RE9U\"},{\"feedbacktargets\":[{\"targetfbid\":\"10151244259254030\",\"entidentifier\":\"10151244259254030\",\"permalink\":\"http:\\/\\/www.facebook.com\\/photo.php?fbid=10151244259254030&set=a.164247464029.118730.143412869029&type=1\",\"commentcount\":4,\"likecount\":57,\"viewercanlike\":false,\"hasviewerliked\":false,\"cancomment\":false,\"showremovemenu\":false,\"actorforpost\":\"\",\"viewerid\":\"0\",\"likesentences\":{\"current\":{\"text\":\"Paul Mccormack, Howard Anderson, Laura McGuinn and 54 others like this.\",\"ranges\":[{\"offset\":0,\"length\":14,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/paul.mccormack.90\",\"id\":100000067745206}]},{\"offset\":16,\"length\":15,\"entities\":[{\"url\":\"\",\"shimhash\":\"CAQF8Bvp8\",\"id\":100000814946993}]},{\"offset\":33,\"length\":13,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/foolofatook\",\"id\":845930579}]}],\"aggregatedranges\":[{\"offset\":51,\"length\":9,\"count\":54}]},\"alternate\":{\"text\":\"You, Paul Mccormack, Howard Anderson, Laura McGuinn and 54 others like this.\",\"ranges\":[{\"offset\":5,\"length\":14,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/paul.mccormack.90\",\"id\":100000067745206}]},{\"offset\":21,\"length\":15,\"entities\":[{\"url\":\"\",\"shimhash\":\"HAQGKveDe\",\"id\":100000814946993}]},{\"offset\":38,\"length\":13,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/foolofatook\",\"id\":845930579}]}],\"aggregatedranges\":[{\"offset\":56,\"length\":9,\"count\":54}]}},\"sharecount\":3,\"ownerid\":\"143412869029\",\"lastseentime\":null,\"canremoveall\":false,\"seencount\":0,\"seenbyall\":false,\"hasunseencollapsed\":false,\"mentionsdatasource\":{\"__m\":\"m885388_157\"},\"isranked\":false,\"isthreaded\":false,\"isownerpage\":true}],\"comments\":[{\"id\":\"10151244259254030_8310946\",\"fbid\":\"10151244303794030\",\"legacyid\":\"8310946\",\"body\":{\"text\":\"Long live the long leaf!\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"1379114029\",\"ftentidentifier\":\"10151244259254030\",\"timestamp\":{\"time\":1352124096,\"text\":\"November 5 at 6:01am\",\"verbose\":\"Monday, November 5, 2012 at 6:01am\"},\"likecount\":0,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":null,\"istranslatable\":false},{\"id\":\"10151244259254030_8311646\",\"fbid\":\"10151244454304030\",\"legacyid\":\"8311646\",\"body\":{\"text\":\"Scott told me he's planning to retire this month, so it's great to see him still educating school students right up to his retirement!  A visit to Weymouth Woods-The Sandhills Nature Preserve will educate any visitor on the prominent role the longleaf pine has played in our state's history. :)\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"100001424529980\",\"ftentidentifier\":\"10151244259254030\",\"timestamp\":{\"time\":1352131085,\"text\":\"November 5 at 7:58am\",\"verbose\":\"Monday, November 5, 2012 at 7:58am\"},\"likecount\":0,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":null,\"istranslatable\":false}],\"profiles\":[{\"id\":\"1379114029\",\"name\":\"Trey McLamb\",\"firstName\":\"Trey\",\"vanity\":\"trey.mclamb.3\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-ash4\\/368893_1379114029_909009035_q.jpg\",\"uri\":\"http:\\/\\/www.facebook.com\\/trey.mclamb.3\",\"gender\":2,\"type\":\"user\"},{\"id\":\"100001424529980\",\"name\":\"Philip Woodward\",\"firstName\":\"Philip\",\"vanity\":\"philip.woodward.33\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-ash3\\/27466_100001424529980_303_q.jpg\",\"uri\":\"http:\\/\\/www.facebook.com\\/philip.woodward.33\",\"gender\":2,\"type\":\"user\"}],\"actions\":[],\"commentlists\":{\"10151244259254030\":{\"range\":{\"offset\":2,\"length\":2},\"values\":[\"10151244259254030_8310946\",\"10151244259254030_8311646\"]}},\"servertime\":1352998853}],1],[\"m885388_126\",[\"UFIController\",\"m885388_125\",\"m885388_158\"],[{\"__m\":\"m885388_125\"},{\"ftentidentifier\":\"10151228371144030\",\"instanceid\":\"u4kzrn2109\",\"source\":0,\"showaddcomment\":true,\"collapseaddcomment\":false,\"markedcomment\":null,\"scrollcomments\":false,\"scrollwidth\":null,\"showshares\":false,\"shownub\":false,\"numberdelimiter\":\",\",\"showtyping\":0,\"logtyping\":0,\"entstream\":false,\"timelinelogdata\":\"AQDxmaeRXL9NZF77OaCWqNbVsV8D3gBFiDCaZ-D1RpMeROpUEi_dej4jyA_WcWnKuvZ9H8UTc31jNFRntInqRzGUldzqVOJ0FigdKiY0je3mmpKh4MkKvX7mIVbjF-HOF72CDOTcknfmSWbDxF3_oK6uXG36vulv2yzJvOGjfKBXKbwb6x8eVToPBPG9fl6oBlIZDShbdIZ7dj5alqWawK5_Y-_vydPAPPrRwU_EFYupjP6bfltCQJFycjoubzzgX38e0B01jzZNm1DbdKZLMCBZxo4I3z7GF4q0xzBFHqVDAOmi-uIHevLM7ma7EMa6mZ5bnqMzDkwDxsHsmVolXTE8qkOVnt9FCzWv5VRoOGZJRFS3UiB7XxnviiGJ8pGgH0c\"},{\"feedbacktargets\":[{\"targetfbid\":\"10151228371144030\",\"entidentifier\":\"10151228371144030\",\"permalink\":\"http:\\/\\/www.facebook.com\\/photo.php?fbid=10151228371144030&set=a.164247464029.118730.143412869029&type=1\",\"commentcount\":1,\"likecount\":60,\"viewercanlike\":false,\"hasviewerliked\":false,\"cancomment\":false,\"showremovemenu\":false,\"actorforpost\":\"\",\"viewerid\":\"0\",\"likesentences\":{\"current\":{\"text\":\"Linda J. Clark Oakley, Jeffery Davidson, Paul Terry and 57 others like this.\",\"ranges\":[{\"offset\":0,\"length\":21,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/linda.j.oakley\",\"id\":1110907259}]},{\"offset\":23,\"length\":16,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/jeffdavidson33\",\"id\":53500988}]},{\"offset\":41,\"length\":10,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/ki4rut\",\"id\":97500862}]}],\"aggregatedranges\":[{\"offset\":56,\"length\":9,\"count\":57}]},\"alternate\":{\"text\":\"You, Linda J. Clark Oakley, Jeffery Davidson, Paul Terry and 57 others like this.\",\"ranges\":[{\"offset\":5,\"length\":21,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/linda.j.oakley\",\"id\":1110907259}]},{\"offset\":28,\"length\":16,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/jeffdavidson33\",\"id\":53500988}]},{\"offset\":46,\"length\":10,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/ki4rut\",\"id\":97500862}]}],\"aggregatedranges\":[{\"offset\":61,\"length\":9,\"count\":57}]}},\"sharecount\":2,\"ownerid\":\"143412869029\",\"lastseentime\":null,\"canremoveall\":false,\"seencount\":0,\"seenbyall\":false,\"hasunseencollapsed\":false,\"mentionsdatasource\":{\"__m\":\"m885388_158\"},\"isranked\":false,\"isthreaded\":false,\"isownerpage\":true}],\"comments\":[{\"id\":\"10151228371144030_8235742\",\"fbid\":\"10151228482749030\",\"legacyid\":\"8235742\",\"body\":{\"text\":\"That sounds great!  Is the vault toilet wheelchair-accessible for people with mobility disabilities?\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"100001424529980\",\"ftentidentifier\":\"10151228371144030\",\"timestamp\":{\"time\":1351259973,\"text\":\"October 26 at 6:59am\",\"verbose\":\"Friday, October 26, 2012 at 6:59am\"},\"likecount\":1,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":null,\"istranslatable\":false}],\"profiles\":[{\"id\":\"100001424529980\",\"name\":\"Philip Woodward\",\"firstName\":\"Philip\",\"vanity\":\"philip.woodward.33\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-ash3\\/27466_100001424529980_303_q.jpg\",\"uri\":\"http:\\/\\/www.facebook.com\\/philip.woodward.33\",\"gender\":2,\"type\":\"user\"}],\"actions\":[],\"commentlists\":{\"10151228371144030\":{\"range\":{\"offset\":0,\"length\":1},\"values\":[\"10151228371144030_8235742\"]}},\"servertime\":1352998853}],1],[\"m885388_128\",[\"UFIController\",\"m885388_127\",\"m885389_159\"],[{\"__m\":\"m885388_127\"},{\"ftentidentifier\":\"10151240100539030\",\"instanceid\":\"u4kzrn2110\",\"source\":0,\"showaddcomment\":true,\"collapseaddcomment\":false,\"markedcomment\":null,\"scrollcomments\":false,\"scrollwidth\":null,\"showshares\":false,\"shownub\":false,\"numberdelimiter\":\",\",\"showtyping\":0,\"logtyping\":0,\"entstream\":false,\"timelinelogdata\":\"AQDThfqlyhON2N8_Gs3WD2OG0WIkl8SHjrupBknwoaqIwyonVKGriKb9oG6FCUOopWXxEyb5jCKhsb5p5doGXthIr0y_ZJ2HxC7q71RLsNNBI6X5KEjNZGIlBR9r7oapQc7Kh1_XS6O3mbsAsc9F3ja8b4B3kM19LifPLlu_1i2GzYVM55nixEhXHzB3PfqiBCznsJlZ_uFXItZo2HOF_Cls53hfRIUOxQ6V-fO55-vOsPYnwClOw3VDaElnW3Pn7Fn0ICBnY3Kav0FBVPKdt9QFOCJYADrh8JTkT-r07kgWpJzy7kFBN-FuAb-ZKqlfXfpp-UT5_-nsD19_E6YpHqF5Lvs9qxypGf5mdp4ThiVKCcHFsR7BgfP8Ed7YGQLME-I\"},{\"feedbacktargets\":[{\"targetfbid\":\"10151240100539030\",\"entidentifier\":\"10151240100539030\",\"permalink\":\"http:\\/\\/www.facebook.com\\/photo.php?fbid=10151240100539030&set=a.164247464029.118730.143412869029&type=1\",\"commentcount\":5,\"likecount\":101,\"viewercanlike\":false,\"hasviewerliked\":false,\"cancomment\":false,\"showremovemenu\":false,\"actorforpost\":\"\",\"viewerid\":\"0\",\"likesentences\":{\"current\":{\"text\":\"Kim Haigler, Michael LaRocca, Jan Bond and 98 others like this.\",\"ranges\":[{\"offset\":0,\"length\":11,\"entities\":[{\"url\":\"\",\"shimhash\":\"ZAQHJsopg\",\"id\":100002631863425}]},{\"offset\":13,\"length\":15,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/michael.larocca.9\",\"id\":100002593149029}]},{\"offset\":30,\"length\":8,\"entities\":[{\"url\":\"\",\"shimhash\":\"OAQEiYkrT\",\"id\":661256400}]}],\"aggregatedranges\":[{\"offset\":43,\"length\":9,\"count\":98}]},\"alternate\":{\"text\":\"You, Kim Haigler, Michael LaRocca, Jan Bond and 98 others like this.\",\"ranges\":[{\"offset\":5,\"length\":11,\"entities\":[{\"url\":\"\",\"shimhash\":\"UAQHMbuxY\",\"id\":100002631863425}]},{\"offset\":18,\"length\":15,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/michael.larocca.9\",\"id\":100002593149029}]},{\"offset\":35,\"length\":8,\"entities\":[{\"url\":\"\",\"shimhash\":\"tAQHkd_nn\",\"id\":661256400}]}],\"aggregatedranges\":[{\"offset\":48,\"length\":9,\"count\":98}]}},\"sharecount\":14,\"ownerid\":\"143412869029\",\"lastseentime\":null,\"canremoveall\":false,\"seencount\":0,\"seenbyall\":false,\"hasunseencollapsed\":false,\"mentionsdatasource\":{\"__m\":\"m885389_159\"},\"isranked\":false,\"isthreaded\":false,\"isownerpage\":true}],\"comments\":[{\"id\":\"10151240100539030_8293657\",\"fbid\":\"10151240528649030\",\"legacyid\":\"8293657\",\"body\":{\"text\":\"So pretty.\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"100001395750440\",\"ftentidentifier\":\"10151240100539030\",\"timestamp\":{\"time\":1351905505,\"text\":\"November 2 at 6:18pm\",\"verbose\":\"Friday, November 2, 2012 at 6:18pm\"},\"likecount\":0,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":null,\"istranslatable\":false},{\"id\":\"10151240100539030_8299512\",\"fbid\":\"10151241856294030\",\"legacyid\":\"8299512\",\"body\":{\"text\":\"Beautiful!\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"100000079142988\",\"ftentidentifier\":\"10151240100539030\",\"timestamp\":{\"time\":1351980224,\"text\":\"November 3 at 3:03pm\",\"verbose\":\"Saturday, November 3, 2012 at 3:03pm\"},\"likecount\":0,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":null,\"istranslatable\":false}],\"profiles\":[{\"id\":\"100001395750440\",\"name\":\"Elaine Marriner Williams\",\"firstName\":\"Elaine\",\"vanity\":\"elaine.m.williams.5\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-ash3\\/573117_100001395750440_181115560_q.jpg\",\"uri\":null,\"gender\":1,\"type\":\"user\"},{\"id\":\"100000079142988\",\"name\":\"Cynthia Self Harper\",\"firstName\":\"Cynthia\",\"vanity\":\"cynthia.harper.9\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-ash3\\/274116_100000079142988_1400389262_q.jpg\",\"uri\":null,\"gender\":1,\"type\":\"user\"}],\"actions\":[],\"commentlists\":{\"10151240100539030\":{\"range\":{\"offset\":3,\"length\":2},\"values\":[\"10151240100539030_8293657\",\"10151240100539030_8299512\"]}},\"servertime\":1352998853}],1],[\"m885388_130\",[\"UFIController\",\"m885388_129\",\"m885389_160\"],[{\"__m\":\"m885388_129\"},{\"ftentidentifier\":\"461599323892770\",\"instanceid\":\"u4kzrn2111\",\"source\":0,\"showaddcomment\":true,\"collapseaddcomment\":false,\"markedcomment\":null,\"scrollcomments\":false,\"scrollwidth\":null,\"showshares\":false,\"shownub\":false,\"numberdelimiter\":\",\",\"showtyping\":0,\"logtyping\":0,\"entstream\":false,\"timelinelogdata\":\"AQAbeWdr_N1EAr2HiUT94k7a1JZGO0EQ0DLF_0Zl6GioIDlOHVbrBteEroOFjwSXCjzHTRxO423i601CAulAHtLgisH6zeihRqyqOho14hNx8s-DfO21Dj0g04J0ircqqICLX2DK9epCFljRNlk0WS0SGuB2STEYDfqT7nG0g6XCx7qZUsAZhT0qAO7V8pGk6VyAWwzJYqT07YDMS5-T6X0MIMdYfItxPzqp9FwqysEN8FcfnrtQ1M_Cjbt4-SHwhEV5bs0yy1doUqghGJa65uvEPrzqyKvWyCkKU8hhkZlhB2QYbK1hFtUYsNNxzdOKlrYgHLYa_q1NeZWCSgVrxgT-5HV5gcnz4vt8437rij--taH689bnXaWXiAOMuc6Ngow\"},{\"feedbacktargets\":[{\"targetfbid\":\"461599323892770\",\"entidentifier\":\"461599323892770\",\"permalink\":\"\\/NorthCarolinaStateParks\\/posts\\/461599323892770\",\"commentcount\":1,\"likecount\":26,\"viewercanlike\":false,\"hasviewerliked\":false,\"cancomment\":false,\"showremovemenu\":false,\"actorforpost\":\"\",\"viewerid\":\"0\",\"likesentences\":{\"current\":{\"text\":\"NatureFind Raleigh-Durham, Laura McGuinn, North Carolina Chapter of the Wildlife Society and 23 others like this.\",\"ranges\":[{\"offset\":0,\"length\":25,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/NatureFindRaleighDurham\",\"id\":171854356200534}]},{\"offset\":27,\"length\":13,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/foolofatook\",\"id\":845930579}]},{\"offset\":42,\"length\":46,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/nctws\",\"id\":112577132162373}]}],\"aggregatedranges\":[{\"offset\":93,\"length\":9,\"count\":23}]},\"alternate\":{\"text\":\"You, NatureFind Raleigh-Durham, Laura McGuinn, North Carolina Chapter of the Wildlife Society and 23 others like this.\",\"ranges\":[{\"offset\":5,\"length\":25,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/NatureFindRaleighDurham\",\"id\":171854356200534}]},{\"offset\":32,\"length\":13,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/foolofatook\",\"id\":845930579}]},{\"offset\":47,\"length\":46,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/nctws\",\"id\":112577132162373}]}],\"aggregatedranges\":[{\"offset\":98,\"length\":9,\"count\":23}]}},\"sharecount\":3,\"ownerid\":\"143412869029\",\"lastseentime\":null,\"canremoveall\":false,\"seencount\":0,\"seenbyall\":false,\"hasunseencollapsed\":false,\"mentionsdatasource\":{\"__m\":\"m885389_160\"},\"isranked\":false,\"isthreaded\":false,\"isownerpage\":true}],\"comments\":[{\"id\":\"461599323892770_5051095\",\"fbid\":\"461726433880059\",\"legacyid\":\"5051095\",\"body\":{\"text\":\"Buy that NC State Parks tag.  I have two.  Helps support our parks!!\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"100001711258648\",\"ftentidentifier\":\"461599323892770\",\"timestamp\":{\"time\":1351883529,\"text\":\"November 2 at 12:12pm\",\"verbose\":\"Friday, November 2, 2012 at 12:12pm\"},\"likecount\":0,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":0,\"istranslatable\":false}],\"profiles\":[{\"id\":\"100001711258648\",\"name\":\"Taffy Wild\",\"firstName\":\"Taffy\",\"vanity\":\"taffy.wild\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-ash4\\/370647_100001711258648_245046137_q.jpg\",\"uri\":\"http:\\/\\/www.facebook.com\\/taffy.wild\",\"gender\":1,\"type\":\"user\"}],\"actions\":[],\"commentlists\":{\"461599323892770\":{\"range\":{\"offset\":0,\"length\":1},\"values\":[\"461599323892770_5051095\"]}},\"servertime\":1352998853}],1],[\"m885388_132\",[\"UFIController\",\"m885388_131\",\"m885389_161\"],[{\"__m\":\"m885388_131\"},{\"ftentidentifier\":\"10151239662119030\",\"instanceid\":\"u4kzrn2112\",\"source\":0,\"showaddcomment\":true,\"collapseaddcomment\":false,\"markedcomment\":null,\"scrollcomments\":false,\"scrollwidth\":null,\"showshares\":false,\"shownub\":false,\"numberdelimiter\":\",\",\"showtyping\":0,\"logtyping\":0,\"entstream\":false,\"timelinelogdata\":\"AQDggTONWQyiumOGxWo2dZ2txSjn9xfZAF3BYV4YKgevjA8kjwh0EY1QCVlyTnPn3RpPJqObQCY6CP_IoGSGtUtaJHrEuLDT4Kwl8bMfdpW92qeDqwHzX5VRCPNS_4K_SujORjPVKDMBKO71-PHI5HnfuSNlHSgH8qI4lkOJenKPJdsZu8OU6b2At_DJkF_uD73pbciBmr0uOhxjFV_YdTr2uFsDn1r6YBsd4j7GAXX_ucpEQ2nBQFwDuq0iIv7ZLaDQfVc3_es3jnftTMRk-SK1UobOa9Sfle4Aq51sZXqsX3OLzj4DZgYPYxWTU9WxBf4kM55O_gEMo3U2KUQtyA3dv9YKiemHQ2IwqHfi8FK21eWAcryGL-cHNHaK9687AXk\"},{\"feedbacktargets\":[{\"targetfbid\":\"10151239662119030\",\"entidentifier\":\"10151239662119030\",\"permalink\":\"http:\\/\\/www.facebook.com\\/photo.php?fbid=10151239662119030&set=a.164247464029.118730.143412869029&type=1\",\"commentcount\":39,\"likecount\":440,\"viewercanlike\":false,\"hasviewerliked\":false,\"cancomment\":false,\"showremovemenu\":false,\"actorforpost\":\"\",\"viewerid\":\"0\",\"likesentences\":{\"current\":{\"text\":\"Bob Ace Wallace, Joannah Hicks, Deborah Norwood Deatherage and 437 others like this.\",\"ranges\":[{\"offset\":0,\"length\":15,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/bob.a.wallace.5\",\"id\":1126958450}]},{\"offset\":17,\"length\":13,\"entities\":[{\"url\":\"\",\"shimhash\":\"YAQHtsUPB\",\"id\":100000144783370}]},{\"offset\":32,\"length\":26,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/deborah.deatherage\",\"id\":1811590930}]}],\"aggregatedranges\":[{\"offset\":63,\"length\":10,\"count\":437}]},\"alternate\":{\"text\":\"You, Bob Ace Wallace, Joannah Hicks, Deborah Norwood Deatherage and 437 others like this.\",\"ranges\":[{\"offset\":5,\"length\":15,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/bob.a.wallace.5\",\"id\":1126958450}]},{\"offset\":22,\"length\":13,\"entities\":[{\"url\":\"\",\"shimhash\":\"SAQEjLIXR\",\"id\":100000144783370}]},{\"offset\":37,\"length\":26,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/deborah.deatherage\",\"id\":1811590930}]}],\"aggregatedranges\":[{\"offset\":68,\"length\":10,\"count\":437}]}},\"sharecount\":133,\"ownerid\":\"143412869029\",\"lastseentime\":null,\"canremoveall\":false,\"seencount\":0,\"seenbyall\":false,\"hasunseencollapsed\":false,\"mentionsdatasource\":{\"__m\":\"m885389_161\"},\"isranked\":false,\"isthreaded\":false,\"isownerpage\":true}],\"comments\":[{\"id\":\"10151239662119030_8311948\",\"fbid\":\"10151244517454030\",\"legacyid\":\"8311948\",\"body\":{\"text\":\"Really beautiful!\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"100003251197199\",\"ftentidentifier\":\"10151239662119030\",\"timestamp\":{\"time\":1352133866,\"text\":\"November 5 at 8:44am\",\"verbose\":\"Monday, November 5, 2012 at 8:44am\"},\"likecount\":0,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":null,\"istranslatable\":false},{\"id\":\"10151239662119030_8372498\",\"fbid\":\"10151256737984030\",\"legacyid\":\"8372498\",\"body\":{\"text\":\"According to the date, the fall festival was 3 days ago?  Doesn't it look more like....this pic out there by now?  It IS certainly breathtaking...bucket list HERE I COME, lol, well no, not yet. but when I do....\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"100001347231698\",\"ftentidentifier\":\"10151239662119030\",\"timestamp\":{\"time\":1352849802,\"text\":\"Tuesday at 3:36pm\",\"verbose\":\"Tuesday, November 13, 2012 at 3:36pm\"},\"likecount\":0,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":null,\"istranslatable\":false}],\"profiles\":[{\"id\":\"100003251197199\",\"name\":\"Betty Anderson\",\"firstName\":\"Betty\",\"vanity\":\"betty.anderson.75\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-snc6\\/260615_100003251197199_975471716_q.jpg\",\"uri\":\"http:\\/\\/www.facebook.com\\/betty.anderson.75\",\"gender\":1,\"type\":\"user\"},{\"id\":\"100001347231698\",\"name\":\"Julie Jo Koehler\",\"firstName\":\"Julie Jo\",\"vanity\":\"juliejokoehler\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-snc7\\/370646_100001347231698_1246012740_q.jpg\",\"uri\":\"http:\\/\\/www.facebook.com\\/juliejokoehler\",\"gender\":1,\"type\":\"user\"}],\"actions\":[],\"commentlists\":{\"10151239662119030\":{\"range\":{\"offset\":37,\"length\":2},\"values\":[\"10151239662119030_8311948\",\"10151239662119030_8372498\"]}},\"servertime\":1352998853}],1],[\"m885388_134\",[\"UFIController\",\"m885388_133\",\"m885389_162\"],[{\"__m\":\"m885388_133\"},{\"ftentidentifier\":\"10151236651299030\",\"instanceid\":\"u4kzrn2113\",\"source\":0,\"showaddcomment\":true,\"collapseaddcomment\":false,\"markedcomment\":null,\"scrollcomments\":false,\"scrollwidth\":null,\"showshares\":false,\"shownub\":false,\"numberdelimiter\":\",\",\"showtyping\":0,\"logtyping\":0,\"entstream\":false,\"timelinelogdata\":\"AQCgytFRPm3iSAqNqVO2XZs-ZSK9sJnkbuF3cHrCtBE09hMzZmSMVfSO7FHdov7EznFPj4jEG-i9_wCAOyoXty4W88h77Hwj2EZD0zEdkkFgIOt5uIVCuj3cpqnGIuZbWv9fpFmWA_vrdGnRQK2W--Oc4ZgEbJXP5DyksKxxykyk0azvOda6HRpY1daAk8mGz1sKpcXgECGma0PX-6HXasotySQfwlOplnoIRdnlJX_KyQxvB_SuzmL94pcYMSmgOQB0JERm8bEpCICV6EL3A8qQkFFbRvMDtc22dx1ZxNxJGfdLWA2YkUKNbo3Sv8W91EGQI2aOeSJOzNYVovxES8NNiQ8Xh-suuFBUEtTyWUvZ4vhXsxShb4ZIi8Ov0tY_PiM\"},{\"feedbacktargets\":[{\"targetfbid\":\"10151236651299030\",\"entidentifier\":\"10151236651299030\",\"permalink\":\"http:\\/\\/www.facebook.com\\/photo.php?fbid=10151236651299030&set=a.164247464029.118730.143412869029&type=1\",\"commentcount\":1,\"likecount\":63,\"viewercanlike\":false,\"hasviewerliked\":false,\"cancomment\":false,\"showremovemenu\":false,\"actorforpost\":\"\",\"viewerid\":\"0\",\"likesentences\":{\"current\":{\"text\":\"Linda J. Clark Oakley, Catawba County Parks, Kang Tu and 60 others like this.\",\"ranges\":[{\"offset\":0,\"length\":21,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/linda.j.oakley\",\"id\":1110907259}]},{\"offset\":23,\"length\":20,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/catawbacountyparks\",\"id\":94474264671}]},{\"offset\":45,\"length\":7,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/kang.tu\",\"id\":1662468753}]}],\"aggregatedranges\":[{\"offset\":57,\"length\":9,\"count\":60}]},\"alternate\":{\"text\":\"You, Linda J. Clark Oakley, Catawba County Parks, Kang Tu and 60 others like this.\",\"ranges\":[{\"offset\":5,\"length\":21,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/linda.j.oakley\",\"id\":1110907259}]},{\"offset\":28,\"length\":20,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/catawbacountyparks\",\"id\":94474264671}]},{\"offset\":50,\"length\":7,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/kang.tu\",\"id\":1662468753}]}],\"aggregatedranges\":[{\"offset\":62,\"length\":9,\"count\":60}]}},\"sharecount\":15,\"ownerid\":\"143412869029\",\"lastseentime\":null,\"canremoveall\":false,\"seencount\":0,\"seenbyall\":false,\"hasunseencollapsed\":false,\"mentionsdatasource\":{\"__m\":\"m885389_162\"},\"isranked\":false,\"isthreaded\":false,\"isownerpage\":true}],\"comments\":[{\"id\":\"10151236651299030_8274395\",\"fbid\":\"10151236679874030\",\"legacyid\":\"8274395\",\"body\":{\"text\":\"Beautiful......\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"1081037282\",\"ftentidentifier\":\"10151236651299030\",\"timestamp\":{\"time\":1351694902,\"text\":\"October 31 at 7:48am\",\"verbose\":\"Wednesday, October 31, 2012 at 7:48am\"},\"likecount\":0,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":null,\"istranslatable\":false}],\"profiles\":[{\"id\":\"1081037282\",\"name\":\"Harold William Yost\",\"firstName\":\"Harold\",\"vanity\":\"harold.w.yost\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-ash3\\/41389_1081037282_6203_q.jpg\",\"uri\":null,\"gender\":2,\"type\":\"user\"}],\"actions\":[],\"commentlists\":{\"10151236651299030\":{\"range\":{\"offset\":0,\"length\":1},\"values\":[\"10151236651299030_8274395\"]}},\"servertime\":1352998853}],1],[\"m885388_136\",[\"UFIController\",\"m885388_135\",\"m885389_163\"],[{\"__m\":\"m885388_135\"},{\"ftentidentifier\":\"399953756744985\",\"instanceid\":\"u4kzrn2114\",\"source\":0,\"showaddcomment\":true,\"collapseaddcomment\":false,\"markedcomment\":null,\"scrollcomments\":false,\"scrollwidth\":null,\"showshares\":false,\"shownub\":false,\"numberdelimiter\":\",\",\"showtyping\":0,\"logtyping\":0,\"entstream\":false,\"timelinelogdata\":\"AQC_Ya7lJvnfPhVvhl6YdJray5UhLUUqnC5IE0B4IqdpvecV8OMXVktoAlg4cD-V9yCVZO6DFY7fJScmm-7BRmEmq9IJjAA3DQruDS8F8pt_fw8y0IWX5xMjBFkfLy2O2h7xAs0yXoTkgkYXpwcHGknOGL-24hK0wxAR-TFhhJe9t0vI87r2kxVI0JGMNrINvL_cQZpS_xf6nS67yQcMWekpyc7squtzJPxOi8DJ7caBPpC3T13Sb1KlNwQzkwFyMr_9bGP4Odn4AYqY_NJI7XQpklcbvxvckw4q2lsqVwO1T_qDykHwkHhVvIwR3tc6IGQW8FxYPzry2v4d6OFlkYIQOXJzA09PboAW0e-Ff8kmFvaV5xT6Wb79W0cmJFxin-c\"},{\"feedbacktargets\":[{\"targetfbid\":\"399953756744985\",\"entidentifier\":\"399953756744985\",\"permalink\":\"\\/NorthCarolinaStateParks\\/posts\\/399953756744985\",\"commentcount\":3,\"likecount\":37,\"viewercanlike\":false,\"hasviewerliked\":false,\"cancomment\":false,\"showremovemenu\":false,\"actorforpost\":\"\",\"viewerid\":\"0\",\"likesentences\":{\"current\":{\"text\":\"Paul T. Brown II, Elaine Williams, Jon Luis and 34 others like this.\",\"ranges\":[{\"offset\":0,\"length\":16,\"entities\":[{\"url\":\"\",\"shimhash\":\"tAQHkd_nn\",\"id\":100003240249568}]},{\"offset\":18,\"length\":15,\"entities\":[{\"url\":\"\",\"shimhash\":\"7AQEijtPg\",\"id\":100000966877835}]},{\"offset\":35,\"length\":8,\"entities\":[{\"url\":\"\",\"shimhash\":\"1AQFaCi3m\",\"id\":1480305390}]}],\"aggregatedranges\":[{\"offset\":48,\"length\":9,\"count\":34}]},\"alternate\":{\"text\":\"You, Paul T. Brown II, Elaine Williams, Jon Luis and 34 others like this.\",\"ranges\":[{\"offset\":5,\"length\":16,\"entities\":[{\"url\":\"\",\"shimhash\":\"FAQHPmOF9\",\"id\":100003240249568}]},{\"offset\":23,\"length\":15,\"entities\":[{\"url\":\"\",\"shimhash\":\"oAQHrW15D\",\"id\":100000966877835}]},{\"offset\":40,\"length\":8,\"entities\":[{\"url\":\"\",\"shimhash\":\"OAQEiYkrT\",\"id\":1480305390}]}],\"aggregatedranges\":[{\"offset\":53,\"length\":9,\"count\":34}]}},\"sharecount\":3,\"ownerid\":\"143412869029\",\"lastseentime\":null,\"canremoveall\":false,\"seencount\":0,\"seenbyall\":false,\"hasunseencollapsed\":false,\"mentionsdatasource\":{\"__m\":\"m885389_163\"},\"isranked\":false,\"isthreaded\":false,\"isownerpage\":true}],\"comments\":[{\"id\":\"399953756744985_77824056\",\"fbid\":\"399954270078267\",\"legacyid\":\"77824056\",\"body\":{\"text\":\"oh man:( now i wanna go\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"100000572201049\",\"ftentidentifier\":\"399953756744985\",\"timestamp\":{\"time\":1351603016,\"text\":\"October 30 at 6:16am\",\"verbose\":\"Tuesday, October 30, 2012 at 6:16am\"},\"likecount\":0,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":0,\"istranslatable\":false},{\"id\":\"399953756744985_77824070\",\"fbid\":\"399955573411470\",\"legacyid\":\"77824070\",\"body\":{\"text\":\"Just learned the park is closed; no commercial electric power.\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"143412869029\",\"ftentidentifier\":\"399953756744985\",\"timestamp\":{\"time\":1351603260,\"text\":\"October 30 at 6:21am\",\"verbose\":\"Tuesday, October 30, 2012 at 6:21am\"},\"likecount\":0,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":0,\"istranslatable\":false},{\"id\":\"399953756744985_77824141\",\"fbid\":\"399960750077619\",\"legacyid\":\"77824141\",\"body\":{\"text\":\"Wow!\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"100000223981444\",\"ftentidentifier\":\"399953756744985\",\"timestamp\":{\"time\":1351604332,\"text\":\"October 30 at 6:38am\",\"verbose\":\"Tuesday, October 30, 2012 at 6:38am\"},\"likecount\":0,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":1,\"istranslatable\":false}],\"profiles\":[{\"id\":\"100000572201049\",\"name\":\"Brandon Swank\",\"firstName\":\"Brandon\",\"vanity\":\"brandon.swank.1257\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-ash4\\/203197_100000572201049_336697534_q.jpg\",\"uri\":\"http:\\/\\/www.facebook.com\\/brandon.swank.1257\",\"gender\":2,\"type\":\"user\"},{\"id\":\"143412869029\",\"name\":\"North Carolina State Parks and Recreation\",\"firstName\":\"North Carolina State Parks and Recreation\",\"vanity\":\"NorthCarolinaStateParks\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-ash3\\/174880_143412869029_133040642_q.jpg\",\"uri\":\"http:\\/\\/www.facebook.com\\/NorthCarolinaStateParks\",\"type\":\"page\"},{\"id\":\"100000223981444\",\"name\":\"Grace Iovine\",\"firstName\":\"Grace\",\"vanity\":\"grace.iovine\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-snc6\\/203137_100000223981444_1456642_q.jpg\",\"uri\":\"http:\\/\\/www.facebook.com\\/grace.iovine\",\"gender\":1,\"type\":\"user\"}],\"actions\":[],\"commentlists\":{\"399953756744985\":{\"range\":{\"offset\":0,\"length\":3},\"values\":[\"399953756744985_77824056\",\"399953756744985_77824070\",\"399953756744985_77824141\"]}},\"servertime\":1352998853}],1],[\"m885388_138\",[\"UFIController\",\"m885388_137\",\"m885389_164\"],[{\"__m\":\"m885388_137\"},{\"ftentidentifier\":\"10151234924589030\",\"instanceid\":\"u4kzrn2115\",\"source\":0,\"showaddcomment\":true,\"collapseaddcomment\":false,\"markedcomment\":null,\"scrollcomments\":false,\"scrollwidth\":null,\"showshares\":false,\"shownub\":false,\"numberdelimiter\":\",\",\"showtyping\":0,\"logtyping\":0,\"entstream\":false,\"timelinelogdata\":\"AQDGs6WfBBsxWWPB21-mGcg-2nqscDD3NDO4Z4uJMrPuyIEQnsKJp1unJaFr21DAvCNLnOMCKbIojSWG3fkkbWiHJXeR3dHR0ROHVVBRb_OSjvBBanrNBi5u56frQvXcxUFEBpaSXjMi2bOiEB2RhAcQ93UPPIAGaZXUPFroemGfOTxYLxEl8h9TRUPwX1_OHtxp070v9N7PozLEpxl34kfJMtcCxOuaIXeRc328qAaSGJRDOBld8nCCF2MeV8tEWz5YBiXQaBivoSDR51TFZvd22v9-_ysaGYodbf3sifaJozfhd-GVBHsMa-VyVNB_d9_84YQ2uCpWceeaZlw5DHBF-FnPmQgKQa7TD1iJVWiHWp1CtopkbOaH7y4EM1T01Xw\"},{\"feedbacktargets\":[{\"targetfbid\":\"10151234924589030\",\"entidentifier\":\"10151234924589030\",\"permalink\":\"http:\\/\\/www.facebook.com\\/photo.php?fbid=10151234924589030&set=a.164247464029.118730.143412869029&type=1\",\"commentcount\":1,\"likecount\":49,\"viewercanlike\":false,\"hasviewerliked\":false,\"cancomment\":false,\"showremovemenu\":false,\"actorforpost\":\"\",\"viewerid\":\"0\",\"likesentences\":{\"current\":{\"text\":\"Denise Antonacci Kenney, Judy King Bailey, Laura McGuinn and 46 others like this.\",\"ranges\":[{\"offset\":0,\"length\":23,\"entities\":[{\"url\":\"\",\"shimhash\":\"JAQEUzQYi\",\"id\":1319934126}]},{\"offset\":25,\"length\":16,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/judy.k.bailey\",\"id\":1518341281}]},{\"offset\":43,\"length\":13,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/foolofatook\",\"id\":845930579}]}],\"aggregatedranges\":[{\"offset\":61,\"length\":9,\"count\":46}]},\"alternate\":{\"text\":\"You, Denise Antonacci Kenney, Judy King Bailey, Laura McGuinn and 46 others like this.\",\"ranges\":[{\"offset\":5,\"length\":23,\"entities\":[{\"url\":\"\",\"shimhash\":\"8AQGAIFeF\",\"id\":1319934126}]},{\"offset\":30,\"length\":16,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/judy.k.bailey\",\"id\":1518341281}]},{\"offset\":48,\"length\":13,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/foolofatook\",\"id\":845930579}]}],\"aggregatedranges\":[{\"offset\":66,\"length\":9,\"count\":46}]}},\"sharecount\":0,\"ownerid\":\"143412869029\",\"lastseentime\":null,\"canremoveall\":false,\"seencount\":0,\"seenbyall\":false,\"hasunseencollapsed\":false,\"mentionsdatasource\":{\"__m\":\"m885389_164\"},\"isranked\":false,\"isthreaded\":false,\"isownerpage\":true}],\"comments\":[{\"id\":\"10151234924589030_8271019\",\"fbid\":\"10151235991799030\",\"legacyid\":\"8271019\",\"body\":{\"text\":\"3 steps behind Jeff.\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"574192498\",\"ftentidentifier\":\"10151234924589030\",\"timestamp\":{\"time\":1351650333,\"text\":\"October 30 at 7:25pm\",\"verbose\":\"Tuesday, October 30, 2012 at 7:25pm\"},\"likecount\":0,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":null,\"istranslatable\":false}],\"profiles\":[{\"id\":\"574192498\",\"name\":\"Neil Pearson\",\"firstName\":\"Neil\",\"vanity\":\"npearson247\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-ash4\\/371568_574192498_897352994_q.jpg\",\"uri\":\"http:\\/\\/www.facebook.com\\/npearson247\",\"gender\":2,\"type\":\"user\"}],\"actions\":[],\"commentlists\":{\"10151234924589030\":{\"range\":{\"offset\":0,\"length\":1},\"values\":[\"10151234924589030_8271019\"]}},\"servertime\":1352998853}],1],[\"m885388_140\",[\"UFIController\",\"m885388_139\",\"m885389_165\"],[{\"__m\":\"m885388_139\"},{\"ftentidentifier\":\"196718357130532\",\"instanceid\":\"u4kzrn2116\",\"source\":0,\"showaddcomment\":true,\"collapseaddcomment\":false,\"markedcomment\":null,\"scrollcomments\":false,\"scrollwidth\":null,\"showshares\":false,\"shownub\":false,\"numberdelimiter\":\",\",\"showtyping\":0,\"logtyping\":0,\"entstream\":false,\"timelinelogdata\":\"AQA5qb40PSKFdE55h_BmyEKwZcs8vPHEt5Gak0aObjV1eOvsbo-vOUhG7tsOaEWeptNifZhDq8l4kSd2HMSvZUyjfq3JhbwRewPmzFEv5fW5hDAdbnGrDq_XTxI5NjjdyegXpMWWBllFXushcO6xb1SowFMg1aDV_I1mpTW6RL4dQJFu73GfKVib4wiQ-DtX0cgfXKLQL4nMAjS9rar7CMj4bcjt4M4CLKXnVdrR9d5fvRibSwFKYZAvCyrziaJ5DpygaWwUqMSIyRMDC64b6ADrZmBEbKdm-gfxci8uC5XqTOgccc9ZQQFk-1VX_GwS4D3rv2t1A0xCTGeHMMRCv-zmA7L90ictdYKeMK09LWzdU-n3tA7xh1kWLrePCT-rKjI\"},{\"feedbacktargets\":[{\"targetfbid\":\"196718357130532\",\"entidentifier\":\"196718357130532\",\"permalink\":\"\\/NorthCarolinaStateParks\\/posts\\/196718357130532\",\"commentcount\":1,\"likecount\":6,\"viewercanlike\":false,\"hasviewerliked\":false,\"cancomment\":false,\"showremovemenu\":false,\"actorforpost\":\"\",\"viewerid\":\"0\",\"likesentences\":{\"current\":{\"text\":\"Mike Kerkado, Joey Pyron, Kevin Stubbs and 3 others like this.\",\"ranges\":[{\"offset\":0,\"length\":12,\"entities\":[{\"url\":\"\",\"shimhash\":\"bAQEN3r6q\",\"id\":645854192}]},{\"offset\":14,\"length\":10,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/joey.pyron\",\"id\":100002897607424}]},{\"offset\":26,\"length\":12,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/kevin.stubbs.14\",\"id\":1691186129}]}],\"aggregatedranges\":[{\"offset\":43,\"length\":8,\"count\":3}]},\"alternate\":{\"text\":\"You, Mike Kerkado, Joey Pyron, Kevin Stubbs and 3 others like this.\",\"ranges\":[{\"offset\":5,\"length\":12,\"entities\":[{\"url\":\"\",\"shimhash\":\"IAQEZ45SG\",\"id\":645854192}]},{\"offset\":19,\"length\":10,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/joey.pyron\",\"id\":100002897607424}]},{\"offset\":31,\"length\":12,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/kevin.stubbs.14\",\"id\":1691186129}]}],\"aggregatedranges\":[{\"offset\":48,\"length\":8,\"count\":3}]}},\"sharecount\":2,\"ownerid\":\"143412869029\",\"lastseentime\":null,\"canremoveall\":false,\"seencount\":0,\"seenbyall\":false,\"hasunseencollapsed\":false,\"mentionsdatasource\":{\"__m\":\"m885389_165\"},\"isranked\":false,\"isthreaded\":false,\"isownerpage\":true}],\"comments\":[{\"id\":\"196718357130532_634807\",\"fbid\":\"196718607130507\",\"legacyid\":\"634807\",\"body\":{\"text\":\"snow! :)\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"1080933129\",\"ftentidentifier\":\"196718357130532\",\"timestamp\":{\"time\":1351519481,\"text\":\"October 29 at 7:04am\",\"verbose\":\"Monday, October 29, 2012 at 7:04am\"},\"likecount\":0,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":1,\"istranslatable\":false}],\"profiles\":[{\"id\":\"1080933129\",\"name\":\"Martha Quillin\",\"firstName\":\"Martha\",\"vanity\":\"martha.quillin\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-snc6\\/274704_1080933129_1594604332_q.jpg\",\"uri\":null,\"gender\":1,\"type\":\"user\"}],\"actions\":[],\"commentlists\":{\"196718357130532\":{\"range\":{\"offset\":0,\"length\":1},\"values\":[\"196718357130532_634807\"]}},\"servertime\":1352998853}],1],[\"m885388_142\",[\"UFIController\",\"m885388_141\",\"m885389_166\"],[{\"__m\":\"m885388_141\"},{\"ftentidentifier\":\"10151228750629030\",\"instanceid\":\"u4kzrn2117\",\"source\":0,\"showaddcomment\":true,\"collapseaddcomment\":false,\"markedcomment\":null,\"scrollcomments\":false,\"scrollwidth\":null,\"showshares\":false,\"shownub\":false,\"numberdelimiter\":\",\",\"showtyping\":0,\"logtyping\":0,\"entstream\":false,\"timelinelogdata\":\"AQDy0RyWEAASiujUkEJO2LK37rsfJ-NYK9j4NTcO80ox0xyCNawOUezRE_SN4QxiKz79Ysf5rsx7WHElt0RczM0NgE6c3IRMVocuQxzYTb5JFDLvDxpv4qyZ9ozv-WeJ_SeyzKbU2BO70EgOg0CrxyqDhwjPAMkRa6elfilpq6Ny_RPgGJTLdniMorR1l5vW0W8YLukNeG-bMM5_oS4Z49P5G9zLnscF-cwLl_vqiqs5oWKbTeRQW8hLxRlieCP32w1O3e5sy8npIP-mkVApOeVANx8BIda1sQVijbwMm_gbhTzTIk5xn7Ch7sTIdU3KkGaJyB0tkac_B4yoCUGFEkmK7L4BnJH-1gfQC0Ms34LXR6KNwuLKeZedl2V49vLxsgI\"},{\"feedbacktargets\":[{\"targetfbid\":\"10151228750629030\",\"entidentifier\":\"10151228750629030\",\"permalink\":\"http:\\/\\/www.facebook.com\\/photo.php?fbid=10151228750629030&set=a.164247464029.118730.143412869029&type=1\",\"commentcount\":3,\"likecount\":23,\"viewercanlike\":false,\"hasviewerliked\":false,\"cancomment\":false,\"showremovemenu\":false,\"actorforpost\":\"\",\"viewerid\":\"0\",\"likesentences\":{\"current\":{\"text\":\"Jeffery Davidson, Amy Pennebaker Cannady, Terri Griffin Taylor and 20 others like this.\",\"ranges\":[{\"offset\":0,\"length\":16,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/jeffdavidson33\",\"id\":53500988}]},{\"offset\":18,\"length\":22,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/amypennebaker.cannady\",\"id\":100000608334199}]},{\"offset\":42,\"length\":20,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/terri.g.taylor.5\",\"id\":100000723810456}]}],\"aggregatedranges\":[{\"offset\":67,\"length\":9,\"count\":20}]},\"alternate\":{\"text\":\"You, Jeffery Davidson, Amy Pennebaker Cannady, Terri Griffin Taylor and 20 others like this.\",\"ranges\":[{\"offset\":5,\"length\":16,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/jeffdavidson33\",\"id\":53500988}]},{\"offset\":23,\"length\":22,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/amypennebaker.cannady\",\"id\":100000608334199}]},{\"offset\":47,\"length\":20,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/terri.g.taylor.5\",\"id\":100000723810456}]}],\"aggregatedranges\":[{\"offset\":72,\"length\":9,\"count\":20}]}},\"sharecount\":2,\"ownerid\":\"143412869029\",\"lastseentime\":null,\"canremoveall\":false,\"seencount\":0,\"seenbyall\":false,\"hasunseencollapsed\":false,\"mentionsdatasource\":{\"__m\":\"m885389_166\"},\"isranked\":false,\"isthreaded\":false,\"isownerpage\":true}],\"comments\":[{\"id\":\"10151228750629030_8240124\",\"fbid\":\"10151229379864030\",\"legacyid\":\"8240124\",\"body\":{\"text\":\"Autumn fronds of cinnamon fern\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"666491233\",\"ftentidentifier\":\"10151228750629030\",\"timestamp\":{\"time\":1351302456,\"text\":\"October 26 at 6:47pm\",\"verbose\":\"Friday, October 26, 2012 at 6:47pm\"},\"likecount\":0,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":null,\"istranslatable\":false},{\"id\":\"10151228750629030_8240148\",\"fbid\":\"10151229383184030\",\"legacyid\":\"8240148\",\"body\":{\"text\":\"Captured Pine Barren Gentians there last week\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"1243046580\",\"ftentidentifier\":\"10151228750629030\",\"timestamp\":{\"time\":1351302722,\"text\":\"October 26 at 6:52pm\",\"verbose\":\"Friday, October 26, 2012 at 6:52pm\"},\"likecount\":1,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":null,\"istranslatable\":false},{\"id\":\"10151228750629030_8240177\",\"fbid\":\"10151229387484030\",\"legacyid\":\"8240177\",\"body\":{\"text\":\"There were some blooming along a roadside near Pink Hill couple weeks ago.\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"666491233\",\"ftentidentifier\":\"10151228750629030\",\"timestamp\":{\"time\":1351303086,\"text\":\"October 26 at 6:58pm\",\"verbose\":\"Friday, October 26, 2012 at 6:58pm\"},\"likecount\":0,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":null,\"istranslatable\":false}],\"profiles\":[{\"id\":\"666491233\",\"name\":\"Kelvin Taylor\",\"firstName\":\"Kelvin\",\"vanity\":\"plantsman\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-snc6\\/275167_666491233_1997685175_q.jpg\",\"uri\":\"http:\\/\\/www.facebook.com\\/plantsman\",\"gender\":2,\"type\":\"user\"},{\"id\":\"1243046580\",\"name\":\"Joe Rone\",\"firstName\":\"Joe\",\"vanity\":\"joe.rone.39\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-snc7\\/371474_1243046580_158145042_q.jpg\",\"uri\":\"http:\\/\\/www.facebook.com\\/joe.rone.39\",\"gender\":2,\"type\":\"user\"}],\"actions\":[],\"commentlists\":{\"10151228750629030\":{\"range\":{\"offset\":0,\"length\":3},\"values\":[\"10151228750629030_8240124\",\"10151228750629030_8240148\",\"10151228750629030_8240177\"]}},\"servertime\":1352998853}],1],[\"m885388_144\",[\"UFIController\",\"m885388_143\",\"m885389_167\"],[{\"__m\":\"m885388_143\"},{\"ftentidentifier\":\"10151228504104030\",\"instanceid\":\"u4kzrn2118\",\"source\":0,\"showaddcomment\":true,\"collapseaddcomment\":false,\"markedcomment\":null,\"scrollcomments\":false,\"scrollwidth\":null,\"showshares\":false,\"shownub\":false,\"numberdelimiter\":\",\",\"showtyping\":0,\"logtyping\":0,\"entstream\":false,\"timelinelogdata\":\"AQAnqduipAgoOm_9IjjltRCBJyhIBv2--eLrjr4l_RRflt2UWyH0EPF2kVqzF4Q12S5pfMKmua5i3Ih93QBatDveVbELZ65qOpzYxXZzGwFtgzfa0fu173jC-rcygo27udXQeX_lMB4R4fV9zUYUpfhC_cdrY4PE-9BLE8Gp5Idd-UGSHP50CUJyyh7gFYJEXsWOfxh7a5iCh6VkqaavWrwXgESYx04gW7S6zKfVC4up7JuXIE6E8YTsPgg3WPB1Rz-g6Y6coNu8nnk7p3McPFiocJiOwjt4Wh7CB7bDVAtAe7E1168oJ7B3R-9lbX53nxlibQl-L6gx0v3C3w-9bqKnhcanSMX6LXUFb-2zwSbivYmm2R3blGuA9ihYamQy7zA\"},{\"feedbacktargets\":[{\"targetfbid\":\"10151228504104030\",\"entidentifier\":\"10151228504104030\",\"permalink\":\"\\/NorthCarolinaStateParks\\/posts\\/10151228504104030\",\"commentcount\":2,\"likecount\":5,\"viewercanlike\":false,\"hasviewerliked\":false,\"cancomment\":false,\"showremovemenu\":false,\"actorforpost\":\"\",\"viewerid\":\"0\",\"likesentences\":{\"current\":{\"text\":\"Patricia Sweeney, Crystal Lloyd, Joey Pyron and 2 others like this.\",\"ranges\":[{\"offset\":0,\"length\":16,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/patricia.sweeney.3367\",\"id\":706894477}]},{\"offset\":18,\"length\":13,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/Crystal.Lloyd89\",\"id\":509951606}]},{\"offset\":33,\"length\":10,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/joey.pyron\",\"id\":100002897607424}]}],\"aggregatedranges\":[{\"offset\":48,\"length\":8,\"count\":2}]},\"alternate\":{\"text\":\"You, Patricia Sweeney, Crystal Lloyd, Joey Pyron and 2 others like this.\",\"ranges\":[{\"offset\":5,\"length\":16,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/patricia.sweeney.3367\",\"id\":706894477}]},{\"offset\":23,\"length\":13,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/Crystal.Lloyd89\",\"id\":509951606}]},{\"offset\":38,\"length\":10,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/joey.pyron\",\"id\":100002897607424}]}],\"aggregatedranges\":[{\"offset\":53,\"length\":8,\"count\":2}]}},\"sharecount\":0,\"ownerid\":\"143412869029\",\"lastseentime\":null,\"canremoveall\":false,\"seencount\":0,\"seenbyall\":false,\"hasunseencollapsed\":false,\"mentionsdatasource\":{\"__m\":\"m885389_167\"},\"isranked\":false,\"isthreaded\":false,\"isownerpage\":true}],\"comments\":[{\"id\":\"10151228504104030_24893771\",\"fbid\":\"10151228505069030\",\"legacyid\":\"24893771\",\"body\":{\"text\":\"Ma Nature can be mean.\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"1235839703\",\"ftentidentifier\":\"10151228504104030\",\"timestamp\":{\"time\":1351261103,\"text\":\"October 26 at 7:18am\",\"verbose\":\"Friday, October 26, 2012 at 7:18am\"},\"likecount\":0,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":0,\"istranslatable\":false},{\"id\":\"10151228504104030_24893839\",\"fbid\":\"10151228513599030\",\"legacyid\":\"24893839\",\"body\":{\"text\":\"Boo!\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"845930579\",\"ftentidentifier\":\"10151228504104030\",\"timestamp\":{\"time\":1351261469,\"text\":\"October 26 at 7:24am\",\"verbose\":\"Friday, October 26, 2012 at 7:24am\"},\"likecount\":0,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":0,\"istranslatable\":false}],\"profiles\":[{\"id\":\"1235839703\",\"name\":\"Robin Proctor\",\"firstName\":\"Robin\",\"vanity\":\"robin.proctor.75\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-ash3\\/41442_1235839703_1087423096_q.jpg\",\"uri\":null,\"gender\":1,\"type\":\"user\"},{\"id\":\"845930579\",\"name\":\"Laura McGuinn\",\"firstName\":\"Laura\",\"vanity\":\"foolofatook\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-ash4\\/187584_845930579_1173473082_q.jpg\",\"uri\":\"http:\\/\\/www.facebook.com\\/foolofatook\",\"gender\":1,\"type\":\"user\"}],\"actions\":[],\"commentlists\":{\"10151228504104030\":{\"range\":{\"offset\":0,\"length\":2},\"values\":[\"10151228504104030_24893771\",\"10151228504104030_24893839\"]}},\"servertime\":1352998853}],1],[\"m885388_146\",[\"UFIController\",\"m885388_145\",\"m885389_168\"],[{\"__m\":\"m885388_145\"},{\"ftentidentifier\":\"10151228392704030\",\"instanceid\":\"u4kzrn2119\",\"source\":0,\"showaddcomment\":true,\"collapseaddcomment\":false,\"markedcomment\":null,\"scrollcomments\":false,\"scrollwidth\":null,\"showshares\":false,\"shownub\":false,\"numberdelimiter\":\",\",\"showtyping\":0,\"logtyping\":0,\"entstream\":false,\"timelinelogdata\":\"AQDkfiAFcMFmC4_ZDIx9bQS1fakojF9Xy_eFn8wrPOXfkINtK4qKeQeu7r2Ao072L_43OkiIzj7jBLEXJufthAR9yH_tEpV6o5KWyFSuIl_qwb8NXyTXCUwi-9t4SAkco3BJa-xRphunCC16bmZyNdNaCoFyK7n767709BoyS7iL4-HldWOzzz6ZVJSeZxhARCS0kDqKYJiNA5P-VUAdmBnGz15lhOR8U3VTj62mZf1TTsRulRWXSk9NB9Ob3502-cARFB-mZuoLZaZshBhfmYHvLxJa9-J1J-YRfDQWi3i70QPxEavoW-uCmsLwkSUjtH4sZcAM9B8yshm2Alg401-3otztTlacqDXV4rt7Kxc2CRzCVwA1_fE68zRBASaotko\"},{\"feedbacktargets\":[{\"targetfbid\":\"10151228392704030\",\"entidentifier\":\"10151228392704030\",\"permalink\":\"\\/NorthCarolinaStateParks\\/posts\\/10151228392704030\",\"commentcount\":1,\"likecount\":4,\"viewercanlike\":false,\"hasviewerliked\":false,\"cancomment\":false,\"showremovemenu\":false,\"actorforpost\":\"\",\"viewerid\":\"0\",\"likesentences\":{\"current\":{\"text\":\"Joey Pyron, Pat Anthony and 2 others like this.\",\"ranges\":[{\"offset\":0,\"length\":10,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/joey.pyron\",\"id\":100002897607424}]},{\"offset\":12,\"length\":11,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/pat.anthony1\",\"id\":528902148}]}],\"aggregatedranges\":[{\"offset\":28,\"length\":8,\"count\":2}]},\"alternate\":{\"text\":\"You, Joey Pyron, Pat Anthony and 2 others like this.\",\"ranges\":[{\"offset\":5,\"length\":10,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/joey.pyron\",\"id\":100002897607424}]},{\"offset\":17,\"length\":11,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/pat.anthony1\",\"id\":528902148}]}],\"aggregatedranges\":[{\"offset\":33,\"length\":8,\"count\":2}]}},\"sharecount\":2,\"ownerid\":\"143412869029\",\"lastseentime\":null,\"canremoveall\":false,\"seencount\":0,\"seenbyall\":false,\"hasunseencollapsed\":false,\"mentionsdatasource\":{\"__m\":\"m885389_168\"},\"isranked\":false,\"isthreaded\":false,\"isownerpage\":true}],\"comments\":[{\"id\":\"10151228392704030_24893689\",\"fbid\":\"10151228497644030\",\"legacyid\":\"24893689\",\"body\":{\"text\":\":( oh well, that's how it goes with outdoor venues.\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"100000400675265\",\"ftentidentifier\":\"10151228392704030\",\"timestamp\":{\"time\":1351260747,\"text\":\"October 26 at 7:12am\",\"verbose\":\"Friday, October 26, 2012 at 7:12am\"},\"likecount\":0,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":0,\"istranslatable\":false}],\"profiles\":[{\"id\":\"100000400675265\",\"name\":\"Judi Nicholson\",\"firstName\":\"Judi\",\"vanity\":\"judi.nicholson.16\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-snc6\\/187254_100000400675265_448802830_q.jpg\",\"uri\":null,\"gender\":1,\"type\":\"user\"}],\"actions\":[],\"commentlists\":{\"10151228392704030\":{\"range\":{\"offset\":0,\"length\":1},\"values\":[\"10151228392704030_24893689\"]}},\"servertime\":1352998853}],1],[\"m885388_148\",[\"UFIController\",\"m885388_147\",\"m885389_169\"],[{\"__m\":\"m885388_147\"},{\"ftentidentifier\":\"10151238196354030\",\"instanceid\":\"u4kzrn2120\",\"source\":0,\"showaddcomment\":true,\"collapseaddcomment\":false,\"markedcomment\":null,\"scrollcomments\":false,\"scrollwidth\":null,\"showshares\":false,\"shownub\":false,\"numberdelimiter\":\",\",\"showtyping\":0,\"logtyping\":0,",
        "\"entstream\":false,\"timelinelogdata\":\"AQDuRVfg_8dXBhIsIw-zeGtZPbdf-zz8-bLQzsP6ye-ibq60RrHOgBae6EDtd7RE7wsQOBU1exBD7U2LnkP6nm9sUFaXey87FvZKHMtYV_nxoT0ODFDWec4scQ_mdTgle8znVNp4Q72dMjt40ONUx7SCCzSSINjJmM6l0dEiQVlmCIyvUcEiXgIuY_e_Yye6273stvLgW2rR7mg1sRhzXAsISrQAJFk3VKoAmHMYvUnRwFkBjrveoFzgtNQFqNSEMK5P0-s_4Z2o242zUrbER8as2JNAXeUxwF5-loVDolpA-RpEppBaLNf4sfG_VgIT5TPnDzN1_KED3SH4oce5beGH1-Lw4L6AlE_E8jgFGIsTzyx-OzWBKMzI0lFNxgBfb48\"},{\"feedbacktargets\":[{\"targetfbid\":\"10151238196354030\",\"entidentifier\":\"10151238196354030\",\"permalink\":\"http:\\/\\/www.facebook.com\\/photo.php?fbid=10151238196354030&set=a.164247464029.118730.143412869029&type=1\",\"commentcount\":1,\"likecount\":65,\"viewercanlike\":false,\"hasviewerliked\":false,\"cancomment\":false,\"showremovemenu\":false,\"actorforpost\":\"\",\"viewerid\":\"0\",\"likesentences\":{\"current\":{\"text\":\"NatureFind Raleigh-Durham, Catawba County Parks, Jay Young and 62 others like this.\",\"ranges\":[{\"offset\":0,\"length\":25,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/NatureFindRaleighDurham\",\"id\":171854356200534}]},{\"offset\":27,\"length\":20,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/catawbacountyparks\",\"id\":94474264671}]},{\"offset\":49,\"length\":9,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/jay.young.1272\",\"id\":100000137846801}]}],\"aggregatedranges\":[{\"offset\":63,\"length\":9,\"count\":62}]},\"alternate\":{\"text\":\"You, NatureFind Raleigh-Durham, Catawba County Parks, Jay Young and 62 others like this.\",\"ranges\":[{\"offset\":5,\"length\":25,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/NatureFindRaleighDurham\",\"id\":171854356200534}]},{\"offset\":32,\"length\":20,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/catawbacountyparks\",\"id\":94474264671}]},{\"offset\":54,\"length\":9,\"entities\":[{\"url\":\"http:\\/\\/www.facebook.com\\/jay.young.1272\",\"id\":100000137846801}]}],\"aggregatedranges\":[{\"offset\":68,\"length\":9,\"count\":62}]}},\"sharecount\":3,\"ownerid\":\"143412869029\",\"lastseentime\":null,\"canremoveall\":false,\"seencount\":0,\"seenbyall\":false,\"hasunseencollapsed\":false,\"mentionsdatasource\":{\"__m\":\"m885389_169\"},\"isranked\":false,\"isthreaded\":false,\"isownerpage\":true}],\"comments\":[{\"id\":\"10151238196354030_8290075\",\"fbid\":\"10151239854259030\",\"legacyid\":\"8290075\",\"body\":{\"text\":\"What a great pic of a great river!\",\"ranges\":[],\"aggregatedranges\":[]},\"author\":\"40401549579\",\"ftentidentifier\":\"10151238196354030\",\"timestamp\":{\"time\":1351869287,\"text\":\"November 2 at 8:14am\",\"verbose\":\"Friday, November 2, 2012 at 8:14am\"},\"likecount\":0,\"hasviewerliked\":false,\"canremove\":false,\"canreport\":false,\"canedit\":false,\"source\":null,\"istranslatable\":false}],\"profiles\":[{\"id\":\"40401549579\",\"name\":\"Neuse Riverkeeper Foundation\",\"firstName\":\"Neuse Riverkeeper Foundation\",\"vanity\":\"NeuseRiverkeeperFoundation\",\"thumbSrc\":\"https:\\/\\/fbcdn-profile-a.akamaihd.net\\/hprofile-ak-snc7\\/373034_40401549579_181859885_q.jpg\",\"uri\":\"http:\\/\\/www.facebook.com\\/NeuseRiverkeeperFoundation\",\"type\":\"page\"}],\"actions\":[],\"commentlists\":{\"10151238196354030\":{\"range\":{\"offset\":0,\"length\":1},\"values\":[\"10151238196354030_8290075\"]}},\"servertime\":1352998853}],1],[\"m885382_91\",[\"MultiBootstrapDataSource\"],[{\"maxResults\":3,\"queryData\":{\"context\":\"topics\",\"viewer\":0,\"filter\":[\"page\"],\"rsp\":\"mentions\"},\"queryEndpoint\":\"\\/ajax\\/typeahead\\/search.php\",\"bootstrapData\":{\"rsp\":\"mentions\"},\"bootstrapEndpoints\":[{\"endpoint\":\"\\/ajax\\/typeahead\\/first_degree.php\",\"data\":{\"context\":\"mentions\",\"viewer\":0,\"token\":\"1352998557-7\",\"filter\":[\"page\",\"group\",\"app\",\"event\"],\"options\":[\"friends_only\",\"nm\"]}}]}],2],[\"m885382_92\",[\"MultiBootstrapDataSource\"],[{\"maxResults\":3,\"queryData\":{\"context\":\"topics\",\"viewer\":0,\"filter\":[\"page\"],\"rsp\":\"mentions\"},\"queryEndpoint\":\"\\/ajax\\/typeahead\\/search.php\",\"bootstrapData\":{\"rsp\":\"mentions\"},\"bootstrapEndpoints\":[{\"endpoint\":\"\\/ajax\\/typeahead\\/first_degree.php\",\"data\":{\"context\":\"mentions\",\"viewer\":0,\"token\":\"1352998557-7\",\"filter\":[\"page\",\"group\",\"app\",\"event\"],\"options\":[\"friends_only\",\"nm\"]}}]}],2],[\"m885382_93\",[\"MultiBootstrapDataSource\"],[{\"maxResults\":3,\"queryData\":{\"context\":\"topics\",\"viewer\":0,\"filter\":[\"page\"],\"rsp\":\"mentions\"},\"queryEndpoint\":\"\\/ajax\\/typeahead\\/search.php\",\"bootstrapData\":{\"rsp\":\"mentions\"},\"bootstrapEndpoints\":[{\"endpoint\":\"\\/ajax\\/typeahead\\/first_degree.php\",\"data\":{\"context\":\"mentions\",\"viewer\":0,\"token\":\"1352998557-7\",\"filter\":[\"page\",\"group\",\"app\",\"event\"],\"options\":[\"friends_only\",\"nm\"]}}]}],2],[\"m885382_94\",[\"MultiBootstrapDataSource\"],[{\"maxResults\":3,\"queryData\":{\"context\":\"topics\",\"viewer\":0,\"filter\":[\"page\"],\"rsp\":\"mentions\"},\"queryEndpoint\":\"\\/ajax\\/typeahead\\/search.php\",\"bootstrapData\":{\"rsp\":\"mentions\"},\"bootstrapEndpoints\":[{\"endpoint\":\"\\/ajax\\/typeahead\\/first_degree.php\",\"data\":{\"context\":\"mentions\",\"viewer\":0,\"token\":\"1352998557-7\",\"filter\":[\"page\",\"group\",\"app\",\"event\"],\"options\":[\"friends_only\",\"nm\"]}}]}],2],[\"m885383_95\",[\"MultiBootstrapDataSource\"],[{\"maxResults\":3,\"queryData\":{\"context\":\"topics\",\"viewer\":0,\"filter\":[\"page\"],\"rsp\":\"mentions\"},\"queryEndpoint\":\"\\/ajax\\/typeahead\\/search.php\",\"bootstrapData\":{\"rsp\":\"mentions\"},\"bootstrapEndpoints\":[{\"endpoint\":\"\\/ajax\\/typeahead\\/first_degree.php\",\"data\":{\"context\":\"mentions\",\"viewer\":0,\"token\":\"1352998557-7\",\"filter\":[\"page\",\"group\",\"app\",\"event\"],\"options\":[\"friends_only\",\"nm\"]}}]}],2],[\"m885383_96\",[\"MultiBootstrapDataSource\"],[{\"maxResults\":3,\"queryData\":{\"context\":\"topics\",\"viewer\":0,\"filter\":[\"page\"],\"rsp\":\"mentions\"},\"queryEndpoint\":\"\\/ajax\\/typeahead\\/search.php\",\"bootstrapData\":{\"rsp\":\"mentions\"},\"bootstrapEndpoints\":[{\"endpoint\":\"\\/ajax\\/typeahead\\/first_degree.php\",\"data\":{\"context\":\"mentions\",\"viewer\":0,\"token\":\"1352998557-7\",\"filter\":[\"page\",\"group\",\"app\",\"event\"],\"options\":[\"friends_only\",\"nm\"]}}]}],2],[\"m885383_97\",[\"MultiBootstrapDataSource\"],[{\"maxResults\":3,\"queryData\":{\"context\":\"topics\",\"viewer\":0,\"filter\":[\"page\"],\"rsp\":\"mentions\"},\"queryEndpoint\":\"\\/ajax\\/typeahead\\/search.php\",\"bootstrapData\":{\"rsp\":\"mentions\"},\"bootstrapEndpoints\":[{\"endpoint\":\"\\/ajax\\/typeahead\\/first_degree.php\",\"data\":{\"context\":\"mentions\",\"viewer\":0,\"token\":\"1352998557-7\",\"filter\":[\"page\",\"group\",\"app\",\"event\"],\"options\":[\"friends_only\",\"nm\"]}}]}],2],[\"m885383_98\",[\"MultiBootstrapDataSource\"],[{\"maxResults\":3,\"queryData\":{\"context\":\"topics\",\"viewer\":0,\"filter\":[\"page\"],\"rsp\":\"mentions\"},\"queryEndpoint\":\"\\/ajax\\/typeahead\\/search.php\",\"bootstrapData\":{\"rsp\":\"mentions\"},\"bootstrapEndpoints\":[{\"endpoint\":\"\\/ajax\\/typeahead\\/first_degree.php\",\"data\":{\"context\":\"mentions\",\"viewer\":0,\"token\":\"1352998557-7\",\"filter\":[\"page\",\"group\",\"app\",\"event\"],\"options\":[\"friends_only\",\"nm\"]}}]}],2],[\"m885383_99\",[\"MultiBootstrapDataSource\"],[{\"maxResults\":3,\"queryData\":{\"context\":\"topics\",\"viewer\":0,\"filter\":[\"page\"],\"rsp\":\"mentions\"},\"queryEndpoint\":\"\\/ajax\\/typeahead\\/search.php\",\"bootstrapData\":{\"rsp\":\"mentions\"},\"bootstrapEndpoints\":[{\"endpoint\":\"\\/ajax\\/typeahead\\/first_degree.php\",\"data\":{\"context\":\"mentions\",\"viewer\":0,\"token\":\"1352998557-7\",\"filter\":[\"page\",\"group\",\"app\",\"event\"],\"options\":[\"friends_only\",\"nm\"]}}]}],2],[\"m885383_100\",[\"MultiBootstrapDataSource\"],[{\"maxResults\":3,\"queryData\":{\"context\":\"topics\",\"viewer\":0,\"filter\":[\"page\"],\"rsp\":\"mentions\"},\"queryEndpoint\":\"\\/ajax\\/typeahead\\/search.php\",\"bootstrapData\":{\"rsp\":\"mentions\"},\"bootstrapEndpoints\":[{\"endpoint\":\"\\/ajax\\/typeahead\\/first_degree.php\",\"data\":{\"context\":\"mentions\",\"viewer\":0,\"token\":\"1352998557-7\",\"filter\":[\"page\",\"group\",\"app\",\"event\"],\"options\":[\"friends_only\",\"nm\"]}}]}],2],[\"m885388_103\",[\"MultiBootstrapDataSource\"],[{\"maxResults\":3,\"queryData\":{\"context\":\"topics\",\"viewer\":0,\"filter\":[\"page\"],\"rsp\":\"mentions\"},\"queryEndpoint\":\"\\/ajax\\/typeahead\\/search.php\",\"bootstrapData\":{\"rsp\":\"mentions\"},\"bootstrapEndpoints\":[{\"endpoint\":\"\\/ajax\\/typeahead\\/first_degree.php\",\"data\":{\"context\":\"mentions\",\"viewer\":0,\"token\":\"1352998557-7\",\"filter\":[\"page\",\"group\",\"app\",\"event\"],\"options\":[\"friends_only\",\"nm\"]}}]}],2],[\"m885388_106\",[\"MultiBootstrapDataSource\"],[{\"maxResults\":3,\"queryData\":{\"context\":\"topics\",\"viewer\":0,\"filter\":[\"page\"],\"rsp\":\"mentions\"},\"queryEndpoint\":\"\\/ajax\\/typeahead\\/search.php\",\"bootstrapData\":{\"rsp\":\"mentions\"},\"bootstrapEndpoints\":[{\"endpoint\":\"\\/ajax\\/typeahead\\/first_degree.php\",\"data\":{\"context\":\"mentions\",\"viewer\":0,\"token\":\"1352998557-7\",\"filter\":[\"page\",\"group\",\"app\",\"event\"],\"options\":[\"friends_only\",\"nm\"]}}]}],2],[\"m885388_149\",[\"MultiBootstrapDataSource\"],[{\"maxResults\":3,\"queryData\":{\"context\":\"topics\",\"viewer\":0,\"filter\":[\"page\"],\"rsp\":\"mentions\"},\"queryEndpoint\":\"\\/ajax\\/typeahead\\/search.php\",\"bootstrapData\":{\"rsp\":\"mentions\"},\"bootstrapEndpoints\":[{\"endpoint\":\"\\/ajax\\/typeahead\\/first_degree.php\",\"data\":{\"context\":\"mentions\",\"viewer\":0,\"token\":\"1352998557-7\",\"filter\":[\"page\",\"group\",\"app\",\"event\"],\"options\":[\"friends_only\",\"nm\"]}}]}],2],[\"m885388_150\",[\"MultiBootstrapDataSource\"],[{\"maxResults\":3,\"queryData\":{\"context\":\"topics\",\"viewer\":0,\"filter\":[\"page\"],\"rsp\":\"mentions\"},\"queryEndpoint\":\"\\/ajax\\/typeahead\\/search.php\",\"bootstrapData\":{\"rsp\":\"mentions\"},\"bootstrapEndpoints\":[{\"endpoint\":\"\\/ajax\\/typeahead\\/first_degree.php\",\"data\":{\"context\":\"mentions\",\"viewer\":0,\"token\":\"1352998557-7\",\"filter\":[\"page\",\"group\",\"app\",\"event\"],\"options\":[\"friends_only\",\"nm\"]}}]}],2],[\"m885388_151\",[\"MultiBootstrapDataSource\"],[{\"maxResults\":3,\"queryData\":{\"context\":\"topics\",\"viewer\":0,\"filter\":[\"page\"],\"rsp\":\"mentions\"},\"queryEndpoint\":\"\\/ajax\\/typeahead\\/search.php\",\"bootstrapData\":{\"rsp\":\"mentions\"},\"bootstrapEndpoints\":[{\"endpoint\":\"\\/ajax\\/typeahead\\/first_degree.php\",\"data\":{\"context\":\"mentions\",\"viewer\":0,\"token\":\"1352998557-7\",\"filter\":[\"page\",\"group\",\"app\",\"event\"],\"options\":[\"friends_only\",\"nm\"]}}]}],2],[\"m885388_152\",[\"MultiBootstrapDataSource\"],[{\"maxResults\":3,\"queryData\":{\"context\":\"topics\",\"viewer\":0,\"filter\":[\"page\"],\"rsp\":\"mentions\"},\"queryEndpoint\":\"\\/ajax\\/typeahead\\/search.php\",\"bootstrapData\":{\"rsp\":\"mentions\"},\"bootstrapEndpoints\":[{\"endpoint\":\"\\/ajax\\/typeahead\\/first_degree.php\",\"data\":{\"context\":\"mentions\",\"viewer\":0,\"token\":\"1352998557-7\",\"filter\":[\"page\",\"group\",\"app\",\"event\"],\"options\":[\"friends_only\",\"nm\"]}}]}],2],[\"m885388_153\",[\"MultiBootstrapDataSource\"],[{\"maxResults\":3,\"queryData\":{\"context\":\"topics\",\"viewer\":0,\"filter\":[\"page\"],\"rsp\":\"mentions\"},\"queryEndpoint\":\"\\/ajax\\/typeahead\\/search.php\",\"bootstrapData\":{\"rsp\":\"mentions\"},\"bootstrapEndpoints\":[{\"endpoint\":\"\\/ajax\\/typeahead\\/first_degree.php\",\"data\":{\"context\":\"mentions\",\"viewer\":0,\"token\":\"1352998557-7\",\"filter\":[\"page\",\"group\",\"app\",\"event\"],\"options\":[\"friends_only\",\"nm\"]}}]}],2],[\"m885388_154\",[\"MultiBootstrapDataSource\"],[{\"maxResults\":3,\"queryData\":{\"context\":\"topics\",\"viewer\":0,\"filter\":[\"page\"],\"rsp\":\"mentions\"},\"queryEndpoint\":\"\\/ajax\\/typeahead\\/search.php\",\"bootstrapData\":{\"rsp\":\"mentions\"},\"bootstrapEndpoints\":[{\"endpoint\":\"\\/ajax\\/typeahead\\/first_degree.php\",\"data\":{\"context\":\"mentions\",\"viewer\":0,\"token\":\"1352998557-7\",\"filter\":[\"page\",\"group\",\"app\",\"event\"],\"options\":[\"friends_only\",\"nm\"]}}]}],2],[\"m885388_155\",[\"MultiBootstrapDataSource\"],[{\"maxResults\":3,\"queryData\":{\"context\":\"topics\",\"viewer\":0,\"filter\":[\"page\"],\"rsp\":\"mentions\"},\"queryEndpoint\":\"\\/ajax\\/typeahead\\/search.php\",\"bootstrapData\":{\"rsp\":\"mentions\"},\"bootstrapEndpoints\":[{\"endpoint\":\"\\/ajax\\/typeahead\\/first_degree.php\",\"data\":{\"context\":\"mentions\",\"viewer\":0,\"token\":\"1352998557-7\",\"filter\":[\"page\",\"group\",\"app\",\"event\"],\"options\":[\"friends_only\",\"nm\"]}}]}],2],[\"m885388_156\",[\"MultiBootstrapDataSource\"],[{\"maxResults\":3,\"queryData\":{\"context\":\"topics\",\"viewer\":0,\"filter\":[\"page\"],\"rsp\":\"mentions\"},\"queryEndpoint\":\"\\/ajax\\/typeahead\\/search.php\",\"bootstrapData\":{\"rsp\":\"mentions\"},\"bootstrapEndpoints\":[{\"endpoint\":\"\\/ajax\\/typeahead\\/first_degree.php\",\"data\":{\"context\":\"mentions\",\"viewer\":0,\"token\":\"1352998557-7\",\"filter\":[\"page\",\"group\",\"app\",\"event\"],\"options\":[\"friends_only\",\"nm\"]}}]}],2],[\"m885388_157\",[\"MultiBootstrapDataSource\"],[{\"maxResults\":3,\"queryData\":{\"context\":\"topics\",\"viewer\":0,\"filter\":[\"page\"],\"rsp\":\"mentions\"},\"queryEndpoint\":\"\\/ajax\\/typeahead\\/search.php\",\"bootstrapData\":{\"rsp\":\"mentions\"},\"bootstrapEndpoints\":[{\"endpoint\":\"\\/ajax\\/typeahead\\/first_degree.php\",\"data\":{\"context\":\"mentions\",\"viewer\":0,\"token\":\"1352998557-7\",\"filter\":[\"page\",\"group\",\"app\",\"event\"],\"options\":[\"friends_only\",\"nm\"]}}]}],2],[\"m885388_158\",[\"MultiBootstrapDataSource\"],[{\"maxResults\":3,\"queryData\":{\"context\":\"topics\",\"viewer\":0,\"filter\":[\"page\"],\"rsp\":\"mentions\"},\"queryEndpoint\":\"\\/ajax\\/typeahead\\/search.php\",\"bootstrapData\":{\"rsp\":\"mentions\"},\"bootstrapEndpoints\":[{\"endpoint\":\"\\/ajax\\/typeahead\\/first_degree.php\",\"data\":{\"context\":\"mentions\",\"viewer\":0,\"token\":\"1352998557-7\",\"filter\":[\"page\",\"group\",\"app\",\"event\"],\"options\":[\"friends_only\",\"nm\"]}}]}],2],[\"m885389_159\",[\"MultiBootstrapDataSource\"],[{\"maxResults\":3,\"queryData\":{\"context\":\"topics\",\"viewer\":0,\"filter\":[\"page\"],\"rsp\":\"mentions\"},\"queryEndpoint\":\"\\/ajax\\/typeahead\\/search.php\",\"bootstrapData\":{\"rsp\":\"mentions\"},\"bootstrapEndpoints\":[{\"endpoint\":\"\\/ajax\\/typeahead\\/first_degree.php\",\"data\":{\"context\":\"mentions\",\"viewer\":0,\"token\":\"1352998557-7\",\"filter\":[\"page\",\"group\",\"app\",\"event\"],\"options\":[\"friends_only\",\"nm\"]}}]}],2],[\"m885389_160\",[\"MultiBootstrapDataSource\"],[{\"maxResults\":3,\"queryData\":{\"context\":\"topics\",\"viewer\":0,\"filter\":[\"page\"],\"rsp\":\"mentions\"},\"queryEndpoint\":\"\\/ajax\\/typeahead\\/search.php\",\"bootstrapData\":{\"rsp\":\"mentions\"},\"bootstrapEndpoints\":[{\"endpoint\":\"\\/ajax\\/typeahead\\/first_degree.php\",\"data\":{\"context\":\"mentions\",\"viewer\":0,\"token\":\"1352998557-7\",\"filter\":[\"page\",\"group\",\"app\",\"event\"],\"options\":[\"friends_only\",\"nm\"]}}]}],2],[\"m885389_161\",[\"MultiBootstrapDataSource\"],[{\"maxResults\":3,\"queryData\":{\"context\":\"topics\",\"viewer\":0,\"filter\":[\"page\"],\"rsp\":\"mentions\"},\"queryEndpoint\":\"\\/ajax\\/typeahead\\/search.php\",\"bootstrapData\":{\"rsp\":\"mentions\"},\"bootstrapEndpoints\":[{\"endpoint\":\"\\/ajax\\/typeahead\\/first_degree.php\",\"data\":{\"context\":\"mentions\",\"viewer\":0,\"token\":\"1352998557-7\",\"filter\":[\"page\",\"group\",\"app\",\"event\"],\"options\":[\"friends_only\",\"nm\"]}}]}],2],[\"m885389_162\",[\"MultiBootstrapDataSource\"],[{\"maxResults\":3,\"queryData\":{\"context\":\"topics\",\"viewer\":0,\"filter\":[\"page\"],\"rsp\":\"mentions\"},\"queryEndpoint\":\"\\/ajax\\/typeahead\\/search.php\",\"bootstrapData\":{\"rsp\":\"mentions\"},\"bootstrapEndpoints\":[{\"endpoint\":\"\\/ajax\\/typeahead\\/first_degree.php\",\"data\":{\"context\":\"mentions\",\"viewer\":0,\"token\":\"1352998557-7\",\"filter\":[\"page\",\"group\",\"app\",\"event\"],\"options\":[\"friends_only\",\"nm\"]}}]}],2],[\"m885389_163\",[\"MultiBootstrapDataSource\"],[{\"maxResults\":3,\"queryData\":{\"context\":\"topics\",\"viewer\":0,\"filter\":[\"page\"],\"rsp\":\"mentions\"},\"queryEndpoint\":\"\\/ajax\\/typeahead\\/search.php\",\"bootstrapData\":{\"rsp\":\"mentions\"},\"bootstrapEndpoints\":[{\"endpoint\":\"\\/ajax\\/typeahead\\/first_degree.php\",\"data\":{\"context\":\"mentions\",\"viewer\":0,\"token\":\"1352998557-7\",\"filter\":[\"page\",\"group\",\"app\",\"event\"],\"options\":[\"friends_only\",\"nm\"]}}]}],2],[\"m885389_164\",[\"MultiBootstrapDataSource\"],[{\"maxResults\":3,\"queryData\":{\"context\":\"topics\",\"viewer\":0,\"filter\":[\"page\"],\"rsp\":\"mentions\"},\"queryEndpoint\":\"\\/ajax\\/typeahead\\/search.php\",\"bootstrapData\":{\"rsp\":\"mentions\"},\"bootstrapEndpoints\":[{\"endpoint\":\"\\/ajax\\/typeahead\\/first_degree.php\",\"data\":{\"context\":\"mentions\",\"viewer\":0,\"token\":\"1352998557-7\",\"filter\":[\"page\",\"group\",\"app\",\"event\"],\"options\":[\"friends_only\",\"nm\"]}}]}],2],[\"m885389_165\",[\"MultiBootstrapDataSource\"],[{\"maxResults\":3,\"queryData\":{\"context\":\"topics\",\"viewer\":0,\"filter\":[\"page\"],\"rsp\":\"mentions\"},\"queryEndpoint\":\"\\/ajax\\/typeahead\\/search.php\",\"bootstrapData\":{\"rsp\":\"mentions\"},\"bootstrapEndpoints\":[{\"endpoint\":\"\\/ajax\\/typeahead\\/first_degree.php\",\"data\":{\"context\":\"mentions\",\"viewer\":0,\"token\":\"1352998557-7\",\"filter\":[\"page\",\"group\",\"app\",\"event\"],\"options\":[\"friends_only\",\"nm\"]}}]}],2],[\"m885389_166\",[\"MultiBootstrapDataSource\"],[{\"maxResults\":3,\"queryData\":{\"context\":\"topics\",\"viewer\":0,\"filter\":[\"page\"],\"rsp\":\"mentions\"},\"queryEndpoint\":\"\\/ajax\\/typeahead\\/search.php\",\"bootstrapData\":{\"rsp\":\"mentions\"},\"bootstrapEndpoints\":[{\"endpoint\":\"\\/ajax\\/typeahead\\/first_degree.php\",\"data\":{\"context\":\"mentions\",\"viewer\":0,\"token\":\"1352998557-7\",\"filter\":[\"page\",\"group\",\"app\",\"event\"],\"options\":[\"friends_only\",\"nm\"]}}]}],2],[\"m885389_167\",[\"MultiBootstrapDataSource\"],[{\"maxResults\":3,\"queryData\":{\"context\":\"topics\",\"viewer\":0,\"filter\":[\"page\"],\"rsp\":\"mentions\"},\"queryEndpoint\":\"\\/ajax\\/typeahead\\/search.php\",\"bootstrapData\":{\"rsp\":\"mentions\"},\"bootstrapEndpoints\":[{\"endpoint\":\"\\/ajax\\/typeahead\\/first_degree.php\",\"data\":{\"context\":\"mentions\",\"viewer\":0,\"token\":\"1352998557-7\",\"filter\":[\"page\",\"group\",\"app\",\"event\"],\"options\":[\"friends_only\",\"nm\"]}}]}],2],[\"m885389_168\",[\"MultiBootstrapDataSource\"],[{\"maxResults\":3,\"queryData\":{\"context\":\"topics\",\"viewer\":0,\"filter\":[\"page\"],\"rsp\":\"mentions\"},\"queryEndpoint\":\"\\/ajax\\/typeahead\\/search.php\",\"bootstrapData\":{\"rsp\":\"mentions\"},\"bootstrapEndpoints\":[{\"endpoint\":\"\\/ajax\\/typeahead\\/first_degree.php\",\"data\":{\"context\":\"mentions\",\"viewer\":0,\"token\":\"1352998557-7\",\"filter\":[\"page\",\"group\",\"app\",\"event\"],\"options\":[\"friends_only\",\"nm\"]}}]}],2],[\"m885389_169\",[\"MultiBootstrapDataSource\"],[{\"maxResults\":3,\"queryData\":{\"context\":\"topics\",\"viewer\":0,\"filter\":[\"page\"],\"rsp\":\"mentions\"},\"queryEndpoint\":\"\\/ajax\\/typeahead\\/search.php\",\"bootstrapData\":{\"rsp\":\"mentions\"},\"bootstrapEndpoints\":[{\"endpoint\":\"\\/ajax\\/typeahead\\/first_degree.php\",\"data\":{\"context\":\"mentions\",\"viewer\":0,\"token\":\"1352998557-7\",\"filter\":[\"page\",\"group\",\"app\",\"event\"],\"options\":[\"friends_only\",\"nm\"]}}]}],2],[\"m885391_170\",[\"XHPTemplate\",\"m885391_171\"],[{\"__m\":\"m885391_171\"}],2]],\"markup\":[[\"m885391_171\",{\"__html\":\"\\u003Cdiv>\\u003Cdiv class=\\\"pvs phm _1yw\\\" data-jsid=\\\"title\\\">\\u003C\\/div>\\u003Cdiv class=\\\"pam _13\\\" data-jsid=\\\"body\\\">\\u003C\\/div>\\u003Cdiv class=\\\"_14\\\">\\u003Cdiv class=\\\"pam uiOverlayFooter uiBoxGray topborder\\\">\\u003Clabel class=\\\"layerConfirm uiOverlayButton uiButton uiButtonConfirm uiButtonLarge\\\" for=\\\"u4kzrn2175\\\">\\u003Cinput value=\\\"Delete\\\" data-jsid=\\\"delete\\\" type=\\\"submit\\\" id=\\\"u4kzrn2175\\\" \\/>\\u003C\\/label>\\u003Ca class=\\\"layerCancel uiOverlayButton uiButton uiButtonLarge\\\" href=\\\"#\\\" role=\\\"button\\\">\\u003Cspan class=\\\"uiButtonText\\\">Cancel\\u003C\\/span>\\u003C\\/a>\\u003C\\/div>\\u003C\\/div>\\u003C\\/div>\"},2]],\"elements\":[[\"m885253_36\",\"u4kzrn230\",2],[\"m885329_37\",\"u4kzrn230\",2],[\"m885382_71\",\"u4kzrn286\",2],[\"m885382_73\",\"u4kzrn287\",2],[\"m885382_75\",\"u4kzrn288\",2],[\"m885382_77\",\"u4kzrn289\",2],[\"m885382_79\",\"u4kzrn290\",2],[\"m885382_81\",\"u4kzrn291\",2],[\"m885382_83\",\"u4kzrn292\",2],[\"m885382_85\",\"u4kzrn293\",2],[\"m885382_87\",\"u4kzrn294\",2],[\"m885382_89\",\"u4kzrn295\",2],[\"m885388_101\",\"u4kzrn297\",2],[\"m885388_104\",\"u4kzrn299\",2],[\"m885388_107\",\"u4kzrn2121\",2],[\"m885388_109\",\"u4kzrn2122\",2],[\"m885388_111\",\"u4kzrn2123\",2],[\"m885388_113\",\"u4kzrn2124\",2],[\"m885388_115\",\"u4kzrn2125\",2],[\"m885388_117\",\"u4kzrn2126\",2],[\"m885388_119\",\"u4kzrn2127\",2],[\"m885388_121\",\"u4kzrn2128\",2],[\"m885388_123\",\"u4kzrn2129\",2],[\"m885388_125\",\"u4kzrn2130\",2],[\"m885388_127\",\"u4kzrn2131\",2],[\"m885388_129\",\"u4kzrn2132\",2],[\"m885388_131\",\"u4kzrn2133\",2],[\"m885388_133\",\"u4kzrn2134\",2],[\"m885388_135\",\"u4kzrn2135\",2],[\"m885388_137\",\"u4kzrn2136\",2],[\"m885388_139\",\"u4kzrn2137\",2],[\"m885388_141\",\"u4kzrn2138\",2],[\"m885388_143\",\"u4kzrn2139\",2],[\"m885388_145\",\"u4kzrn2140\",2],[\"m885388_147\",\"u4kzrn2141\",2],[\"m885361_38\",\"u4kzrn2142\",2],[\"m885361_39\",\"u4kzrn2143\",2],[\"m885361_40\",\"u4kzrn2144\",2],[\"m885361_41\",\"u4kzrn2145\",2],[\"m885361_42\",\"u4kzrn2146\",2],[\"m885361_43\",\"u4kzrn2147\",2],[\"m885362_44\",\"u4kzrn2148\",2],[\"m885362_45\",\"u4kzrn2149\",2],[\"m885362_46\",\"u4kzrn2150\",2],[\"m885362_47\",\"u4kzrn2151\",2],[\"m885362_48\",\"u4kzrn2152\",2],[\"m885362_49\",\"u4kzrn2153\",2],[\"m885362_50\",\"u4kzrn2154\",2],[\"m885362_51\",\"u4kzrn2155\",2],[\"m885362_52\",\"u4kzrn2156\",2],[\"m885362_53\",\"u4kzrn2157\",2],[\"m885363_54\",\"u4kzrn2158\",2],[\"m885363_55\",\"u4kzrn2159\",2],[\"m885363_56\",\"u4kzrn2160\",2],[\"m885363_57\",\"u4kzrn2161\",2],[\"m885363_58\",\"u4kzrn2162\",2],[\"m885363_59\",\"u4kzrn2163\",2],[\"m885363_60\",\"u4kzrn2164\",2],[\"m885363_61\",\"u4kzrn2165\",2],[\"m885363_62\",\"u4kzrn2166\",2],[\"m885363_63\",\"u4kzrn2167\",2],[\"m885363_64\",\"u4kzrn2168\",2],[\"m885364_65\",\"u4kzrn2169\",2],[\"m885364_66\",\"u4kzrn2170\",2],[\"m885364_67\",\"u4kzrn2171\",2],[\"m885364_68\",\"u4kzrn2172\",2],[\"m885364_69\",\"u4kzrn2173\",2],[\"m885364_70\",\"u4kzrn2174\",2]],\"define\":[[\"UFISpamCountImpl\",[],{\"module\":null}],[\"UFIConfig\",[],{\"renderEmoji\":false,\"renderEmoticons\":1}],[\"UFICommentTemplates\",[\"m885391_170\"],{\":fb:ufi:hide-dialog-template\":{\"__m\":\"m885391_170\"}}]],\"require\":[[\"TimelineContentLoader\",\"setParallelLoadConfig\",[],[{\"required_units\":5,\"max_parallelism\":3}]],[\"TimelineContentLoader\",\"markSectionAsLoaded\",[],[\"u4kzrn231\",\"recent\",\"0\"]],[\"TimelineCapsule\",\"balanceCapsule\",[\"m885253_36\"],[{\"__m\":\"m885253_36\"},{\"tail_balance\":\"both\",\"heights_action\":null}]],[\"TimelineCapsule\",\"balanceCapsule\",[\"m885329_37\"],[{\"__m\":\"m885329_37\"},{\"tail_balance\":\"both\",\"heights_action\":null}]],[\"m885382_72\"],[\"Hovercard\"],[\"Tooltip\"],[\"m885382_74\"],[\"m885382_76\"],[\"m885382_78\"],[\"m885382_80\"],[\"m885382_82\"],[\"m885382_84\"],[\"m885382_86\"],[\"m885382_88\"],[\"m885382_90\"],[\"m885388_102\"],[\"m885388_105\"],[\"m885388_108\"],[\"m885388_110\"],[\"m885388_112\"],[\"m885388_114\"],[\"m885388_116\"],[\"m885388_118\"],[\"m885388_120\"],[\"m885388_122\"],[\"m885388_124\"],[\"m885388_126\"],[\"m885388_128\"],[\"m885388_130\"],[\"m885388_132\"],[\"m885388_134\"],[\"m885388_136\"],[\"m885388_138\"],[\"m885388_140\"],[\"m885388_142\"],[\"m885388_144\"],[\"m885388_146\"],[\"m885388_148\"],[\"TimelineContentLoader\",\"setTimeCutoff\",[],[1351254247]],[\"TimelineContentLoader\",\"removeSection\",[],[\"month_2012_11\"]],[\"TimelineContentLoader\",\"enableScrollLoad\",[],[\"u4kzrn231_scroll_trigger\",\"recent\",\"0\",250]],[\"TimelineLogging\",\"init\",[],[143412869029]],[\"TimelineController\"],[\"Arbiter\",\"inform\",[],[\"TimelineConstants\\/sectionLoaded\",{\"key\":\"recent\",\"hideSubSections\":true}]]]},\"css\":[\"gYYej\",\"FjEEx\",\"M7bAm\",\"GhKAb\",\"+tEeI\"],\"bootloadable\":{\"ufi-tracking-js\":{\"resources\":[\"JU6Q1\",\"gM1Bs\"]},\"legacy:dom-collect\":{\"resources\":[\"JU6Q1\",\"S0jPm\"]},\"ScrollAwareDOM\":{\"resources\":[\"JU6Q1\",\"MX2xp\",\"rKoQM\"],\"module\":true},\"DialogX\":{\"resources\":[\"JU6Q1\",\"MX2xp\",\"rKoQM\",\"M7bAm\"],\"module\":true},\"UFIRetryActions\":{\"resources\":[\"JU6Q1\",\"gM1Bs\",\"8ay4F\"],\"module\":true},\"ScrollableArea\":{\"resources\":[\"JU6Q1\",\"MX2xp\",\"rKoQM\",\"M7bAm\"],\"module\":true},\"CompactTypeaheadRenderer\":{\"resources\":[\"JU6Q1\",\"M7bAm\",\"rKoQM\"],\"module\":true},\"ContextualTypeaheadView\":{\"resources\":[\"JU6Q1\",\"rKoQM\",\"MX2xp\",\"M7bAm\"],\"module\":true},\"InputSelection\":{\"resources\":[\"JU6Q1\",\"MX2xp\"],\"module\":true},\"MentionsInput\":{\"resources\":[\"JU6Q1\",\"MX2xp\",\"M7bAm\",\"rKoQM\",\"\\/eRBf\"],\"module\":true},\"TextAreaControl\":{\"resources\":[\"JU6Q1\",\"rKoQM\",\"M7bAm\"],\"module\":true},\"Typeahead\":{\"resources\":[\"JU6Q1\",\"MX2xp\",\"M7bAm\",\"rKoQM\"],\"module\":true},\"TypeaheadAreaCore\":{\"resources\":[\"JU6Q1\",\"MX2xp\",\"rKoQM\",\"\\/eRBf\"],\"module\":true},\"TypeaheadBestName\":{\"resources\":[\"JU6Q1\",\"rKoQM\"],\"module\":true},\"TypeaheadHoistFriends\":{\"resources\":[\"\\/eRBf\"],\"module\":true},\"TypingDetector\":{\"resources\":[\"JU6Q1\",\"XQYFa\"],\"module\":true},\"ContextualDialogX\":{\"resources\":[\"JU6Q1\",\"rKoQM\",\"MX2xp\",\"M7bAm\"],\"module\":true},\"DOMScroll\":{\"resources\":[\"JU6Q1\",\"MX2xp\"],\"module\":true},\"UFILegacyUtils\":{\"resources\":[\"JU6Q1\",\"rKoQM\",\"MX2xp\",\"M7bAm\",\"GnDzR\",\"RehkB\",\"4dUxD\"],\"module\":true},\"UFIScrollHighlight\":{\"resources\":[\"JU6Q1\",\"MX2xp\",\"gM1Bs\",\"rSdpp\"],\"module\":true},\"UFICommentRemovalControls\":{\"resources\":[\"gM1Bs\",\"JU6Q1\",\"M7bAm\",\"zDD1Q\"],\"module\":true},\"legacy:live-timer\":{\"resources\":[\"JU6Q1\"]}},\"resource_map\":{\"GhKAb\":{\"type\":\"css\",\"nonblocking\":1,\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/yT\\/r\\/0u_EQbkO7v5.css\"},\"+tEeI\":{\"type\":\"css\",\"permanent\":1,\"nonblocking\":1,\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/yG\\/r\\/_T6pUZoTIzY.css\"},\"S0jPm\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/y3\\/r\\/pnfMB7xvR7T.js\"},\"8ay4F\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/yP\\/r\\/QwvlOnFPoWJ.js\"},\"\\/eRBf\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/yO\\/r\\/0NMoB5FKoxZ.js\"},\"GnDzR\":{\"type\":\"css\",\"permanent\":1,\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/yE\\/r\\/6yM77Pd7IOM.css\"},\"RehkB\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/yp\\/r\\/1BwvjpU5pi8.js\"},\"4dUxD\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/yF\\/r\\/DppFVi35Rfx.js\"},\"rSdpp\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/yc\\/r\\/2PkTFjdVPzc.js\"},\"zDD1Q\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/yS\\/r\\/6zZtZHIyeXO.js\"}},\"js\":[\"JU6Q1\",\"MX2xp\",\"xjH7g\",\"rKoQM\",\"gM1Bs\",\"M2kwM\"],\"displayJS\":[\"gM1Bs\",\"JU6Q1\"],\"onafterload\":[\"Bootloader.loadComponents([\\\"ufi-tracking-js\\\"], function(){ ufi_add_all_link_data(); });\",\"Bootloader.loadComponents([\\\"legacy:live-timer\\\"], function(){ LiveTimer.restart(1352998853); });\"],\"id\":\"pagelet_timeline_recent\",\"phase\":2,\"tti_phase\":2})</script>\n", 
        "\n", 
        "<script>bigPipe.onPageletArrive({\"content\":{\"page_nax_callout\":\"\"},\"id\":\"page_nax_callout\",\"phase\":3})</script>\n", 
        "\n", 
        "<code class=\"hidden_elem\" id=\"u4kzrn2180\"><!-- <div class=\"fbTimelineStickyHeader fixed_elem fbTimelineStickyHeaderHidden\" aria-hidden=\"true\" id=\"u4kzrn2179\"><div class=\"stickyHeaderWrap clearfix\"><div class=\"back\"></div><div class=\"name\"><a class=\"profileThumb\" href=\"http://www.facebook.com/NorthCarolinaStateParks\"><img class=\"_s0 _rw img\" src=\"https://fbcdn-profile-a.akamaihd.net/hprofile-ak-ash3/174880_143412869029_133040642_q.jpg\" alt=\"\" /></a><span class=\"uiButtonGroup fbStickyHeaderBreadcrumb uiButtonGroupOverlay\" id=\"u4kzrn2177\"><span class=\"firstItem uiButtonGroupItem buttonItem\"><a class=\"nameButton uiButton uiButtonOverlay\" href=\"http://www.facebook.com/NorthCarolinaStateParks\" role=\"button\"><span class=\"uiButtonText\">North Carolina State Parks and Recreation</span></a></span><span class=\"uiButtonGroupItem selectorItem\"><div class=\"uiSelector inlineBlock pageMenu uiSelectorNormal uiSelectorDynamicLabel\"><div class=\"wrap\"><a class=\"pageMenuButton uiSelectorButton uiButton uiButtonOverlay\" href=\"#\" role=\"button\" aria-haspopup=\"1\" data-label=\"Timeline\" data-length=\"30\" rel=\"toggle\"><span class=\"uiButtonText\">Timeline</span></a><div class=\"uiSelectorMenuWrapper uiToggleFlyout\"><div role=\"menu\" class=\"uiMenu uiSelectorMenu\"><ul class=\"uiMenuInner\"><li class=\"uiMenuItem uiMenuItemCheckbox checked\" data-label=\"Timeline\"><a class=\"itemAnchor itemWithIcon\" role=\"menuitemcheckbox\" tabindex=\"0\" href=\"http://www.facebook.com/NorthCarolinaStateParks\" aria-checked=\"true\"><i class=\"mrs itemIcon img sp_2qck9m sx_3d437a\"></i><span class=\"itemLabel fsm\">Timeline</span></a></li><li class=\"uiMenuItem uiMenuItemCheckbox\" data-label=\"About\"><a class=\"itemAnchor itemWithIcon\" role=\"menuitemcheckbox\" tabindex=\"-1\" href=\"http://www.facebook.com/NorthCarolinaStateParks/info\" aria-checked=\"false\"><i class=\"mrs itemIcon img sp_2vmry3 sx_1c39cb\"></i><span class=\"itemLabel fsm\">About</span></a></li><li class=\"uiMenuSeparator separator hidden_elem\"></li></ul></div></div></div></div></span><span class=\"lastItem uiButtonGroupItem selectorItem\"><div class=\"uiSelector inlineBlock sectionMenu uiSelectorNormal uiSelectorDynamicLabel\"><div class=\"wrap\"><a class=\"uiSelectorButton uiButton uiButtonOverlay\" href=\"#\" role=\"button\" aria-haspopup=\"1\" data-length=\"30\" rel=\"toggle\"><span class=\"uiButtonText\">Now</span></a><div class=\"uiSelectorMenuWrapper uiToggleFlyout\"><div role=\"menu\" class=\"uiMenu uiSelectorMenu\"><ul class=\"uiMenuInner\"><li class=\"uiMenuItem uiMenuItemRadio uiSelectorOption checked\" data-label=\"Now\"><a class=\"itemAnchor\" role=\"menuitemradio\" tabindex=\"0\" href=\"#\" aria-checked=\"true\" data-key=\"recent\"><span class=\"itemLabel fsm\">Now</span></a></li><li class=\"uiMenuItem uiMenuItemRadio uiSelectorOption\" data-label=\"November\"><a class=\"itemAnchor\" role=\"menuitemradio\" tabindex=\"-1\" href=\"#\" aria-checked=\"false\" data-key=\"month_2012_11\"><span class=\"itemLabel fsm\">November</span></a></li><li class=\"uiMenuItem uiMenuItemRadio uiSelectorOption\" data-label=\"October\"><a class=\"itemAnchor\" role=\"menuitemradio\" tabindex=\"-1\" href=\"#\" aria-checked=\"false\" data-key=\"month_2012_10\"><span class=\"itemLabel fsm\">October</span></a></li><li class=\"uiMenuItem uiMenuItemRadio uiSelectorOption\" data-label=\"2012\"><a class=\"itemAnchor\" role=\"menuitemradio\" tabindex=\"-1\" href=\"#\" aria-checked=\"false\" data-key=\"year_2012\"><span class=\"itemLabel fsm\">2012</span></a></li><li class=\"uiMenuItem uiMenuItemRadio uiSelectorOption\" data-label=\"2011\"><a class=\"itemAnchor\" role=\"menuitemradio\" tabindex=\"-1\" href=\"#\" aria-checked=\"false\" data-key=\"year_2011\"><span class=\"itemLabel fsm\">2011</span></a></li><li class=\"uiMenuItem uiMenuItemRadio uiSelectorOption\" data-label=\"2010\"><a class=\"itemAnchor\" role=\"menuitemradio\" tabindex=\"-1\" href=\"#\" aria-checked=\"false\" data-key=\"year_2010\"><span class=\"itemLabel fsm\">2010</span></a></li><li class=\"uiMenuItem uiMenuItemRadio uiSelectorOption\" data-label=\"Joined Facebook\"><a class=\"itemAnchor\" role=\"menuitemradio\" tabindex=\"-1\" href=\"#\" aria-checked=\"false\" data-key=\"year_2009\"><span class=\"itemLabel fsm\">Joined Facebook</span></a></li></ul></div></div></div><select><option value=\"\"></option><option value=\"recent\" selected=\"1\">Now</option><option value=\"month_2012_11\">November</option><option value=\"month_2012_10\">October</option><option value=\"year_2012\">2012</option><option value=\"year_2011\">2011</option><option value=\"year_2010\">2010</option><option value=\"year_2009\">Joined Facebook</option></select></div></span><span class=\"uiButtonGroupItem selectorItem hidden_elem\"><div class=\"uiSelector inlineBlock subsectionMenu uiSelectorNormal uiSelectorDynamicLabel\"><div class=\"wrap\"><a class=\"uiSelectorButton uiButton uiButtonOverlay\" href=\"#\" role=\"button\" aria-haspopup=\"1\" data-length=\"30\" rel=\"toggle\"><span class=\"uiButtonText\">Highlights</span></a><div class=\"uiSelectorMenuWrapper uiToggleFlyout\"><div role=\"menu\" class=\"uiMenu uiSelectorMenu\"><ul class=\"uiMenuInner\"><li class=\"uiMenuItem uiMenuItemRadio uiSelectorOption highlights checked\" data-label=\"Highlights\"><a class=\"itemAnchor\" role=\"menuitemradio\" tabindex=\"0\" href=\"#\" aria-checked=\"true\"><span class=\"itemLabel fsm\">Highlights</span></a></li><li class=\"uiMenuItem uiMenuItemRadio uiSelectorOption allStories\" data-label=\"All Stories\"><a class=\"itemAnchor\" role=\"menuitemradio\" tabindex=\"-1\" href=\"#\" aria-checked=\"false\"><span class=\"itemLabel fsm\">All Stories</span></a></li><li class=\"uiMenuSeparator separator hidden_elem\"></li></ul></div></div></div><select><option value=\"\"></option><option value=\"highlights\" selected=\"1\">Highlights</option><option value=\"allStories\">All Stories</option></select></div></span></span></div><div class=\"actions\"><span class=\"uiButtonGroup fbTimelineConnectButtonGroup uiButtonGroupOverlay\" id=\"u4kzrn2178\"><span class=\"firstItem uiButtonGroupItem buttonItem\"><a class=\"uiButton uiButtonOverlay uiButtonLarge\" href=\"/ajax/timeline/sign_up_dialog/?page_id=143412869029&amp;next=http%3A%2F%2Fwww.facebook.com%2FNorthCarolinaStateParks\" role=\"button\" rel=\"dialog\"><i class=\"mrs img sp_2qck9m sx_e3c960\"></i><span class=\"uiButtonText\">Like</span></a></span><span class=\"lastItem uiButtonGroupItem buttonItem\"></span></span></div></div></div> --></code>\n", 
        "<script>bigPipe.onPageletArrive({\"content\":{\"timeline_sticky_header\":{\"container_id\":\"u4kzrn2180\"}},\"jsmods\":{\"elements\":[[\"m885394_173\",\"u4kzrn2177\",4],[\"m885395_174\",\"u4kzrn2178\",2],[\"m885394_172\",\"u4kzrn2179\",2]],\"require\":[[\"TimelineStickyHeader\",\"init\",[\"m885394_172\"],[{\"__m\":\"m885394_172\"}]],[\"TimelineStickyHeaderNav\",\"init\",[\"m885394_173\"],[{\"__m\":\"m885394_173\"},{\"custom_subsection_menu\":true,\"pull_from_scrubber\":false}]]]},\"css\":[\"M7bAm\",\"BbtG6\"],\"bootloadable\":{\"StickyController\":{\"resources\":[\"JU6Q1\",\"MX2xp\",\"M7bAm\",\"rKoQM\"],\"module\":true}},\"js\":[\"JU6Q1\",\"MX2xp\",\"rKoQM\",\"M2kwM\"],\"id\":\"timeline_sticky_header\",\"phase\":3})</script>\n", 
        "\n", 
        "<code class=\"hidden_elem\" id=\"u4kzrn2187\"><!-- <div class=\"fbTimelineTimePeriod fbTimelineTimePeriodUnexpanded\" id=\"pagelet_timeline_earlier_this_month_all\"><div class=\"fbTimelineSection fbTimelineCompactSection fbTimelineSectionTransparent\"><div class=\"fbTimelinePeriodPlaceholder\"><div class=\"uiHeader fbTimelineContentHeader sectionHeader\"><div class=\"clearfix uiHeaderTop\"><div><h3 class=\"uiHeaderTitle\">Earlier in November</h3></div></div></div><div class=\"loadingContainer\"><a class=\"mbm phm forceLoad\" href=\"#\" id=\"u4kzrn2181\"><span class=\"sectionLabel fwb\" data-year=\"2012\" data-month=\"November\">Show November</span></a></div></div></div><img class=\"ptl loadingIndicator img\" src=\"https://s-static.ak.facebook.com/rsrc.php/v2/yb/r/GsNJNwuI-UM.gif\" alt=\"\" width=\"16\" height=\"11\" /></div><div class=\"fbTimelineTimePeriod fbTimelineTimePeriodUnexpanded\" id=\"pagelet_timeline_month_all_last\"><div class=\"fbTimelineSection fbTimelineCompactSection fbTimelineSectionTransparent\"><div class=\"fbTimelinePeriodPlaceholder\"><div class=\"uiHeader fbTimelineContentHeader sectionHeader\"><div class=\"clearfix uiHeaderTop\"><div><h3 class=\"uiHeaderTitle\">Earlier in October</h3></div></div></div><div class=\"loadingContainer\"><a class=\"mbm phm forceLoad\" href=\"#\" id=\"u4kzrn2182\"><span class=\"sectionLabel fwb\" data-year=\"2012\" data-month=\"October\">Show October</span></a></div></div></div><img class=\"ptl loadingIndicator img\" src=\"https://s-static.ak.facebook.com/rsrc.php/v2/yb/r/GsNJNwuI-UM.gif\" alt=\"\" width=\"16\" height=\"11\" /></div><div class=\"fbTimelineTimePeriod fbTimelineTimePeriodUnexpanded\" id=\"pagelet_timeline_year_current\"><div class=\"fbTimelineSection fbTimelineCompactSection fbTimelineSectionTransparent\"><div class=\"fbTimelinePeriodPlaceholder\"><div class=\"uiHeader fbTimelineContentHeader sectionHeader\"><div class=\"clearfix uiHeaderTop\"><div><h3 class=\"uiHeaderTitle\">Earlier in 2012</h3></div></div></div><div class=\"loadingContainer\"><a class=\"mbm phm forceLoad\" href=\"#\" id=\"u4kzrn2183\"><span class=\"sectionLabel fwb\" data-year=\"2012\">Show 2012</span></a></div></div></div><img class=\"ptl loadingIndicator img\" src=\"https://s-static.ak.facebook.com/rsrc.php/v2/yb/r/GsNJNwuI-UM.gif\" alt=\"\" width=\"16\" height=\"11\" /></div><div class=\"fbTimelineTimePeriod fbTimelineTimePeriodUnexpanded\" id=\"pagelet_timeline_year_last\"><div class=\"fbTimelineSection fbTimelineCompactSection fbTimelineSectionTransparent\"><div class=\"fbTimelinePeriodPlaceholder\"><div class=\"uiHeader fbTimelineContentHeader sectionHeader\"><div class=\"clearfix uiHeaderTop\"><div><h3 class=\"uiHeaderTitle\">2011</h3></div></div></div><div class=\"loadingContainer\"><a class=\"mbm phm forceLoad\" href=\"#\" id=\"u4kzrn2184\"><span class=\"sectionLabel fwb\" data-year=\"2011\">Show 2011</span></a></div></div></div><img class=\"ptl loadingIndicator img\" src=\"https://s-static.ak.facebook.com/rsrc.php/v2/yb/r/GsNJNwuI-UM.gif\" alt=\"\" width=\"16\" height=\"11\" /></div><div class=\"fbTimelineTimePeriod fbTimelineTimePeriodUnexpanded\" id=\"pagelet_timeline_year_2010\"><div class=\"fbTimelineSection fbTimelineCompactSection fbTimelineSectionTransparent\"><div class=\"fbTimelinePeriodPlaceholder\"><div class=\"uiHeader fbTimelineContentHeader sectionHeader\"><div class=\"clearfix uiHeaderTop\"><div><h3 class=\"uiHeaderTitle\">2010</h3></div></div></div><div class=\"loadingContainer\"><a class=\"mbm phm forceLoad\" href=\"#\" id=\"u4kzrn2185\"><span class=\"sectionLabel fwb\" data-year=\"2010\">Show 2010</span></a></div></div></div><img class=\"ptl loadingIndicator img\" src=\"https://s-static.ak.facebook.com/rsrc.php/v2/yb/r/GsNJNwuI-UM.gif\" alt=\"\" width=\"16\" height=\"11\" /></div><div class=\"fbTimelineTimePeriod fbTimelineTimePeriodUnexpanded\" id=\"pagelet_timeline_year_2009\"><div class=\"fbTimelineSection fbTimelineCompactSection fbTimelineSectionTransparent\"><div class=\"fbTimelinePeriodPlaceholder\"><div class=\"uiHeader fbTimelineContentHeader sectionHeader\"><div class=\"clearfix uiHeaderTop\"><div><h3 class=\"uiHeaderTitle\">Joined Facebook</h3></div></div></div><div class=\"loadingContainer\"><a class=\"mbm phm forceLoad\" href=\"#\" id=\"u4kzrn2186\"><span class=\"sectionLabel fwb\" data-year=\"2009\">Show Joined Facebook</span></a></div></div></div><img class=\"ptl loadingIndicator img\" src=\"https://s-static.ak.facebook.com/rsrc.php/v2/yb/r/GsNJNwuI-UM.gif\" alt=\"\" width=\"16\" height=\"11\" /></div> --></code>\n", 
        "<script>bigPipe.onPageletArrive({\"append\":\"timeline_tab_content_extra\",\"content\":{\"timeline_section_placeholders\":{\"container_id\":\"u4kzrn2187\"}},\"jsmods\":{\"elements\":[[\"m885396_175\",\"u4kzrn2181\",2],[\"m885396_177\",\"u4kzrn2182\",2],[\"m885396_179\",\"u4kzrn2183\",2],[\"m885396_181\",\"u4kzrn2184\",2],[\"m885396_183\",\"u4kzrn2185\",2],[\"m885396_185\",\"u4kzrn2186\",2],[\"m885396_176\",\"pagelet_timeline_earlier_this_month_all\",2],[\"m885396_178\",\"pagelet_timeline_month_all_last\",2],[\"m885396_180\",\"pagelet_timeline_year_current\",2],[\"m885396_182\",\"pagelet_timeline_year_last\",2],[\"m885396_184\",\"pagelet_timeline_year_2010\",2],[\"m885396_186\",\"pagelet_timeline_year_2009\",2]],\"require\":[[\"TimelineContentLoader\",\"loadSectionOnClick\",[\"m885396_175\"],[{\"__m\":\"m885396_175\"},\"month_2012_11\"]],[\"TimelineContentLoader\",\"registerTimePeriod\",[\"m885396_176\"],[{\"__m\":\"m885396_176\"},\"month_2012_11\",{\"profile_id\":143412869029,\"start\":1351753200,\"end\":1354348799,\"query_type\":31,\"section_pagelet_id\":\"pagelet_timeline_earlier_this_month_all\",\"load_immediately\":false},false,null,1,-1]],[\"TimelineContentLoader\",\"loadSectionOnClick\",[\"m885396_177\"],[{\"__m\":\"m885396_177\"},\"month_2012_10\"]],[\"TimelineContentLoader\",\"registerTimePeriod\",[\"m885396_178\"],[{\"__m\":\"m885396_178\"},\"month_2012_10\",{\"profile_id\":143412869029,\"start\":1349074800,\"end\":1351753199,\"query_type\":25,\"section_pagelet_id\":\"pagelet_timeline_month_all_last\",\"load_immediately\":false},false,null,2,-1]],[\"TimelineContentLoader\",\"loadSectionOnClick\",[\"m885396_179\"],[{\"__m\":\"m885396_179\"},\"year_2012\"]],[\"TimelineContentLoader\",\"registerTimePeriod\",[\"m885396_180\"],[{\"__m\":\"m885396_180\"},\"year_2012\",{\"profile_id\":143412869029,\"start\":1325404800,\"end\":1357027199,\"query_type\":8,\"filter_after_timestamp\":1349074799,\"section_pagelet_id\":\"pagelet_timeline_year_current\",\"load_immediately\":false},false,null,3,\"76.611952328237\"]],[\"TimelineContentLoader\",\"setExpandLoadDataForSection\",[],[\"year_2012\",{\"profile_id\":143412869029,\"start\":1325404800,\"end\":1357027199,\"query_type\":9}]],[\"TimelineContentLoader\",\"loadSectionOnClick\",[\"m885396_181\"],[{\"__m\":\"m885396_181\"},\"year_2011\"]],[\"TimelineContentLoader\",\"registerTimePeriod\",[\"m885396_182\"],[{\"__m\":\"m885396_182\"},\"year_2011\",{\"profile_id\":143412869029,\"start\":1293868800,\"end\":1325404799,\"query_type\":8,\"section_pagelet_id\":\"pagelet_timeline_year_last\",\"load_immediately\":false},false,null,4,\"68.172027046572\"]],[\"TimelineContentLoader\",\"setExpandLoadDataForSection\",[],[\"year_2011\",{\"profile_id\":143412869029,\"start\":1293868800,\"end\":1325404799,\"query_type\":9}]],[\"TimelineContentLoader\",\"loadSectionOnClick\",[\"m885396_183\"],[{\"__m\":\"m885396_183\"},\"year_2010\"]],[\"TimelineContentLoader\",\"registerTimePeriod\",[\"m885396_184\"],[{\"__m\":\"m885396_184\"},\"year_2010\",{\"profile_id\":143412869029,\"start\":1262332800,\"end\":1293868799,\"query_type\":8,\"section_pagelet_id\":\"pagelet_timeline_year_2010\",\"load_immediately\":false},false,null,5,\"68.372609818806\"]],[\"TimelineContentLoader\",\"setExpandLoadDataForSection\",[],[\"year_2010\",{\"profile_id\":143412869029,\"start\":1262332800,\"end\":1293868799,\"query_type\":9}]],[\"TimelineContentLoader\",\"loadSectionOnClick\",[\"m885396_185\"],[{\"__m\":\"m885396_185\"},\"year_2009\"]],[\"TimelineContentLoader\",\"registerTimePeriod\",[\"m885396_186\"],[{\"__m\":\"m885396_186\"},\"year_2009\",{\"profile_id\":143412869029,\"start\":1230796800,\"end\":1262332799,\"query_type\":8,\"section_pagelet_id\":\"pagelet_timeline_year_2009\",\"load_immediately\":false},false,null,6,\"31.350507030832\"]],[\"TimelineContentLoader\",\"setExpandLoadDataForSection\",[],[\"year_2009\",{\"profile_id\":143412869029,\"start\":1230796800,\"end\":1262332799,\"query_type\":9}]]]},\"css\":[\"gYYej\",\"M7bAm\"],\"js\":[\"JU6Q1\",\"MX2xp\",\"xjH7g\"],\"id\":\"timeline_section_placeholders\",\"phase\":3})</script>\n", 
        "\n", 
        "<code class=\"hidden_elem\" id=\"u4kzrn2188\"><!-- <div></div> --></code>\n", 
        "<script>bigPipe.onPageletArrive({\"content\":{\"pagelet_timeline_recent_more_pager\":{\"container_id\":\"u4kzrn2188\"}},\"id\":\"pagelet_timeline_recent_more_pager\",\"phase\":3})</script>\n", 
        "\n", 
        "<code class=\"hidden_elem\" id=\"u4kzrn2190\"><!-- <ul class=\"fbTimelineScrubber hidden_elem\" role=\"navigation\" data-gt=\"&#123;&quot;timeline_scrubber&quot;:&quot;1&quot;,&quot;profile_id&quot;:&quot;143412869029&quot;&#125;\" id=\"u4kzrn2189\"><li data-key=\"recent\" class=\"selected\"><a href=\"/NorthCarolinaStateParks/timeline\" rel=\"ignore\">Now</a></li><li data-key=\"month_2012_10\"><a href=\"/NorthCarolinaStateParks/timeline/2012/10\" rel=\"ignore\" tabindex=\"-1\">October</a></li><li data-key=\"year_2012\" class=\"clearfix\" data-rollup=\"2010s\" data-year=\"2012\"><a href=\"/NorthCarolinaStateParks/timeline/2012\" rel=\"ignore\" tabindex=\"-1\">2012</a><ul class=\"clearfix\"><li data-key=\"month_2012_9\"><a href=\"/NorthCarolinaStateParks/timeline/2012/9\" rel=\"ignore\">September</a></li><li data-key=\"month_2012_8\"><a href=\"/NorthCarolinaStateParks/timeline/2012/8\" rel=\"ignore\">August</a></li><li data-key=\"month_2012_7\"><a href=\"/NorthCarolinaStateParks/timeline/2012/7\" rel=\"ignore\">July</a></li><li data-key=\"month_2012_6\"><a href=\"/NorthCarolinaStateParks/timeline/2012/6\" rel=\"ignore\">June</a></li><li data-key=\"month_2012_5\"><a href=\"/NorthCarolinaStateParks/timeline/2012/5\" rel=\"ignore\">May</a></li><li data-key=\"month_2012_4\"><a href=\"/NorthCarolinaStateParks/timeline/2012/4\" rel=\"ignore\">April</a></li><li data-key=\"month_2012_3\"><a href=\"/NorthCarolinaStateParks/timeline/2012/3\" rel=\"ignore\">March</a></li><li data-key=\"month_2012_2\"><a href=\"/NorthCarolinaStateParks/timeline/2012/2\" rel=\"ignore\">February</a></li><li data-key=\"month_2012_1\"><a href=\"/NorthCarolinaStateParks/timeline/2012/1\" rel=\"ignore\">January</a></li></ul></li><li data-key=\"year_2011\" class=\"clearfix\" data-rollup=\"2010s\" data-year=\"2011\"><a href=\"/NorthCarolinaStateParks/timeline/2011\" rel=\"ignore\" tabindex=\"-1\">2011</a><ul class=\"clearfix\"><li data-key=\"month_2011_12\"><a href=\"/NorthCarolinaStateParks/timeline/2011/12\" rel=\"ignore\">December</a></li><li data-key=\"month_2011_11\"><a href=\"/NorthCarolinaStateParks/timeline/2011/11\" rel=\"ignore\">November</a></li><li data-key=\"month_2011_10\"><a href=\"/NorthCarolinaStateParks/timeline/2011/10\" rel=\"ignore\">October</a></li><li data-key=\"month_2011_9\"><a href=\"/NorthCarolinaStateParks/timeline/2011/9\" rel=\"ignore\">September</a></li><li data-key=\"month_2011_8\"><a href=\"/NorthCarolinaStateParks/timeline/2011/8\" rel=\"ignore\">August</a></li><li data-key=\"month_2011_7\"><a href=\"/NorthCarolinaStateParks/timeline/2011/7\" rel=\"ignore\">July</a></li><li data-key=\"month_2011_6\"><a href=\"/NorthCarolinaStateParks/timeline/2011/6\" rel=\"ignore\">June</a></li><li data-key=\"month_2011_5\"><a href=\"/NorthCarolinaStateParks/timeline/2011/5\" rel=\"ignore\">May</a></li><li data-key=\"month_2011_4\"><a href=\"/NorthCarolinaStateParks/timeline/2011/4\" rel=\"ignore\">April</a></li><li data-key=\"month_2011_3\"><a href=\"/NorthCarolinaStateParks/timeline/2011/3\" rel=\"ignore\">March</a></li><li data-key=\"month_2011_2\"><a href=\"/NorthCarolinaStateParks/timeline/2011/2\" rel=\"ignore\">February</a></li><li data-key=\"month_2011_1\"><a href=\"/NorthCarolinaStateParks/timeline/2011/1\" rel=\"ignore\">January</a></li></ul></li><li data-key=\"year_2010\" class=\"clearfix\" data-rollup=\"2010s\" data-year=\"2010\"><a href=\"/NorthCarolinaStateParks/timeline/2010\" rel=\"ignore\" tabindex=\"-1\">2010</a><ul class=\"clearfix\"><li data-key=\"month_2010_12\"><a href=\"/NorthCarolinaStateParks/timeline/2010/12\" rel=\"ignore\">December</a></li><li data-key=\"month_2010_11\"><a href=\"/NorthCarolinaStateParks/timeline/2010/11\" rel=\"ignore\">November</a></li><li data-key=\"month_2010_10\"><a href=\"/NorthCarolinaStateParks/timeline/2010/10\" rel=\"ignore\">October</a></li><li data-key=\"month_2010_9\"><a href=\"/NorthCarolinaStateParks/timeline/2010/9\" rel=\"ignore\">September</a></li><li data-key=\"month_2010_8\"><a href=\"/NorthCarolinaStateParks/timeline/2010/8\" rel=\"ignore\">August</a></li><li data-key=\"month_2010_7\"><a href=\"/NorthCarolinaStateParks/timeline/2010/7\" rel=\"ignore\">July</a></li><li data-key=\"month_2010_6\"><a href=\"/NorthCarolinaStateParks/timeline/2010/6\" rel=\"ignore\">June</a></li><li data-key=\"month_2010_5\"><a href=\"/NorthCarolinaStateParks/timeline/2010/5\" rel=\"ignore\">May</a></li><li data-key=\"month_2010_4\"><a href=\"/NorthCarolinaStateParks/timeline/2010/4\" rel=\"ignore\">April</a></li><li data-key=\"month_2010_3\"><a href=\"/NorthCarolinaStateParks/timeline/2010/3\" rel=\"ignore\">March</a></li><li data-key=\"month_2010_2\"><a href=\"/NorthCarolinaStateParks/timeline/2010/2\" rel=\"ignore\">February</a></li><li data-key=\"month_2010_1\"><a href=\"/NorthCarolinaStateParks/timeline/2010/1\" rel=\"ignore\">January</a></li></ul></li><li data-key=\"year_2009\" class=\"clearfix\"><a href=\"/NorthCarolinaStateParks/timeline/2009\" rel=\"ignore\" tabindex=\"-1\">Joined Facebook</a><ul class=\"clearfix\"><li data-key=\"month_2009_12\"><a href=\"/NorthCarolinaStateParks/timeline/2009/12\" rel=\"ignore\">December</a></li><li data-key=\"month_2009_11\"><a href=\"/NorthCarolinaStateParks/timeline/2009/11\" rel=\"ignore\">November</a></li><li data-key=\"month_2009_10\"><a href=\"/NorthCarolinaStateParks/timeline/2009/10\" rel=\"ignore\">October</a></li><li data-key=\"month_2009_9\"><a href=\"/NorthCarolinaStateParks/timeline/2009/9\" rel=\"ignore\">September</a></li></ul></li></ul> --></code>\n", 
        "<script>bigPipe.onPageletArrive({\"append\":\"rightColContent\",\"content\":{\"pagelet_timeline_scrubber\":{\"container_id\":\"u4kzrn2190\"}},\"css\":[\"gYYej\"],\"resource_map\":{\"Xta22\":{\"type\":\"js\",\"src\":\"https:\\/\\/s-static.ak.fbcdn.net\\/rsrc.php\\/v2\\/yM\\/r\\/h-EyXSwUkCW.js\"}},\"js\":[\"JU6Q1\",\"MX2xp\",\"rKoQM\",\"vVPJj\",\"Xta22\"],\"jscc_map\":\"({\\\"jR1G1Ekgzo4YUtuI0Ne0\\\":function(){return new PagesTimelineScrubber($(\\\"u4kzrn2189\\\"));}})\",\"onload\":[\"JSCC.get(\\\"jR1G1Ekgzo4YUtuI0Ne0\\\")\"],\"id\":\"pagelet_timeline_scrubber\",\"phase\":3})</script>\n", 
        "\n", 
        "<script>bigPipe.onPageletArrive({\"is_last\":true,\"css\":[\"M7bAm\",\"jg0IV\",\"zmU4p\"],\"js\":[\"JU6Q1\",\"MX2xp\",\"mCr\\/S\",\"UjXTQ\",\"T8gbI\"],\"id\":\"\",\"phase\":3,\"the_end\":true})</script>                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   </body></html>\n", 
    };

    /*
     * Settings for extracting scroll-down ajax urls from facebook, as they are
     * constructed as of Nov 14 2012.
     * 
     * <bean class="org.archive.modules.extractor.ExtractorMultipleRegex">
     *  <property name="uriRegex" value="^https?://(?:www\.)?facebook\.com/[^/?]+$" />
     *  <property name="contentRegexes">
     *   <map>
     *    <entry key="jsonBlob" value='\{("profile_id":\d+,[^}]+)\}' />
     *    <entry key="ajaxpipeToken" value='"ajaxpipe_token":"([^"]+)"' />
     *    <entry key="timeCutoff" value='"setTimeCutoff",[^,]*,\[(\d+)\]\]' />
     *   </map>
     *  </property>
     *  <property name="template">
     *   <value>/ajax/pagelet/generic.php/ProfileTimelineSectionPagelet?ajaxpipe=1&amp;ajaxpipe_token=${ajaxpipeToken[1]}&amp;no_script_path=1&amp;data=${java.net.URLEncoder.encode('{' + jsonBlob[1] + ',"time_cutoff":' + timeCutoff[1] + ',"force_no_friend_activity":false}', 'UTF-8')}&amp;__user=0&amp;__a=1&amp;__adt=${jsonBlobIndex}</value>
     *   </property>
     * </bean>
     */
    @Override
    protected Extractor makeExtractor() {
        ExtractorMultipleRegex e = new ExtractorMultipleRegex();
        UriErrorLoggerModule ulm = new UnitTestUriLoggerModule();  
        e.setLoggerModule(ulm);
        
        e.setUriRegex("^https?://(?:www\\.)?facebook\\.com/[^/?]+$");
        
        LinkedHashMap<String, String> contentRegexes = new LinkedHashMap<String,String>();
        contentRegexes.put("jsonBlob", "\\[\"TimelineContentLoader\",\"registerTimePeriod\",[^,]+,[^,]+,[^,]+,\\{(\"profile_id\":[^}]+)\\},false,null,(\\d+),");
        contentRegexes.put("ajaxpipeToken", "\"ajaxpipe_token\":\"([^\"]+)\"");
        contentRegexes.put("timeCutoff", "\"setTimeCutoff\",[^,]*,\\[(\\d+)\\]\\]");
        e.setContentRegexes(contentRegexes);
        
        e.setTemplate("/ajax/pagelet/generic.php/ProfileTimelineSectionPagelet"
                        + "?ajaxpipe=1&ajaxpipe_token=${ajaxpipeToken[1]}&no_script_path=1"
                        + "&data=${java.net.URLEncoder.encode('{' + jsonBlob[1] + ',\"time_cutoff\":' + timeCutoff[1] + ',\"force_no_friend_activity\":false}', 'UTF-8')}"
                        + "&__user=0&__a=1&__adt=${jsonBlob[2]}");

        return e;
    }

    public void testFacebookScrollExample() throws Exception {
        UURI testUuri = UURIFactory.getInstance(TEST_URI);
        CrawlURI testUri = new CrawlURI(testUuri, null, null, LinkContext.NAVLINK_MISC);
        StringBuilder content = new StringBuilder();
        for (String chunk: TEST_CONTENT_CHUNKS) {
            content.append(chunk);
        }
        Recorder recorder = createRecorder(content.toString(), "UTF-8");
        testUri.setContentType("text/html");
        testUri.setRecorder(recorder);
        testUri.setContentSize(content.length());
        
        extractor.process(testUri);
        
        for (String expectedLinkString: EXPECTED_OUTLINKS) {
            Link expectedLink = new Link(testUri.getUURI(), 
                    UURIFactory.getInstance(expectedLinkString),
                    HTMLLinkContext.INFERRED_MISC, Hop.INFERRED);
            assertTrue(testUri.getOutLinks().contains(expectedLink));
        }
    }
}
